/* -*- mode: c++; indent-tabs-mode: nil -*- */
/* @file QC_Datasource.qpp Datasource class definition

  Qore Programming Language

  Copyright 2003 - 2014 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>
#include <qore/intern/QC_Datasource.h>

static const char *DSC_ERR = "DATASOURCE-CONSTRUCTOR-ERROR";

static int ds_set_options(ManagedDatasource* ds, const QoreHashNode* opts, ExceptionSink* xsink) {
   ConstHashIterator hi(opts);
   while (hi.next()) {
      // skip "min" and "max" options
      if (!strcmp(hi.getKey(), "min") || !strcmp(hi.getKey(), "max"))
         continue;

      if (ds->setOption(hi.getKey(), hi.getValue(), xsink))
         return -1;
   }
   return 0;
}

static void ds_constructor_hash(const QoreHashNode* opts, QoreObject* self, Queue* queue, const AbstractQoreNode* arg, ExceptionSink* xsink) {
   if (arg && !queue)
      arg = 0;

   ReferenceHolder<Queue> q(queue, xsink);

   const char *str = check_hash_key(opts, "type", DSC_ERR, xsink);
   if (*xsink) return;

   if (!str) {
      xsink->raiseException(DSC_ERR, "expecting a string value with the 'type' key giving the driver name");
      return;
   }

   DBIDriver *db_driver = DBI.find(str, xsink);
   if (!db_driver) {
      if (!*xsink)
         xsink->raiseException("DATASOURCE-UNSUPPORTED-DATABASE", "DBI driver '%s' cannot be loaded", str);
      return;
   }

   ReferenceHolder<ManagedDatasource> ds(new ManagedDatasource(db_driver), xsink);

   str = check_hash_key(opts, "user", DSC_ERR, xsink);
   if (*xsink) return;
   if (str) ds->setPendingUsername(str);

   str = check_hash_key(opts, "pass", DSC_ERR, xsink);
   if (*xsink) return;
   if (str) ds->setPendingPassword(str);

   str = check_hash_key(opts, "db", DSC_ERR, xsink);
   if (*xsink) return;
   if (str) ds->setPendingDBName(str);

   str = check_hash_key(opts, "charset", DSC_ERR, xsink);
   if (*xsink) return;
   if (str) ds->setPendingDBEncoding(str);

   str = check_hash_key(opts, "host", DSC_ERR, xsink);
   if (*xsink) return;
   if (str) ds->setPendingHostName(str);

   bool found;
   int port = (int)opts->getKeyAsBigInt("port", found);
   if (port) {
      if (port < 0) {
         xsink->raiseException(DSC_ERR, "port value must be zero (meaning use the default port) or positive (value given: %d)", port);
         return;
      }

      ds->setPendingPort(port);
   }

   {
      const AbstractQoreNode* p = opts->getKeyValue("options");

      if (!is_nothing(p)) {
         if (get_node_type(p) != NT_HASH) {
            xsink->raiseException(DSC_ERR, "'options' key is not type 'hash' but is type '%s'", get_type_name(p));
            return;
         }
         const QoreHashNode *oh = reinterpret_cast<const QoreHashNode*>(p);
         if (ds_set_options(*ds, oh, xsink))
            return;
      }
   }

   if (queue)
      ds->setEventQueue(q.release(), arg ? arg->refSelf() : 0, xsink);

   self->setPrivate(CID_DATASOURCE, ds.release());
}
/* Qore class Qore::SQL::Datasource */

qore_classid_t CID_DATASOURCE;
QoreClass* QC_DATASOURCE;

// nothing Datasource::beginTransaction(){}
static AbstractQoreNode* Datasource_beginTransaction(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 801 "QC_Datasource.qpp"
ds->beginTransaction(xsink);
   return 0;
}

// nothing Datasource::clearEventQueue(){}
static AbstractQoreNode* Datasource_clearEventQueue(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 1275 "QC_Datasource.qpp"
ds->setEventQueue(0, 0, xsink);
   return 0;
}

// nothing Datasource::close(){}
static AbstractQoreNode* Datasource_close(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 441 "QC_Datasource.qpp"
ds->close(xsink);
   return 0;
}

// nothing Datasource::commit(){}
static AbstractQoreNode* Datasource_commit(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 456 "QC_Datasource.qpp"
ds->commit(xsink);
   return 0;
}

// Datasource::constructor(string driver, *string user, *string pass, *string db, *string encoding, *string host, *softint port, *hash options, *Qore::Thread::Queue queue, any arg) {}
static void Datasource_constructor_VsNsNsNsNsNsniNhC5QueueVa(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* driver = HARD_QORE_STRING(args, 0);
   const QoreStringNode* user = reinterpret_cast<const QoreStringNode*>(get_param(args, 1));
   const QoreStringNode* pass = reinterpret_cast<const QoreStringNode*>(get_param(args, 2));
   const QoreStringNode* db = reinterpret_cast<const QoreStringNode*>(get_param(args, 3));
   const QoreStringNode* encoding = reinterpret_cast<const QoreStringNode*>(get_param(args, 4));
   const QoreStringNode* host = reinterpret_cast<const QoreStringNode*>(get_param(args, 5));
   const QoreBigIntNode* tmp_port = reinterpret_cast<const QoreBigIntNode*>(get_param(args, 6));
   int64 port = tmp_port ? tmp_port->val : 0;
   const QoreHashNode* options = reinterpret_cast<const QoreHashNode*>(get_param(args, 7));
   HARD_QORE_OBJ_OR_NOTHING_DATA(queue, Queue, args, 8, CID_QUEUE, xsink);
   if (*xsink)
      return;
   const AbstractQoreNode* arg = get_param(args, 9);
# 279 "QC_Datasource.qpp"
ReferenceHolder<Queue> q(queue, xsink);

   if (!port && driver->find(':') != -1) {
      ReferenceHolder<QoreHashNode> h(parseDatasource(driver->getBuffer(), xsink), xsink);
      if (!h) {
         assert(*xsink);
         return;
      }
      ds_constructor_hash(*h, self, q.release(), arg, xsink);
      return;
   }

   if (arg && !queue)
      arg = 0;

   DBIDriver *db_driver = DBI.find(driver->getBuffer(), xsink);
   if (!db_driver) {
      if (!*xsink)
	 xsink->raiseException("DATASOURCE-UNSUPPORTED-DATABASE", "DBI driver '%s' cannot be loaded", driver->getBuffer());
      return;
   }

   ReferenceHolder<ManagedDatasource> ds(new ManagedDatasource(db_driver), xsink);

   if (user && !user->empty())
      ds->setPendingUsername(user->getBuffer());

   if (pass && !pass->empty())
      ds->setPendingPassword(pass->getBuffer());

   if (db && !db->empty())
      ds->setPendingDBName(db->getBuffer());

   if (encoding && !encoding->empty())
      ds->setPendingDBEncoding(encoding->getBuffer());
   
   if (host && !host->empty())
      ds->setPendingHostName(host->getBuffer());
   
   if (port) {
      if (port < 0) {
	 xsink->raiseException(DSC_ERR, "port value must be zero (meaning use the default port) or positive (value given: %d)", port);
	 return;
      }
      ds->setPendingPort((int)port);
   }

   if (options)
      ds_set_options(*ds, options, xsink);

   if (queue)
      ds->setEventQueue(q.release(), arg ? arg->refSelf() : 0, xsink);

   self->setPrivate(CID_DATASOURCE, ds.release());
}

// Datasource::constructor(string desc, *Qore::Thread::Queue queue, any arg) {}
static void Datasource_constructor_VsC5QueueVa(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* desc = HARD_QORE_STRING(args, 0);
   HARD_QORE_OBJ_OR_NOTHING_DATA(queue, Queue, args, 1, CID_QUEUE, xsink);
   if (*xsink)
      return;
   const AbstractQoreNode* arg = get_param(args, 2);
# 352 "QC_Datasource.qpp"
ReferenceHolder<Queue> q(queue, xsink);
   ReferenceHolder<QoreHashNode> h(parseDatasource(desc->getBuffer(), xsink), xsink);
   if (!h) {
      assert(*xsink);
      return;
   }
   ds_constructor_hash(*h, self, q.release(), arg, xsink);
   return;
}

// Datasource::constructor(hash opts, *Qore::Thread::Queue queue, any arg) {}
static void Datasource_constructor_VhC5QueueVa(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreHashNode* opts = HARD_QORE_HASH(args, 0);
   HARD_QORE_OBJ_OR_NOTHING_DATA(queue, Queue, args, 1, CID_QUEUE, xsink);
   if (*xsink)
      return;
   const AbstractQoreNode* arg = get_param(args, 2);
# 385 "QC_Datasource.qpp"
ds_constructor_hash(opts, self, queue, arg, xsink);
}

// Datasource::copy() {}
static void Datasource_copy(QoreObject* self, QoreObject* old, ManagedDatasource* ds, ExceptionSink* xsink) {
# 410 "QC_Datasource.qpp"
self->setPrivate(CID_DATASOURCE, ds->copy());
}

// bool Datasource::currentThreadInTransaction(){}
static bool Datasource_currentThreadInTransaction(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 1155 "QC_Datasource.qpp"
return ds->currentThreadInTransaction();
}

// *hash Datasource::describe(string sql, ...){}
static AbstractQoreNode* Datasource_describe_VsVV(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
# 781 "QC_Datasource.qpp"
ReferenceHolder<QoreListNode> argl((args->size() > 1 ? args->copyListFrom(1) : 0), xsink);
   return ds->describe(sql, *argl, xsink);
}

// Datasource::destructor() {}
static void Datasource_destructor(QoreObject* self, ManagedDatasource* ds, ExceptionSink* xsink) {
# 398 "QC_Datasource.qpp"
ds->destructor(xsink);
   ds->deref(xsink);
}

// any Datasource::exec(string sql, ...){}
static AbstractQoreNode* Datasource_exec_VsVV(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
# 515 "QC_Datasource.qpp"
ReferenceHolder<QoreListNode> argl((args->size() > 1 ? args->copyListFrom(1) : 0), xsink);
   return ds->exec(sql, *argl, xsink);
}

// any Datasource::execRaw(string sql){}
static AbstractQoreNode* Datasource_execRaw_Vs(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
# 562 "QC_Datasource.qpp"
return ds->execRaw(sql, xsink);
}

// bool Datasource::getAutoCommit(){}
static bool Datasource_getAutoCommit(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 494 "QC_Datasource.qpp"
return ds->getAutoCommit();
}

// int Datasource::getCapabilities(){}
static int64 Datasource_getCapabilities(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 830 "QC_Datasource.qpp"
return ds->getCapabilities();
}

// list Datasource::getCapabilityList(){}
static AbstractQoreNode* Datasource_getCapabilityList(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 844 "QC_Datasource.qpp"
return ds->getCapabilityList();
}

// any Datasource::getClientVersion(){}
static AbstractQoreNode* Datasource_getClientVersion(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 1127 "QC_Datasource.qpp"
return ds->getClientVersion(xsink);
}

// hash Datasource::getConfigHash(){}
static AbstractQoreNode* Datasource_getConfigHash(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 1233 "QC_Datasource.qpp"
return ds->getConfigHash();
}

// string Datasource::getConfigString(){}
static AbstractQoreNode* Datasource_getConfigString(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 1247 "QC_Datasource.qpp"
return ds->getConfigString();
}

// string Datasource::getDBCharset(){}
static AbstractQoreNode* Datasource_getDBCharset(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 994 "QC_Datasource.qpp"
return ds->getPendingDBEncoding();
}

// string Datasource::getDBEncoding(){}
static AbstractQoreNode* Datasource_getDBEncoding(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 1008 "QC_Datasource.qpp"
return ds->getPendingDBEncoding();
}

// *string Datasource::getDBName(){}
static AbstractQoreNode* Datasource_getDBName(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 980 "QC_Datasource.qpp"
return ds->getPendingDBName();
}

// string Datasource::getDriverName(){}
static AbstractQoreNode* Datasource_getDriverName(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 1097 "QC_Datasource.qpp"
return new QoreStringNode(ds->getDriver()->getName());
}

// *string Datasource::getHostName(){}
static AbstractQoreNode* Datasource_getHostName(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 1048 "QC_Datasource.qpp"
return ds->getPendingHostName();
}

// string Datasource::getOSCharset(){}
static AbstractQoreNode* Datasource_getOSCharset(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 1022 "QC_Datasource.qpp"
const QoreEncoding *enc = ds->getQoreEncoding();
   return new QoreStringNode(enc ? enc->getCode() : "(unknown)");
}

// *string Datasource::getOSEncoding(){}
static AbstractQoreNode* Datasource_getOSEncoding(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 1035 "QC_Datasource.qpp"
const QoreEncoding *enc = ds->getQoreEncoding();
   return enc ? new QoreStringNode(enc->getCode()) : 0;
}

// any Datasource::getOption(string opt){}
static AbstractQoreNode* Datasource_getOption_Vs(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* opt = HARD_QORE_STRING(args, 0);
# 1219 "QC_Datasource.qpp"
return ds->getOption(opt->getBuffer(), xsink);
}

// hash Datasource::getOptionHash(){}
static AbstractQoreNode* Datasource_getOptionHash(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 1190 "QC_Datasource.qpp"
return ds->getOptionHash(xsink);
}

// *string Datasource::getPassword(){}
static AbstractQoreNode* Datasource_getPassword(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 968 "QC_Datasource.qpp"
return ds->getPendingPassword();
}

// *int Datasource::getPort(){}
static AbstractQoreNode* Datasource_getPort(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 1060 "QC_Datasource.qpp"
int port = ds->getPort();
   return port ? new QoreBigIntNode(port) : 0;
}

// any Datasource::getServerVersion(){}
static AbstractQoreNode* Datasource_getServerVersion(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 1113 "QC_Datasource.qpp"
return ds->getServerVersion(xsink);
}

// int Datasource::getTransactionLockTimeout(){}
static int64 Datasource_getTransactionLockTimeout(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 1085 "QC_Datasource.qpp"
return ds->getTransactionLockTimeout();
}

// *string Datasource::getUserName(){}
static AbstractQoreNode* Datasource_getUserName(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 956 "QC_Datasource.qpp"
return ds->getPendingUsername();
}

// bool Datasource::inTransaction(){}
static bool Datasource_inTransaction(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 1139 "QC_Datasource.qpp"
return ds->isInTransaction();
}

// nothing Datasource::open(){}
static AbstractQoreNode* Datasource_open(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 426 "QC_Datasource.qpp"
ds->open(xsink);
   return 0;
}

// nothing Datasource::reset(){}
static AbstractQoreNode* Datasource_reset(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 816 "QC_Datasource.qpp"
ds->reset(xsink);
   return 0;
}

// nothing Datasource::rollback(){}
static AbstractQoreNode* Datasource_rollback(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 471 "QC_Datasource.qpp"
ds->rollback(xsink);
   return 0;
}

// any Datasource::select(string sql, ...){}
static AbstractQoreNode* Datasource_select_VsVV(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
# 598 "QC_Datasource.qpp"
ReferenceHolder<QoreListNode> argl((args->size() > 1 ? args->copyListFrom(1) : 0), xsink);
   return ds->select(sql, *argl, xsink);
}

// *hash Datasource::selectRow(string sql, ...){}
static AbstractQoreNode* Datasource_selectRow_VsVV(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
# 625 "QC_Datasource.qpp"
ReferenceHolder<QoreListNode> argl((args->size() > 1 ? args->copyListFrom(1) : 0), xsink);
   return ds->selectRow(sql, *argl, xsink);
}

// any Datasource::selectRows(string sql, ...){}
static AbstractQoreNode* Datasource_selectRows_VsVV(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
# 657 "QC_Datasource.qpp"
ReferenceHolder<QoreListNode> argl((args->size() > 1 ? args->copyListFrom(1) : 0), xsink);
   return ds->selectRows(sql, *argl, xsink);
}

// nothing Datasource::setAutoCommit(bool ac = True){}
static AbstractQoreNode* Datasource_setAutoCommit_Vb(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   bool ac = HARD_QORE_BOOL(args, 0);
# 482 "QC_Datasource.qpp"
ds->setAutoCommit(ac, xsink);
   return 0;
}

// nothing Datasource::setDBCharset(string encoding){}
static AbstractQoreNode* Datasource_setDBCharset_Vs(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* encoding = HARD_QORE_STRING(args, 0);
# 902 "QC_Datasource.qpp"
ds->setPendingDBEncoding(encoding->getBuffer());
   return 0;
}

// nothing Datasource::setDBEncoding(string encoding){}
static AbstractQoreNode* Datasource_setDBEncoding_Vs(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* encoding = HARD_QORE_STRING(args, 0);
# 916 "QC_Datasource.qpp"
ds->setPendingDBEncoding(encoding->getBuffer());
   return 0;
}

// nothing Datasource::setDBName(string db){}
static AbstractQoreNode* Datasource_setDBName_Vs(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* db = HARD_QORE_STRING(args, 0);
# 886 "QC_Datasource.qpp"
ds->setPendingDBName(db->getBuffer());
   return 0;
}

// nothing Datasource::setEventQueue(Qore::Thread::Queue queue, any arg){}
static AbstractQoreNode* Datasource_setEventQueue_C5QueueVa(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   HARD_QORE_OBJ_DATA(queue, Queue, args, 0, CID_QUEUE, "Datasource::setEventQueue()", "Qore::Thread::Queue", xsink);
   if (*xsink)
      return 0;
   const AbstractQoreNode* arg = get_param(args, 1);
# 1260 "QC_Datasource.qpp"
ReferenceHolder<Queue> q(queue, xsink);

   if (queue->getMax() != -1) {
      xsink->raiseException("QUEUE-ERROR", "the Queue object passed as a maximum size of %d entr%s, which could cause internal I/O to freeze in the Datasource object if used; use a Queue object with no maximum size for datasource event monitoring", queue->getMax(), queue->getMax() == 1 ? "y" : "ies");
      return 0;
   }

   ds->setEventQueue(q.release(), arg ? arg->refSelf() : 0, xsink);
   return 0;
}

// nothing Datasource::setHostName(string host){}
static AbstractQoreNode* Datasource_setHostName_Vs(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* host = HARD_QORE_STRING(args, 0);
# 930 "QC_Datasource.qpp"
ds->setPendingHostName(host->getBuffer());
   return 0;
}

// nothing Datasource::setOption(string opt, any val){}
static AbstractQoreNode* Datasource_setOption_VsVa(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* opt = HARD_QORE_STRING(args, 0);
   const AbstractQoreNode* val = get_param(args, 1);
# 1205 "QC_Datasource.qpp"
ds->setOption(opt->getBuffer(), val, xsink);
   return 0;
}

// nothing Datasource::setPassword(string pass){}
static AbstractQoreNode* Datasource_setPassword_Vs(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* pass = HARD_QORE_STRING(args, 0);
# 872 "QC_Datasource.qpp"
ds->setPendingPassword(pass->getBuffer());
   return 0;
}

// nothing Datasource::setPort(softint port = 0){}
static AbstractQoreNode* Datasource_setPort_vi(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   int64 port = HARD_QORE_INT(args, 0);
# 944 "QC_Datasource.qpp"
ds->setPendingPort((int)port);
   return 0;
}

// nothing Datasource::setTransactionLockTimeout(timeout timeout_ms = 0){}
static AbstractQoreNode* Datasource_setTransactionLockTimeout_Vt(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 1073 "QC_Datasource.qpp"
ds->setTransactionLockTimeout((int)timeout_ms);
   return 0;
}

// nothing Datasource::setUserName(string user){}
static AbstractQoreNode* Datasource_setUserName_Vs(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* user = HARD_QORE_STRING(args, 0);
# 858 "QC_Datasource.qpp"
ds->setPendingUsername(user->getBuffer());
   return 0;
}

// int Datasource::transactionTid(){}
static int64 Datasource_transactionTid(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 1174 "QC_Datasource.qpp"
return ds->transactionTid();
}

// any Datasource::vexec(string sql, *softlist vargs){}
static AbstractQoreNode* Datasource_vexec_Vsnl(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
   const QoreListNode* vargs = reinterpret_cast<const QoreListNode*>(get_param(args, 1));
# 539 "QC_Datasource.qpp"
return ds->exec(sql, vargs, xsink);
}

// any Datasource::vselect(string sql, *softlist vargs){}
static AbstractQoreNode* Datasource_vselect_Vsnl(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
   const QoreListNode* vargs = reinterpret_cast<const QoreListNode*>(get_param(args, 1));
# 691 "QC_Datasource.qpp"
return ds->select(sql, vargs, xsink);
}

// *hash Datasource::vselectRow(string sql, *softlist vargs){}
static AbstractQoreNode* Datasource_vselectRow_Vsnl(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
   const QoreListNode* vargs = reinterpret_cast<const QoreListNode*>(get_param(args, 1));
# 718 "QC_Datasource.qpp"
return ds->selectRow(sql, vargs, xsink);
}

// any Datasource::vselectRows(string sql, *softlist vargs){}
static AbstractQoreNode* Datasource_vselectRows_Vsnl(QoreObject* self, ManagedDatasource* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
   const QoreListNode* vargs = reinterpret_cast<const QoreListNode*>(get_param(args, 1));
# 751 "QC_Datasource.qpp"
return ds->selectRows(sql, vargs, xsink);
}

DLLLOCAL QoreClass* initDatasourceClass(QoreNamespace& ns) {
   QC_DATASOURCE = new QoreClass("Datasource", QDOM_DATABASE);
   CID_DATASOURCE = QC_DATASOURCE->getID();

   // set parent class
   assert(QC_ABSTRACTDATASOURCE);
   QC_DATASOURCE->addBuiltinVirtualBaseClass(QC_ABSTRACTDATASOURCE);

   // nothing Datasource::beginTransaction(){}
   QC_DATASOURCE->addMethodExtended3("beginTransaction", (q_method_t)Datasource_beginTransaction, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // nothing Datasource::clearEventQueue(){}
   QC_DATASOURCE->addMethodExtended3("clearEventQueue", (q_method_t)Datasource_clearEventQueue, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // nothing Datasource::close(){}
   QC_DATASOURCE->addMethodExtended3("close", (q_method_t)Datasource_close, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // nothing Datasource::commit(){}
   QC_DATASOURCE->addMethodExtended3("commit", (q_method_t)Datasource_commit, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // Datasource::constructor(string driver, *string user, *string pass, *string db, *string encoding, *string host, *softint port, *hash options, *Qore::Thread::Queue queue, any arg) {}
   QC_DATASOURCE->setConstructorExtended3(Datasource_constructor_VsNsNsNsNsNsniNhC5QueueVa, false, QC_NO_FLAGS, QDOM_DEFAULT, 10, stringTypeInfo, NULL, "driver", stringOrNothingTypeInfo, NULL, "user", stringOrNothingTypeInfo, NULL, "pass", stringOrNothingTypeInfo, NULL, "db", stringOrNothingTypeInfo, NULL, "encoding", stringOrNothingTypeInfo, NULL, "host", softBigIntOrNothingTypeInfo, NULL, "port", hashOrNothingTypeInfo, NULL, "options", QC_QUEUE->getOrNothingTypeInfo(), NULL, "queue", anyTypeInfo, NULL, "arg");

   // Datasource::constructor(string desc, *Qore::Thread::Queue queue, any arg) {}
   QC_DATASOURCE->setConstructorExtended3(Datasource_constructor_VsC5QueueVa, false, QC_NO_FLAGS, QDOM_DEFAULT, 3, stringTypeInfo, NULL, "desc", QC_QUEUE->getOrNothingTypeInfo(), NULL, "queue", anyTypeInfo, NULL, "arg");

   // Datasource::constructor(hash opts, *Qore::Thread::Queue queue, any arg) {}
   QC_DATASOURCE->setConstructorExtended3(Datasource_constructor_VhC5QueueVa, false, QC_NO_FLAGS, QDOM_DEFAULT, 3, hashTypeInfo, NULL, "opts", QC_QUEUE->getOrNothingTypeInfo(), NULL, "queue", anyTypeInfo, NULL, "arg");

   // Datasource::copy() {}
   QC_DATASOURCE->setCopy((q_copy_t)Datasource_copy);

   // bool Datasource::currentThreadInTransaction(){}
   QC_DATASOURCE->addMethodExtended3("currentThreadInTransaction", (q_method_bool_t)Datasource_currentThreadInTransaction, false, QC_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

   // *hash Datasource::describe(string sql, ...){}
   QC_DATASOURCE->addMethodExtended3("describe", (q_method_t)Datasource_describe_VsVV, false, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, hashOrNothingTypeInfo, 1, stringTypeInfo, NULL, "sql");

   // Datasource::destructor() {}
   QC_DATASOURCE->setDestructor((q_destructor_t)Datasource_destructor);

   // any Datasource::exec(string sql, ...){}
   QC_DATASOURCE->addMethodExtended3("exec", (q_method_t)Datasource_exec_VsVV, false, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, anyTypeInfo, 1, stringTypeInfo, NULL, "sql");

   // any Datasource::execRaw(string sql){}
   QC_DATASOURCE->addMethodExtended3("execRaw", (q_method_t)Datasource_execRaw_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 1, stringTypeInfo, NULL, "sql");

   // bool Datasource::getAutoCommit(){}
   QC_DATASOURCE->addMethodExtended3("getAutoCommit", (q_method_bool_t)Datasource_getAutoCommit, false, QC_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

   // int Datasource::getCapabilities(){}
   QC_DATASOURCE->addMethodExtended3("getCapabilities", (q_method_int64_t)Datasource_getCapabilities, false, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

   // list Datasource::getCapabilityList(){}
   QC_DATASOURCE->addMethodExtended3("getCapabilityList", (q_method_t)Datasource_getCapabilityList, false, QC_CONSTANT, QDOM_DEFAULT, listTypeInfo);

   // any Datasource::getClientVersion(){}
   QC_DATASOURCE->addMethodExtended3("getClientVersion", (q_method_t)Datasource_getClientVersion, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo);

   // hash Datasource::getConfigHash(){}
   QC_DATASOURCE->addMethodExtended3("getConfigHash", (q_method_t)Datasource_getConfigHash, false, QC_CONSTANT, QDOM_DEFAULT, hashTypeInfo);

   // string Datasource::getConfigString(){}
   QC_DATASOURCE->addMethodExtended3("getConfigString", (q_method_t)Datasource_getConfigString, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

   // string Datasource::getDBCharset(){}
   QC_DATASOURCE->addMethodExtended3("getDBCharset", (q_method_t)Datasource_getDBCharset, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

   // string Datasource::getDBEncoding(){}
   QC_DATASOURCE->addMethodExtended3("getDBEncoding", (q_method_t)Datasource_getDBEncoding, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

   // *string Datasource::getDBName(){}
   QC_DATASOURCE->addMethodExtended3("getDBName", (q_method_t)Datasource_getDBName, false, QC_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

   // string Datasource::getDriverName(){}
   QC_DATASOURCE->addMethodExtended3("getDriverName", (q_method_t)Datasource_getDriverName, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

   // *string Datasource::getHostName(){}
   QC_DATASOURCE->addMethodExtended3("getHostName", (q_method_t)Datasource_getHostName, false, QC_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

   // string Datasource::getOSCharset(){}
   QC_DATASOURCE->addMethodExtended3("getOSCharset", (q_method_t)Datasource_getOSCharset, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

   // *string Datasource::getOSEncoding(){}
   QC_DATASOURCE->addMethodExtended3("getOSEncoding", (q_method_t)Datasource_getOSEncoding, false, QC_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

   // any Datasource::getOption(string opt){}
   QC_DATASOURCE->addMethodExtended3("getOption", (q_method_t)Datasource_getOption_Vs, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, anyTypeInfo, 1, stringTypeInfo, NULL, "opt");

   // hash Datasource::getOptionHash(){}
   QC_DATASOURCE->addMethodExtended3("getOptionHash", (q_method_t)Datasource_getOptionHash, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, hashTypeInfo);

   // *string Datasource::getPassword(){}
   QC_DATASOURCE->addMethodExtended3("getPassword", (q_method_t)Datasource_getPassword, false, QC_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

   // *int Datasource::getPort(){}
   QC_DATASOURCE->addMethodExtended3("getPort", (q_method_t)Datasource_getPort, false, QC_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo);

   // any Datasource::getServerVersion(){}
   QC_DATASOURCE->addMethodExtended3("getServerVersion", (q_method_t)Datasource_getServerVersion, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo);

   // int Datasource::getTransactionLockTimeout(){}
   QC_DATASOURCE->addMethodExtended3("getTransactionLockTimeout", (q_method_int64_t)Datasource_getTransactionLockTimeout, false, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

   // *string Datasource::getUserName(){}
   QC_DATASOURCE->addMethodExtended3("getUserName", (q_method_t)Datasource_getUserName, false, QC_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

   // bool Datasource::inTransaction(){}
   QC_DATASOURCE->addMethodExtended3("inTransaction", (q_method_bool_t)Datasource_inTransaction, false, QC_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

   // nothing Datasource::open(){}
   QC_DATASOURCE->addMethodExtended3("open", (q_method_t)Datasource_open, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // nothing Datasource::reset(){}
   QC_DATASOURCE->addMethodExtended3("reset", (q_method_t)Datasource_reset, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // nothing Datasource::rollback(){}
   QC_DATASOURCE->addMethodExtended3("rollback", (q_method_t)Datasource_rollback, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // any Datasource::select(string sql, ...){}
   QC_DATASOURCE->addMethodExtended3("select", (q_method_t)Datasource_select_VsVV, false, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, anyTypeInfo, 1, stringTypeInfo, NULL, "sql");

   // *hash Datasource::selectRow(string sql, ...){}
   QC_DATASOURCE->addMethodExtended3("selectRow", (q_method_t)Datasource_selectRow_VsVV, false, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, hashOrNothingTypeInfo, 1, stringTypeInfo, NULL, "sql");

   // any Datasource::selectRows(string sql, ...){}
   QC_DATASOURCE->addMethodExtended3("selectRows", (q_method_t)Datasource_selectRows_VsVV, false, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, anyTypeInfo, 1, stringTypeInfo, NULL, "sql");

   // nothing Datasource::setAutoCommit(bool ac = True){}
   QC_DATASOURCE->addMethodExtended3("setAutoCommit", (q_method_t)Datasource_setAutoCommit_Vb, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, boolTypeInfo, &True, "ac");

   // nothing Datasource::setDBCharset(string encoding){}
   QC_DATASOURCE->addMethodExtended3("setDBCharset", (q_method_t)Datasource_setDBCharset_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, NULL, "encoding");

   // nothing Datasource::setDBEncoding(string encoding){}
   QC_DATASOURCE->addMethodExtended3("setDBEncoding", (q_method_t)Datasource_setDBEncoding_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, NULL, "encoding");

   // nothing Datasource::setDBName(string db){}
   QC_DATASOURCE->addMethodExtended3("setDBName", (q_method_t)Datasource_setDBName_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, NULL, "db");

   // nothing Datasource::setEventQueue(Qore::Thread::Queue queue, any arg){}
   QC_DATASOURCE->addMethodExtended3("setEventQueue", (q_method_t)Datasource_setEventQueue_C5QueueVa, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, QC_QUEUE->getTypeInfo(), NULL, "queue", anyTypeInfo, NULL, "arg");

   // nothing Datasource::setHostName(string host){}
   QC_DATASOURCE->addMethodExtended3("setHostName", (q_method_t)Datasource_setHostName_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, NULL, "host");

   // nothing Datasource::setOption(string opt, any val){}
   QC_DATASOURCE->addMethodExtended3("setOption", (q_method_t)Datasource_setOption_VsVa, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, NULL, "opt", anyTypeInfo, NULL, "val");

   // nothing Datasource::setPassword(string pass){}
   QC_DATASOURCE->addMethodExtended3("setPassword", (q_method_t)Datasource_setPassword_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, NULL, "pass");

   // nothing Datasource::setPort(softint port = 0){}
   QC_DATASOURCE->addMethodExtended3("setPort", (q_method_t)Datasource_setPort_vi, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, softBigIntTypeInfo, zero(), "port");

   // nothing Datasource::setTransactionLockTimeout(timeout timeout_ms = 0){}
   QC_DATASOURCE->addMethodExtended3("setTransactionLockTimeout", (q_method_t)Datasource_setTransactionLockTimeout_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, timeoutTypeInfo, zero(), "timeout_ms");

   // nothing Datasource::setUserName(string user){}
   QC_DATASOURCE->addMethodExtended3("setUserName", (q_method_t)Datasource_setUserName_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, NULL, "user");

   // int Datasource::transactionTid(){}
   QC_DATASOURCE->addMethodExtended3("transactionTid", (q_method_int64_t)Datasource_transactionTid, false, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

   // any Datasource::vexec(string sql, *softlist vargs){}
   QC_DATASOURCE->addMethodExtended3("vexec", (q_method_t)Datasource_vexec_Vsnl, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 2, stringTypeInfo, NULL, "sql", softListOrNothingTypeInfo, NULL, "vargs");

   // any Datasource::vselect(string sql, *softlist vargs){}
   QC_DATASOURCE->addMethodExtended3("vselect", (q_method_t)Datasource_vselect_Vsnl, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 2, stringTypeInfo, NULL, "sql", softListOrNothingTypeInfo, NULL, "vargs");

   // *hash Datasource::vselectRow(string sql, *softlist vargs){}
   QC_DATASOURCE->addMethodExtended3("vselectRow", (q_method_t)Datasource_vselectRow_Vsnl, false, QC_NO_FLAGS, QDOM_DEFAULT, hashOrNothingTypeInfo, 2, stringTypeInfo, NULL, "sql", softListOrNothingTypeInfo, NULL, "vargs");

   // any Datasource::vselectRows(string sql, *softlist vargs){}
   QC_DATASOURCE->addMethodExtended3("vselectRows", (q_method_t)Datasource_vselectRows_Vsnl, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 2, stringTypeInfo, NULL, "sql", softListOrNothingTypeInfo, NULL, "vargs");
# 116 "QC_Datasource.qpp"
# 123 "QC_Datasource.qpp"
# 126 "QC_Datasource.qpp"
# 129 "QC_Datasource.qpp"
# 132 "QC_Datasource.qpp"
# 135 "QC_Datasource.qpp"
# 138 "QC_Datasource.qpp"
# 141 "QC_Datasource.qpp"
// the following have no drivers yet
//const DSInformix = "informix";
//const DSTimesTen = "timesten";
   ns.addConstant("DSDB2", new QoreStringNode("db2"));
   ns.addConstant("DSFreeTDS", new QoreStringNode("freetds"));
   ns.addConstant("DSMSSQL", new QoreStringNode("freetds"));
   ns.addConstant("DSMySQL", new QoreStringNode("mysql"));
   ns.addConstant("DSOracle", new QoreStringNode("oracle"));
   ns.addConstant("DSPGSQL", new QoreStringNode("pgsql"));
   ns.addConstant("DSSQLite3", new QoreStringNode("sqlite3"));
   ns.addConstant("DSSybase", new QoreStringNode("sybase"));
# 149 "QC_Datasource.qpp"
# 156 "QC_Datasource.qpp"
# 159 "QC_Datasource.qpp"
# 162 "QC_Datasource.qpp"
# 165 "QC_Datasource.qpp"
# 168 "QC_Datasource.qpp"
# 171 "QC_Datasource.qpp"
# 174 "QC_Datasource.qpp"
# 177 "QC_Datasource.qpp"
# 180 "QC_Datasource.qpp"
# 183 "QC_Datasource.qpp"
# 186 "QC_Datasource.qpp"
# 189 "QC_Datasource.qpp"
# 192 "QC_Datasource.qpp"
# 195 "QC_Datasource.qpp"
   ns.addConstant("DBI_CAP_AUTORECONNECT", new QoreBigIntNode(DBI_CAP_AUTORECONNECT));
   ns.addConstant("DBI_CAP_BIND_BY_PLACEHOLDER", new QoreBigIntNode(DBI_CAP_BIND_BY_PLACEHOLDER));
   ns.addConstant("DBI_CAP_BIND_BY_VALUE", new QoreBigIntNode(DBI_CAP_BIND_BY_VALUE));
   ns.addConstant("DBI_CAP_CHARSET_SUPPORT", new QoreBigIntNode(DBI_CAP_CHARSET_SUPPORT));
   ns.addConstant("DBI_CAP_EVENTS", new QoreBigIntNode(DBI_CAP_EVENTS));
   ns.addConstant("DBI_CAP_HAS_DESCRIBE", new QoreBigIntNode(DBI_CAP_HAS_DESCRIBE));
   ns.addConstant("DBI_CAP_HAS_EXECRAW", new QoreBigIntNode(DBI_CAP_HAS_EXECRAW));
   ns.addConstant("DBI_CAP_HAS_NUMBER_SUPPORT", new QoreBigIntNode(DBI_CAP_HAS_NUMBER_SUPPORT));
   ns.addConstant("DBI_CAP_HAS_OPTION_SUPPORT", new QoreBigIntNode(DBI_CAP_HAS_OPTION_SUPPORT));
   ns.addConstant("DBI_CAP_HAS_SELECT_ROW", new QoreBigIntNode(DBI_CAP_HAS_SELECT_ROW));
   ns.addConstant("DBI_CAP_HAS_STATEMENT", new QoreBigIntNode(DBI_CAP_HAS_STATEMENT));
   ns.addConstant("DBI_CAP_LOB_SUPPORT", new QoreBigIntNode(DBI_CAP_LOB_SUPPORT));
   ns.addConstant("DBI_CAP_SERVER_TIME_ZONE", new QoreBigIntNode(DBI_CAP_SERVER_TIME_ZONE));
   ns.addConstant("DBI_CAP_STORED_PROCEDURES", new QoreBigIntNode(DBI_CAP_STORED_PROCEDURES));
   ns.addConstant("DBI_CAP_TIME_ZONE_SUPPORT", new QoreBigIntNode(DBI_CAP_TIME_ZONE_SUPPORT));
   ns.addConstant("DBI_CAP_TRANSACTION_MANAGEMENT", new QoreBigIntNode(DBI_CAP_TRANSACTION_MANAGEMENT));

   return QC_DATASOURCE;
}
