/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
  QC_DatasourcePool.cpp

  Qore Programming Language

  Copyright (C) 2003 - 2014 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>
#include <qore/DBI.h>

#include <qore/intern/QC_DatasourcePool.h>

#define DP_MIN 3
#define DP_MAX 10

static const char *DSPC_ERR = "DATASOURCEPOOL-CONSTRUCTOR-ERROR";

static void dsp_constructor_hash(const QoreHashNode* opts, QoreObject* self, Queue* queue, const AbstractQoreNode* arg, ExceptionSink* xsink) {
   if (arg && !queue)
      arg = 0;

   ReferenceHolder<Queue> q(queue, xsink);

   const char *str = check_hash_key(opts, "type", DSPC_ERR, xsink);
   if (*xsink) return;

   if (!str) {
      xsink->raiseException(DSPC_ERR, "expecting a string value with the 'type' key giving the driver name");
      return;
   }

   DBIDriver *db_driver = DBI.find(str);
   if (!db_driver) {
      if (!*xsink)
         xsink->raiseException("DATASOURCEPOOL-UNSUPPORTED-DATABASE", "no DBI driver can be found for database type '%s'", str);
      return;
   }

   const char *user = check_hash_key(opts, "user", DSPC_ERR, xsink);
   if (*xsink) return;

   const char *pass = check_hash_key(opts, "pass", DSPC_ERR, xsink);
   if (*xsink) return;

   const char *db = check_hash_key(opts, "db", DSPC_ERR, xsink);
   if (*xsink) return;

   const char *charset = check_hash_key(opts, "charset", DSPC_ERR, xsink);
   if (*xsink) return;

   const char *host = check_hash_key(opts, "host", DSPC_ERR, xsink);
   if (*xsink) return;

   bool found;
   int port = (int)opts->getKeyAsBigInt("port", found);
   if (port < 0) {
      xsink->raiseException(DSPC_ERR, "port value must be zero (meaning use the default port) or positive (value given: %d)", port);
      return;
   }

   // get options
   int min = 0, max = 0;

   const AbstractQoreNode *p = opts->getKeyValue("options");
   if (!is_nothing(p)) {
      if (p->getType() != NT_HASH) {
         xsink->raiseException(DSPC_ERR, "'options' key is not hash, instead got type '%s'", p->getTypeName());
         return;
      }

      const QoreHashNode* h = reinterpret_cast<const QoreHashNode *>(p);

      min = (int)h->getKeyAsBigInt("min", found);
      if (found) {
         if (min < 0) {
            xsink->raiseException(DSPC_ERR, "minimum connections must be > 0 (value given: %d)", min);
            return;
         }
      }

      max = (int)h->getKeyAsBigInt("max", found);
      if (found) {
         if (max < min) {
            xsink->raiseException(DSPC_ERR, "maximum connections must be >= min(%d) (value given: %d)", min, max);
            return;
         }
      }
   }
   // here we cast either a non-hash NULL pointer to a hash or a hash to a hash
   const QoreHashNode* oh = reinterpret_cast<const QoreHashNode*>(p);

   if (!min)
      min = DP_MIN;

   if (!max)
      max = DP_MAX;

   SimpleRefHolder<DatasourcePool> ds(new DatasourcePool(xsink, db_driver, user, pass, db, charset, host, min, max, port, oh, q.release(), arg ? arg->refSelf() : 0));
   if (!*xsink)
      self->setPrivate(CID_DATASOURCEPOOL, ds.release());
}
/* Qore class Qore::SQL::DatasourcePool */

qore_classid_t CID_DATASOURCEPOOL;
QoreClass* QC_DATASOURCEPOOL;

// nothing DatasourcePool::beginTransaction(){}
static AbstractQoreNode* DatasourcePool_beginTransaction(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 562 "QC_DatasourcePool.qpp"
ds->beginTransaction(xsink);
   return 0;
}

// nothing DatasourcePool::clearEventQueue(){}
static AbstractQoreNode* DatasourcePool_clearEventQueue(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 947 "QC_DatasourcePool.qpp"
ds->setEventQueue(0, 0, xsink);
   return 0;
}

// nothing DatasourcePool::clearWarningCallback(){}
static AbstractQoreNode* DatasourcePool_clearWarningCallback(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 848 "QC_DatasourcePool.qpp"
ds->clearWarningCallback(xsink);
   return 0;
}

// nothing DatasourcePool::commit(){}
static AbstractQoreNode* DatasourcePool_commit(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 298 "QC_DatasourcePool.qpp"
ds->commit(xsink);
   return 0;
}

// DatasourcePool::constructor(string driver, *string user, *string pass, *string db, *string encoding, *string host, softint min = 3, softint max = 10, softint port = 0, *Qore::Thread::Queue queue, any arg) {}
static void DatasourcePool_constructor_VsNsNsNsNsNsviviviC5QueueVa(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* driver = HARD_QORE_STRING(args, 0);
   const QoreStringNode* user = reinterpret_cast<const QoreStringNode*>(get_param(args, 1));
   const QoreStringNode* pass = reinterpret_cast<const QoreStringNode*>(get_param(args, 2));
   const QoreStringNode* db = reinterpret_cast<const QoreStringNode*>(get_param(args, 3));
   const QoreStringNode* encoding = reinterpret_cast<const QoreStringNode*>(get_param(args, 4));
   const QoreStringNode* host = reinterpret_cast<const QoreStringNode*>(get_param(args, 5));
   int64 min = HARD_QORE_INT(args, 6);
   int64 max = HARD_QORE_INT(args, 7);
   int64 port = HARD_QORE_INT(args, 8);
   HARD_QORE_OBJ_OR_NOTHING_DATA(queue, Queue, args, 9, CID_QUEUE, xsink);
   if (*xsink)
      return;
   const AbstractQoreNode* arg = get_param(args, 10);
# 172 "QC_DatasourcePool.qpp"
if (!port && min == 3 && max == 10 && driver->find(':') != -1) {
      ReferenceHolder<QoreHashNode> h(parseDatasource(driver->getBuffer(), xsink), xsink);
      if (!h) {
         assert(*xsink);
         return;
      }
      dsp_constructor_hash(*h, self, queue, arg, xsink);
      return;
   }

   ReferenceHolder<Queue> q(queue, xsink);

   DBIDriver *db_driver = DBI.find(driver->getBuffer());
   if (!db_driver) {
      xsink->raiseException("DATASOURCEPOOL-UNSUPPORTED-DATABASE", "no DBI driver can be found for database type '%s'", driver->getBuffer());
      return;
   }

   if (min <= 0) {
      xsink->raiseException(DSPC_ERR, "minimum connections must be > 0 (value given: "QLLD")", min);
      return;
   }

   if (max < min) {
      xsink->raiseException(DSPC_ERR, "maximum connections must be >= min(%d) (value given: "QLLD")", min, max);
      return;
   }
   
   if (port < 0) {
      xsink->raiseException(DSPC_ERR, "port value must be zero (meaning use the default port) or positive (value given: "QLLD")", port);
      return;
   }

   SimpleRefHolder<DatasourcePool> ds(new DatasourcePool(xsink, db_driver, user ? user->getBuffer() : 0, pass ? pass->getBuffer() : 0,
							 db ? db->getBuffer() : 0, encoding ? encoding->getBuffer() : 0, host ? host->getBuffer() : 0, 
							 min, max, port, 0, q.release(), arg ? arg->refSelf() : 0));
   if (!*xsink)
      self->setPrivate(CID_DATASOURCEPOOL, ds.release());
}

// DatasourcePool::constructor(string desc, *Qore::Thread::Queue queue, any arg) {}
static void DatasourcePool_constructor_VsC5QueueVa(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* desc = HARD_QORE_STRING(args, 0);
   HARD_QORE_OBJ_OR_NOTHING_DATA(queue, Queue, args, 1, CID_QUEUE, xsink);
   if (*xsink)
      return;
   const AbstractQoreNode* arg = get_param(args, 2);
# 229 "QC_DatasourcePool.qpp"
ReferenceHolder<Queue> q(queue, xsink);
   ReferenceHolder<QoreHashNode> h(parseDatasource(desc->getBuffer(), xsink), xsink);
   if (!h) {
      assert(*xsink);
      return;
   }
   dsp_constructor_hash(*h, self, q.release(), arg, xsink);
}

// DatasourcePool::constructor(hash opts, *Qore::Thread::Queue queue, any arg) {}
static void DatasourcePool_constructor_VhC5QueueVa(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreHashNode* opts = HARD_QORE_HASH(args, 0);
   HARD_QORE_OBJ_OR_NOTHING_DATA(queue, Queue, args, 1, CID_QUEUE, xsink);
   if (*xsink)
      return;
   const AbstractQoreNode* arg = get_param(args, 2);
# 261 "QC_DatasourcePool.qpp"
dsp_constructor_hash(opts, self, queue, arg, xsink);
}

// DatasourcePool::copy() {}
static void DatasourcePool_copy(QoreObject* self, QoreObject* old, DatasourcePool* ds, ExceptionSink* xsink) {
# 286 "QC_DatasourcePool.qpp"
self->setPrivate(CID_DATASOURCEPOOL, new DatasourcePool(*ds, xsink));
}

// bool DatasourcePool::currentThreadInTransaction(){}
static bool DatasourcePool_currentThreadInTransaction(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 785 "QC_DatasourcePool.qpp"
return ds->currentThreadInTransaction();
}

// DatasourcePool::destructor() {}
static void DatasourcePool_destructor(QoreObject* self, DatasourcePool* ds, ExceptionSink* xsink) {
# 274 "QC_DatasourcePool.qpp"
ds->destructor(xsink);
   ds->deref();
}

// any DatasourcePool::exec(string sql, ...){}
static AbstractQoreNode* DatasourcePool_exec_VsVV(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
# 329 "QC_DatasourcePool.qpp"
ReferenceHolder<QoreListNode> vargs(args->size() > 1 ? args->copyListFrom(1) : 0, xsink);
   return ds->exec(sql, *vargs, xsink);
}

// any DatasourcePool::execRaw(string sql){}
static AbstractQoreNode* DatasourcePool_execRaw_Vs(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
# 372 "QC_DatasourcePool.qpp"
return ds->execRaw(sql, xsink);
}

// any DatasourcePool::getClientVersion(){}
static AbstractQoreNode* DatasourcePool_getClientVersion(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 759 "QC_DatasourcePool.qpp"
return ds->getClientVersion(xsink);
}

// hash DatasourcePool::getConfigHash(){}
static AbstractQoreNode* DatasourcePool_getConfigHash(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 822 "QC_DatasourcePool.qpp"
return ds->getConfigHash();
}

// string DatasourcePool::getConfigString(){}
static AbstractQoreNode* DatasourcePool_getConfigString(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 836 "QC_DatasourcePool.qpp"
return ds->getConfigString();
}

// *string DatasourcePool::getDBCharset(){}
static AbstractQoreNode* DatasourcePool_getDBCharset(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 612 "QC_DatasourcePool.qpp"
return ds->getPendingDBEncoding();
}

// string DatasourcePool::getDBEncoding(){}
static AbstractQoreNode* DatasourcePool_getDBEncoding(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 626 "QC_DatasourcePool.qpp"
return ds->getPendingDBEncoding();
}

// *string DatasourcePool::getDBName(){}
static AbstractQoreNode* DatasourcePool_getDBName(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 598 "QC_DatasourcePool.qpp"
return ds->getPendingDBName();
}

// string DatasourcePool::getDriverName(){}
static AbstractQoreNode* DatasourcePool_getDriverName(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 691 "QC_DatasourcePool.qpp"
return new QoreStringNode(ds->getDriverName());
}

// int DatasourcePool::getErrorTimeout(){}
static int64 DatasourcePool_getErrorTimeout(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 919 "QC_DatasourcePool.qpp"
return ds->getErrorTimeout();
}

// *string DatasourcePool::getHostName(){}
static AbstractQoreNode* DatasourcePool_getHostName(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 666 "QC_DatasourcePool.qpp"
return ds->getPendingHostName();
}

// int DatasourcePool::getMaximum(){}
static int64 DatasourcePool_getMaximum(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 718 "QC_DatasourcePool.qpp"
return ds->getMax();
}

// int DatasourcePool::getMinimum(){}
static int64 DatasourcePool_getMinimum(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 705 "QC_DatasourcePool.qpp"
return ds->getMin();
}

// string DatasourcePool::getOSCharset(){}
static AbstractQoreNode* DatasourcePool_getOSCharset(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 640 "QC_DatasourcePool.qpp"
const QoreEncoding *enc = ds->getQoreEncoding();
   return new QoreStringNode(enc ? enc->getCode() : "(unknown)");
}

// *string DatasourcePool::getOSEncoding(){}
static AbstractQoreNode* DatasourcePool_getOSEncoding(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 653 "QC_DatasourcePool.qpp"
const QoreEncoding *enc = ds->getQoreEncoding();
   return enc ? new QoreStringNode(enc->getCode()) : 0;
}

// any DatasourcePool::getOption(string opt){}
static AbstractQoreNode* DatasourcePool_getOption_Vs(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* opt = HARD_QORE_STRING(args, 0);
# 808 "QC_DatasourcePool.qpp"
return ds->getOption(opt->getBuffer(), xsink);
}

// hash DatasourcePool::getOptionHash(){}
static AbstractQoreNode* DatasourcePool_getOptionHash(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 797 "QC_DatasourcePool.qpp"
return ds->getOptionHash();
}

// *string DatasourcePool::getPassword(){}
static AbstractQoreNode* DatasourcePool_getPassword(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 586 "QC_DatasourcePool.qpp"
return ds->getPendingPassword();
}

// *int DatasourcePool::getPort(){}
static AbstractQoreNode* DatasourcePool_getPort(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 678 "QC_DatasourcePool.qpp"
int port = ds->getPendingPort();
   return port ? new QoreBigIntNode(port) : 0;
}

// any DatasourcePool::getServerVersion(){}
static AbstractQoreNode* DatasourcePool_getServerVersion(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 745 "QC_DatasourcePool.qpp"
return ds->getServerVersion(xsink);
}

// *hash DatasourcePool::getUsageInfo(){}
static AbstractQoreNode* DatasourcePool_getUsageInfo(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 891 "QC_DatasourcePool.qpp"
return ds->getUsageInfo();
}

// *string DatasourcePool::getUserName(){}
static AbstractQoreNode* DatasourcePool_getUserName(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 574 "QC_DatasourcePool.qpp"
return ds->getPendingUsername();
}

// bool DatasourcePool::inTransaction(){}
static bool DatasourcePool_inTransaction(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 771 "QC_DatasourcePool.qpp"
return ds->inTransaction();
}

// nothing DatasourcePool::rollback(){}
static AbstractQoreNode* DatasourcePool_rollback(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 310 "QC_DatasourcePool.qpp"
ds->rollback(xsink);
   return 0;
}

// any DatasourcePool::select(string sql, ...){}
static AbstractQoreNode* DatasourcePool_select_VsVV(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
# 406 "QC_DatasourcePool.qpp"
ReferenceHolder<QoreListNode> vargs(args->size() > 1 ? args->copyListFrom(1) : 0, xsink);
   return ds->select(sql, *vargs, xsink);
}

// any DatasourcePool::selectRow(string sql, ...){}
static AbstractQoreNode* DatasourcePool_selectRow_VsVV(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
# 432 "QC_DatasourcePool.qpp"
ReferenceHolder<QoreListNode> vargs(args->size() > 1 ? args->copyListFrom(1) : 0, xsink);
   return ds->selectRow(sql, *vargs, xsink);
}

// any DatasourcePool::selectRows(string sql, ...){}
static AbstractQoreNode* DatasourcePool_selectRows_VsVV(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
# 460 "QC_DatasourcePool.qpp"
ReferenceHolder<QoreListNode> vargs(args->size() > 1 ? args->copyListFrom(1) : 0, xsink);
   return ds->selectRows(sql, *vargs, xsink);
}

// nothing DatasourcePool::setErrorTimeout(timeout ts){}
static AbstractQoreNode* DatasourcePool_setErrorTimeout_Vt(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
   int64 ts = HARD_QORE_INT(args, 0);
# 903 "QC_DatasourcePool.qpp"
if (ts < 0)
      ts = 0;
   ds->setErrorTimeout(ts);
   return 0;
}

// nothing DatasourcePool::setEventQueue(Qore::Thread::Queue queue, any arg){}
static AbstractQoreNode* DatasourcePool_setEventQueue_C5QueueVa(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
   HARD_QORE_OBJ_DATA(queue, Queue, args, 0, CID_QUEUE, "DatasourcePool::setEventQueue()", "Qore::Thread::Queue", xsink);
   if (*xsink)
      return 0;
   const AbstractQoreNode* arg = get_param(args, 1);
# 932 "QC_DatasourcePool.qpp"
ReferenceHolder<Queue> q(queue, xsink);

   if (queue->getMax() != -1) {
      xsink->raiseException("QUEUE-ERROR", "the Queue object passed as a maximum size of %d entr%s, which could cause internal I/O to freeze in the Datasource object if used; use a Queue object with no maximum size for datasource event monitoring", queue->getMax(), queue->getMax() == 1 ? "y" : "ies");
      return 0;
   }

   ds->setEventQueue(q.release(), arg ? arg->refSelf() : 0, xsink);
   return 0;
}

// nothing DatasourcePool::setWarningCallback(timeout ms, code callback, any arg){}
static AbstractQoreNode* DatasourcePool_setWarningCallback_VtVcVa(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
   int64 ms = HARD_QORE_INT(args, 0);
   const ResolvedCallReferenceNode* callback = reinterpret_cast<const ResolvedCallReferenceNode*>(get_param(args, 1));
   const AbstractQoreNode* arg = get_param(args, 2);
# 869 "QC_DatasourcePool.qpp"
ds->setWarningCallback(ms, callback->refRefSelf(), arg ? arg->refSelf() : 0, xsink);
   return 0;
}

// string DatasourcePool::toString(){}
static AbstractQoreNode* DatasourcePool_toString(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
# 731 "QC_DatasourcePool.qpp"
return ds->toString();
}

// any DatasourcePool::vexec(string sql, *softlist vargs){}
static AbstractQoreNode* DatasourcePool_vexec_Vsnl(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
   const QoreListNode* vargs = reinterpret_cast<const QoreListNode*>(get_param(args, 1));
# 351 "QC_DatasourcePool.qpp"
return ds->exec(sql, vargs, xsink);
}

// any DatasourcePool::vselect(string sql, *softlist vargs){}
static AbstractQoreNode* DatasourcePool_vselect_Vsnl(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
   const QoreListNode* vargs = reinterpret_cast<const QoreListNode*>(get_param(args, 1));
# 492 "QC_DatasourcePool.qpp"
return ds->select(sql, vargs, xsink);
}

// any DatasourcePool::vselectRow(string sql, *softlist vargs){}
static AbstractQoreNode* DatasourcePool_vselectRow_Vsnl(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
   const QoreListNode* vargs = reinterpret_cast<const QoreListNode*>(get_param(args, 1));
# 517 "QC_DatasourcePool.qpp"
return ds->selectRow(sql, vargs, xsink);
}

// any DatasourcePool::vselectRows(string sql, *softlist vargs){}
static AbstractQoreNode* DatasourcePool_vselectRows_Vsnl(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sql = HARD_QORE_STRING(args, 0);
   const QoreListNode* vargs = reinterpret_cast<const QoreListNode*>(get_param(args, 1));
# 546 "QC_DatasourcePool.qpp"
return ds->selectRows(sql, vargs, xsink);
}

DLLLOCAL QoreClass* initDatasourcePoolClass(QoreNamespace& ns) {
   QC_DATASOURCEPOOL = new QoreClass("DatasourcePool", QDOM_DATABASE);
   CID_DATASOURCEPOOL = QC_DATASOURCEPOOL->getID();

   // set parent class
   assert(QC_ABSTRACTDATASOURCE);
   QC_DATASOURCEPOOL->addBuiltinVirtualBaseClass(QC_ABSTRACTDATASOURCE);

   // nothing DatasourcePool::beginTransaction(){}
   QC_DATASOURCEPOOL->addMethodExtended3("beginTransaction", (q_method_t)DatasourcePool_beginTransaction, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // nothing DatasourcePool::clearEventQueue(){}
   QC_DATASOURCEPOOL->addMethodExtended3("clearEventQueue", (q_method_t)DatasourcePool_clearEventQueue, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // nothing DatasourcePool::clearWarningCallback(){}
   QC_DATASOURCEPOOL->addMethodExtended3("clearWarningCallback", (q_method_t)DatasourcePool_clearWarningCallback, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // nothing DatasourcePool::commit(){}
   QC_DATASOURCEPOOL->addMethodExtended3("commit", (q_method_t)DatasourcePool_commit, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // DatasourcePool::constructor(string driver, *string user, *string pass, *string db, *string encoding, *string host, softint min = 3, softint max = 10, softint port = 0, *Qore::Thread::Queue queue, any arg) {}
   QC_DATASOURCEPOOL->setConstructorExtended3(DatasourcePool_constructor_VsNsNsNsNsNsviviviC5QueueVa, false, QC_NO_FLAGS, QDOM_DEFAULT, 11, stringTypeInfo, NULL, "driver", stringOrNothingTypeInfo, NULL, "user", stringOrNothingTypeInfo, NULL, "pass", stringOrNothingTypeInfo, NULL, "db", stringOrNothingTypeInfo, NULL, "encoding", stringOrNothingTypeInfo, NULL, "host", softBigIntTypeInfo, new QoreBigIntNode(3), "min", softBigIntTypeInfo, new QoreBigIntNode(10), "max", softBigIntTypeInfo, zero(), "port", QC_QUEUE->getOrNothingTypeInfo(), NULL, "queue", anyTypeInfo, NULL, "arg");

   // DatasourcePool::constructor(string desc, *Qore::Thread::Queue queue, any arg) {}
   QC_DATASOURCEPOOL->setConstructorExtended3(DatasourcePool_constructor_VsC5QueueVa, false, QC_NO_FLAGS, QDOM_DEFAULT, 3, stringTypeInfo, NULL, "desc", QC_QUEUE->getOrNothingTypeInfo(), NULL, "queue", anyTypeInfo, NULL, "arg");

   // DatasourcePool::constructor(hash opts, *Qore::Thread::Queue queue, any arg) {}
   QC_DATASOURCEPOOL->setConstructorExtended3(DatasourcePool_constructor_VhC5QueueVa, false, QC_NO_FLAGS, QDOM_DEFAULT, 3, hashTypeInfo, NULL, "opts", QC_QUEUE->getOrNothingTypeInfo(), NULL, "queue", anyTypeInfo, NULL, "arg");

   // DatasourcePool::copy() {}
   QC_DATASOURCEPOOL->setCopy((q_copy_t)DatasourcePool_copy);

   // bool DatasourcePool::currentThreadInTransaction(){}
   QC_DATASOURCEPOOL->addMethodExtended3("currentThreadInTransaction", (q_method_bool_t)DatasourcePool_currentThreadInTransaction, false, QC_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

   // DatasourcePool::destructor() {}
   QC_DATASOURCEPOOL->setDestructor((q_destructor_t)DatasourcePool_destructor);

   // any DatasourcePool::exec(string sql, ...){}
   QC_DATASOURCEPOOL->addMethodExtended3("exec", (q_method_t)DatasourcePool_exec_VsVV, false, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, anyTypeInfo, 1, stringTypeInfo, NULL, "sql");

   // any DatasourcePool::execRaw(string sql){}
   QC_DATASOURCEPOOL->addMethodExtended3("execRaw", (q_method_t)DatasourcePool_execRaw_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 1, stringTypeInfo, NULL, "sql");

   // any DatasourcePool::getClientVersion(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getClientVersion", (q_method_t)DatasourcePool_getClientVersion, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo);

   // hash DatasourcePool::getConfigHash(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getConfigHash", (q_method_t)DatasourcePool_getConfigHash, false, QC_CONSTANT, QDOM_DEFAULT, hashTypeInfo);

   // string DatasourcePool::getConfigString(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getConfigString", (q_method_t)DatasourcePool_getConfigString, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

   // *string DatasourcePool::getDBCharset(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getDBCharset", (q_method_t)DatasourcePool_getDBCharset, false, QC_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

   // string DatasourcePool::getDBEncoding(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getDBEncoding", (q_method_t)DatasourcePool_getDBEncoding, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

   // *string DatasourcePool::getDBName(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getDBName", (q_method_t)DatasourcePool_getDBName, false, QC_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

   // string DatasourcePool::getDriverName(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getDriverName", (q_method_t)DatasourcePool_getDriverName, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

   // int DatasourcePool::getErrorTimeout(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getErrorTimeout", (q_method_int64_t)DatasourcePool_getErrorTimeout, false, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

   // *string DatasourcePool::getHostName(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getHostName", (q_method_t)DatasourcePool_getHostName, false, QC_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

   // int DatasourcePool::getMaximum(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getMaximum", (q_method_int64_t)DatasourcePool_getMaximum, false, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

   // int DatasourcePool::getMinimum(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getMinimum", (q_method_int64_t)DatasourcePool_getMinimum, false, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

   // string DatasourcePool::getOSCharset(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getOSCharset", (q_method_t)DatasourcePool_getOSCharset, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

   // *string DatasourcePool::getOSEncoding(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getOSEncoding", (q_method_t)DatasourcePool_getOSEncoding, false, QC_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

   // any DatasourcePool::getOption(string opt){}
   QC_DATASOURCEPOOL->addMethodExtended3("getOption", (q_method_t)DatasourcePool_getOption_Vs, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, anyTypeInfo, 1, stringTypeInfo, NULL, "opt");

   // hash DatasourcePool::getOptionHash(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getOptionHash", (q_method_t)DatasourcePool_getOptionHash, false, QC_CONSTANT, QDOM_DEFAULT, hashTypeInfo);

   // *string DatasourcePool::getPassword(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getPassword", (q_method_t)DatasourcePool_getPassword, false, QC_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

   // *int DatasourcePool::getPort(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getPort", (q_method_t)DatasourcePool_getPort, false, QC_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo);

   // any DatasourcePool::getServerVersion(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getServerVersion", (q_method_t)DatasourcePool_getServerVersion, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo);

   // *hash DatasourcePool::getUsageInfo(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getUsageInfo", (q_method_t)DatasourcePool_getUsageInfo, false, QC_CONSTANT, QDOM_DEFAULT, hashOrNothingTypeInfo);

   // *string DatasourcePool::getUserName(){}
   QC_DATASOURCEPOOL->addMethodExtended3("getUserName", (q_method_t)DatasourcePool_getUserName, false, QC_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

   // bool DatasourcePool::inTransaction(){}
   QC_DATASOURCEPOOL->addMethodExtended3("inTransaction", (q_method_bool_t)DatasourcePool_inTransaction, false, QC_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

   // nothing DatasourcePool::rollback(){}
   QC_DATASOURCEPOOL->addMethodExtended3("rollback", (q_method_t)DatasourcePool_rollback, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // any DatasourcePool::select(string sql, ...){}
   QC_DATASOURCEPOOL->addMethodExtended3("select", (q_method_t)DatasourcePool_select_VsVV, false, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, anyTypeInfo, 1, stringTypeInfo, NULL, "sql");

   // any DatasourcePool::selectRow(string sql, ...){}
   QC_DATASOURCEPOOL->addMethodExtended3("selectRow", (q_method_t)DatasourcePool_selectRow_VsVV, false, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, anyTypeInfo, 1, stringTypeInfo, NULL, "sql");

   // any DatasourcePool::selectRows(string sql, ...){}
   QC_DATASOURCEPOOL->addMethodExtended3("selectRows", (q_method_t)DatasourcePool_selectRows_VsVV, false, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, anyTypeInfo, 1, stringTypeInfo, NULL, "sql");

   // nothing DatasourcePool::setErrorTimeout(timeout ts){}
   QC_DATASOURCEPOOL->addMethodExtended3("setErrorTimeout", (q_method_t)DatasourcePool_setErrorTimeout_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, timeoutTypeInfo, NULL, "ts");

   // nothing DatasourcePool::setEventQueue(Qore::Thread::Queue queue, any arg){}
   QC_DATASOURCEPOOL->addMethodExtended3("setEventQueue", (q_method_t)DatasourcePool_setEventQueue_C5QueueVa, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, QC_QUEUE->getTypeInfo(), NULL, "queue", anyTypeInfo, NULL, "arg");

   // nothing DatasourcePool::setWarningCallback(timeout ms, code callback, any arg){}
   QC_DATASOURCEPOOL->addMethodExtended3("setWarningCallback", (q_method_t)DatasourcePool_setWarningCallback_VtVcVa, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 3, timeoutTypeInfo, NULL, "ms", codeTypeInfo, NULL, "callback", anyTypeInfo, NULL, "arg");

   // string DatasourcePool::toString(){}
   QC_DATASOURCEPOOL->addMethodExtended3("toString", (q_method_t)DatasourcePool_toString, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

   // any DatasourcePool::vexec(string sql, *softlist vargs){}
   QC_DATASOURCEPOOL->addMethodExtended3("vexec", (q_method_t)DatasourcePool_vexec_Vsnl, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 2, stringTypeInfo, NULL, "sql", softListOrNothingTypeInfo, NULL, "vargs");

   // any DatasourcePool::vselect(string sql, *softlist vargs){}
   QC_DATASOURCEPOOL->addMethodExtended3("vselect", (q_method_t)DatasourcePool_vselect_Vsnl, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 2, stringTypeInfo, NULL, "sql", softListOrNothingTypeInfo, NULL, "vargs");

   // any DatasourcePool::vselectRow(string sql, *softlist vargs){}
   QC_DATASOURCEPOOL->addMethodExtended3("vselectRow", (q_method_t)DatasourcePool_vselectRow_Vsnl, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 2, stringTypeInfo, NULL, "sql", softListOrNothingTypeInfo, NULL, "vargs");

   // any DatasourcePool::vselectRows(string sql, *softlist vargs){}
   QC_DATASOURCEPOOL->addMethodExtended3("vselectRows", (q_method_t)DatasourcePool_vselectRows_Vsnl, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 2, stringTypeInfo, NULL, "sql", softListOrNothingTypeInfo, NULL, "vargs");

   return QC_DATASOURCEPOOL;
}
