/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
  QC_File.qpp

  Qore Programming Language

  Copyright 2003 - 2013 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>
#include <qore/intern/QC_File.h>
#include <qore/intern/QC_Queue.h>
#ifdef HAVE_TERMIOS_H
#include <qore/intern/QC_TermIOS.h>
#endif

#ifndef TCSANOW
#define TCSANOW 0
#endif

QoreFileHelper::QoreFileHelper(QoreObject* obj, ExceptionSink* xs) : QorePrivateObjectAccessHelper(xs) {
   File* f = reinterpret_cast<File*>(obj->getReferencedPrivateData(CID_FILE, xs));
   if (!f) {
      if (!*xsink)
         xsink->raiseException("FILE-ERROR", "expecting an object derived from File; got class '%s' instead", obj->getClassName());
      return;
   }
   ptr = (void*)f;
}

QoreFileHelper::~QoreFileHelper() {
   if (ptr)
      reinterpret_cast<File*>(ptr)->deref(xsink);
}

QoreFile* QoreFileHelper::operator*() const {
   return reinterpret_cast<File*>(ptr);
}

QoreFile* QoreFileHelper::operator->() const {
   return reinterpret_cast<File*>(ptr);
}

static void FILE_system_constructor(QoreObject *self, int fd, va_list args) {
   //printd(5, "FILE_system_constructor() self=%p, descriptor=%d\n", self, fd);
   File* f = new File(QCS_DEFAULT);
   f->makeSpecial(fd);
   self->setPrivate(CID_FILE, f);
}

#ifdef HAVE_STRUCT_FLOCK
static int lock_intern(struct flock &fl, const QoreListNode *args, ExceptionSink *xsink) {
   fl.l_type = (short)HARD_QORE_INT(args, 0);
   fl.l_start = (off_t)HARD_QORE_INT(args, 1);
   fl.l_len = (off_t)HARD_QORE_INT(args, 2);
   if (fl.l_len < 0) {
      xsink->raiseException("FILE-LOCK-ERROR", "length of locked area cannot be negative (value passed=%d)", fl.l_len);
      return -1;
   }
   fl.l_whence = (short)HARD_QORE_INT(args, 3);
   return 0;
}
#else
static int file_lock_error(const char *f, ExceptionSink *xsink) {
   xsink->raiseException("MISSING-FEATURE-ERROR", "this platform does not support UNIX-style file locking with fnctl(), therefore the File::%s() method is not available; for maximum portability, check Option::HAVE_FILE_LOCKING before calling this method", f);
   return 0;
}
#ifndef F_RDLCK
#define F_RDLCK 0
#endif
#ifndef F_UNLCK
#define F_UNLCK 0
#endif
#ifndef F_WRLCK
#define F_WRLCK 0
#endif
#endif

AbstractQoreNode *missing_method_error(const char *meth, const char *feat, ExceptionSink *xsink) {
   xsink->raiseException("MISSING-FEATURE-ERROR", "the %s() method is not available on this build; for maximum portability, check Option::HAVE_%s before calling this method", meth, feat);
   return 0;
}
/* Qore class Qore::File */

qore_classid_t CID_FILE;
QoreClass* QC_FILE;

// nothing File::chown(softint uid, softint gid = -1){}
static AbstractQoreNode* File_chown_vivi(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
   int64 uid = HARD_QORE_INT(args, 0);
   int64 gid = HARD_QORE_INT(args, 1);
# 941 "QC_File.qpp"
#ifdef HAVE_CHOWN
   if (check_terminal_io(self, "File::chown", xsink))
      return 0;

   f->chown( (uid_t)uid, (gid_t)gid, xsink);
   return 0;
#else
   return missing_method_error("File::chown", "UNIX_FILEMGT", xsink);
#endif
   return 0;
}

// File::constructor(*string encoding) {}
static void File_constructor_Ns(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* encoding = reinterpret_cast<const QoreStringNode*>(get_param(args, 0));
# 224 "QC_File.qpp"
const QoreEncoding *qe = encoding ? QEM.findCreate(encoding) : QCS_DEFAULT;

   SimpleRefHolder<File> f(new File(qe));
   if (f->isTty()) {
      xsink->raiseException("ILLEGAL-EXPRESSION", "File::constructor() cannot be called with a TTY target when 'no-terminal-io' is set");
      return;
   }

   self->setPrivate(CID_FILE, f.release());
}

// File::copy() {}
static void File_copy(QoreObject* self, QoreObject* old, File* f, ExceptionSink* xsink) {
# 250 "QC_File.qpp"
self->setPrivate(CID_FILE, new File(f->getEncoding()));
}

// int File::f_printf(string fmt, ...){}
static int64 File_f_printf_VsVV(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
# 708 "QC_File.qpp"
if (check_terminal_io(self, "File::f_printf", xsink))
      return 0;

   QoreStringNodeHolder str(q_sprintf(args, 1, 0, xsink));
   if (!str)
      return 0;

   return f->write(*str, xsink);
}

// int File::f_printf(){}
static int64 File_f_printf(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
# 722 "QC_File.qpp"
return 0;
}

// int File::f_vprintf(string fmt, any fmt_args){}
static int64 File_f_vprintf_VsVa(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
# 749 "QC_File.qpp"
if (check_terminal_io(self, "File::f_vprintf", xsink))
      return 0;

   QoreStringNodeHolder str(q_vsprintf(args, 1, 0, xsink));
   if (!str)
      return 0;

   return f->write(*str, xsink);
}

// int File::f_vprintf(){}
static int64 File_f_vprintf(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
# 763 "QC_File.qpp"
return 0;
}

// string File::getCharset(){}
static AbstractQoreNode* File_getCharset(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
# 786 "QC_File.qpp"
return new QoreStringNode(f->getEncoding()->getCode());
}

// hash File::getLockInfo(){}
static AbstractQoreNode* File_getLockInfo(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
# 902 "QC_File.qpp"
#ifdef HAVE_STRUCT_FLOCK
   if (check_terminal_io(self, "File::getLockInfo", xsink))
      return 0;

   struct flock fl;
   if (f->getLockInfo(fl, xsink))
      return 0;

   QoreHashNode *h = new QoreHashNode;
   h->setKeyValue("start", new QoreBigIntNode(fl.l_start), xsink);
   h->setKeyValue("len", new QoreBigIntNode(fl.l_len), xsink);
   h->setKeyValue("pid", new QoreBigIntNode(fl.l_pid), xsink);
   h->setKeyValue("type", new QoreBigIntNode(fl.l_type), xsink);
   h->setKeyValue("whence", new QoreBigIntNode(fl.l_whence), xsink);

   return h;
#else
   file_lock_error("getLockInfo", xsink);
   return 0;
#endif
   return 0;
}

// nothing File::getTerminalAttributes(TermIOS termios){}
static AbstractQoreNode* File_getTerminalAttributes_C7TermIOS(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
   HARD_QORE_OBJ_DATA(termios, QoreTermIOS, args, 0, CID_TERMIOS, "File::getTerminalAttributes()", "TermIOS", xsink);
   if (*xsink)
      return 0;
# 972 "QC_File.qpp"
#ifdef HAVE_TERMIOS_H
   ReferenceHolder<QoreTermIOS> holder(termios, xsink);
   if (check_terminal_io(self, "File::getTerminalAttributes", xsink))
      return 0;

   f->getTerminalAttributes(termios, xsink);
   return 0;
#else
   return missing_method_error("File::getTerminalAttributes", "TERMIOS", xsink);
#endif
   return 0;
}

// TermIOS File::getTerminalAttributes(){}
static QoreObject* File_getTerminalAttributes(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
# 1003 "QC_File.qpp"
#ifdef HAVE_TERMIOS_H
   if (check_terminal_io(self, "File::getTerminalAttributes", xsink))
      return 0;

   ReferenceHolder<QoreTermIOS> holder(new QoreTermIOS, xsink);
   f->getTerminalAttributes(*holder, xsink);
   if (*xsink)
      return 0;
   return new QoreObject(QC_TERMIOS, getProgram(), holder.release());
#else
   return (QoreObject*)missing_method_error("File::getTerminalAttributes", "TERMIOS", xsink);
#endif
   return 0;
}

// int File::lock(softint type = F_RDLCK, softint start = 0, softint len = 0, softint whence = SEEK_SET){}
static int64 File_lock_vivivivi(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
# 821 "QC_File.qpp"
#ifdef HAVE_STRUCT_FLOCK
   if (check_terminal_io(self, "lock", xsink))
      return 0;

   struct flock fl;
   if (lock_intern(fl, args, xsink))
      return 0;
   return f->lock(fl, xsink);
#else
   file_lock_error("lock", xsink);
   return 0;
#endif
   return 0;
}

// nothing File::lockBlocking(softint type = F_RDLCK, softint start = 0, softint len = 0, softint whence = SEEK_SET){}
static AbstractQoreNode* File_lockBlocking_vivivivi(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
# 865 "QC_File.qpp"
#ifdef HAVE_STRUCT_FLOCK
   if (check_terminal_io(self, "File::lockBlocking", xsink))
      return 0;

   struct flock fl;
   if (lock_intern(fl, args, xsink))
      return 0;

   f->lockBlocking(fl, xsink);
   return 0;
#else
   file_lock_error("lockBlocking", xsink);
   return 0;
#endif
   return 0;
}

// int File::open(string path, softint flags = O_RDONLY, softint mode = 0666, *string encoding){}
static int64 File_open_VsviviNs(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
   int64 flags = HARD_QORE_INT(args, 1);
   int64 mode = HARD_QORE_INT(args, 2);
   const QoreStringNode* encoding = reinterpret_cast<const QoreStringNode*>(get_param(args, 3));
# 282 "QC_File.qpp"
if (check_terminal_io(self, "File::open", xsink))
      return 0;

   const QoreEncoding *qe = encoding ? QEM.findCreate(encoding) : f->getEncoding();
   return f->open(path->getBuffer(), flags, mode, qe);
}

// nothing File::open2(string path, softint flags = O_RDONLY, softint mode = 0666, *string encoding){}
static AbstractQoreNode* File_open2_VsviviNs(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
   int64 flags = HARD_QORE_INT(args, 1);
   int64 mode = HARD_QORE_INT(args, 2);
   const QoreStringNode* encoding = reinterpret_cast<const QoreStringNode*>(get_param(args, 3));
# 323 "QC_File.qpp"
if (check_terminal_io(self, "File::open2", xsink))
      return 0;

   const QoreEncoding *qe = encoding ? QEM.findCreate(encoding) : f->getEncoding();
   f->open2(xsink, path->getBuffer(), flags, mode, qe);
   return 0;
}

// int File::print(string data){}
static int64 File_print_Vs(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* data = HARD_QORE_STRING(args, 0);
# 423 "QC_File.qpp"
if (check_terminal_io(self, "File::print", xsink))
      return 0;

   return f->write(data, xsink);
}

// int File::printf(string fmt, ...){}
static int64 File_printf_VsVV(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
# 627 "QC_File.qpp"
if (check_terminal_io(self, "File::printf", xsink))
      return 0;

   QoreStringNodeHolder str(q_sprintf(args, 0, 0, xsink));
   if (!str)
      return 0;

   return f->write(*str, xsink);
}

// int File::printf(){}
static int64 File_printf(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
# 641 "QC_File.qpp"
return 0;
}

// nothing File::setCharset(*string encoding){}
static AbstractQoreNode* File_setCharset_Ns(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* encoding = reinterpret_cast<const QoreStringNode*>(get_param(args, 0));
# 773 "QC_File.qpp"
if (check_terminal_io(self, "File::setCharset", xsink))
      return 0;

   const QoreEncoding *qe = encoding ? QEM.findCreate(encoding) : QCS_DEFAULT;
   f->setEncoding(qe);
   return 0;
}

// nothing File::setTerminalAttributes(softint action = TCSANOW, TermIOS termios){}
static AbstractQoreNode* File_setTerminalAttributes_viC7TermIOS(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
   int64 action = HARD_QORE_INT(args, 0);
   HARD_QORE_OBJ_DATA(termios, QoreTermIOS, args, 1, CID_TERMIOS, "File::setTerminalAttributes()", "TermIOS", xsink);
   if (*xsink)
      return 0;
# 1052 "QC_File.qpp"
#ifdef HAVE_TERMIOS_H
   ReferenceHolder<QoreTermIOS> holder(termios, xsink);
   if (check_terminal_io(self, "File::setTerminalAttributes", xsink))
      return 0;

   f->setTerminalAttributes(action, termios, xsink);
   return 0;
#else
   return missing_method_error("File::setTerminalAttributes", "TERMIOS", xsink);
#endif
   return 0;
}

// int File::sync(){}
static int64 File_sync(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
# 342 "QC_File.qpp"
if (check_terminal_io(self, "File::sync", xsink))
      return 0;

   return f->sync();
}

// int File::vprintf(string fmt, any fmt_args){}
static int64 File_vprintf_VsVa(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
# 668 "QC_File.qpp"
if (check_terminal_io(self, "File::vprintf", xsink))
      return 0;

   QoreStringNodeHolder str(q_vsprintf(args, 0, 0, xsink));
   if (!str)
      return 0;

   return f->write(*str, xsink);
}

// int File::vprintf(){}
static int64 File_vprintf(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
# 682 "QC_File.qpp"
return 0;
}

// int File::write(binary data){}
static int64 File_write_Vy(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* data = HARD_QORE_BINARY(args, 0);
# 367 "QC_File.qpp"
if (check_terminal_io(self, "File::write", xsink))
      return 0;

   return f->write(data, xsink);
}

// int File::write(string data){}
static int64 File_write_Vs(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* data = HARD_QORE_STRING(args, 0);
# 395 "QC_File.qpp"
if (check_terminal_io(self, "File::write", xsink))
      return 0;

   return f->write(data, xsink);
}

// int File::writei1(int c){}
static int64 File_writei1_Vi(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
   int64 c = HARD_QORE_INT(args, 0);
# 448 "QC_File.qpp"
if (check_terminal_io(self, "File::writei1", xsink))
      return 0;

   return f->writei1(c, xsink);
}

// int File::writei2(int s){}
static int64 File_writei2_Vi(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
   int64 s = HARD_QORE_INT(args, 0);
# 473 "QC_File.qpp"
if (check_terminal_io(self, "File::writei2", xsink))
      return 0;

   return f->writei2(s, xsink);
}

// int File::writei2LSB(int s){}
static int64 File_writei2LSB_Vi(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
   int64 s = HARD_QORE_INT(args, 0);
# 548 "QC_File.qpp"
if (check_terminal_io(self, "File::writei2LSB", xsink))
      return 0;

   return f->writei2LSB(s, xsink);
}

// int File::writei4(int i){}
static int64 File_writei4_Vi(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
   int64 i = HARD_QORE_INT(args, 0);
# 498 "QC_File.qpp"
if (check_terminal_io(self, "File::writei4", xsink))
      return 0;

   return f->writei4(i, xsink);
}

// int File::writei4LSB(int i){}
static int64 File_writei4LSB_Vi(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
   int64 i = HARD_QORE_INT(args, 0);
# 573 "QC_File.qpp"
if (check_terminal_io(self, "File::writei4LSB", xsink))
      return 0;

   return f->writei4LSB(i, xsink);
}

// int File::writei8(int i){}
static int64 File_writei8_Vi(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
   int64 i = HARD_QORE_INT(args, 0);
# 523 "QC_File.qpp"
if (check_terminal_io(self, "File::writei8", xsink))
      return 0;

   return f->writei8(i, xsink);
}

// int File::writei8LSB(int i){}
static int64 File_writei8LSB_Vi(QoreObject* self, File* f, const QoreListNode* args, ExceptionSink* xsink) {
   int64 i = HARD_QORE_INT(args, 0);
# 598 "QC_File.qpp"
if (check_terminal_io(self, "File::writei8LSB", xsink))
      return 0;

   return f->writei8LSB(i, xsink);
}

// static hash File::hlstat(string path){}
static AbstractQoreNode* static_File_hlstat_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 1169 "QC_File.qpp"
struct stat sbuf;
#ifdef HAVE_LSTAT
   if (lstat(path->getBuffer(), &sbuf))
#else
   if (stat(path->getBuffer(), &sbuf))
#endif
   {
      xsink->raiseErrnoException("FILE-HLSTAT-ERROR", errno, "lstat() command failed");
      return 0;
   }

   return stat_to_hash(sbuf);
}

// static hash File::hstat(string path){}
static AbstractQoreNode* static_File_hstat_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 1142 "QC_File.qpp"
struct stat sbuf;
   if (stat(path->getBuffer(), &sbuf)) {
      xsink->raiseErrnoException("FILE-HSTAT-ERROR", errno, "stat() command failed");
      return 0;
   }

   return stat_to_hash(sbuf);
}

// static list File::lstat(string path){}
static AbstractQoreNode* static_File_lstat_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 1110 "QC_File.qpp"
struct stat sbuf;
#ifdef HAVE_LSTAT
   if (lstat(path->getBuffer(), &sbuf))
#else
   if (stat(path->getBuffer(), &sbuf))
#endif
   {
      xsink->raiseErrnoException("FILE-LSTAT-ERROR", errno, "lstat() command failed");
      return 0;
   }

   return stat_to_list(sbuf);
}

// static list File::stat(string path){}
static AbstractQoreNode* static_File_stat_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 1082 "QC_File.qpp"
struct stat sbuf;
   if (stat(path->getBuffer(), &sbuf)) {
      xsink->raiseErrnoException("FILE-STAT-ERROR", errno, "stat() command failed");
      return 0;
   }

   return stat_to_list(sbuf);
}

// static hash File::statvfs(string path){}
static AbstractQoreNode* static_File_statvfs_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 1201 "QC_File.qpp"
#ifdef HAVE_SYS_STATVFS_H
   struct statvfs vfs;
   if (statvfs(path->getBuffer(), &vfs)) {
      xsink->raiseErrnoException("FILE-STATVFS-ERROR", errno, "statvfs() call failed");
      return 0;
   }

   return statvfs_to_hash(vfs);
#else
   file_lock_error("statvfs", xsink);
   return 0;
#endif
   return 0;
}

DLLLOCAL QoreClass* initFileClass(QoreNamespace& ns) {
   QC_FILE = new QoreClass("File", QDOM_FILESYSTEM);
   CID_FILE = QC_FILE->getID();

   // set parent class
   assert(QC_READONLYFILE);
   QC_FILE->addBuiltinVirtualBaseClass(QC_READONLYFILE);

   // set system constructor
   QC_FILE->setSystemConstructor(FILE_system_constructor);

   // nothing File::chown(softint uid, softint gid = -1){}
   QC_FILE->addMethodExtended3("chown", (q_method_t)File_chown_vivi, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, softBigIntTypeInfo, NULL, "uid", softBigIntTypeInfo, new QoreBigIntNode(-1), "gid");

   // File::constructor(*string encoding) {}
   QC_FILE->setConstructorExtended3(File_constructor_Ns, false, QC_NO_FLAGS, QDOM_DEFAULT, 1, stringOrNothingTypeInfo, NULL, "encoding");

   // File::copy() {}
   QC_FILE->setCopy((q_copy_t)File_copy);

   // int File::f_printf(string fmt, ...){}
   QC_FILE->addMethodExtended3("f_printf", (q_method_int64_t)File_f_printf_VsVV, false, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, bigIntTypeInfo, 1, stringTypeInfo, NULL, "fmt");

   // int File::f_printf(){}
   QC_FILE->addMethodExtended3("f_printf", (q_method_int64_t)File_f_printf, false, QC_RUNTIME_NOOP, QDOM_DEFAULT, bigIntTypeInfo);

   // int File::f_vprintf(string fmt, any fmt_args){}
   QC_FILE->addMethodExtended3("f_vprintf", (q_method_int64_t)File_f_vprintf_VsVa, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, stringTypeInfo, NULL, "fmt", anyTypeInfo, NULL, "fmt_args");

   // int File::f_vprintf(){}
   QC_FILE->addMethodExtended3("f_vprintf", (q_method_int64_t)File_f_vprintf, false, QC_RUNTIME_NOOP, QDOM_DEFAULT, bigIntTypeInfo);

   // string File::getCharset(){}
   QC_FILE->addMethodExtended3("getCharset", (q_method_t)File_getCharset, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

   // hash File::getLockInfo(){}
   QC_FILE->addMethodExtended3("getLockInfo", (q_method_t)File_getLockInfo, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, hashTypeInfo);

   // nothing File::getTerminalAttributes(TermIOS termios){}
   QC_FILE->addMethodExtended3("getTerminalAttributes", (q_method_t)File_getTerminalAttributes_C7TermIOS, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, QC_TERMIOS->getTypeInfo(), NULL, "termios");

   // TermIOS File::getTerminalAttributes(){}
   QC_FILE->addMethodExtended3("getTerminalAttributes", (q_method_t)File_getTerminalAttributes, false, QC_NO_FLAGS, QDOM_DEFAULT, QC_TERMIOS->getTypeInfo());

   // int File::lock(softint type = F_RDLCK, softint start = 0, softint len = 0, softint whence = SEEK_SET){}
   QC_FILE->addMethodExtended3("lock", (q_method_int64_t)File_lock_vivivivi, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 4, softBigIntTypeInfo, new QoreBigIntNode(F_RDLCK), "type", softBigIntTypeInfo, zero(), "start", softBigIntTypeInfo, zero(), "len", softBigIntTypeInfo, new QoreBigIntNode(SEEK_SET), "whence");

   // nothing File::lockBlocking(softint type = F_RDLCK, softint start = 0, softint len = 0, softint whence = SEEK_SET){}
   QC_FILE->addMethodExtended3("lockBlocking", (q_method_t)File_lockBlocking_vivivivi, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 4, softBigIntTypeInfo, new QoreBigIntNode(F_RDLCK), "type", softBigIntTypeInfo, zero(), "start", softBigIntTypeInfo, zero(), "len", softBigIntTypeInfo, new QoreBigIntNode(SEEK_SET), "whence");

   // int File::open(string path, softint flags = O_RDONLY, softint mode = 0666, *string encoding){}
   QC_FILE->addMethodExtended3("open", (q_method_int64_t)File_open_VsviviNs, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 4, stringTypeInfo, NULL, "path", softBigIntTypeInfo, new QoreBigIntNode(O_RDONLY), "flags", softBigIntTypeInfo, new QoreBigIntNode(0666), "mode", stringOrNothingTypeInfo, NULL, "encoding");

   // nothing File::open2(string path, softint flags = O_RDONLY, softint mode = 0666, *string encoding){}
   QC_FILE->addMethodExtended3("open2", (q_method_t)File_open2_VsviviNs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 4, stringTypeInfo, NULL, "path", softBigIntTypeInfo, new QoreBigIntNode(O_RDONLY), "flags", softBigIntTypeInfo, new QoreBigIntNode(0666), "mode", stringOrNothingTypeInfo, NULL, "encoding");

   // int File::print(string data){}
   QC_FILE->addMethodExtended3("print", (q_method_int64_t)File_print_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, stringTypeInfo, NULL, "data");

   // int File::printf(string fmt, ...){}
   QC_FILE->addMethodExtended3("printf", (q_method_int64_t)File_printf_VsVV, false, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, bigIntTypeInfo, 1, stringTypeInfo, NULL, "fmt");

   // int File::printf(){}
   QC_FILE->addMethodExtended3("printf", (q_method_int64_t)File_printf, false, QC_RUNTIME_NOOP, QDOM_DEFAULT, bigIntTypeInfo);

   // nothing File::setCharset(*string encoding){}
   QC_FILE->addMethodExtended3("setCharset", (q_method_t)File_setCharset_Ns, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringOrNothingTypeInfo, NULL, "encoding");

   // nothing File::setTerminalAttributes(softint action = TCSANOW, TermIOS termios){}
   QC_FILE->addMethodExtended3("setTerminalAttributes", (q_method_t)File_setTerminalAttributes_viC7TermIOS, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, softBigIntTypeInfo, new QoreBigIntNode(TCSANOW), "action", QC_TERMIOS->getTypeInfo(), NULL, "termios");

   // int File::sync(){}
   QC_FILE->addMethodExtended3("sync", (q_method_int64_t)File_sync, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo);

   // int File::vprintf(string fmt, any fmt_args){}
   QC_FILE->addMethodExtended3("vprintf", (q_method_int64_t)File_vprintf_VsVa, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, stringTypeInfo, NULL, "fmt", anyTypeInfo, NULL, "fmt_args");

   // int File::vprintf(){}
   QC_FILE->addMethodExtended3("vprintf", (q_method_int64_t)File_vprintf, false, QC_RUNTIME_NOOP, QDOM_DEFAULT, bigIntTypeInfo);

   // int File::write(binary data){}
   QC_FILE->addMethodExtended3("write", (q_method_int64_t)File_write_Vy, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, binaryTypeInfo, NULL, "data");

   // int File::write(string data){}
   QC_FILE->addMethodExtended3("write", (q_method_int64_t)File_write_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, stringTypeInfo, NULL, "data");

   // int File::writei1(int c){}
   QC_FILE->addMethodExtended3("writei1", (q_method_int64_t)File_writei1_Vi, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, bigIntTypeInfo, NULL, "c");

   // int File::writei2(int s){}
   QC_FILE->addMethodExtended3("writei2", (q_method_int64_t)File_writei2_Vi, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, bigIntTypeInfo, NULL, "s");

   // int File::writei2LSB(int s){}
   QC_FILE->addMethodExtended3("writei2LSB", (q_method_int64_t)File_writei2LSB_Vi, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, bigIntTypeInfo, NULL, "s");

   // int File::writei4(int i){}
   QC_FILE->addMethodExtended3("writei4", (q_method_int64_t)File_writei4_Vi, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, bigIntTypeInfo, NULL, "i");

   // int File::writei4LSB(int i){}
   QC_FILE->addMethodExtended3("writei4LSB", (q_method_int64_t)File_writei4LSB_Vi, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, bigIntTypeInfo, NULL, "i");

   // int File::writei8(int i){}
   QC_FILE->addMethodExtended3("writei8", (q_method_int64_t)File_writei8_Vi, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, bigIntTypeInfo, NULL, "i");

   // int File::writei8LSB(int i){}
   QC_FILE->addMethodExtended3("writei8LSB", (q_method_int64_t)File_writei8LSB_Vi, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, bigIntTypeInfo, NULL, "i");

   // static hash File::hlstat(string path){}
   QC_FILE->addStaticMethodExtended3("hlstat", (q_func_t)static_File_hlstat_Vs, false, QC_NO_FLAGS, QDOM_FILESYSTEM, hashTypeInfo, 1, stringTypeInfo, NULL, "path");

   // static hash File::hstat(string path){}
   QC_FILE->addStaticMethodExtended3("hstat", (q_func_t)static_File_hstat_Vs, false, QC_NO_FLAGS, QDOM_FILESYSTEM, hashTypeInfo, 1, stringTypeInfo, NULL, "path");

   // static list File::lstat(string path){}
   QC_FILE->addStaticMethodExtended3("lstat", (q_func_t)static_File_lstat_Vs, false, QC_NO_FLAGS, QDOM_FILESYSTEM, listTypeInfo, 1, stringTypeInfo, NULL, "path");

   // static list File::stat(string path){}
   QC_FILE->addStaticMethodExtended3("stat", (q_func_t)static_File_stat_Vs, false, QC_NO_FLAGS, QDOM_FILESYSTEM, listTypeInfo, 1, stringTypeInfo, NULL, "path");

   // static hash File::statvfs(string path){}
   QC_FILE->addStaticMethodExtended3("statvfs", (q_func_t)static_File_statvfs_Vs, false, QC_NO_FLAGS, QDOM_FILESYSTEM, hashTypeInfo, 1, stringTypeInfo, NULL, "path");
# 99 "QC_File.qpp"
// the following constants were defined earlier but are system-specific; they are included in the build
// in case anyone is using them on the platforms where they are supported
#ifndef O_NOCTTY
#define O_NOCTTY 0
#endif
#ifndef O_NONBLOCK
#define O_NONBLOCK 0
#endif
#ifndef O_NDELAY
#define O_NDELAY 0
#endif
#ifndef O_SYNC
#define O_SYNC 0
#endif
#ifndef O_DIRECT
#define O_DIRECT 0
#endif
#ifndef O_DIRECTORY
#define O_DIRECTORY 0
#endif
#ifndef O_NOFOLLOW
#define O_NOFOLLOW 0
#endif
   ns.addConstant("O_ACCMODE", new QoreBigIntNode(O_ACCMODE));
   ns.addConstant("O_APPEND", new QoreBigIntNode(O_APPEND));
   ns.addConstant("O_CREAT", new QoreBigIntNode(O_CREAT));
   ns.addConstant("O_DIRECT", new QoreBigIntNode(O_DIRECT));
   ns.addConstant("O_DIRECTORY", new QoreBigIntNode(O_DIRECTORY));
   ns.addConstant("O_EXCL", new QoreBigIntNode(O_EXCL));
   ns.addConstant("O_NDELAY", new QoreBigIntNode(O_NDELAY));
   ns.addConstant("O_NOCTTY", new QoreBigIntNode(O_NOCTTY));
   ns.addConstant("O_NOFOLLOW", new QoreBigIntNode(O_NOFOLLOW));
   ns.addConstant("O_NONBLOCK", new QoreBigIntNode(O_NONBLOCK));
   ns.addConstant("O_RDONLY", new QoreBigIntNode(O_RDONLY));
   ns.addConstant("O_RDWR", new QoreBigIntNode(O_RDWR));
   ns.addConstant("O_SYNC", new QoreBigIntNode(O_SYNC));
   ns.addConstant("O_TRUNC", new QoreBigIntNode(O_TRUNC));
   ns.addConstant("O_WRONLY", new QoreBigIntNode(O_WRONLY));
   ns.addConstant("F_RDLCK", new QoreBigIntNode(F_RDLCK));
   ns.addConstant("F_UNLCK", new QoreBigIntNode(F_UNLCK));
   ns.addConstant("F_WRLCK", new QoreBigIntNode(F_WRLCK));
   ns.addConstant("SEEK_CUR", new QoreBigIntNode(SEEK_CUR));
   ns.addConstant("SEEK_END", new QoreBigIntNode(SEEK_END));
   ns.addConstant("SEEK_SET", new QoreBigIntNode(SEEK_SET));

   return QC_FILE;
}
