/* -*- mode: c++; indent-tabs-mode: nil -*- */
/** @file QC_FileLineIterator.qpp FileLineIterator class definition */
/*
  Qore Programming Language

  Copyright 2003 - 2013 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>

class FileLineIterator : public QoreIteratorBase, public QoreFile {
protected:
   QoreStringNode* line,
      * eol;
   int64 num;
   bool validp;
   bool trim;

   DLLLOCAL ~FileLineIterator() {
      if (eol)
         eol->deref();
      if (line)
         line->deref();
   }

public:
   DLLLOCAL FileLineIterator(ExceptionSink* xsink, const char* name, const QoreEncoding* enc = QCS_DEFAULT, QoreStringNode* n_eol = 0, bool n_trim = true)
   : QoreFile(enc), line(new QoreStringNode(enc)), eol(n_eol), num(0), validp(false), trim(n_trim) {
      if (open(name, O_RDONLY, 0, enc))
         xsink->raiseErrnoException("FILELINEITERATOR-OPEN-ERROR", errno, "cannot open '%s'", name);
   }

   DLLLOCAL FileLineIterator(ExceptionSink* xsink, const FileLineIterator& old)
   : QoreFile(old.getEncoding()), line(new QoreStringNode(old.getEncoding())), eol(old.eol->stringRefSelf()), num(old.num), validp(old.validp), trim(old.trim) {
      // reopen file
      std::string fn = old.getFileNameStr();
      if (open(fn.c_str(), O_RDONLY, 0, old.getEncoding()))
         xsink->raiseErrnoException("FILELINEITERATOR-COPY-ERROR", errno, "cannot reopen '%s'", fn.c_str());
      else if (validp) {
         // set file in same position
         setPos(old.getPos());
      }
   }

   DLLLOCAL bool next() {
      // make sure and use a new string if the iterator was already valid
      if (validp) {
         line->deref();
         line = new QoreStringNode(getEncoding());
      }
      validp = eol ? !readUntil(eol->getBuffer(), *line, !trim) : !readLine(*line, !trim);
      if (validp) {
         // increment line number
         ++num;
      }
      else {
         // reset iterator
         num = 0;
         // reset file position
         setPos(0);
      }
      //printd(5, "FileLineIterator::next() this: %p line: %d validp: %d '%s'\n", this, num, validp, line->getBuffer());
      return validp;
   }

   DLLLOCAL int64 index() const {
      return num;
   }

   DLLLOCAL QoreStringNode* getValue() {
      assert(validp);
      return line->stringRefSelf();
   }

   DLLLOCAL bool valid() const {
      return validp;
   }

   DLLLOCAL int checkValid(ExceptionSink* xsink) const {
      if (!validp) {
         xsink->raiseException("ITERATOR-ERROR", "the %s is not pointing at a valid element; make sure %s::next() returns True before calling this method", getName(), getName());
         return -1;
      }
      return 0;
   }

   DLLLOCAL void reset() {
      if (validp) {
         // reset line number
         num = 0;
         // mark as invalid
         validp = false;
         // reset line buffer
         line->deref();
         line = new QoreStringNode(getEncoding());
         // reset file position
         setPos(0);
      }
   }

   DLLLOCAL virtual void deref() {
      if (ROdereference())
         delete this;
   }

   DLLLOCAL virtual const char* getName() const { return "FileLineIterator"; }
};
/* Qore class Qore::FileLineIterator */

qore_classid_t CID_FILELINEITERATOR;
QoreClass* QC_FILELINEITERATOR;

// FileLineIterator::constructor(string path, *string encoding, *string eol, bool trim = True) {}
static void FileLineIterator_constructor_VsNsNsVb(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
   const QoreStringNode* encoding = reinterpret_cast<const QoreStringNode*>(get_param(args, 1));
   const QoreStringNode* eol = reinterpret_cast<const QoreStringNode*>(get_param(args, 2));
   bool trim = HARD_QORE_BOOL(args, 3);
# 159 "QC_FileLineIterator.qpp"
const QoreEncoding *qe = encoding ? QEM.findCreate(encoding) : QCS_DEFAULT;
   if (eol && eol->empty())
      eol = 0;
   // make sure that we use the same encoding for the EOL marker as for the file, convert if necessary
   SimpleRefHolder<QoreStringNode> neol(eol ? new QoreStringNode(qe) : 0);
   if (eol) {
      neol->concat(eol, xsink);
      if (*xsink)
         return;
   }

   SimpleRefHolder<FileLineIterator> fli(new FileLineIterator(xsink, path->getBuffer(), qe, neol.release(), trim));
   if (*xsink)
      return;

   if (fli->isTty() && runtime_check_parse_option(PO_NO_TERMINAL_IO)) {
      xsink->raiseException("ILLEGAL-EXPRESSION", "FileLineIterator::constructor() cannot be called with a TTY target when 'no-terminal-io' is set");
      return;
   }

   self->setPrivate(CID_FILELINEITERATOR, fli.release());
}

// FileLineIterator::copy() {}
static void FileLineIterator_copy(QoreObject* self, QoreObject* old, FileLineIterator* i, ExceptionSink* xsink) {
# 192 "QC_FileLineIterator.qpp"
SimpleRefHolder<FileLineIterator> fli(new FileLineIterator(xsink, *i));

   if (fli->isTty() && runtime_check_parse_option(PO_NO_TERMINAL_IO)) {
      xsink->raiseException("ILLEGAL-EXPRESSION", "FileLineIterator::copy() cannot be called with a TTY target when 'no-terminal-io' is set");
      return;
   }

   if (!*xsink)
      self->setPrivate(CID_FILELINEITERATOR, fli.release());
}

// string FileLineIterator::getEncoding(){}
static AbstractQoreNode* FileLineIterator_getEncoding(QoreObject* self, FileLineIterator* i, const QoreListNode* args, ExceptionSink* xsink) {
# 298 "QC_FileLineIterator.qpp"
return new QoreStringNode(i->getEncoding()->getCode());
}

// string FileLineIterator::getFileName(){}
static AbstractQoreNode* FileLineIterator_getFileName(QoreObject* self, FileLineIterator* i, const QoreListNode* args, ExceptionSink* xsink) {
# 338 "QC_FileLineIterator.qpp"
return i->getFileName();
}

// string FileLineIterator::getLine(){}
static AbstractQoreNode* FileLineIterator_getLine(QoreObject* self, FileLineIterator* i, const QoreListNode* args, ExceptionSink* xsink) {
# 259 "QC_FileLineIterator.qpp"
return i->checkValid(xsink) ? 0 : i->getValue();
}

// int FileLineIterator::getPos(){}
static int64 FileLineIterator_getPos(QoreObject* self, FileLineIterator* i, const QoreListNode* args, ExceptionSink* xsink) {
# 314 "QC_FileLineIterator.qpp"
return i->getPos();
}

// string FileLineIterator::getValue(){}
static AbstractQoreNode* FileLineIterator_getValue(QoreObject* self, FileLineIterator* i, const QoreListNode* args, ExceptionSink* xsink) {
# 240 "QC_FileLineIterator.qpp"
return i->checkValid(xsink) ? 0 : i->getValue();
}

// int FileLineIterator::index(){}
static int64 FileLineIterator_index(QoreObject* self, FileLineIterator* i, const QoreListNode* args, ExceptionSink* xsink) {
# 286 "QC_FileLineIterator.qpp"
return i->index();
}

// bool FileLineIterator::isTty(){}
static bool FileLineIterator_isTty(QoreObject* self, FileLineIterator* i, const QoreListNode* args, ExceptionSink* xsink) {
# 326 "QC_FileLineIterator.qpp"
return i->isTty();
}

// bool FileLineIterator::next(){}
static bool FileLineIterator_next(QoreObject* self, FileLineIterator* i, const QoreListNode* args, ExceptionSink* xsink) {
# 219 "QC_FileLineIterator.qpp"
if (i->check(xsink))
      return false;
   return i->next();
}

// nothing FileLineIterator::reset(){}
static AbstractQoreNode* FileLineIterator_reset(QoreObject* self, FileLineIterator* i, const QoreListNode* args, ExceptionSink* xsink) {
# 352 "QC_FileLineIterator.qpp"
if (i->check(xsink))
       return 0;
   i->reset();
   return 0;
}

// bool FileLineIterator::valid(){}
static bool FileLineIterator_valid(QoreObject* self, FileLineIterator* i, const QoreListNode* args, ExceptionSink* xsink) {
# 272 "QC_FileLineIterator.qpp"
return i->valid();
}

DLLLOCAL QoreClass* initFileLineIteratorClass(QoreNamespace& ns) {
   QC_FILELINEITERATOR = new QoreClass("FileLineIterator", QDOM_FILESYSTEM);
   CID_FILELINEITERATOR = QC_FILELINEITERATOR->getID();

   // set parent class
   assert(QC_ABSTRACTITERATOR);
   QC_FILELINEITERATOR->addBuiltinVirtualBaseClass(QC_ABSTRACTITERATOR);

   // FileLineIterator::constructor(string path, *string encoding, *string eol, bool trim = True) {}
   QC_FILELINEITERATOR->setConstructorExtended3(FileLineIterator_constructor_VsNsNsVb, false, QC_NO_FLAGS, QDOM_DEFAULT, 4, stringTypeInfo, NULL, "path", stringOrNothingTypeInfo, NULL, "encoding", stringOrNothingTypeInfo, NULL, "eol", boolTypeInfo, &True, "trim");

   // FileLineIterator::copy() {}
   QC_FILELINEITERATOR->setCopy((q_copy_t)FileLineIterator_copy);

   // string FileLineIterator::getEncoding(){}
   QC_FILELINEITERATOR->addMethodExtended3("getEncoding", (q_method_t)FileLineIterator_getEncoding, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

   // string FileLineIterator::getFileName(){}
   QC_FILELINEITERATOR->addMethodExtended3("getFileName", (q_method_t)FileLineIterator_getFileName, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

   // string FileLineIterator::getLine(){}
   QC_FILELINEITERATOR->addMethodExtended3("getLine", (q_method_t)FileLineIterator_getLine, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo);

   // int FileLineIterator::getPos(){}
   QC_FILELINEITERATOR->addMethodExtended3("getPos", (q_method_int64_t)FileLineIterator_getPos, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, bigIntTypeInfo);

   // string FileLineIterator::getValue(){}
   QC_FILELINEITERATOR->addMethodExtended3("getValue", (q_method_t)FileLineIterator_getValue, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo);

   // int FileLineIterator::index(){}
   QC_FILELINEITERATOR->addMethodExtended3("index", (q_method_int64_t)FileLineIterator_index, false, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

   // bool FileLineIterator::isTty(){}
   QC_FILELINEITERATOR->addMethodExtended3("isTty", (q_method_bool_t)FileLineIterator_isTty, false, QC_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

   // bool FileLineIterator::next(){}
   QC_FILELINEITERATOR->addMethodExtended3("next", (q_method_bool_t)FileLineIterator_next, false, QC_NO_FLAGS, QDOM_DEFAULT, boolTypeInfo);

   // nothing FileLineIterator::reset(){}
   QC_FILELINEITERATOR->addMethodExtended3("reset", (q_method_t)FileLineIterator_reset, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // bool FileLineIterator::valid(){}
   QC_FILELINEITERATOR->addMethodExtended3("valid", (q_method_bool_t)FileLineIterator_valid, false, QC_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

   return QC_FILELINEITERATOR;
}
