/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
  QC_Program.qpp

  Qore Programming Language

  Copyright 2003 - 2013 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>
#include <qore/intern/QC_Program.h>
#include <qore/intern/QC_TimeZone.h>
#include <qore/intern/qore_program_private.h>
/* Qore class Qore::Program */

qore_classid_t CID_PROGRAM;
QoreClass* QC_PROGRAM;

// any Program::callFunction(string name, ...){}
static AbstractQoreNode* Program_callFunction_VsVV(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* name = HARD_QORE_STRING(args, 0);
# 666 "QC_Program.qpp"
TempEncodingHelper tmp(name, QCS_DEFAULT, xsink);
   if (!tmp)
      return 0;

   ReferenceHolder<QoreListNode> vargs(xsink);
   if (args->size() > 1)
      vargs = args->copyListFrom(1);

   return p->callFunction(tmp->getBuffer(), *vargs, xsink);
}

// any Program::callFunctionArgs(string name, *softlist vargs){}
static AbstractQoreNode* Program_callFunctionArgs_Vsnl(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* name = HARD_QORE_STRING(args, 0);
   const QoreListNode* vargs = reinterpret_cast<const QoreListNode*>(get_param(args, 1));
# 697 "QC_Program.qpp"
TempEncodingHelper tmp(name, QCS_DEFAULT, xsink);
   if (!tmp)
      return 0;
   return p->callFunction(tmp->getBuffer(), vargs, xsink);
}

// Program::constructor(softint po = PO_DEFAULT) {}
static void Program_constructor_vi(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
   int64 po = HARD_QORE_INT(args, 0);
# 454 "QC_Program.qpp"
if (po & PO_SYSTEM_OPS) {
      xsink->raiseException("PROGRAM-OPTION-ERROR", "parse options (0x" QLLX ") contain restricted options that can only be set by the system", po);
      return;
   }

   QoreProgram* pgm = new QoreProgram(getProgram(), po);
   self->setPrivate(CID_PROGRAM, pgm);
}

// Program::copy() {}
static void Program_copy(QoreObject* self, QoreObject* old, QoreProgram* p, ExceptionSink* xsink) {
# 468 "QC_Program.qpp"
xsink->raiseException("PROGRAM-COPY-ERROR", "copying Program objects is currently unsupported");
}

// nothing Program::define(string def, any val){}
static AbstractQoreNode* Program_define_VsVa(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* def = HARD_QORE_STRING(args, 0);
   const AbstractQoreNode* val = get_param(args, 1);
# 1097 "QC_Program.qpp"
TempEncodingHelper t(def, QCS_DEFAULT, xsink);
   if (t)
      qore_program_private::runTimeDefine(p, t->getBuffer(), val ? val->refSelf() : 0, xsink);
   return 0;
}

// Program::destructor() {}
static void Program_destructor(QoreObject* self, QoreProgram* p, ExceptionSink* xsink) {
# 475 "QC_Program.qpp"
p->waitForTerminationAndDeref(xsink);
}

// nothing Program::disableParseOptions(softint opt){}
static AbstractQoreNode* Program_disableParseOptions_vi(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   int64 opt = HARD_QORE_INT(args, 0);
# 894 "QC_Program.qpp"
if (opt & PO_SYSTEM_OPS) {
      xsink->raiseException("PROGRAM-OPTION-ERROR", "parse options (0x" QLLX ") contain restricted options that can only be set by the system", opt);
      return 0;
   }

   p->disableParseOptions(opt, xsink);
   return 0;
}

// bool Program::existsFunction(string name){}
static bool Program_existsFunction_Vs(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* name = HARD_QORE_STRING(args, 0);
# 717 "QC_Program.qpp"
TempEncodingHelper tmp(name, QCS_DEFAULT, xsink);
   if (!tmp)
      return false;
   return p->existsFunction(tmp->getBuffer());
}

// any Program::getDefine(string def){}
static AbstractQoreNode* Program_getDefine_Vs(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* def = HARD_QORE_STRING(args, 0);
# 1142 "QC_Program.qpp"
TempEncodingHelper t(def, QCS_DEFAULT, xsink);
   return t ? qore_program_private::runTimeGetDefine(p, t->getBuffer()) : 0;
}

// any Program::getGlobalVariable(string varname, *reference rexists){}
static AbstractQoreNode* Program_getGlobalVariable_VsNr(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* varname = HARD_QORE_STRING(args, 0);
   const ReferenceNode* rexists = reinterpret_cast<const ReferenceNode*>(get_param(args, 1));
# 1000 "QC_Program.qpp"
TempEncodingHelper t(varname, QCS_DEFAULT, xsink);
   if (!t)
      return 0;

    bool found;
    ReferenceHolder<AbstractQoreNode> rv(p->getGlobalVariableValue(t->getBuffer(), found), xsink);

    if (rexists) {
	QoreTypeSafeReferenceHelper r(rexists, xsink);
	if (!r)
	    return 0;

	if (r.assign(get_bool_node(found), xsink))
	    return 0;
    }

    return rv.release();
}

// int Program::getParseOptions(){}
static int64 Program_getParseOptions(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
# 849 "QC_Program.qpp"
return p->getParseOptions64();
}

// *string Program::getScriptDir(){}
static AbstractQoreNode* Program_getScriptDir(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
# 944 "QC_Program.qpp"
return p->getScriptDir();
}

// *string Program::getScriptName(){}
static AbstractQoreNode* Program_getScriptName(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
# 958 "QC_Program.qpp"
return p->getScriptName();
}

// *string Program::getScriptPath(){}
static AbstractQoreNode* Program_getScriptPath(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
# 972 "QC_Program.qpp"
return p->getScriptPath();
}

// TimeZone Program::getTimeZone(){}
static QoreObject* Program_getTimeZone(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
# 1084 "QC_Program.qpp"
return new QoreObject(QC_TIMEZONE, 0, new TimeZoneData(p->currentTZ()));
}

// list Program::getUserFunctionList(){}
static AbstractQoreNode* Program_getUserFunctionList(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
# 837 "QC_Program.qpp"
return p->getUserFunctionList();
}

// nothing Program::importClass(string cls){}
static AbstractQoreNode* Program_importClass_Vs(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* cls = HARD_QORE_STRING(args, 0);
# 822 "QC_Program.qpp"
TempEncodingHelper tmp(cls, QCS_DEFAULT, xsink);
   if (!tmp)
      return 0;
   qore_program_private::importClass(*p, *(getProgram()), tmp->getBuffer(), xsink);
   return 0;
}

// nothing Program::importFunction(string func){}
static AbstractQoreNode* Program_importFunction_Vs(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* func = HARD_QORE_STRING(args, 0);
# 752 "QC_Program.qpp"
TempEncodingHelper tmp(func, QCS_DEFAULT, xsink);
   if (!tmp)
      return 0;
   qore_program_private::exportFunction(getProgram(), xsink, p, tmp->getBuffer());
   return 0;
}

// nothing Program::importFunction(string func, string new_name){}
static AbstractQoreNode* Program_importFunction_VsVs(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* func = HARD_QORE_STRING(args, 0);
   const QoreStringNode* new_name = HARD_QORE_STRING(args, 1);
# 776 "QC_Program.qpp"
TempEncodingHelper tfunc(func, QCS_DEFAULT, xsink);
   if (!tfunc)
      return 0;
   TempEncodingHelper tnew_name(new_name, QCS_DEFAULT, xsink);
   if (!tnew_name)
      return 0;
   qore_program_private::exportFunction(getProgram(), xsink, p, tfunc->getBuffer(), tnew_name->getBuffer());
   return 0;
}

// nothing Program::importGlobalVariable(string varname, bool readonly = False){}
static AbstractQoreNode* Program_importGlobalVariable_VsVb(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* varname = HARD_QORE_STRING(args, 0);
   bool readonly = HARD_QORE_BOOL(args, 1);
# 800 "QC_Program.qpp"
TempEncodingHelper tmp(varname, QCS_DEFAULT, xsink);
   if (!tmp)
      return 0;
   qore_program_private::exportGlobalVariable(getProgram(), tmp->getBuffer(), readonly, p, xsink);
   return 0;
}

// bool Program::isDefined(string def){}
static bool Program_isDefined_Vs(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* def = HARD_QORE_STRING(args, 0);
# 1127 "QC_Program.qpp"
TempEncodingHelper t(def, QCS_DEFAULT, xsink);
   return t ? qore_program_private::runTimeIsDefined(p, t->getBuffer()) : false;
}

// nothing Program::loadModule(string name){}
static AbstractQoreNode* Program_loadModule_Vs(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* name = HARD_QORE_STRING(args, 0);
# 1165 "QC_Program.qpp"
QMM.runTimeLoadModule(name->getBuffer(), p, *xsink);
   return 0;
}

// nothing Program::lockOptions(){}
static AbstractQoreNode* Program_lockOptions(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
# 983 "QC_Program.qpp"
p->lockOptions();
   return 0;
}

// *hash Program::parse(string code, string label, *softint warning_mask, *string source, *softint offset, softbool format_label = True){}
static AbstractQoreNode* Program_parse_VsVsniNsnivb(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* code = HARD_QORE_STRING(args, 0);
   const QoreStringNode* label = HARD_QORE_STRING(args, 1);
   const QoreBigIntNode* tmp_warning_mask = reinterpret_cast<const QoreBigIntNode*>(get_param(args, 2));
   int64 warning_mask = tmp_warning_mask ? tmp_warning_mask->val : 0;
   const QoreStringNode* source = reinterpret_cast<const QoreStringNode*>(get_param(args, 3));
   const QoreBigIntNode* tmp_offset = reinterpret_cast<const QoreBigIntNode*>(get_param(args, 4));
   int64 offset = tmp_offset ? tmp_offset->val : 0;
   bool format_label = HARD_QORE_BOOL(args, 5);
# 512 "QC_Program.qpp"
QoreStringMaker nlabel(label->getEncoding(), format_label ? "<run-time-loaded: %s>" : "%s", label->getBuffer());

   if (warning_mask) {
      ExceptionSink wsink;
      p->parse(code, &nlabel, xsink, &wsink, warning_mask, source, offset);
      if (!wsink.isException())
         return 0;

      QoreException *e = wsink.catchException();
      return e->makeExceptionObjectAndDelete(xsink);
   }

   p->parse(code, &nlabel, xsink, 0, 0, source, offset);
   return 0;
}

// nothing Program::parseCommit(){}
static AbstractQoreNode* Program_parseCommit(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
# 596 "QC_Program.qpp"
p->parseCommit(xsink);
   return 0;
}

// *hash Program::parseCommit(int warning_mask){}
static AbstractQoreNode* Program_parseCommit_Vi(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   int64 warning_mask = HARD_QORE_INT(args, 0);
# 620 "QC_Program.qpp"
ExceptionSink wsink;
   p->parseCommit(xsink, &wsink, warning_mask);
   if (!wsink.isException())
      return 0;

   QoreException *e = wsink.catchException();
   return e->makeExceptionObjectAndDelete(xsink);
}

// *hash Program::parsePending(string code, string label, *softint warning_mask, *string source, *softint offset, softbool format_label = True){}
static AbstractQoreNode* Program_parsePending_VsVsniNsnivb(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* code = HARD_QORE_STRING(args, 0);
   const QoreStringNode* label = HARD_QORE_STRING(args, 1);
   const QoreBigIntNode* tmp_warning_mask = reinterpret_cast<const QoreBigIntNode*>(get_param(args, 2));
   int64 warning_mask = tmp_warning_mask ? tmp_warning_mask->val : 0;
   const QoreStringNode* source = reinterpret_cast<const QoreStringNode*>(get_param(args, 3));
   const QoreBigIntNode* tmp_offset = reinterpret_cast<const QoreBigIntNode*>(get_param(args, 4));
   int64 offset = tmp_offset ? tmp_offset->val : 0;
   bool format_label = HARD_QORE_BOOL(args, 5);
# 564 "QC_Program.qpp"
QoreStringMaker nlabel(label->getEncoding(), format_label ? "<run-time-loaded: %s>" : "%s", label->getBuffer());

   if (warning_mask) {
      ExceptionSink wsink;
      p->parsePending(code, &nlabel, xsink, &wsink, warning_mask, source, offset);
      if (!wsink.isException())
         return 0;

      QoreException *e = wsink.catchException();
      return e->makeExceptionObjectAndDelete(xsink);
   }

   p->parsePending(code, &nlabel, xsink, 0, 0, source, offset);
   return 0;
}

// nothing Program::parseRollback(){}
static AbstractQoreNode* Program_parseRollback(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
# 643 "QC_Program.qpp"
p->parseRollback();
   return 0;
}

// nothing Program::replaceParseOptions(softint opt){}
static AbstractQoreNode* Program_replaceParseOptions_vi(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   int64 opt = HARD_QORE_INT(args, 0);
# 918 "QC_Program.qpp"
p->replaceParseOptions(opt, xsink);
   return 0;
}

// any Program::run(){}
static AbstractQoreNode* Program_run(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
# 732 "QC_Program.qpp"
return p->run(xsink);
}

// nothing Program::setParseOptions(softint opt = PO_DEFAULT){}
static AbstractQoreNode* Program_setParseOptions_vi(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   int64 opt = HARD_QORE_INT(args, 0);
# 869 "QC_Program.qpp"
if (opt & PO_SYSTEM_OPS) {
      xsink->raiseException("PROGRAM-OPTION-ERROR", "parse options (0x" QLLX ") contain restricted options that can only be set by the system", opt);
      return 0;
   }

   p->setParseOptions(opt, xsink);
   return 0;
}

// nothing Program::setScriptPath(*string path){}
static AbstractQoreNode* Program_setScriptPath_Ns(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = reinterpret_cast<const QoreStringNode*>(get_param(args, 0));
# 930 "QC_Program.qpp"
p->setScriptPath(path ? path->getBuffer() : 0);
   return 0;
}

// nothing Program::setTimeZone(TimeZone zone){}
static AbstractQoreNode* Program_setTimeZone_C8TimeZone(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   HARD_QORE_OBJ_DATA(zone, TimeZoneData, args, 0, CID_TIMEZONE, "Program::setTimeZone()", "TimeZone", xsink);
   if (*xsink)
      return 0;
# 1072 "QC_Program.qpp"
p->setTZ(zone->get());
   return 0;
}

// nothing Program::setTimeZoneRegion(string region){}
static AbstractQoreNode* Program_setTimeZoneRegion_Vs(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* region = HARD_QORE_STRING(args, 0);
# 1032 "QC_Program.qpp"
const AbstractQoreZoneInfo *zone = QTZM.findLoadRegion(region->getBuffer(), xsink);
   if (!*xsink)
      p->setTZ(zone);
   return 0;
}

// nothing Program::setTimeZoneUTCOffset(softint seconds_east){}
static AbstractQoreNode* Program_setTimeZoneUTCOffset_vi(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   int64 seconds_east = HARD_QORE_INT(args, 0);
# 1057 "QC_Program.qpp"
const AbstractQoreZoneInfo *zone = QTZM.findCreateOffsetZone(seconds_east);
   p->setTZ(zone);
   return 0;
}

// nothing Program::undefine(string def){}
static AbstractQoreNode* Program_undefine_Vs(QoreObject* self, QoreProgram* p, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* def = HARD_QORE_STRING(args, 0);
# 1111 "QC_Program.qpp"
TempEncodingHelper t(def, QCS_DEFAULT, xsink);
   if (t)
      qore_program_private::runTimeUnDefine(p, t->getBuffer(), xsink);
   return 0;
}

DLLLOCAL QoreClass* initProgramClass(QoreNamespace& ns) {
   QC_PROGRAM = new QoreClass("Program", QDOM_DEFAULT);
   CID_PROGRAM = QC_PROGRAM->getID();

   // any Program::callFunction(string name, ...){}
   QC_PROGRAM->addMethodExtended3("callFunction", (q_method_t)Program_callFunction_VsVV, false, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, anyTypeInfo, 1, stringTypeInfo, NULL, "name");

   // any Program::callFunctionArgs(string name, *softlist vargs){}
   QC_PROGRAM->addMethodExtended3("callFunctionArgs", (q_method_t)Program_callFunctionArgs_Vsnl, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 2, stringTypeInfo, NULL, "name", softListOrNothingTypeInfo, NULL, "vargs");

   // Program::constructor(softint po = PO_DEFAULT) {}
   QC_PROGRAM->setConstructorExtended3(Program_constructor_vi, false, QC_NO_FLAGS, QDOM_EMBEDDED_LOGIC, 1, softBigIntTypeInfo, new QoreBigIntNode(PO_DEFAULT), "po");

   // Program::copy() {}
   QC_PROGRAM->setCopy((q_copy_t)Program_copy);

   // nothing Program::define(string def, any val){}
   QC_PROGRAM->addMethodExtended3("define", (q_method_t)Program_define_VsVa, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, NULL, "def", anyTypeInfo, NULL, "val");

   // Program::destructor() {}
   QC_PROGRAM->setDestructor((q_destructor_t)Program_destructor);

   // nothing Program::disableParseOptions(softint opt){}
   QC_PROGRAM->addMethodExtended3("disableParseOptions", (q_method_t)Program_disableParseOptions_vi, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, softBigIntTypeInfo, NULL, "opt");

   // bool Program::existsFunction(string name){}
   QC_PROGRAM->addMethodExtended3("existsFunction", (q_method_bool_t)Program_existsFunction_Vs, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, boolTypeInfo, 1, stringTypeInfo, NULL, "name");

   // any Program::getDefine(string def){}
   QC_PROGRAM->addMethodExtended3("getDefine", (q_method_t)Program_getDefine_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 1, stringTypeInfo, NULL, "def");

   // any Program::getGlobalVariable(string varname, *reference rexists){}
   QC_PROGRAM->addMethodExtended3("getGlobalVariable", (q_method_t)Program_getGlobalVariable_VsNr, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 2, stringTypeInfo, NULL, "varname", anyTypeInfo, NULL, "rexists");

   // int Program::getParseOptions(){}
   QC_PROGRAM->addMethodExtended3("getParseOptions", (q_method_int64_t)Program_getParseOptions, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo);

   // *string Program::getScriptDir(){}
   QC_PROGRAM->addMethodExtended3("getScriptDir", (q_method_t)Program_getScriptDir, false, QC_NO_FLAGS, QDOM_DEFAULT, stringOrNothingTypeInfo);

   // *string Program::getScriptName(){}
   QC_PROGRAM->addMethodExtended3("getScriptName", (q_method_t)Program_getScriptName, false, QC_NO_FLAGS, QDOM_DEFAULT, stringOrNothingTypeInfo);

   // *string Program::getScriptPath(){}
   QC_PROGRAM->addMethodExtended3("getScriptPath", (q_method_t)Program_getScriptPath, false, QC_NO_FLAGS, QDOM_DEFAULT, stringOrNothingTypeInfo);

   // TimeZone Program::getTimeZone(){}
   QC_PROGRAM->addMethodExtended3("getTimeZone", (q_method_t)Program_getTimeZone, false, QC_NO_FLAGS, QDOM_DEFAULT, QC_TIMEZONE->getTypeInfo());

   // list Program::getUserFunctionList(){}
   QC_PROGRAM->addMethodExtended3("getUserFunctionList", (q_method_t)Program_getUserFunctionList, false, QC_NO_FLAGS, QDOM_DEFAULT, listTypeInfo);

   // nothing Program::importClass(string cls){}
   QC_PROGRAM->addMethodExtended3("importClass", (q_method_t)Program_importClass_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, NULL, "cls");

   // nothing Program::importFunction(string func){}
   QC_PROGRAM->addMethodExtended3("importFunction", (q_method_t)Program_importFunction_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, NULL, "func");

   // nothing Program::importFunction(string func, string new_name){}
   QC_PROGRAM->addMethodExtended3("importFunction", (q_method_t)Program_importFunction_VsVs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, NULL, "func", stringTypeInfo, NULL, "new_name");

   // nothing Program::importGlobalVariable(string varname, bool readonly = False){}
   QC_PROGRAM->addMethodExtended3("importGlobalVariable", (q_method_t)Program_importGlobalVariable_VsVb, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, NULL, "varname", boolTypeInfo, &False, "readonly");

   // bool Program::isDefined(string def){}
   QC_PROGRAM->addMethodExtended3("isDefined", (q_method_bool_t)Program_isDefined_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, boolTypeInfo, 1, stringTypeInfo, NULL, "def");

   // nothing Program::loadModule(string name){}
   QC_PROGRAM->addMethodExtended3("loadModule", (q_method_t)Program_loadModule_Vs, false, QC_NO_FLAGS, QDOM_MODULES, nothingTypeInfo, 1, stringTypeInfo, NULL, "name");

   // nothing Program::lockOptions(){}
   QC_PROGRAM->addMethodExtended3("lockOptions", (q_method_t)Program_lockOptions, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // *hash Program::parse(string code, string label, *softint warning_mask, *string source, *softint offset, softbool format_label = True){}
   QC_PROGRAM->addMethodExtended3("parse", (q_method_t)Program_parse_VsVsniNsnivb, false, QC_NO_FLAGS, QDOM_EMBEDDED_LOGIC, hashOrNothingTypeInfo, 6, stringTypeInfo, NULL, "code", stringTypeInfo, NULL, "label", softBigIntOrNothingTypeInfo, NULL, "warning_mask", stringOrNothingTypeInfo, NULL, "source", softBigIntOrNothingTypeInfo, NULL, "offset", softBoolTypeInfo, &True, "format_label");

   // nothing Program::parseCommit(){}
   QC_PROGRAM->addMethodExtended3("parseCommit", (q_method_t)Program_parseCommit, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // *hash Program::parseCommit(int warning_mask){}
   QC_PROGRAM->addMethodExtended3("parseCommit", (q_method_t)Program_parseCommit_Vi, false, QC_NO_FLAGS, QDOM_DEFAULT, hashOrNothingTypeInfo, 1, bigIntTypeInfo, NULL, "warning_mask");

   // *hash Program::parsePending(string code, string label, *softint warning_mask, *string source, *softint offset, softbool format_label = True){}
   QC_PROGRAM->addMethodExtended3("parsePending", (q_method_t)Program_parsePending_VsVsniNsnivb, false, QC_NO_FLAGS, QDOM_EMBEDDED_LOGIC, hashOrNothingTypeInfo, 6, stringTypeInfo, NULL, "code", stringTypeInfo, NULL, "label", softBigIntOrNothingTypeInfo, NULL, "warning_mask", stringOrNothingTypeInfo, NULL, "source", softBigIntOrNothingTypeInfo, NULL, "offset", softBoolTypeInfo, &True, "format_label");

   // nothing Program::parseRollback(){}
   QC_PROGRAM->addMethodExtended3("parseRollback", (q_method_t)Program_parseRollback, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // nothing Program::replaceParseOptions(softint opt){}
   QC_PROGRAM->addMethodExtended3("replaceParseOptions", (q_method_t)Program_replaceParseOptions_vi, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, softBigIntTypeInfo, NULL, "opt");

   // any Program::run(){}
   QC_PROGRAM->addMethodExtended3("run", (q_method_t)Program_run, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo);

   // nothing Program::setParseOptions(softint opt = PO_DEFAULT){}
   QC_PROGRAM->addMethodExtended3("setParseOptions", (q_method_t)Program_setParseOptions_vi, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, softBigIntTypeInfo, new QoreBigIntNode(PO_DEFAULT), "opt");

   // nothing Program::setScriptPath(*string path){}
   QC_PROGRAM->addMethodExtended3("setScriptPath", (q_method_t)Program_setScriptPath_Ns, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringOrNothingTypeInfo, NULL, "path");

   // nothing Program::setTimeZone(TimeZone zone){}
   QC_PROGRAM->addMethodExtended3("setTimeZone", (q_method_t)Program_setTimeZone_C8TimeZone, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, QC_TIMEZONE->getTypeInfo(), NULL, "zone");

   // nothing Program::setTimeZoneRegion(string region){}
   QC_PROGRAM->addMethodExtended3("setTimeZoneRegion", (q_method_t)Program_setTimeZoneRegion_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, NULL, "region");

   // nothing Program::setTimeZoneUTCOffset(softint seconds_east){}
   QC_PROGRAM->addMethodExtended3("setTimeZoneUTCOffset", (q_method_t)Program_setTimeZoneUTCOffset_vi, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, softBigIntTypeInfo, NULL, "seconds_east");

   // nothing Program::undefine(string def){}
   QC_PROGRAM->addMethodExtended3("undefine", (q_method_t)Program_undefine_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, NULL, "def");
# 30 "QC_Program.qpp"
# 41 "QC_Program.qpp"
# 48 "QC_Program.qpp"
# 51 "QC_Program.qpp"
# 56 "QC_Program.qpp"
# 61 "QC_Program.qpp"
# 64 "QC_Program.qpp"
# 67 "QC_Program.qpp"
# 73 "QC_Program.qpp"
# 80 "QC_Program.qpp"
# 87 "QC_Program.qpp"
# 93 "QC_Program.qpp"
# 99 "QC_Program.qpp"
# 105 "QC_Program.qpp"
# 111 "QC_Program.qpp"
# 117 "QC_Program.qpp"
# 120 "QC_Program.qpp"
# 127 "QC_Program.qpp"
# 133 "QC_Program.qpp"
# 139 "QC_Program.qpp"
# 145 "QC_Program.qpp"
# 151 "QC_Program.qpp"
# 157 "QC_Program.qpp"
# 164 "QC_Program.qpp"
# 170 "QC_Program.qpp"
# 176 "QC_Program.qpp"
# 182 "QC_Program.qpp"
# 192 "QC_Program.qpp"
# 198 "QC_Program.qpp"
# 204 "QC_Program.qpp"
# 207 "QC_Program.qpp"
# 210 "QC_Program.qpp"
# 216 "QC_Program.qpp"
# 222 "QC_Program.qpp"
# 228 "QC_Program.qpp"
# 234 "QC_Program.qpp"
# 240 "QC_Program.qpp"
# 243 "QC_Program.qpp"
# 250 "QC_Program.qpp"
# 255 "QC_Program.qpp"
# 261 "QC_Program.qpp"
# 270 "QC_Program.qpp"
# 278 "QC_Program.qpp"
   ns.addConstant("PO_ALLOW_BARE_REFS", new QoreBigIntNode(PO_ALLOW_BARE_REFS));
   ns.addConstant("PO_ASSUME_LOCAL", new QoreBigIntNode(PO_ASSUME_LOCAL));
   ns.addConstant("PO_DEFAULT", new QoreBigIntNode(PO_DEFAULT));
   ns.addConstant("PO_FREE_OPTIONS", new QoreBigIntNode(PO_FREE_OPTIONS));
   ns.addConstant("PO_IN_MODULE", new QoreBigIntNode(PO_IN_MODULE));
   ns.addConstant("PO_LOCKDOWN", new QoreBigIntNode(PO_LOCKDOWN));
   ns.addConstant("PO_LOCK_WARNINGS", new QoreBigIntNode(PO_LOCK_WARNINGS));
   ns.addConstant("PO_NEW_STYLE", new QoreBigIntNode(PO_NEW_STYLE));
   ns.addConstant("PO_NO_CHILD_PO_RESTRICTIONS", new QoreBigIntNode(PO_NO_CHILD_PO_RESTRICTIONS));
   ns.addConstant("PO_NO_CLASS_DEFS", new QoreBigIntNode(PO_NO_CLASS_DEFS));
   ns.addConstant("PO_NO_CONSTANT_DEFS", new QoreBigIntNode(PO_NO_CONSTANT_DEFS));
   ns.addConstant("PO_NO_DATABASE", new QoreBigIntNode(PO_NO_DATABASE));
   ns.addConstant("PO_NO_EMBEDDED_LOGIC", new QoreBigIntNode(PO_NO_EMBEDDED_LOGIC));
   ns.addConstant("PO_NO_EXTERNAL_ACCESS", new QoreBigIntNode(PO_NO_EXTERNAL_ACCESS));
   ns.addConstant("PO_NO_EXTERNAL_INFO", new QoreBigIntNode(PO_NO_EXTERNAL_INFO));
   ns.addConstant("PO_NO_EXTERNAL_PROCESS", new QoreBigIntNode(PO_NO_EXTERNAL_PROCESS));
   ns.addConstant("PO_NO_FILESYSTEM", new QoreBigIntNode(PO_NO_FILESYSTEM));
   ns.addConstant("PO_NO_GLOBAL_VARS", new QoreBigIntNode(PO_NO_GLOBAL_VARS));
   ns.addConstant("PO_NO_GUI", new QoreBigIntNode(PO_NO_GUI));
   ns.addConstant("PO_NO_INHERIT_GLOBAL_VARS", new QoreBigIntNode(PO_NO_INHERIT_GLOBAL_VARS));
   ns.addConstant("PO_NO_INHERIT_USER_FUNC_VARIANTS", new QoreBigIntNode(PO_NO_INHERIT_USER_FUNC_VARIANTS));
   ns.addConstant("PO_NO_IO", new QoreBigIntNode(PO_NO_IO));
   ns.addConstant("PO_NO_LOCALE_CONTROL", new QoreBigIntNode(PO_NO_LOCALE_CONTROL));
   ns.addConstant("PO_NO_MODULES", new QoreBigIntNode(PO_NO_MODULES));
   ns.addConstant("PO_NO_NAMESPACE_DEFS", new QoreBigIntNode(PO_NO_NAMESPACE_DEFS));
   ns.addConstant("PO_NO_NETWORK", new QoreBigIntNode(PO_NO_NETWORK));
   ns.addConstant("PO_NO_NEW", new QoreBigIntNode(PO_NO_NEW));
   ns.addConstant("PO_NO_PROCESS_CONTROL", new QoreBigIntNode(PO_NO_PROCESS_CONTROL));
   ns.addConstant("PO_NO_SUBROUTINE_DEFS", new QoreBigIntNode(PO_NO_SUBROUTINE_DEFS));
   ns.addConstant("PO_NO_SYSTEM_CLASSES", new QoreBigIntNode(PO_NO_SYSTEM_CLASSES));
   ns.addConstant("PO_NO_SYSTEM_FUNC_VARIANTS", new QoreBigIntNode(PO_NO_SYSTEM_FUNC_VARIANTS));
   ns.addConstant("PO_NO_TERMINAL_IO", new QoreBigIntNode(PO_NO_TERMINAL_IO));
   ns.addConstant("PO_NO_THREADS", new QoreBigIntNode(PO_NO_THREADS));
   ns.addConstant("PO_NO_THREAD_CLASSES", new QoreBigIntNode(PO_NO_THREAD_CLASSES));
   ns.addConstant("PO_NO_THREAD_CONTROL", new QoreBigIntNode(PO_NO_THREAD_CONTROL));
   ns.addConstant("PO_NO_THREAD_INFO", new QoreBigIntNode(PO_NO_THREAD_INFO));
   ns.addConstant("PO_NO_TOP_LEVEL_STATEMENTS", new QoreBigIntNode(PO_NO_TOP_LEVEL_STATEMENTS));
   ns.addConstant("PO_NO_USER_CLASSES", new QoreBigIntNode(PO_NO_USER_CLASSES));
   ns.addConstant("PO_POSITIVE_OPTIONS", new QoreBigIntNode(PO_POSITIVE_OPTIONS));
   ns.addConstant("PO_REQUIRE_OUR", new QoreBigIntNode(PO_REQUIRE_OUR));
   ns.addConstant("PO_REQUIRE_PROTOTYPES", new QoreBigIntNode(PO_REQUIRE_PROTOTYPES));
   ns.addConstant("PO_REQUIRE_TYPES", new QoreBigIntNode(PO_REQUIRE_TYPES));
   ns.addConstant("PO_STRICT_ARGS", new QoreBigIntNode(PO_STRICT_ARGS));
   ns.addConstant("PO_STRICT_BOOLEAN_EVAL", new QoreBigIntNode(PO_STRICT_BOOLEAN_EVAL));
# 293 "QC_Program.qpp"
# 300 "QC_Program.qpp"
# 305 "QC_Program.qpp"
# 319 "QC_Program.qpp"
# 324 "QC_Program.qpp"
# 329 "QC_Program.qpp"
# 334 "QC_Program.qpp"
# 339 "QC_Program.qpp"
# 344 "QC_Program.qpp"
# 349 "QC_Program.qpp"
# 354 "QC_Program.qpp"
# 366 "QC_Program.qpp"
# 369 "QC_Program.qpp"
# 376 "QC_Program.qpp"
# 381 "QC_Program.qpp"
# 386 "QC_Program.qpp"
# 391 "QC_Program.qpp"
# 396 "QC_Program.qpp"
   ns.addConstant("WARN_ALL", new QoreBigIntNode(QP_WARN_ALL));
   ns.addConstant("WARN_CALL_WITH_TYPE_ERRORS", new QoreBigIntNode(QP_WARN_CALL_WITH_TYPE_ERRORS));
   ns.addConstant("WARN_DEFAULT", new QoreBigIntNode(QP_WARN_DEFAULT));
   ns.addConstant("WARN_DEPRECATED", new QoreBigIntNode(QP_WARN_DEPRECATED));
   ns.addConstant("WARN_DUPLICATE_BLOCK_VARS", new QoreBigIntNode(QP_WARN_DUPLICATE_BLOCK_VARS));
   ns.addConstant("WARN_DUPLICATE_GLOBAL_VARS", new QoreBigIntNode(QP_WARN_DUPLICATE_GLOBAL_VARS));
   ns.addConstant("WARN_DUPLICATE_HASH_KEY", new QoreBigIntNode(QP_WARN_DUPLICATE_HASH_KEY));
   ns.addConstant("WARN_DUPLICATE_LOCAL_VARS", new QoreBigIntNode(QP_WARN_DUPLICATE_LOCAL_VARS));
   ns.addConstant("WARN_EXCESS_ARGS", new QoreBigIntNode(QP_WARN_EXCESS_ARGS));
   ns.addConstant("WARN_INVALID_OPERATION", new QoreBigIntNode(QP_WARN_INVALID_OPERATION));
   ns.addConstant("WARN_MODULES", new QoreBigIntNode(QP_WARN_MODULES));
   ns.addConstant("WARN_NONE", new QoreBigIntNode(QP_WARN_NONE));
   ns.addConstant("WARN_NONEXISTENT_METHOD_CALL", new QoreBigIntNode(QP_WARN_NONEXISTENT_METHOD_CALL));
   ns.addConstant("WARN_RETURN_VALUE_IGNORED", new QoreBigIntNode(QP_WARN_RETURN_VALUE_IGNORED));
   ns.addConstant("WARN_UNDECLARED_VAR", new QoreBigIntNode(QP_WARN_UNDECLARED_VAR));
   ns.addConstant("WARN_UNKNOWN_WARNING", new QoreBigIntNode(QP_WARN_UNKNOWN_WARNING));
   ns.addConstant("WARN_UNREACHABLE_CODE", new QoreBigIntNode(QP_WARN_UNREACHABLE_CODE));
   ns.addConstant("WARN_UNREFERENCED_VARIABLE", new QoreBigIntNode(QP_WARN_UNREFERENCED_VARIABLE));
   ns.addConstant("WARN_WARNING_MASK_UNCHANGED", new QoreBigIntNode(QP_WARN_WARNING_MASK_UNCHANGED));

   return QC_PROGRAM;
}
