/* -*- mode: c++; indent-tabs-mode: nil -*- */
/* @file QC_Queue.qpp Queue class definition

  Qore Programming Language
  
  Copyright 2003 - 2013 David Nichols
  
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>
#include <qore/intern/QC_Queue.h>
#include <qore/intern/QoreQueueIntern.h>
/* Qore class Qore::Thread::Queue */

qore_classid_t CID_QUEUE;
QoreClass* QC_QUEUE;

// nothing Queue::clear(){}
static AbstractQoreNode* Queue_clear(QoreObject* self, Queue *q, const QoreListNode* args, ExceptionSink* xsink) {
# 173 "QC_Queue.qpp"
q->clear(xsink);
   return 0;
}

// Queue::constructor(int max = -1) {}
static void Queue_constructor_Vi(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
   int64 max = HARD_QORE_INT(args, 0);
# 49 "QC_Queue.qpp"
if (!max || (max < 0 && max != -1) || max > 0x7fffffff)
      xsink->raiseException("QUEUE-SIZE-ERROR", QLLD" is an invalid size for a Queue", max);
   else
      self->setPrivate(CID_QUEUE, new Queue(max));
}

// Queue::copy() {}
static void Queue_copy(QoreObject* self, QoreObject* old, Queue *q, ExceptionSink* xsink) {
# 69 "QC_Queue.qpp"
self->setPrivate(CID_QUEUE, new Queue(*q));
}

// Queue::destructor() {}
static void Queue_destructor(QoreObject* self, Queue *q, ExceptionSink* xsink) {
# 61 "QC_Queue.qpp"
qore_queue_private::destructor(*q, xsink);
   q->deref(xsink);
}

// bool Queue::empty(){}
static bool Queue_empty(QoreObject* self, Queue *q, const QoreListNode* args, ExceptionSink* xsink) {
# 199 "QC_Queue.qpp"
return q->empty();
}

// any Queue::get(timeout timeout_ms = 0){}
static AbstractQoreNode* Queue_get_Vt(QoreObject* self, Queue *q, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 124 "QC_Queue.qpp"
AbstractQoreNode *rv;

   if (timeout_ms) {
      bool to;
      rv = q->shift(xsink, timeout_ms, &to);
      if (to)
	 xsink->raiseException("QUEUE-TIMEOUT", "timed out after %d ms", timeout_ms);
   }
   else
      rv = q->shift(xsink);

   return rv;
}

// int Queue::getReadWaiting(){}
static int64 Queue_getReadWaiting(QoreObject* self, Queue *q, const QoreListNode* args, ExceptionSink* xsink) {
# 241 "QC_Queue.qpp"
return q->getReadWaiting();
}

// int Queue::getWaiting(){}
static int64 Queue_getWaiting(QoreObject* self, Queue *q, const QoreListNode* args, ExceptionSink* xsink) {
# 225 "QC_Queue.qpp"
return q->getReadWaiting();
}

// int Queue::getWriteWaiting(){}
static int64 Queue_getWriteWaiting(QoreObject* self, Queue *q, const QoreListNode* args, ExceptionSink* xsink) {
# 255 "QC_Queue.qpp"
return q->getWriteWaiting();
}

// nothing Queue::insert(any arg, timeout timeout_ms = 0){}
static AbstractQoreNode* Queue_insert_VaVt(QoreObject* self, Queue *q, const QoreListNode* args, ExceptionSink* xsink) {
   const AbstractQoreNode* arg = get_param(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 104 "QC_Queue.qpp"
bool to;
   q->insert(xsink, arg, timeout_ms, &to);
   if (to)
      xsink->raiseException("QUEUE-TIMEOUT", "timed out after %d ms", timeout_ms);
   return 0;
}

// int Queue::max(){}
static int64 Queue_max(QoreObject* self, Queue *q, const QoreListNode* args, ExceptionSink* xsink) {
# 211 "QC_Queue.qpp"
return q->getMax();
}

// any Queue::pop(timeout timeout_ms = 0){}
static AbstractQoreNode* Queue_pop_Vt(QoreObject* self, Queue *q, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 152 "QC_Queue.qpp"
AbstractQoreNode *rv;

   if (timeout_ms) {
      bool to;
      rv = q->pop(xsink, timeout_ms, &to);
      if (to)
	 xsink->raiseException("QUEUE-TIMEOUT", "timed out after %d ms", timeout_ms);
   }
   else
      rv = q->pop(xsink);

   return rv;
}

// nothing Queue::push(any arg, timeout timeout_ms = 0){}
static AbstractQoreNode* Queue_push_VaVt(QoreObject* self, Queue *q, const QoreListNode* args, ExceptionSink* xsink) {
   const AbstractQoreNode* arg = get_param(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 85 "QC_Queue.qpp"
bool to;
   q->push(xsink, arg, timeout_ms, &to);
   if (to)
      xsink->raiseException("QUEUE-TIMEOUT", "timed out after %d ms", timeout_ms);
   return 0;
}

// int Queue::size(){}
static int64 Queue_size(QoreObject* self, Queue *q, const QoreListNode* args, ExceptionSink* xsink) {
# 185 "QC_Queue.qpp"
return q->size();
}

DLLLOCAL QoreClass* initQueueClass(QoreNamespace& ns) {
   QC_QUEUE = new QoreClass("Queue", QDOM_THREAD_CLASS);
   CID_QUEUE = QC_QUEUE->getID();

   // nothing Queue::clear(){}
   QC_QUEUE->addMethodExtended3("clear", (q_method_t)Queue_clear, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // Queue::constructor(int max = -1) {}
   QC_QUEUE->setConstructorExtended3(Queue_constructor_Vi, false, QC_NO_FLAGS, QDOM_DEFAULT, 1, bigIntTypeInfo, new QoreBigIntNode(-1), "max");

   // Queue::copy() {}
   QC_QUEUE->setCopy((q_copy_t)Queue_copy);

   // Queue::destructor() {}
   QC_QUEUE->setDestructor((q_destructor_t)Queue_destructor);

   // bool Queue::empty(){}
   QC_QUEUE->addMethodExtended3("empty", (q_method_bool_t)Queue_empty, false, QC_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

   // any Queue::get(timeout timeout_ms = 0){}
   QC_QUEUE->addMethodExtended3("get", (q_method_t)Queue_get_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 1, timeoutTypeInfo, zero(), "timeout_ms");

   // int Queue::getReadWaiting(){}
   QC_QUEUE->addMethodExtended3("getReadWaiting", (q_method_int64_t)Queue_getReadWaiting, false, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

   // int Queue::getWaiting(){}
   QC_QUEUE->addMethodExtended3("getWaiting", (q_method_int64_t)Queue_getWaiting, false, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

   // int Queue::getWriteWaiting(){}
   QC_QUEUE->addMethodExtended3("getWriteWaiting", (q_method_int64_t)Queue_getWriteWaiting, false, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

   // nothing Queue::insert(any arg, timeout timeout_ms = 0){}
   QC_QUEUE->addMethodExtended3("insert", (q_method_t)Queue_insert_VaVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, anyTypeInfo, NULL, "arg", timeoutTypeInfo, zero(), "timeout_ms");

   // int Queue::max(){}
   QC_QUEUE->addMethodExtended3("max", (q_method_int64_t)Queue_max, false, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

   // any Queue::pop(timeout timeout_ms = 0){}
   QC_QUEUE->addMethodExtended3("pop", (q_method_t)Queue_pop_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 1, timeoutTypeInfo, zero(), "timeout_ms");

   // nothing Queue::push(any arg, timeout timeout_ms = 0){}
   QC_QUEUE->addMethodExtended3("push", (q_method_t)Queue_push_VaVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, anyTypeInfo, NULL, "arg", timeoutTypeInfo, zero(), "timeout_ms");

   // int Queue::size(){}
   QC_QUEUE->addMethodExtended3("size", (q_method_int64_t)Queue_size, false, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

   return QC_QUEUE;
}
