/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
  QC_Socket.qpp

  Socket class definition

  Qore Programming Language

  Copyright (C) 2003 - 2014 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>
#include <qore/intern/QC_Socket.h>
#include <qore/intern/ssl_constants.h>
#include <qore/intern/QC_Queue.h>

#include <errno.h>
#include <string.h>

static void hash_set_int_key(QoreHashNode& h, int k, const char* str) {
   char buf[15];
   sprintf(buf, "%d", k);
   h.setKeyValue(buf, new QoreStringNode(str), 0);
}

static QoreHashNode* get_network_address_family_map() {
   QoreHashNode* rv = new QoreHashNode;

   hash_set_int_key(*rv, AF_INET, "AF_INET");
   hash_set_int_key(*rv, AF_INET6, "AF_INET6");
   hash_set_int_key(*rv, AF_UNIX, "AF_UNIX");
   hash_set_int_key(*rv, AF_UNSPEC, "AF_UNSPEC");

   return rv;
}

static int64 checkOpenResult(int rc, const char *method_name, ExceptionSink *xsink) {
   if (*xsink)
      return 0;
   if (rc == -2) {
      xsink->raiseException("SOCKET-NOT-OPEN", "socket must be open before Socket::%s() call", method_name);
      return 0;
   }
   return rc;
}
/* Qore class Qore::Socket */

qore_classid_t CID_SOCKET;
QoreClass* QC_SOCKET;

// Socket Socket::accept(){}
static QoreObject* Socket_accept(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 685 "QC_Socket.qpp"
QoreSocketObject *n = s->accept(-1, xsink);
   if (!n) {
      assert(*xsink);
      return 0;
   }

   assert(n);

   // ensure that a socket object is returned (and not a subclass)
   QoreObject *ns = new QoreObject(QC_SOCKET, getProgram(), n);

   // save backwards-compatible peer parameters as members in new object (deprecated: Socket::getPeerInfo() should be used in the future)
   my_socket_priv::setAccept(*n, ns);

   return ns;
}

// *Socket Socket::accept(timeout timeout_ms){}
static QoreObject* Socket_accept_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 720 "QC_Socket.qpp"
QoreSocketObject *n = s->accept(timeout_ms, xsink);
   if (!n)
      return 0;

   // ensure that a socket object is returned (and not a subclass)
   QoreObject *ns = new QoreObject(QC_SOCKET, getProgram(), n);

   // save backwards-compatible peer parameters as members in new object (deprecated: Socket::getPeerInfo() should be used in the future)
   my_socket_priv::setAccept(*n, ns);

   return ns;
}

// Socket Socket::acceptSSL(){}
static QoreObject* Socket_acceptSSL(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 749 "QC_Socket.qpp"
QoreSocketObject *n = s->acceptSSL(-1, xsink);
   if (!n) {
      assert(*xsink);
      return 0;
   }

   // ensure that a socket object is returned (and not a subclass)
   QoreObject *ns = new QoreObject(QC_SOCKET, getProgram(), n);

   // save backwards-compatible peer parameters as members in new object (deprecated: Socket::getPeerInfo() should be used in the future)
   my_socket_priv::setAccept(*n, ns);
   
   return ns;
}

// *Socket Socket::acceptSSL(timeout timeout_ms){}
static QoreObject* Socket_acceptSSL_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 782 "QC_Socket.qpp"
QoreSocketObject* n = s->acceptSSL(timeout_ms, xsink);
   if (!n)
      return 0;

   // ensure that a socket object is returned (and not a subclass)
   QoreObject* ns = new QoreObject(QC_SOCKET, getProgram(), n);

   // save backwards-compatible peer parameters as members in new object (deprecated: Socket::getPeerInfo() should be used in the future)
   my_socket_priv::setAccept(*n, ns);
   
   return ns;
}

// int Socket::bind(string str, softbool reuseaddr = False){}
static int64 Socket_bind_Vsvb(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   bool reuseaddr = HARD_QORE_BOOL(args, 1);
# 586 "QC_Socket.qpp"
return s->bind(str->getBuffer(), reuseaddr);
}

// int Socket::bind(int port, softbool reuseaddr = False){}
static int64 Socket_bind_Vivb(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 port = HARD_QORE_INT(args, 0);
   bool reuseaddr = HARD_QORE_BOOL(args, 1);
# 605 "QC_Socket.qpp"
return s->bind(port, reuseaddr);
}

// nothing Socket::bindINET(*string iface, *softstring service, softbool reuseaddr = False, softint family = AF_UNSPEC, softint socktype = SOCK_STREAM, softint protocol = 0){}
static AbstractQoreNode* Socket_bindINET_Nsnsvbvivivi(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* iface = reinterpret_cast<const QoreStringNode*>(get_param(args, 0));
   const QoreStringNode* service = reinterpret_cast<const QoreStringNode*>(get_param(args, 1));
   bool reuseaddr = HARD_QORE_BOOL(args, 2);
   int64 family = HARD_QORE_INT(args, 3);
   int64 socktype = HARD_QORE_INT(args, 4);
   int64 protocol = HARD_QORE_INT(args, 5);
# 659 "QC_Socket.qpp"
if ((!iface || iface->empty())
       && (!service || service->empty())) {
      xsink->raiseException("SOCKET-BIND-ERROR", "both interace (first parameter) and service (second parameter) were either not present or empty strings; at least one of the first 2 parameters must be present for a successful call to Socket::bindINET()");
      return 0;
   }

   s->bindINET(iface ? iface->getBuffer() : 0, service ? service->getBuffer() : 0, reuseaddr, family, socktype, protocol, xsink);
   return 0;
}

// nothing Socket::bindUNIX(string path, softint socktype = SOCK_STREAM, softint protocol = 0){}
static AbstractQoreNode* Socket_bindUNIX_Vsvivi(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
   int64 socktype = HARD_QORE_INT(args, 1);
   int64 protocol = HARD_QORE_INT(args, 2);
# 626 "QC_Socket.qpp"
s->bindUNIX(path->getBuffer(), socktype, protocol, xsink);
   return 0;
}

// nothing Socket::clearStats(){}
static AbstractQoreNode* Socket_clearStats(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2636 "QC_Socket.qpp"
s->clearStats();
   return 0;
}

// nothing Socket::clearWarningQueue(){}
static AbstractQoreNode* Socket_clearWarningQueue(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2558 "QC_Socket.qpp"
s->clearWarningQueue(xsink);
   return 0;
}

// int Socket::close(){}
static int64 Socket_close(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2079 "QC_Socket.qpp"
return s->close();
}

// nothing Socket::connect(string target, timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_connect_VsVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* target = HARD_QORE_STRING(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 420 "QC_Socket.qpp"
s->connect(target->getBuffer(), timeout_ms, xsink);
   return 0;
}

// nothing Socket::connectINET(string host, softstring service, timeout timeout_ms = -1, softint family = AF_UNSPEC, softint socktype = SOCK_STREAM, softint protocol = 0){}
static AbstractQoreNode* Socket_connectINET_VsvsVtvivivi(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* host = HARD_QORE_STRING(args, 0);
   const QoreStringNode* service = HARD_QORE_STRING(args, 1);
   int64 timeout_ms = HARD_QORE_INT(args, 2);
   int64 family = HARD_QORE_INT(args, 3);
   int64 socktype = HARD_QORE_INT(args, 4);
   int64 protocol = HARD_QORE_INT(args, 5);
# 449 "QC_Socket.qpp"
s->connectINET2(host->getBuffer(), service->getBuffer(), family, socktype, protocol, timeout_ms, xsink);
   return 0;
}

// nothing Socket::connectINETSSL(string host, softstring service, timeout timeout_ms = -1, softint family = AF_UNSPEC, softint socktype = SOCK_STREAM, softint protocol = 0){}
static AbstractQoreNode* Socket_connectINETSSL_VsvsVtvivivi(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* host = HARD_QORE_STRING(args, 0);
   const QoreStringNode* service = HARD_QORE_STRING(args, 1);
   int64 timeout_ms = HARD_QORE_INT(args, 2);
   int64 family = HARD_QORE_INT(args, 3);
   int64 socktype = HARD_QORE_INT(args, 4);
   int64 protocol = HARD_QORE_INT(args, 5);
# 536 "QC_Socket.qpp"
s->connectINET2SSL(host->getBuffer(), service->getBuffer(), family, socktype, protocol, timeout_ms, xsink);
   return 0;
}

// nothing Socket::connectSSL(string target, timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_connectSSL_VsVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* target = HARD_QORE_STRING(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 506 "QC_Socket.qpp"
s->connectSSL(target->getBuffer(), timeout_ms, xsink);
   return 0;
}

// nothing Socket::connectUNIX(string path, softint socktype = SOCK_STREAM, softint protocol = 0){}
static AbstractQoreNode* Socket_connectUNIX_Vsvivi(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
   int64 socktype = HARD_QORE_INT(args, 1);
   int64 protocol = HARD_QORE_INT(args, 2);
# 473 "QC_Socket.qpp"
s->connectUNIX(path->getBuffer(), socktype, protocol, xsink);
   return 0;
}

// nothing Socket::connectUNIXSSL(string path, softint socktype = SOCK_STREAM, softint protocol = 0){}
static AbstractQoreNode* Socket_connectUNIXSSL_Vsvivi(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
   int64 socktype = HARD_QORE_INT(args, 1);
   int64 protocol = HARD_QORE_INT(args, 2);
# 561 "QC_Socket.qpp"
s->connectUNIXSSL(path->getBuffer(), socktype, protocol, xsink);
   return 0;
}

// Socket::constructor() {}
static void Socket_constructor(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
# 382 "QC_Socket.qpp"
self->setPrivate(CID_SOCKET, new QoreSocketObject);
}

// Socket::copy() {}
static void Socket_copy(QoreObject* self, QoreObject* old, QoreSocketObject* s, ExceptionSink* xsink) {
# 390 "QC_Socket.qpp"
self->setPrivate(CID_SOCKET, new QoreSocketObject);
}

// string Socket::getCharset(){}
static AbstractQoreNode* Socket_getCharset(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2212 "QC_Socket.qpp"
return new QoreStringNode(s->getEncoding()->getCode());
}

// string Socket::getEncoding(){}
static AbstractQoreNode* Socket_getEncoding(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2197 "QC_Socket.qpp"
return new QoreStringNode(s->getEncoding()->getCode());
}

// bool Socket::getNoDelay(){}
static bool Socket_getNoDelay(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2484 "QC_Socket.qpp"
return s->getNoDelay();
}

// hash Socket::getPeerInfo(bool host_lookup = True){}
static AbstractQoreNode* Socket_getPeerInfo_Vb(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   bool host_lookup = HARD_QORE_BOOL(args, 0);
# 2504 "QC_Socket.qpp"
return s->getPeerInfo(xsink, host_lookup);
}

// int Socket::getPort(){}
static int64 Socket_getPort(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2120 "QC_Socket.qpp"
return s->getPort();
}

// int Socket::getRecvTimeout(){}
static int64 Socket_getRecvTimeout(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2190 "QC_Socket.qpp"
return s->getRecvTimeout();
}

// *string Socket::getSSLCipherName(){}
static AbstractQoreNode* Socket_getSSLCipherName(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2281 "QC_Socket.qpp"
const char *str = s->getSSLCipherName();
   return str ? new QoreStringNode(str) : 0;
}

// *string Socket::getSSLCipherVersion(){}
static AbstractQoreNode* Socket_getSSLCipherVersion(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2294 "QC_Socket.qpp"
const char *str = s->getSSLCipherVersion();
   return str ? new QoreStringNode(str) : 0;
}

// int Socket::getSendTimeout(){}
static int64 Socket_getSendTimeout(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2178 "QC_Socket.qpp"
return s->getSendTimeout();
}

// int Socket::getSocket(){}
static int64 Socket_getSocket(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2132 "QC_Socket.qpp"
return s->getSocket();
}

// hash Socket::getSocketInfo(bool host_lookup = True){}
static AbstractQoreNode* Socket_getSocketInfo_Vb(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   bool host_lookup = HARD_QORE_BOOL(args, 0);
# 2524 "QC_Socket.qpp"
return s->getSocketInfo(xsink, host_lookup);
}

// hash Socket::getUsageInfo(){}
static AbstractQoreNode* Socket_getUsageInfo(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2622 "QC_Socket.qpp"
return s->getUsageInfo();
}

// bool Socket::isDataAvailable(timeout timeout_ms = 0){}
static bool Socket_isDataAvailable_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 2238 "QC_Socket.qpp"
return s->isDataAvailable(xsink, (int)timeout_ms);
}

// bool Socket::isOpen(){}
static bool Socket_isOpen(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2269 "QC_Socket.qpp"
return s->isOpen();
}

// bool Socket::isSecure(){}
static bool Socket_isSecure(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2307 "QC_Socket.qpp"
return s->isSecure();
}

// bool Socket::isWriteFinished(timeout timeout_ms = 0){}
static bool Socket_isWriteFinished_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 2257 "QC_Socket.qpp"
return s->isWriteFinished(xsink, (int)timeout_ms);
}

// int Socket::listen(int backlog = 20){}
static int64 Socket_listen_Vi(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 backlog = HARD_QORE_INT(args, 0);
# 810 "QC_Socket.qpp"
return checkOpenResult(s->listen((int)backlog), "listen", xsink);
}

// bool Socket::pendingHttpChunkedBody(){}
static bool Socket_pendingHttpChunkedBody(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2648 "QC_Socket.qpp"
return s->pendingHttpChunkedBody();
}

// hash Socket::readHTTPChunkedBody(timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_readHTTPChunkedBody_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 1967 "QC_Socket.qpp"
// when rc = -3 it's a timeout, but rv will be NULL anyway, so we do nothing
   return s->readHTTPChunkedBody(timeout_ms, xsink);
}

// hash Socket::readHTTPChunkedBodyBinary(timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_readHTTPChunkedBodyBinary_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 1994 "QC_Socket.qpp"
// when rc = -3 it's a timeout, but rv will be NULL anyway, so we do nothing
   return s->readHTTPChunkedBodyBinary(timeout_ms, xsink);
}

// nothing Socket::readHTTPChunkedBodyBinaryWithCallback(code rcb, timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_readHTTPChunkedBodyBinaryWithCallback_VcVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const ResolvedCallReferenceNode* rcb = reinterpret_cast<const ResolvedCallReferenceNode*>(get_param(args, 0));
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 2061 "QC_Socket.qpp"
s->readHTTPChunkedBodyBinaryWithCallback(*rcb, self, timeout_ms, xsink);
   return 0;
}

// nothing Socket::readHTTPChunkedBodyWithCallback(code rcb, timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_readHTTPChunkedBodyWithCallback_VcVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const ResolvedCallReferenceNode* rcb = reinterpret_cast<const ResolvedCallReferenceNode*>(get_param(args, 0));
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 2028 "QC_Socket.qpp"
s->readHTTPChunkedBodyWithCallback(*rcb, self, timeout_ms, xsink);
   return 0;
}

// hash Socket::readHTTPHeader(timeout timeout_ms = -1, *reference info){}
static AbstractQoreNode* Socket_readHTTPHeader_VtNr(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
   const ReferenceNode* info = reinterpret_cast<const ReferenceNode*>(get_param(args, 1));
# 1912 "QC_Socket.qpp"
OptHashRefHelper ohrh(info, xsink);
   return s->readHTTPHeader(xsink, *ohrh, timeout_ms);
}

// string Socket::readHTTPHeaderString(timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_readHTTPHeaderString_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 1941 "QC_Socket.qpp"
return s->readHTTPHeaderString(xsink, timeout_ms);
}

// string Socket::recv(softint size = 0, timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_recv_viVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 size = HARD_QORE_INT(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 1278 "QC_Socket.qpp"
return size > 0 ? s->recv(size, timeout_ms, xsink) : s->recv(timeout_ms, xsink);
}

// binary Socket::recvBinary(softint size = 0, timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_recvBinary_viVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 size = HARD_QORE_INT(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 1306 "QC_Socket.qpp"
return size > 0 ? s->recvBinary(size, timeout_ms, xsink) : s->recvBinary(timeout_ms, xsink);
}

// int Socket::recvi1(timeout timeout_ms = -1){}
static int64 Socket_recvi1_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 1333 "QC_Socket.qpp"
char b;
   s->recvi1(timeout_ms, &b, xsink);
   return b;
}

// int Socket::recvi2(timeout timeout_ms = -1){}
static int64 Socket_recvi2_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 1362 "QC_Socket.qpp"
short b;
   s->recvi2(timeout_ms, &b, xsink);
   return b;
}

// int Socket::recvi2LSB(timeout timeout_ms = -1){}
static int64 Socket_recvi2LSB_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 1449 "QC_Socket.qpp"
short b;
   s->recvi2LSB(timeout_ms, &b, xsink);
   return b;
}

// int Socket::recvi4(timeout timeout_ms = -1){}
static int64 Socket_recvi4_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 1391 "QC_Socket.qpp"
int b;
   s->recvi4(timeout_ms, &b, xsink);
   return b;
}

// int Socket::recvi4LSB(timeout timeout_ms = -1){}
static int64 Socket_recvi4LSB_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 1478 "QC_Socket.qpp"
int b;
   s->recvi4LSB(timeout_ms, &b, xsink);
   return b;
}

// int Socket::recvi8(timeout timeout_ms = -1){}
static int64 Socket_recvi8_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 1420 "QC_Socket.qpp"
int64 b;
   s->recvi8(timeout_ms, &b, xsink);
   return b;
}

// int Socket::recvi8LSB(timeout timeout_ms = -1){}
static int64 Socket_recvi8LSB_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 1507 "QC_Socket.qpp"
int64 b;
   s->recvi8LSB(timeout_ms, &b, xsink);
   return b;
}

// int Socket::recvu1(timeout timeout_ms = -1){}
static int64 Socket_recvu1_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 1536 "QC_Socket.qpp"
unsigned char b;
   s->recvu1(timeout_ms, &b, xsink);
   return b;
}

// int Socket::recvu2(timeout timeout_ms = -1){}
static int64 Socket_recvu2_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 1565 "QC_Socket.qpp"
unsigned short b;
   s->recvu2(timeout_ms, &b, xsink);
   return b;
}

// int Socket::recvu2LSB(timeout timeout_ms = -1){}
static int64 Socket_recvu2LSB_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 1623 "QC_Socket.qpp"
unsigned short b;
   s->recvu2LSB(timeout_ms, &b, xsink);
   return b;
}

// int Socket::recvu4(timeout timeout_ms = -1){}
static int64 Socket_recvu4_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 1594 "QC_Socket.qpp"
unsigned int b;
   s->recvu4(timeout_ms, &b, xsink);
   return b;
}

// int Socket::recvu4LSB(timeout timeout_ms = -1){}
static int64 Socket_recvu4LSB_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 1652 "QC_Socket.qpp"
unsigned b;
   s->recvu4LSB(timeout_ms, &b, xsink);
   return b;
}

// int Socket::send(binary bin, int timeout_ms = -1){}
static int64 Socket_send_VyVi(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 842 "QC_Socket.qpp"
return s->send(bin, timeout_ms, xsink);
}

// int Socket::send(string str, timeout timeout_ms = -1){}
static int64 Socket_send_VsVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 875 "QC_Socket.qpp"
return s->send(str, timeout_ms, xsink);
}

// nothing Socket::send2(binary bin, timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_send2_VyVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 966 "QC_Socket.qpp"
s->send(bin, timeout_ms, xsink);
   return 0;
}

// nothing Socket::send2(string str, timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_send2_VsVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 996 "QC_Socket.qpp"
s->send(str, timeout_ms, xsink);
   return 0;
}

// int Socket::sendBinary(string str, timeout timeout_ms = -1){}
static int64 Socket_sendBinary_VsVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 906 "QC_Socket.qpp"
return s->send(str->getBuffer(), str->strlen(), timeout_ms, xsink);
}

// int Socket::sendBinary(binary bin, timeout timeout_ms = -1){}
static int64 Socket_sendBinary_VyVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 937 "QC_Socket.qpp"
return s->send(bin, timeout_ms, xsink);
}

// nothing Socket::sendBinary2(string str, timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_sendBinary2_VsVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 1024 "QC_Socket.qpp"
s->send(str->getBuffer(), str->strlen(), timeout_ms, xsink);
   return 0;
}

// nothing Socket::sendBinary2(binary bin, timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_sendBinary2_VyVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 1052 "QC_Socket.qpp"
s->send(bin, timeout_ms, xsink);
   return 0;
}

// nothing Socket::sendHTTPMessage(string method, string path, string http_version, hash headers, *string body, *reference info, timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_sendHTTPMessage_VsVsVsVhNsNrVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* method = HARD_QORE_STRING(args, 0);
   const QoreStringNode* path = HARD_QORE_STRING(args, 1);
   const QoreStringNode* http_version = HARD_QORE_STRING(args, 2);
   const QoreHashNode* headers = HARD_QORE_HASH(args, 3);
   const QoreStringNode* body = reinterpret_cast<const QoreStringNode*>(get_param(args, 4));
   const ReferenceNode* info = reinterpret_cast<const ReferenceNode*>(get_param(args, 5));
   int64 timeout_ms = HARD_QORE_INT(args, 6);
# 1687 "QC_Socket.qpp"
TempEncodingHelper tmp_data;
   if (body && !tmp_data.set(body, s->getEncoding(), xsink))
      return 0;

   OptHashRefHelper ohrh(info, xsink);
   s->sendHTTPMessage(xsink, *ohrh, method->getBuffer(), path->getBuffer(), http_version->getBuffer(), headers, body ? tmp_data->getBuffer() : 0, body ? tmp_data->strlen() : 0, QORE_SOURCE_SOCKET, timeout_ms);
   return 0;
}

// nothing Socket::sendHTTPMessage(string method, string path, string http_version, hash headers, binary body, *reference info, timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_sendHTTPMessage_VsVsVsVhVyNrVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* method = HARD_QORE_STRING(args, 0);
   const QoreStringNode* path = HARD_QORE_STRING(args, 1);
   const QoreStringNode* http_version = HARD_QORE_STRING(args, 2);
   const QoreHashNode* headers = HARD_QORE_HASH(args, 3);
   const BinaryNode* body = HARD_QORE_BINARY(args, 4);
   const ReferenceNode* info = reinterpret_cast<const ReferenceNode*>(get_param(args, 5));
   int64 timeout_ms = HARD_QORE_INT(args, 6);
# 1723 "QC_Socket.qpp"
OptHashRefHelper ohrh(info, xsink);
   s->sendHTTPMessage(xsink, *ohrh, method->getBuffer(), path->getBuffer(), http_version->getBuffer(), headers, body->getPtr(), body->size(), QORE_SOURCE_SOCKET, timeout_ms);
   return 0;
}

// nothing Socket::sendHTTPMessageWithCallback(code scb, string method, string path, string http_version, hash headers, *reference info, timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_sendHTTPMessageWithCallback_VcVsVsVsVhNrVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const ResolvedCallReferenceNode* scb = reinterpret_cast<const ResolvedCallReferenceNode*>(get_param(args, 0));
   const QoreStringNode* method = HARD_QORE_STRING(args, 1);
   const QoreStringNode* path = HARD_QORE_STRING(args, 2);
   const QoreStringNode* http_version = HARD_QORE_STRING(args, 3);
   const QoreHashNode* headers = HARD_QORE_HASH(args, 4);
   const ReferenceNode* info = reinterpret_cast<const ReferenceNode*>(get_param(args, 5));
   int64 timeout_ms = HARD_QORE_INT(args, 6);
# 1757 "QC_Socket.qpp"
OptHashRefHelper ohrh(info, xsink);
   s->sendHTTPMessageWithCallback(xsink, *ohrh, method->getBuffer(), path->getBuffer(), http_version->getBuffer(), headers, *scb, QORE_SOURCE_SOCKET, timeout_ms);
   return 0;
}

// nothing Socket::sendHTTPResponse(softint status_code, string status_desc, string http_version, hash headers, *string body, timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_sendHTTPResponse_viVsVsVhNsVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 status_code = HARD_QORE_INT(args, 0);
   const QoreStringNode* status_desc = HARD_QORE_STRING(args, 1);
   const QoreStringNode* http_version = HARD_QORE_STRING(args, 2);
   const QoreHashNode* headers = HARD_QORE_HASH(args, 3);
   const QoreStringNode* body = reinterpret_cast<const QoreStringNode*>(get_param(args, 4));
   int64 timeout_ms = HARD_QORE_INT(args, 5);
# 1787 "QC_Socket.qpp"
if (status_code < 100 || status_code >= 600) {
      xsink->raiseException("SOCKET-SENDHTTPRESPONSE-STATUS-ERROR", "expecting valid HTTP status code between 100 and 599 as first parameter of Socket::sendHTTPResponse() call, got value %d instead", status_code);
      return 0;
   }

   TempEncodingHelper tmp_data;
   if (body && !tmp_data.set(body, s->getEncoding(), xsink))
      return 0;

   s->sendHTTPResponse(xsink, status_code, status_desc->getBuffer(), http_version->getBuffer(), headers, body ? tmp_data->getBuffer() : 0, body ? tmp_data->strlen() : 0, QORE_SOURCE_SOCKET, timeout_ms);
   return 0;
}

// nothing Socket::sendHTTPResponse(softint status_code, string status_desc, string http_version, hash headers, binary body, timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_sendHTTPResponse_viVsVsVhVyVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 status_code = HARD_QORE_INT(args, 0);
   const QoreStringNode* status_desc = HARD_QORE_STRING(args, 1);
   const QoreStringNode* http_version = HARD_QORE_STRING(args, 2);
   const QoreHashNode* headers = HARD_QORE_HASH(args, 3);
   const BinaryNode* body = HARD_QORE_BINARY(args, 4);
   int64 timeout_ms = HARD_QORE_INT(args, 5);
# 1823 "QC_Socket.qpp"
if (status_code < 100 || status_code >= 600) {
      xsink->raiseException("SOCKET-SENDHTTPRESPONSE-STATUS-ERROR", "expecting valid HTTP status code between 100 and 599 as first parameter of Socket::sendHTTPResponse() call, got value %d instead", status_code);
      return 0;
   }

   s->sendHTTPResponse(xsink, status_code, status_desc->getBuffer(), http_version->getBuffer(), headers, body ? body->getPtr() : 0, body ? body->size() : 0, QORE_SOURCE_SOCKET, timeout_ms);
   return 0;
}

// nothing Socket::sendHTTPResponseWithCallback(code scb, softint status_code, string status_desc, string http_version, hash headers, timeout timeout_ms = -1){}
static AbstractQoreNode* Socket_sendHTTPResponseWithCallback_VcviVsVsVhVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const ResolvedCallReferenceNode* scb = reinterpret_cast<const ResolvedCallReferenceNode*>(get_param(args, 0));
   int64 status_code = HARD_QORE_INT(args, 1);
   const QoreStringNode* status_desc = HARD_QORE_STRING(args, 2);
   const QoreStringNode* http_version = HARD_QORE_STRING(args, 3);
   const QoreHashNode* headers = HARD_QORE_HASH(args, 4);
   int64 timeout_ms = HARD_QORE_INT(args, 5);
# 1857 "QC_Socket.qpp"
if (status_code < 100 || status_code >= 600) {
      xsink->raiseException("SOCKET-SENDHTTPRESPONSE-STATUS-ERROR", "expecting valid HTTP status code between 100 and 599 as second parameter of Socket::sendHTTPResponseWithCallback() call, got value %d instead", status_code);
      return 0;
   }

   s->sendHTTPResponseWithCallback(xsink, status_code, status_desc->getBuffer(), http_version->getBuffer(), headers, *scb, QORE_SOURCE_SOCKET, timeout_ms);
   return 0;
}

// int Socket::sendi1(softint i = 0, timeout timeout_ms = -1){}
static int64 Socket_sendi1_viVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 i = HARD_QORE_INT(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 1080 "QC_Socket.qpp"
return s->sendi1((char)i, timeout_ms, xsink);
}

// int Socket::sendi2(softint i = 0, timeout timeout_ms = -1){}
static int64 Socket_sendi2_viVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 i = HARD_QORE_INT(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 1108 "QC_Socket.qpp"
return s->sendi2((short)i, timeout_ms, xsink);
}

// int Socket::sendi2LSB(softint i = 0, timeout timeout_ms = -1){}
static int64 Socket_sendi2LSB_viVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 i = HARD_QORE_INT(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 1192 "QC_Socket.qpp"
return s->sendi2LSB((short)i, timeout_ms, xsink);
}

// int Socket::sendi4(softint i = 0, timeout timeout_ms = -1){}
static int64 Socket_sendi4_viVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 i = HARD_QORE_INT(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 1136 "QC_Socket.qpp"
return s->sendi4((int)i, timeout_ms, xsink);
}

// int Socket::sendi4LSB(softint i = 0, timeout timeout_ms = -1){}
static int64 Socket_sendi4LSB_viVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 i = HARD_QORE_INT(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 1220 "QC_Socket.qpp"
return s->sendi4LSB((int)i, timeout_ms, xsink);
}

// int Socket::sendi8(softint i = 0, timeout timeout_ms = -1){}
static int64 Socket_sendi8_viVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 i = HARD_QORE_INT(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 1164 "QC_Socket.qpp"
return s->sendi8(i, timeout_ms, xsink);
}

// int Socket::sendi8LSB(softint i = 0, timeout timeout_ms = -1){}
static int64 Socket_sendi8LSB_viVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 i = HARD_QORE_INT(args, 0);
   int64 timeout_ms = HARD_QORE_INT(args, 1);
# 1248 "QC_Socket.qpp"
return s->sendi8LSB(i, timeout_ms, xsink);
}

// nothing Socket::setCertificate(SSLCertificate cert){}
static AbstractQoreNode* Socket_setCertificate_C14SSLCertificate(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   HARD_QORE_OBJ_DATA(cert, QoreSSLCertificate, args, 0, CID_SSLCERTIFICATE, "Socket::setCertificate()", "SSLCertificate", xsink);
   if (*xsink)
      return 0;
# 2332 "QC_Socket.qpp"
// pass reference from QoreObject::getReferencedPrivateData() to QoreSocketObject::setCertificate()
   s->setCertificate(cert);
   return 0;
}

// nothing Socket::setCertificate(string cert_pem){}
static AbstractQoreNode* Socket_setCertificate_Vs(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* cert_pem = HARD_QORE_STRING(args, 0);
# 2345 "QC_Socket.qpp"
SimpleRefHolder<QoreSSLCertificate> cert(new QoreSSLCertificate(cert_pem, xsink));
   if (*xsink)
      return 0;

   // QoreSocketObject::setCertificate() takes over ownership of certificate reference
   s->setCertificate(cert.release());
   return 0;
}

// nothing Socket::setCertificate(binary cert_der){}
static AbstractQoreNode* Socket_setCertificate_Vy(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* cert_der = HARD_QORE_BINARY(args, 0);
# 2362 "QC_Socket.qpp"
SimpleRefHolder<QoreSSLCertificate> cert(new QoreSSLCertificate(cert_der, xsink));
   if (*xsink)
      return 0;

   // QoreSocketObject::setCertificate() takes over ownership of certificate reference
   s->setCertificate(cert.release());
   return 0;
}

// nothing Socket::setCharset(string encoding){}
static AbstractQoreNode* Socket_setCharset_Vs(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* encoding = HARD_QORE_STRING(args, 0);
# 2219 "QC_Socket.qpp"
s->setEncoding(QEM.findCreate(encoding));
   return 0;
}

// nothing Socket::setEncoding(string encoding){}
static AbstractQoreNode* Socket_setEncoding_Vs(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* encoding = HARD_QORE_STRING(args, 0);
# 2204 "QC_Socket.qpp"
s->setEncoding(QEM.findCreate(encoding));
   return 0;
}

// nothing Socket::setEventQueue(){}
static AbstractQoreNode* Socket_setEventQueue(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2425 "QC_Socket.qpp"
s->setEventQueue(0, xsink);
   return 0;
}

// nothing Socket::setEventQueue(Queue queue){}
static AbstractQoreNode* Socket_setEventQueue_C5Queue(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   HARD_QORE_OBJ_DATA(queue, Queue, args, 0, CID_QUEUE, "Socket::setEventQueue()", "Queue", xsink);
   if (*xsink)
      return 0;
# 2441 "QC_Socket.qpp"
ReferenceHolder<Queue> q(queue, xsink);
   if (queue->getMax() != -1) {
      xsink->raiseException("QUEUE-ERROR", "the Queue object passed as a maximum size of %d entr%s, which could cause internal I/O to freeze in the Socket object if used; use a Queue object with no maximum size for I/O monitoring", queue->getMax(), queue->getMax() == 1 ? "y" : "ies");
      return 0;
   }
   // pass reference from QoreObject::getReferencedPrivateData() to function
   s->setEventQueue(q.release(), xsink);
   return 0;
}

// int Socket::setNoDelay(bool nd = True){}
static int64 Socket_setNoDelay_Vb(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   bool nd = HARD_QORE_BOOL(args, 0);
# 2470 "QC_Socket.qpp"
return s->setNoDelay(nd);
}

// nothing Socket::setPrivateKey(SSLPrivateKey key){}
static AbstractQoreNode* Socket_setPrivateKey_C13SSLPrivateKey(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   HARD_QORE_OBJ_DATA(key, QoreSSLPrivateKey, args, 0, CID_SSLPRIVATEKEY, "Socket::setPrivateKey()", "SSLPrivateKey", xsink);
   if (*xsink)
      return 0;
# 2379 "QC_Socket.qpp"
// pass reference from QoreObject::getReferencedPrivateData() to QoreSocketObject::setPrivateKey()
   s->setPrivateKey(key);
   return 0;
}

// nothing Socket::setPrivateKey(string key_pem, *string pass){}
static AbstractQoreNode* Socket_setPrivateKey_VsNs(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* key_pem = HARD_QORE_STRING(args, 0);
   const QoreStringNode* pass = reinterpret_cast<const QoreStringNode*>(get_param(args, 1));
# 2393 "QC_Socket.qpp"
SimpleRefHolder<QoreSSLPrivateKey> key(new QoreSSLPrivateKey(key_pem, pass ? pass->getBuffer() : 0, xsink));
   if (*xsink)
      return 0;

   s->setPrivateKey(key.release());
   return 0;
}

// nothing Socket::setPrivateKey(binary key_der){}
static AbstractQoreNode* Socket_setPrivateKey_Vy(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* key_der = HARD_QORE_BINARY(args, 0);
# 2409 "QC_Socket.qpp"
SimpleRefHolder<QoreSSLPrivateKey> key(new QoreSSLPrivateKey(key_der, xsink));
   if (*xsink)
      return 0;

   s->setPrivateKey(key.release());
   return 0;
}

// int Socket::setRecvTimeout(timeout timeout_ms){}
static int64 Socket_setRecvTimeout_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 2166 "QC_Socket.qpp"
return s->setRecvTimeout(timeout_ms);
}

// int Socket::setSendTimeout(timeout timeout_ms){}
static int64 Socket_setSendTimeout_Vt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 timeout_ms = HARD_QORE_INT(args, 0);
# 2149 "QC_Socket.qpp"
return s->setSendTimeout(timeout_ms);
}

// nothing Socket::setWarningQueue(int warning_ms, int warning_bs, Queue queue, any arg, timeout min_ms = 1s){}
static AbstractQoreNode* Socket_setWarningQueue_ViViC5QueueVaVt(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
   int64 warning_ms = HARD_QORE_INT(args, 0);
   int64 warning_bs = HARD_QORE_INT(args, 1);
   HARD_QORE_OBJ_DATA(queue, Queue, args, 2, CID_QUEUE, "Socket::setWarningQueue()", "Queue", xsink);
   if (*xsink)
      return 0;
   const AbstractQoreNode* arg = get_param(args, 3);
   int64 min_ms = HARD_QORE_INT(args, 4);
# 2593 "QC_Socket.qpp"
ReferenceHolder<Queue> q(queue, xsink);
   if (queue->getMax() != -1) {
      xsink->raiseException("QUEUE-ERROR", "the Queue object passed as a maximum size of %d entr%s, which could cause internal I/O to freeze in the Socket object if used; use a Queue object with no maximum size for I/O monitoring", queue->getMax(), queue->getMax() == 1 ? "y" : "ies");
      return 0;
   }
   // pass reference from QoreObject::getReferencedPrivateData() to function
   s->setWarningQueue(xsink, warning_ms, warning_bs, q.release(), arg ? arg->refSelf() : 0, min_ms);
   return 0;
}

// int Socket::shutdown(){}
static int64 Socket_shutdown(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2094 "QC_Socket.qpp"
return s->shutdown();
}

// nothing Socket::shutdownSSL(){}
static AbstractQoreNode* Socket_shutdownSSL(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2108 "QC_Socket.qpp"
s->shutdownSSL(xsink);
   return 0;
}

// nothing Socket::upgradeClientToSSL(){}
static AbstractQoreNode* Socket_upgradeClientToSSL(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2534 "QC_Socket.qpp"
s->upgradeClientToSSL(xsink);
   return 0;
}

// nothing Socket::upgradeServerToSSL(){}
static AbstractQoreNode* Socket_upgradeServerToSSL(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2544 "QC_Socket.qpp"
s->upgradeServerToSSL(xsink);
   return 0;
}

// *string Socket::verifyPeerCertificate(){}
static AbstractQoreNode* Socket_verifyPeerCertificate(QoreObject* self, QoreSocketObject* s, const QoreListNode* args, ExceptionSink* xsink) {
# 2319 "QC_Socket.qpp"
const char *c = getSSLCVCode(s->verifyPeerCertificate());
   return c ? new QoreStringNode(c) : 0;
}

DLLLOCAL QoreClass* initSocketClass(QoreNamespace& ns) {
   QC_SOCKET = new QoreClass("Socket", QDOM_NETWORK);
   CID_SOCKET = QC_SOCKET->getID();

   // public members
   QC_SOCKET->addPublicMember("source", stringOrNothingTypeInfo);
   QC_SOCKET->addPublicMember("source_host", stringOrNothingTypeInfo);

   QC_SOCKET->unsetPublicMemberFlag();

   // Socket Socket::accept(){}
   QC_SOCKET->addMethodExtended3("accept", (q_method_t)Socket_accept, false, QC_NO_FLAGS, QDOM_DEFAULT, QC_SOCKET->getTypeInfo());

   // *Socket Socket::accept(timeout timeout_ms){}
   QC_SOCKET->addMethodExtended3("accept", (q_method_t)Socket_accept_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, QC_SOCKET->getOrNothingTypeInfo(), 1, timeoutTypeInfo, NULL, "timeout_ms");

   // Socket Socket::acceptSSL(){}
   QC_SOCKET->addMethodExtended3("acceptSSL", (q_method_t)Socket_acceptSSL, false, QC_NO_FLAGS, QDOM_DEFAULT, QC_SOCKET->getTypeInfo());

   // *Socket Socket::acceptSSL(timeout timeout_ms){}
   QC_SOCKET->addMethodExtended3("acceptSSL", (q_method_t)Socket_acceptSSL_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, QC_SOCKET->getOrNothingTypeInfo(), 1, timeoutTypeInfo, NULL, "timeout_ms");

   // int Socket::bind(string str, softbool reuseaddr = False){}
   QC_SOCKET->addMethodExtended3("bind", (q_method_int64_t)Socket_bind_Vsvb, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, stringTypeInfo, NULL, "str", softBoolTypeInfo, &False, "reuseaddr");

   // int Socket::bind(int port, softbool reuseaddr = False){}
   QC_SOCKET->addMethodExtended3("bind", (q_method_int64_t)Socket_bind_Vivb, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, bigIntTypeInfo, NULL, "port", softBoolTypeInfo, &False, "reuseaddr");

   // nothing Socket::bindINET(*string iface, *softstring service, softbool reuseaddr = False, softint family = AF_UNSPEC, softint socktype = SOCK_STREAM, softint protocol = 0){}
   QC_SOCKET->addMethodExtended3("bindINET", (q_method_t)Socket_bindINET_Nsnsvbvivivi, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 6, stringOrNothingTypeInfo, NULL, "iface", softStringOrNothingTypeInfo, NULL, "service", softBoolTypeInfo, &False, "reuseaddr", softBigIntTypeInfo, new QoreBigIntNode(AF_UNSPEC), "family", softBigIntTypeInfo, new QoreBigIntNode(SOCK_STREAM), "socktype", softBigIntTypeInfo, zero(), "protocol");

   // nothing Socket::bindUNIX(string path, softint socktype = SOCK_STREAM, softint protocol = 0){}
   QC_SOCKET->addMethodExtended3("bindUNIX", (q_method_t)Socket_bindUNIX_Vsvivi, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 3, stringTypeInfo, NULL, "path", softBigIntTypeInfo, new QoreBigIntNode(SOCK_STREAM), "socktype", softBigIntTypeInfo, zero(), "protocol");

   // nothing Socket::clearStats(){}
   QC_SOCKET->addMethodExtended3("clearStats", (q_method_t)Socket_clearStats, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // nothing Socket::clearWarningQueue(){}
   QC_SOCKET->addMethodExtended3("clearWarningQueue", (q_method_t)Socket_clearWarningQueue, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // int Socket::close(){}
   QC_SOCKET->addMethodExtended3("close", (q_method_int64_t)Socket_close, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo);

   // nothing Socket::connect(string target, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("connect", (q_method_t)Socket_connect_VsVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, NULL, "target", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // nothing Socket::connectINET(string host, softstring service, timeout timeout_ms = -1, softint family = AF_UNSPEC, softint socktype = SOCK_STREAM, softint protocol = 0){}
   QC_SOCKET->addMethodExtended3("connectINET", (q_method_t)Socket_connectINET_VsvsVtvivivi, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 6, stringTypeInfo, NULL, "host", softStringTypeInfo, NULL, "service", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms", softBigIntTypeInfo, new QoreBigIntNode(AF_UNSPEC), "family", softBigIntTypeInfo, new QoreBigIntNode(SOCK_STREAM), "socktype", softBigIntTypeInfo, zero(), "protocol");

   // nothing Socket::connectINETSSL(string host, softstring service, timeout timeout_ms = -1, softint family = AF_UNSPEC, softint socktype = SOCK_STREAM, softint protocol = 0){}
   QC_SOCKET->addMethodExtended3("connectINETSSL", (q_method_t)Socket_connectINETSSL_VsvsVtvivivi, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 6, stringTypeInfo, NULL, "host", softStringTypeInfo, NULL, "service", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms", softBigIntTypeInfo, new QoreBigIntNode(AF_UNSPEC), "family", softBigIntTypeInfo, new QoreBigIntNode(SOCK_STREAM), "socktype", softBigIntTypeInfo, zero(), "protocol");

   // nothing Socket::connectSSL(string target, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("connectSSL", (q_method_t)Socket_connectSSL_VsVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, NULL, "target", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // nothing Socket::connectUNIX(string path, softint socktype = SOCK_STREAM, softint protocol = 0){}
   QC_SOCKET->addMethodExtended3("connectUNIX", (q_method_t)Socket_connectUNIX_Vsvivi, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 3, stringTypeInfo, NULL, "path", softBigIntTypeInfo, new QoreBigIntNode(SOCK_STREAM), "socktype", softBigIntTypeInfo, zero(), "protocol");

   // nothing Socket::connectUNIXSSL(string path, softint socktype = SOCK_STREAM, softint protocol = 0){}
   QC_SOCKET->addMethodExtended3("connectUNIXSSL", (q_method_t)Socket_connectUNIXSSL_Vsvivi, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 3, stringTypeInfo, NULL, "path", softBigIntTypeInfo, new QoreBigIntNode(SOCK_STREAM), "socktype", softBigIntTypeInfo, zero(), "protocol");

   // Socket::constructor() {}
   QC_SOCKET->setConstructorExtended3(Socket_constructor, false, QC_NO_FLAGS, QDOM_DEFAULT);

   // Socket::copy() {}
   QC_SOCKET->setCopy((q_copy_t)Socket_copy);

   // string Socket::getCharset(){}
   QC_SOCKET->addMethodExtended3("getCharset", (q_method_t)Socket_getCharset, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

   // string Socket::getEncoding(){}
   QC_SOCKET->addMethodExtended3("getEncoding", (q_method_t)Socket_getEncoding, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

   // bool Socket::getNoDelay(){}
   QC_SOCKET->addMethodExtended3("getNoDelay", (q_method_bool_t)Socket_getNoDelay, false, QC_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

   // hash Socket::getPeerInfo(bool host_lookup = True){}
   QC_SOCKET->addMethodExtended3("getPeerInfo", (q_method_t)Socket_getPeerInfo_Vb, false, QC_NO_FLAGS, QDOM_DEFAULT, hashTypeInfo, 1, boolTypeInfo, &True, "host_lookup");

   // int Socket::getPort(){}
   QC_SOCKET->addMethodExtended3("getPort", (q_method_int64_t)Socket_getPort, false, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

   // int Socket::getRecvTimeout(){}
   QC_SOCKET->addMethodExtended3("getRecvTimeout", (q_method_int64_t)Socket_getRecvTimeout, false, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

   // *string Socket::getSSLCipherName(){}
   QC_SOCKET->addMethodExtended3("getSSLCipherName", (q_method_t)Socket_getSSLCipherName, false, QC_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

   // *string Socket::getSSLCipherVersion(){}
   QC_SOCKET->addMethodExtended3("getSSLCipherVersion", (q_method_t)Socket_getSSLCipherVersion, false, QC_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

   // int Socket::getSendTimeout(){}
   QC_SOCKET->addMethodExtended3("getSendTimeout", (q_method_int64_t)Socket_getSendTimeout, false, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

   // int Socket::getSocket(){}
   QC_SOCKET->addMethodExtended3("getSocket", (q_method_int64_t)Socket_getSocket, false, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

   // hash Socket::getSocketInfo(bool host_lookup = True){}
   QC_SOCKET->addMethodExtended3("getSocketInfo", (q_method_t)Socket_getSocketInfo_Vb, false, QC_NO_FLAGS, QDOM_DEFAULT, hashTypeInfo, 1, boolTypeInfo, &True, "host_lookup");

   // hash Socket::getUsageInfo(){}
   QC_SOCKET->addMethodExtended3("getUsageInfo", (q_method_t)Socket_getUsageInfo, false, QC_CONSTANT, QDOM_DEFAULT, hashTypeInfo);

   // bool Socket::isDataAvailable(timeout timeout_ms = 0){}
   QC_SOCKET->addMethodExtended3("isDataAvailable", (q_method_bool_t)Socket_isDataAvailable_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, boolTypeInfo, 1, timeoutTypeInfo, zero(), "timeout_ms");

   // bool Socket::isOpen(){}
   QC_SOCKET->addMethodExtended3("isOpen", (q_method_bool_t)Socket_isOpen, false, QC_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

   // bool Socket::isSecure(){}
   QC_SOCKET->addMethodExtended3("isSecure", (q_method_bool_t)Socket_isSecure, false, QC_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

   // bool Socket::isWriteFinished(timeout timeout_ms = 0){}
   QC_SOCKET->addMethodExtended3("isWriteFinished", (q_method_bool_t)Socket_isWriteFinished_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, boolTypeInfo, 1, timeoutTypeInfo, zero(), "timeout_ms");

   // int Socket::listen(int backlog = 20){}
   QC_SOCKET->addMethodExtended3("listen", (q_method_int64_t)Socket_listen_Vi, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, bigIntTypeInfo, new QoreBigIntNode(20), "backlog");

   // bool Socket::pendingHttpChunkedBody(){}
   QC_SOCKET->addMethodExtended3("pendingHttpChunkedBody", (q_method_bool_t)Socket_pendingHttpChunkedBody, false, QC_NO_FLAGS, QDOM_DEFAULT, boolTypeInfo);

   // hash Socket::readHTTPChunkedBody(timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("readHTTPChunkedBody", (q_method_t)Socket_readHTTPChunkedBody_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, hashTypeInfo, 1, timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // hash Socket::readHTTPChunkedBodyBinary(timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("readHTTPChunkedBodyBinary", (q_method_t)Socket_readHTTPChunkedBodyBinary_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, hashTypeInfo, 1, timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // nothing Socket::readHTTPChunkedBodyBinaryWithCallback(code rcb, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("readHTTPChunkedBodyBinaryWithCallback", (q_method_t)Socket_readHTTPChunkedBodyBinaryWithCallback_VcVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, codeTypeInfo, NULL, "rcb", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // nothing Socket::readHTTPChunkedBodyWithCallback(code rcb, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("readHTTPChunkedBodyWithCallback", (q_method_t)Socket_readHTTPChunkedBodyWithCallback_VcVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, codeTypeInfo, NULL, "rcb", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // hash Socket::readHTTPHeader(timeout timeout_ms = -1, *reference info){}
   QC_SOCKET->addMethodExtended3("readHTTPHeader", (q_method_t)Socket_readHTTPHeader_VtNr, false, QC_NO_FLAGS, QDOM_DEFAULT, hashTypeInfo, 2, timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms", anyTypeInfo, NULL, "info");

   // string Socket::readHTTPHeaderString(timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("readHTTPHeaderString", (q_method_t)Socket_readHTTPHeaderString_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 1, timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // string Socket::recv(softint size = 0, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("recv", (q_method_t)Socket_recv_viVt, false, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, softBigIntTypeInfo, zero(), "size", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // binary Socket::recvBinary(softint size = 0, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("recvBinary", (q_method_t)Socket_recvBinary_viVt, false, QC_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo, 2, softBigIntTypeInfo, zero(), "size", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::recvi1(timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("recvi1", (q_method_int64_t)Socket_recvi1_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::recvi2(timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("recvi2", (q_method_int64_t)Socket_recvi2_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::recvi2LSB(timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("recvi2LSB", (q_method_int64_t)Socket_recvi2LSB_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::recvi4(timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("recvi4", (q_method_int64_t)Socket_recvi4_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::recvi4LSB(timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("recvi4LSB", (q_method_int64_t)Socket_recvi4LSB_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::recvi8(timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("recvi8", (q_method_int64_t)Socket_recvi8_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::recvi8LSB(timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("recvi8LSB", (q_method_int64_t)Socket_recvi8LSB_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::recvu1(timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("recvu1", (q_method_int64_t)Socket_recvu1_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::recvu2(timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("recvu2", (q_method_int64_t)Socket_recvu2_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::recvu2LSB(timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("recvu2LSB", (q_method_int64_t)Socket_recvu2LSB_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::recvu4(timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("recvu4", (q_method_int64_t)Socket_recvu4_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::recvu4LSB(timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("recvu4LSB", (q_method_int64_t)Socket_recvu4LSB_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::send(binary bin, int timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("send", (q_method_int64_t)Socket_send_VyVi, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, binaryTypeInfo, NULL, "bin", bigIntTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::send(string str, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("send", (q_method_int64_t)Socket_send_VsVt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, stringTypeInfo, NULL, "str", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // nothing Socket::send2(binary bin, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("send2", (q_method_t)Socket_send2_VyVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, binaryTypeInfo, NULL, "bin", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // nothing Socket::send2(string str, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("send2", (q_method_t)Socket_send2_VsVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, NULL, "str", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::sendBinary(string str, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendBinary", (q_method_int64_t)Socket_sendBinary_VsVt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, stringTypeInfo, NULL, "str", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::sendBinary(binary bin, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendBinary", (q_method_int64_t)Socket_sendBinary_VyVt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, binaryTypeInfo, NULL, "bin", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // nothing Socket::sendBinary2(string str, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendBinary2", (q_method_t)Socket_sendBinary2_VsVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, NULL, "str", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // nothing Socket::sendBinary2(binary bin, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendBinary2", (q_method_t)Socket_sendBinary2_VyVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, binaryTypeInfo, NULL, "bin", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // nothing Socket::sendHTTPMessage(string method, string path, string http_version, hash headers, *string body, *reference info, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendHTTPMessage", (q_method_t)Socket_sendHTTPMessage_VsVsVsVhNsNrVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 7, stringTypeInfo, NULL, "method", stringTypeInfo, NULL, "path", stringTypeInfo, NULL, "http_version", hashTypeInfo, NULL, "headers", stringOrNothingTypeInfo, NULL, "body", anyTypeInfo, NULL, "info", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // nothing Socket::sendHTTPMessage(string method, string path, string http_version, hash headers, binary body, *reference info, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendHTTPMessage", (q_method_t)Socket_sendHTTPMessage_VsVsVsVhVyNrVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 7, stringTypeInfo, NULL, "method", stringTypeInfo, NULL, "path", stringTypeInfo, NULL, "http_version", hashTypeInfo, NULL, "headers", binaryTypeInfo, NULL, "body", anyTypeInfo, NULL, "info", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // nothing Socket::sendHTTPMessageWithCallback(code scb, string method, string path, string http_version, hash headers, *reference info, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendHTTPMessageWithCallback", (q_method_t)Socket_sendHTTPMessageWithCallback_VcVsVsVsVhNrVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 7, codeTypeInfo, NULL, "scb", stringTypeInfo, NULL, "method", stringTypeInfo, NULL, "path", stringTypeInfo, NULL, "http_version", hashTypeInfo, NULL, "headers", anyTypeInfo, NULL, "info", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // nothing Socket::sendHTTPResponse(softint status_code, string status_desc, string http_version, hash headers, *string body, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendHTTPResponse", (q_method_t)Socket_sendHTTPResponse_viVsVsVhNsVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 6, softBigIntTypeInfo, NULL, "status_code", stringTypeInfo, NULL, "status_desc", stringTypeInfo, NULL, "http_version", hashTypeInfo, NULL, "headers", stringOrNothingTypeInfo, NULL, "body", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // nothing Socket::sendHTTPResponse(softint status_code, string status_desc, string http_version, hash headers, binary body, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendHTTPResponse", (q_method_t)Socket_sendHTTPResponse_viVsVsVhVyVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 6, softBigIntTypeInfo, NULL, "status_code", stringTypeInfo, NULL, "status_desc", stringTypeInfo, NULL, "http_version", hashTypeInfo, NULL, "headers", binaryTypeInfo, NULL, "body", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // nothing Socket::sendHTTPResponseWithCallback(code scb, softint status_code, string status_desc, string http_version, hash headers, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendHTTPResponseWithCallback", (q_method_t)Socket_sendHTTPResponseWithCallback_VcviVsVsVhVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 6, codeTypeInfo, NULL, "scb", softBigIntTypeInfo, NULL, "status_code", stringTypeInfo, NULL, "status_desc", stringTypeInfo, NULL, "http_version", hashTypeInfo, NULL, "headers", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::sendi1(softint i = 0, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendi1", (q_method_int64_t)Socket_sendi1_viVt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, softBigIntTypeInfo, zero(), "i", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::sendi2(softint i = 0, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendi2", (q_method_int64_t)Socket_sendi2_viVt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, softBigIntTypeInfo, zero(), "i", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::sendi2LSB(softint i = 0, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendi2LSB", (q_method_int64_t)Socket_sendi2LSB_viVt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, softBigIntTypeInfo, zero(), "i", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::sendi4(softint i = 0, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendi4", (q_method_int64_t)Socket_sendi4_viVt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, softBigIntTypeInfo, zero(), "i", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::sendi4LSB(softint i = 0, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendi4LSB", (q_method_int64_t)Socket_sendi4LSB_viVt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, softBigIntTypeInfo, zero(), "i", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::sendi8(softint i = 0, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendi8", (q_method_int64_t)Socket_sendi8_viVt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, softBigIntTypeInfo, zero(), "i", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // int Socket::sendi8LSB(softint i = 0, timeout timeout_ms = -1){}
   QC_SOCKET->addMethodExtended3("sendi8LSB", (q_method_int64_t)Socket_sendi8LSB_viVt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, softBigIntTypeInfo, zero(), "i", timeoutTypeInfo, new QoreBigIntNode(-1), "timeout_ms");

   // nothing Socket::setCertificate(SSLCertificate cert){}
   QC_SOCKET->addMethodExtended3("setCertificate", (q_method_t)Socket_setCertificate_C14SSLCertificate, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, QC_SSLCERTIFICATE->getTypeInfo(), NULL, "cert");

   // nothing Socket::setCertificate(string cert_pem){}
   QC_SOCKET->addMethodExtended3("setCertificate", (q_method_t)Socket_setCertificate_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, NULL, "cert_pem");

   // nothing Socket::setCertificate(binary cert_der){}
   QC_SOCKET->addMethodExtended3("setCertificate", (q_method_t)Socket_setCertificate_Vy, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, binaryTypeInfo, NULL, "cert_der");

   // nothing Socket::setCharset(string encoding){}
   QC_SOCKET->addMethodExtended3("setCharset", (q_method_t)Socket_setCharset_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, NULL, "encoding");

   // nothing Socket::setEncoding(string encoding){}
   QC_SOCKET->addMethodExtended3("setEncoding", (q_method_t)Socket_setEncoding_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, NULL, "encoding");

   // nothing Socket::setEventQueue(){}
   QC_SOCKET->addMethodExtended3("setEventQueue", (q_method_t)Socket_setEventQueue, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // nothing Socket::setEventQueue(Queue queue){}
   QC_SOCKET->addMethodExtended3("setEventQueue", (q_method_t)Socket_setEventQueue_C5Queue, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, QC_QUEUE->getTypeInfo(), NULL, "queue");

   // int Socket::setNoDelay(bool nd = True){}
   QC_SOCKET->addMethodExtended3("setNoDelay", (q_method_int64_t)Socket_setNoDelay_Vb, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, boolTypeInfo, &True, "nd");

   // nothing Socket::setPrivateKey(SSLPrivateKey key){}
   QC_SOCKET->addMethodExtended3("setPrivateKey", (q_method_t)Socket_setPrivateKey_C13SSLPrivateKey, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, QC_SSLPRIVATEKEY->getTypeInfo(), NULL, "key");

   // nothing Socket::setPrivateKey(string key_pem, *string pass){}
   QC_SOCKET->addMethodExtended3("setPrivateKey", (q_method_t)Socket_setPrivateKey_VsNs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, NULL, "key_pem", stringOrNothingTypeInfo, NULL, "pass");

   // nothing Socket::setPrivateKey(binary key_der){}
   QC_SOCKET->addMethodExtended3("setPrivateKey", (q_method_t)Socket_setPrivateKey_Vy, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, binaryTypeInfo, NULL, "key_der");

   // int Socket::setRecvTimeout(timeout timeout_ms){}
   QC_SOCKET->addMethodExtended3("setRecvTimeout", (q_method_int64_t)Socket_setRecvTimeout_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, timeoutTypeInfo, NULL, "timeout_ms");

   // int Socket::setSendTimeout(timeout timeout_ms){}
   QC_SOCKET->addMethodExtended3("setSendTimeout", (q_method_int64_t)Socket_setSendTimeout_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, timeoutTypeInfo, NULL, "timeout_ms");

   // nothing Socket::setWarningQueue(int warning_ms, int warning_bs, Queue queue, any arg, timeout min_ms = 1s){}
   QC_SOCKET->addMethodExtended3("setWarningQueue", (q_method_t)Socket_setWarningQueue_ViViC5QueueVaVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 5, bigIntTypeInfo, NULL, "warning_ms", bigIntTypeInfo, NULL, "warning_bs", QC_QUEUE->getTypeInfo(), NULL, "queue", anyTypeInfo, NULL, "arg", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 1, 0), "min_ms");

   // int Socket::shutdown(){}
   QC_SOCKET->addMethodExtended3("shutdown", (q_method_int64_t)Socket_shutdown, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo);

   // nothing Socket::shutdownSSL(){}
   QC_SOCKET->addMethodExtended3("shutdownSSL", (q_method_t)Socket_shutdownSSL, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // nothing Socket::upgradeClientToSSL(){}
   QC_SOCKET->addMethodExtended3("upgradeClientToSSL", (q_method_t)Socket_upgradeClientToSSL, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // nothing Socket::upgradeServerToSSL(){}
   QC_SOCKET->addMethodExtended3("upgradeServerToSSL", (q_method_t)Socket_upgradeServerToSSL, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // *string Socket::verifyPeerCertificate(){}
   QC_SOCKET->addMethodExtended3("verifyPeerCertificate", (q_method_t)Socket_verifyPeerCertificate, false, QC_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);
# 62 "QC_Socket.qpp"
# 72 "QC_Socket.qpp"
# 75 "QC_Socket.qpp"
# 78 "QC_Socket.qpp"
# 81 "QC_Socket.qpp"
# 84 "QC_Socket.qpp"
# 87 "QC_Socket.qpp"
# 90 "QC_Socket.qpp"
# 93 "QC_Socket.qpp"
# 96 "QC_Socket.qpp"
# 99 "QC_Socket.qpp"
# 102 "QC_Socket.qpp"
# 105 "QC_Socket.qpp"
# 108 "QC_Socket.qpp"
# 111 "QC_Socket.qpp"
# 114 "QC_Socket.qpp"
# 117 "QC_Socket.qpp"
# 120 "QC_Socket.qpp"
# 123 "QC_Socket.qpp"
# 126 "QC_Socket.qpp"
# 129 "QC_Socket.qpp"
# 132 "QC_Socket.qpp"
# 135 "QC_Socket.qpp"
# 138 "QC_Socket.qpp"
# 141 "QC_Socket.qpp"
# 144 "QC_Socket.qpp"
# 147 "QC_Socket.qpp"
# 150 "QC_Socket.qpp"
# 153 "QC_Socket.qpp"
# 156 "QC_Socket.qpp"
# 159 "QC_Socket.qpp"
# 162 "QC_Socket.qpp"
# 165 "QC_Socket.qpp"
# 168 "QC_Socket.qpp"
   ns.addConstant("X509_V_ERR_AKID_ISSUER_SERIAL_MISMATCH", new QoreStringNode("X509_V_ERR_AKID_ISSUER_SERIAL_MISMATCH"));
   ns.addConstant("X509_V_ERR_AKID_SKID_MISMATCH", new QoreStringNode("X509_V_ERR_AKID_SKID_MISMATCH"));
   ns.addConstant("X509_V_ERR_APPLICATION_VERIFICATION", new QoreStringNode("X509_V_ERR_APPLICATION_VERIFICATION"));
   ns.addConstant("X509_V_ERR_CERT_CHAIN_TOO_LONG", new QoreStringNode("X509_V_ERR_CERT_CHAIN_TOO_LONG"));
   ns.addConstant("X509_V_ERR_CERT_HAS_EXPIRED", new QoreStringNode("X509_V_ERR_CERT_HAS_EXPIRED"));
   ns.addConstant("X509_V_ERR_CERT_NOT_YET_VALID", new QoreStringNode("X509_V_ERR_CERT_NOT_YET_VALID"));
   ns.addConstant("X509_V_ERR_CERT_REJECTED", new QoreStringNode("X509_V_ERR_CERT_REJECTED"));
   ns.addConstant("X509_V_ERR_CERT_REVOKED", new QoreStringNode("X509_V_ERR_CERT_REVOKED"));
   ns.addConstant("X509_V_ERR_CERT_SIGNATURE_FAILURE", new QoreStringNode("X509_V_ERR_CERT_SIGNATURE_FAILURE"));
   ns.addConstant("X509_V_ERR_CERT_UNTRUSTED", new QoreStringNode("X509_V_ERR_CERT_UNTRUSTED"));
   ns.addConstant("X509_V_ERR_CRL_HAS_EXPIRED", new QoreStringNode("X509_V_ERR_CRL_HAS_EXPIRED"));
   ns.addConstant("X509_V_ERR_CRL_NOT_YET_VALID", new QoreStringNode("X509_V_ERR_CRL_NOT_YET_VALID"));
   ns.addConstant("X509_V_ERR_CRL_SIGNATURE_FAILURE", new QoreStringNode("X509_V_ERR_CRL_SIGNATURE_FAILURE"));
   ns.addConstant("X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT", new QoreStringNode("X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT"));
   ns.addConstant("X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD", new QoreStringNode("X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD"));
   ns.addConstant("X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD", new QoreStringNode("X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD"));
   ns.addConstant("X509_V_ERR_ERROR_IN_CRL_LAST_UPDATE_FIELD", new QoreStringNode("X509_V_ERR_ERROR_IN_CRL_LAST_UPDATE_FIELD"));
   ns.addConstant("X509_V_ERR_ERROR_IN_CRL_NEXT_UPDATE_FIELD", new QoreStringNode("X509_V_ERR_ERROR_IN_CRL_NEXT_UPDATE_FIELD"));
   ns.addConstant("X509_V_ERR_INVALID_CA", new QoreStringNode("X509_V_ERR_INVALID_CA"));
   ns.addConstant("X509_V_ERR_INVALID_PURPOSE", new QoreStringNode("X509_V_ERR_INVALID_PURPOSE"));
   ns.addConstant("X509_V_ERR_KEYUSAGE_NO_CERTSIGN", new QoreStringNode("X509_V_ERR_KEYUSAGE_NO_CERTSIGN"));
   ns.addConstant("X509_V_ERR_OUT_OF_MEM", new QoreStringNode("X509_V_ERR_OUT_OF_MEM"));
   ns.addConstant("X509_V_ERR_PATH_LENGTH_EXCEEDED", new QoreStringNode("X509_V_ERR_PATH_LENGTH_EXCEEDED"));
   ns.addConstant("X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN", new QoreStringNode("X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN"));
   ns.addConstant("X509_V_ERR_SUBJECT_ISSUER_MISMATCH", new QoreStringNode("X509_V_ERR_SUBJECT_ISSUER_MISMATCH"));
   ns.addConstant("X509_V_ERR_UNABLE_TO_DECODE_ISSUER_PUBLIC_KEY", new QoreStringNode("X509_V_ERR_UNABLE_TO_DECODE_ISSUER_PUBLIC_KEY"));
   ns.addConstant("X509_V_ERR_UNABLE_TO_DECRYPT_CERT_SIGNATURE", new QoreStringNode("X509_V_ERR_UNABLE_TO_DECRYPT_CERT_SIGNATURE"));
   ns.addConstant("X509_V_ERR_UNABLE_TO_DECRYPT_CRL_SIGNATURE", new QoreStringNode("X509_V_ERR_UNABLE_TO_DECRYPT_CRL_SIGNATURE"));
   ns.addConstant("X509_V_ERR_UNABLE_TO_GET_CRL", new QoreStringNode("X509_V_ERR_UNABLE_TO_GET_CRL"));
   ns.addConstant("X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT", new QoreStringNode("X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT"));
   ns.addConstant("X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY", new QoreStringNode("X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY"));
   ns.addConstant("X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE", new QoreStringNode("X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE"));
   ns.addConstant("X509_V_OK", new QoreStringNode("X509_V_OK"));
   QoreHashNode* hash_X509_VerificationReasons = new QoreHashNode;
   hash_X509_VerificationReasons->setKeyValue("X509_V_OK", new QoreStringNode("OK"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT", new QoreStringNode("Unable to get issuer certificate"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_UNABLE_TO_GET_CRL", new QoreStringNode("Unable to get certificate CRL"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_UNABLE_TO_DECRYPT_CERT_SIGNATURE", new QoreStringNode("Unable to decrypt certificate's signature. This means that the actual signature value could not be determined rather than it not matching the expected value; this is only meaningful for RSA"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_UNABLE_TO_DECRYPT_CRL_SIGNATURE", new QoreStringNode("Unable to decrypt CRL's signature"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_UNABLE_TO_DECODE_ISSUER_PUBLIC_KEY", new QoreStringNode("Unable to decode issuer public key (SubjectPublicKeyInfo)"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_CERT_SIGNATURE_FAILURE", new QoreStringNode("Certificate signature failure; the signature of the certificate is invalid"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_CRL_SIGNATURE_FAILURE", new QoreStringNode("CRL signature failure; the signature of the certificate is invalid"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_CERT_NOT_YET_VALID", new QoreStringNode("Certificate is not yet valid"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_CERT_HAS_EXPIRED", new QoreStringNode("Certificate has expired"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_CRL_NOT_YET_VALID", new QoreStringNode("CRL is not yet valid"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_CRL_HAS_EXPIRED", new QoreStringNode("CRL has expired"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD", new QoreStringNode("Format error in certificate's notBefore field (invalid time)"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD", new QoreStringNode("Format error in certificate's notAfter field (invalid time)"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_ERROR_IN_CRL_LAST_UPDATE_FIELD", new QoreStringNode("Format error in CRL's lastUpdate field (invalid time)"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_ERROR_IN_CRL_NEXT_UPDATE_FIELD", new QoreStringNode("Format error in CRL's nextUpdate field (invalid time)"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_OUT_OF_MEM", new QoreStringNode("Out of memory error"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT", new QoreStringNode("Certificate is self-signed and cannot be found in the trusted list"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN", new QoreStringNode("Self signed certificate in certificate chain"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY", new QoreStringNode("Unable to get local issuer certificate. This normally means the list of trusted certificates is not complete"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE", new QoreStringNode("Unable to verify the first certificate"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_CERT_CHAIN_TOO_LONG", new QoreStringNode("Certificate chain too long"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_CERT_REVOKED", new QoreStringNode("Certificate has been revoked"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_INVALID_CA", new QoreStringNode("Invalid CA certificate"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_PATH_LENGTH_EXCEEDED", new QoreStringNode("The basicConstraints pathlength parameter has been exceeded"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_INVALID_PURPOSE", new QoreStringNode("The certificate cannot be used for the specified purpose"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_CERT_UNTRUSTED", new QoreStringNode("Root CA is not marked as trusted for the specified purpose"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_CERT_REJECTED", new QoreStringNode("Root CA is marked to reject the specified purpose"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_SUBJECT_ISSUER_MISMATCH", new QoreStringNode("The current candidate issuer certificate was rejected because its subject name did not match the issuer name of the current certificate"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_AKID_SKID_MISMATCH", new QoreStringNode("The current candidate issuer certificate was rejected because its subject key identifier was present and did not match the authority key identifier of the current certificate"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_AKID_ISSUER_SERIAL_MISMATCH", new QoreStringNode("Issuer name and serial number of candidate certificate do not match the authority key identifier of the current certificate"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_KEYUSAGE_NO_CERTSIGN", new QoreStringNode("The keyUsage extension does not permit certificate signing"), 0);
   hash_X509_VerificationReasons->setKeyValue("X509_V_ERR_APPLICATION_VERIFICATION", new QoreStringNode("Verification failure"), 0);

   ns.addConstant("X509_VerificationReasons", hash_X509_VerificationReasons);
# 206 "QC_Socket.qpp"
# 213 "QC_Socket.qpp"
# 216 "QC_Socket.qpp"
#ifndef AF_LOCAL
#define AF_LOCAL AF_UNIX
#endif
# 223 "QC_Socket.qpp"
# 226 "QC_Socket.qpp"
# 229 "QC_Socket.qpp"
# 232 "QC_Socket.qpp"
   ns.addConstant("AFMap", (get_network_address_family_map()));
   QoreHashNode* hash_AFStrMap = new QoreHashNode;
   hash_AFStrMap->setKeyValue("AF_INET", new QoreBigIntNode(AF_INET), 0);
   hash_AFStrMap->setKeyValue("AF_INET6", new QoreBigIntNode(AF_INET6), 0);
   hash_AFStrMap->setKeyValue("AF_UNIX", new QoreBigIntNode(AF_UNIX), 0);
   hash_AFStrMap->setKeyValue("AF_LOCAL", new QoreBigIntNode(AF_UNIX), 0);
   hash_AFStrMap->setKeyValue("AF_UNSPEC", new QoreBigIntNode(AF_UNSPEC), 0);

   ns.addConstant("AFStrMap", hash_AFStrMap);
   ns.addConstant("AF_INET", new QoreBigIntNode(AF_INET));
   ns.addConstant("AF_INET6", new QoreBigIntNode(AF_INET6));
   ns.addConstant("AF_LOCAL", new QoreBigIntNode(AF_LOCAL));
   ns.addConstant("AF_UNIX", new QoreBigIntNode(AF_UNIX));
   ns.addConstant("AF_UNSPEC", new QoreBigIntNode(AF_UNSPEC));
# 243 "QC_Socket.qpp"
#ifndef AI_ADDRCONFIG
#define AI_ADDRCONFIG 0
#endif
#ifndef AI_ALL
#define AI_ALL 0
#endif
#ifndef AI_CANONNAME
#define AI_CANONNAME 0
#endif
#ifndef AI_NUMERICHOST
#define AI_NUMERICHOST 0
#endif
#ifndef AI_NUMERICSERV
#define AI_NUMERICSERV 0
#endif
#ifndef AI_PASSIVE
#define AI_PASSIVE 0
#endif
#ifndef AI_V4MAPPED
#define AI_V4MAPPED 0
#endif
# 270 "QC_Socket.qpp"
# 273 "QC_Socket.qpp"
# 276 "QC_Socket.qpp"
# 279 "QC_Socket.qpp"
# 282 "QC_Socket.qpp"
# 288 "QC_Socket.qpp"
   ns.addConstant("AI_ADDRCONFIG", new QoreBigIntNode(AI_ADDRCONFIG));
   ns.addConstant("AI_ALL", new QoreBigIntNode(AI_ALL));
   ns.addConstant("AI_CANONNAME", new QoreBigIntNode(AI_CANONNAME));
   ns.addConstant("AI_NUMERICHOST", new QoreBigIntNode(AI_NUMERICHOST));
   ns.addConstant("AI_NUMERICSERV", new QoreBigIntNode(AI_NUMERICSERV));
   ns.addConstant("AI_PASSIVE", new QoreBigIntNode(AI_PASSIVE));
   ns.addConstant("AI_V4MAPPED", new QoreBigIntNode(AI_V4MAPPED));
# 303 "QC_Socket.qpp"
   ns.addConstant("IPPROTO_TCP", new QoreBigIntNode(IPPROTO_TCP));
   ns.addConstant("IPPROTO_UDP", new QoreBigIntNode(IPPROTO_UDP));
# 314 "QC_Socket.qpp"
# 321 "QC_Socket.qpp"
   ns.addConstant("SOCK_DGRAM", new QoreBigIntNode(SOCK_DGRAM));
   ns.addConstant("SOCK_RAW", new QoreBigIntNode(SOCK_RAW));
   ns.addConstant("SOCK_STREAM", new QoreBigIntNode(SOCK_STREAM));

   return QC_SOCKET;
}
