/* -*- mode: c++; indent-tabs-mode: nil -*- */
/** @file QC_ThreadPool.qpp ThreadPool class definition */
/*
  Qore Programming Language

  Copyright (C) 2003 - 2014 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>
#include <qore/intern/ThreadPool.h>

static void tpt_start_thread(ExceptionSink* xsink, ThreadPoolThread* tpt) {
   tpt->worker(xsink);
}

ThreadPoolThread::ThreadPoolThread(ThreadPool& n_tp, ExceptionSink* xsink) : tp(n_tp), task(0), stopCond(0), stopflag(false), stopped(false) {
   id = q_start_thread(xsink, (q_thread_t)tpt_start_thread, this);
   if (id > 0)
      tp.ref();
}

void ThreadPoolThread::worker(ExceptionSink* xsink) {
   SafeLocker sl(m);
      
   while (!stopflag || task) {
      if (!task) {
         //printd(5, "ThreadPoolThread::worker() id %d about to wait stopflag: %d task: %p\n", id, stopflag, task);
         c.wait(m);
         if (stopflag && !task)
            break;
      }
      
      assert(task);
      
      sl.unlock();
      discard(task->run(xsink), xsink);
      sl.lock();
      task->del(xsink);
      task = 0;
      
      if (stopflag || tp.done(this))
         break;
   }

   //printd(5, "ThreadPoolThread::worker() stopping id %d: %s\n", id, stopCond ? "wait" : "after detach");

   if (stopCond) {
      stopped = true;
      stopCond->signal();
   }
   else {
      sl.unlock();
      finalize(xsink);
   }
}

void ThreadPoolThread::finalize(ExceptionSink* xsink) {
   tp.deref(xsink);
   delete this;
}

static void tp_start_thread(ExceptionSink* xsink, ThreadPool* tp) {
   tp->worker(xsink);
}

ThreadPool::ThreadPool(ExceptionSink* xsink, int n_max, int n_minidle, int n_maxidle, int n_release_ms) : 
   max(n_max), minidle(n_minidle), maxidle(n_maxidle), release_ms(n_release_ms), quit(false), waiting(false), stopflag(false), stopped(false), confirm(false) {
   if (max < 0)
      max = 0;
   if (minidle < 0)
      minidle = 0;
   if (maxidle <= 0)
      maxidle = minidle;
   if (q_start_thread(xsink, (q_thread_t)tp_start_thread, this) == -1) {
      assert(*xsink);
      stopped = true;
   }
}

void ThreadPool::worker(ExceptionSink* xsink) {
   SafeLocker sl(m);

   for (int i = 0; i < minidle; ++i) {
      if (addIdleWorker(xsink)) {
         xsink->handleExceptions();
         break;
      }
   }

   while (!stopflag) {
      if (q.empty()) {
         if (release_ms && (int)fh.size() > minidle) {
            if (cond.wait(m, release_ms) && q.empty()) {
               // timeout occurred: terminate an idle thread
               ThreadPoolThread* tpt = fh.front();
               //printd(5, "ThreadPool::worker() this: %p release_ms: %d timeout - stopping idle thread %p (minidle: %d maxidle: %d fh.size(): %ld)\n", this, release_ms, tpt, minidle, maxidle, fh.size());
               fh.pop_front();
               tpt->stop();
               continue;
            }
         }
         else
            cond.wait(m);
      }

      if (stopflag)
         break;

      while (!q.empty()) {
         ThreadPoolThread* tpt = getThreadUnlocked(xsink);
         if (!tpt) {
            xsink->handleExceptions();
            break;
         }
         tpt->submit(q.front());
         q.pop_front();
      }

      while ((int)fh.size() < minidle && (!max || ((int)fh.size() + (int)ah.size() < max))) {
         if (addIdleWorker(xsink)) {
            xsink->handleExceptions();
            break;
         }
      }
   }

   // idle threads can be terminated with stop() in all cases
   for (tplist_t::iterator i = fh.begin(), e = fh.end(); i != e; ++i)
      (*i)->stop();

   sl.unlock();

   // terminate all worker threads
   if (confirm) {
      for (tplist_t::iterator i = ah.begin(), e = ah.end(); i != e; ++i)
         (*i)->stopWait();

      for (tplist_t::iterator i = ah.begin(), e = ah.end(); i != e; ++i)
         (*i)->stopConfirm(xsink);

   }
   else {
      for (tplist_t::iterator i = ah.begin(), e = ah.end(); i != e; ++i)
         (*i)->stop();
   }

   sl.lock();

#ifdef DEBUG
   ah.clear();
   fh.clear();
#endif

   stopped = true;
   stopCond.broadcast();

   sl.unlock();

   // cancel tasks if any
   for (taskq_t::iterator i = q.begin(), e = q.end(); i != e; ++i) {
      (*i)->cancel(xsink);
      (*i)->del(xsink);
   }

#ifdef DEBUG
   q.clear();;
#endif
}
/* Qore class Qore::Thread::ThreadPool */

qore_classid_t CID_THREADPOOL;
QoreClass* QC_THREADPOOL;

// ThreadPool::constructor(int max = 0, int minidle = 0, int maxidle = 0, timeout release_ms = 5s) {}
static void ThreadPool_constructor_ViViViVt(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
   int64 max = HARD_QORE_INT(args, 0);
   int64 minidle = HARD_QORE_INT(args, 1);
   int64 maxidle = HARD_QORE_INT(args, 2);
   int64 release_ms = HARD_QORE_INT(args, 3);
# 238 "QC_ThreadPool.qpp"
if (max > 0) {
      if (minidle > max) {
         xsink->raiseException("THREADPOOL-ERROR", "cannot create a ThreadPool object with minidle (%d) > max (%d)", minidle, max);
         return;
      }
      if (maxidle > max) {
         xsink->raiseException("THREADPOOL-ERROR", "cannot create a ThreadPool object with maxidle (%d) > max (%d)", maxidle, max);
         return;
      }
   }
   if (maxidle > 0) {
      if (minidle > maxidle) {
         xsink->raiseException("THREADPOOL-ERROR", "cannot create a ThreadPool object with minidle (%d) > maxidle (%d)", minidle, maxidle);
         return;
      }
   }
   if (release_ms < 0) {
      xsink->raiseException("THREADPOOL-ERROR", "cannot create a ThreadPool object with release_ms < 0 (value passed: %d)", release_ms);
      return;
   }

   ReferenceHolder<ThreadPool> tp(new ThreadPool(xsink, max, minidle, maxidle, release_ms), xsink);
   if (*xsink)
      return;

   self->setPrivate(CID_THREADPOOL, tp.release());
}

// ThreadPool::destructor() {}
static void ThreadPool_destructor(QoreObject* self, ThreadPool* tp, ExceptionSink* xsink) {
# 273 "QC_ThreadPool.qpp"
tp->stop();
   tp->deref();
}

// nothing ThreadPool::stop(){}
static AbstractQoreNode* ThreadPool_stop(QoreObject* self, ThreadPool* tp, const QoreListNode* args, ExceptionSink* xsink) {
# 290 "QC_ThreadPool.qpp"
tp->stop();
   return 0;
}

// nothing ThreadPool::stopWait(){}
static AbstractQoreNode* ThreadPool_stopWait(QoreObject* self, ThreadPool* tp, const QoreListNode* args, ExceptionSink* xsink) {
# 304 "QC_ThreadPool.qpp"
tp->stopWait(xsink);
   return 0;
}

// nothing ThreadPool::submit(code task, *code cancel){}
static AbstractQoreNode* ThreadPool_submit_VcNc(QoreObject* self, ThreadPool* tp, const QoreListNode* args, ExceptionSink* xsink) {
   const ResolvedCallReferenceNode* task = reinterpret_cast<const ResolvedCallReferenceNode*>(get_param(args, 0));
   const ResolvedCallReferenceNode* cancel = reinterpret_cast<const ResolvedCallReferenceNode*>(get_param(args, 1));
# 317 "QC_ThreadPool.qpp"
tp->submit(task->refRefSelf(), cancel ? cancel->refRefSelf() : 0, xsink);
   return 0;
}

// string ThreadPool::toString(){}
static AbstractQoreNode* ThreadPool_toString(QoreObject* self, ThreadPool* tp, const QoreListNode* args, ExceptionSink* xsink) {
# 329 "QC_ThreadPool.qpp"
QoreStringNode* str = new QoreStringNode;
   tp->toString(*str);
   return str;
}

DLLLOCAL QoreClass* initThreadPoolClass(QoreNamespace& ns) {
   QC_THREADPOOL = new QoreClass("ThreadPool", QDOM_THREAD_CLASS);
   CID_THREADPOOL = QC_THREADPOOL->getID();

   // ThreadPool::constructor(int max = 0, int minidle = 0, int maxidle = 0, timeout release_ms = 5s) {}
   QC_THREADPOOL->setConstructorExtended3(ThreadPool_constructor_ViViViVt, false, QC_NO_FLAGS, QDOM_DEFAULT, 4, bigIntTypeInfo, zero(), "max", bigIntTypeInfo, zero(), "minidle", bigIntTypeInfo, zero(), "maxidle", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 5, 0), "release_ms");

   // ThreadPool::destructor() {}
   QC_THREADPOOL->setDestructor((q_destructor_t)ThreadPool_destructor);

   // nothing ThreadPool::stop(){}
   QC_THREADPOOL->addMethodExtended3("stop", (q_method_t)ThreadPool_stop, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // nothing ThreadPool::stopWait(){}
   QC_THREADPOOL->addMethodExtended3("stopWait", (q_method_t)ThreadPool_stopWait, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // nothing ThreadPool::submit(code task, *code cancel){}
   QC_THREADPOOL->addMethodExtended3("submit", (q_method_t)ThreadPool_submit_VcNc, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, codeTypeInfo, NULL, "task", codeOrNothingTypeInfo, NULL, "cancel");

   // string ThreadPool::toString(){}
   QC_THREADPOOL->addMethodExtended3("toString", (q_method_t)ThreadPool_toString, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

   return QC_THREADPOOL;
}
