/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
  ql_compression.qpp

  Qore Programming Language

  Copyright 2003 - 2013 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>

#include <zlib.h>

#include <qore/intern/ql_compression.h>

#include <bzlib.h>

#include <errno.h>

#ifndef QORE_BZ2_WORK_FACTOR
#define QORE_BZ2_WORK_FACTOR 30
#endif

#ifndef QORE_BZ2_VERBOSITY
#define QORE_BZ2_VERBOSITY 0
#endif

#ifndef BZ2_DEFAULT_COMPRESSION
#define BZ2_DEFAULT_COMPRESSION 9
#endif

class qore_bz_stream : public bz_stream {
public:
   DLLLOCAL qore_bz_stream() {
      bzalloc = 0;
      bzfree = 0;
      opaque = 0;
   }
};

class qore_bz_compressor : public qore_bz_stream {
   bool ok;

public:
   DLLLOCAL qore_bz_compressor(int level, ExceptionSink *xsink) {
      int rc = BZ2_bzCompressInit(this, level, QORE_BZ2_VERBOSITY, QORE_BZ2_WORK_FACTOR);
      ok = (rc == BZ_OK);
      if (!ok)
	 xsink->raiseException("BZIP2-COMPRESS-ERROR", "code %d returned from BZ2_bzCompressInit()", rc);
   }
   
   DLLLOCAL ~qore_bz_compressor() {
      if (ok)
	 BZ2_bzCompressEnd(this);
   }
   DLLLOCAL operator bool() const { return ok; }

   DLLLOCAL BinaryNode *compress(const void *ptr, unsigned long len, ExceptionSink *xsink) {
      char *p = (char *)ptr;
      
      int bs = len >> 3;
      if (!bs)
	 bs = 1;

      next_in = p;
      avail_in = len;

      SimpleRefHolder<BinaryNode> b(new BinaryNode());

      if (b->preallocate(bs)) {
	 xsink->outOfMemory();
	 return 0;      
      }

      while (true) {
	 int64 done = (int64)total_out_lo32 + (((int64)total_out_hi32) << 32);
	 next_out = ((char *)b->getPtr()) + done;
	 avail_out = (unsigned)(bs - done);

	 int rc = BZ2_bzCompress(this, BZ_FINISH);
	 //printd(5, "bs=%d, done=%lld, avail_out=%d, rc=%d, total_out_lo32=%d, total_out_hi32 = %d\n", bs, done, avail_out, rc, total_out_lo32, total_out_hi32);
	 if (rc == BZ_STREAM_END)
	    break;
	 if (rc != BZ_FINISH_OK) {
	    xsink->raiseException("BZIP2-COMPRESS-ERROR", "error code %d returned from BZ2_bzCompress()", rc);
	    return 0;
	 }
	 bs *= 2;
	 if (b->preallocate(bs)) {
	    xsink->outOfMemory();
	    return 0;      
	 }
      }
      b->setSize((int64)total_out_lo32 + (((int64)total_out_hi32) << 32));

      return b.release();
   }
};

class qore_bz_decompressor : public qore_bz_stream {
   bool ok;

public:
   DLLLOCAL qore_bz_decompressor(ExceptionSink *xsink) {
      int rc = BZ2_bzDecompressInit(this, QORE_BZ2_VERBOSITY, 0);
      ok = (rc == BZ_OK);
      if (!ok)
	 xsink->raiseException("BZIP2-DECOMPRESS-ERROR", "code %d returned from BZ2_bzDecompressInit()", rc);
   }

   DLLLOCAL ~qore_bz_decompressor() {
      if (ok)
	 BZ2_bzDecompressEnd(this);
   }
   DLLLOCAL operator bool() const { return ok; }

   DLLLOCAL BinaryNode *decompress(const void *ptr, unsigned long len, ExceptionSink *xsink) {
      char *p = (char *)ptr;

      int bs = len << 1;

      next_in = p;
      avail_in = len;

      SimpleRefHolder<BinaryNode> b(new BinaryNode());

      if (b->preallocate(bs)) {
	 xsink->outOfMemory();
	 return 0;      
      }

      while (true) {
	 int64 done = (int64)total_out_lo32 + (((int64)total_out_hi32) << 32);
	 next_out = ((char *)b->getPtr()) + done;
	 avail_out = (unsigned)(bs - done);

	 int rc = BZ2_bzDecompress(this);
	 //printd(5, "bs=%d, done=%lld, avail_out=%d, rc=%d, total_out_lo32=%d, total_out_hi32 = %d\n", bs, done, avail_out, rc, total_out_lo32, total_out_hi32);
	 if (rc == BZ_STREAM_END)
	    break;
	 if (rc != BZ_OK) {
	    xsink->raiseException("BZIP2-DECOMPRESS-ERROR", "error code %d returned from BZ2_bzDecompress()", rc);
	    return 0;
	 }
	 bs *= 2;
	 if (b->preallocate(bs)) {
	    xsink->outOfMemory();
	    return 0;      
	 }
      }
      b->setSize((int64)total_out_lo32 + (((int64)total_out_hi32) << 32));
      
      return b.release();
   }

   DLLLOCAL QoreStringNode *decompress_to_string(const void *ptr, unsigned long len, const QoreEncoding *enc, ExceptionSink *xsink) {
      static char np[] = {'\0'};

      SimpleRefHolder<BinaryNode> b(decompress(ptr, len, xsink));
      if (!b)
	 return 0;

      len = b->size();

      // terminate the string
      b->append(np, 1);

      return new QoreStringNode((char *)b->giveBuffer(), len, len + 1, enc);
   }
};

BinaryNode *qore_bzip2(const void *ptr, unsigned long len, int level, ExceptionSink *xsink) {
   qore_bz_compressor c(level, xsink);
   if (!c)
      return 0;

   return c.compress(ptr, len, xsink);
}

BinaryNode *qore_bunzip2_to_binary(const BinaryNode *b, ExceptionSink *xsink) {
   qore_bz_decompressor c(xsink);
   if (!c)
      return 0;

   return c.decompress(b->getPtr(), b->size(), xsink);
}

QoreStringNode *qore_bunzip2_to_string(const BinaryNode *b, const QoreEncoding *enc, ExceptionSink *xsink) {
   qore_bz_decompressor c(xsink);
   if (!c)
      return 0;

   return c.decompress_to_string(b->getPtr(), b->size(), enc, xsink);
}

#ifdef HAVE_GZ_HEADER
class qore_gz_header : public gz_header {
   DLLLOCAL qore_gz_header(bool n_text, char *n_name, char *n_comment) {
      text = n_text;
      time = ::time(0);
      xflags = 0;
      os = 3; // always set to UNIX
      extra = Z_NULL;
      extra_len = 0;
      extra_max = 0;
      name = (Bytef *)n_name;
      name_max = 0;
      comment = (Bytef *)n_comment;
      comm_max = 0;
      hcrc = 0;
      done = 0;
   }
};
#endif

static void do_zlib_exception(int rc, const char *func, ExceptionSink *xsink) {
   QoreStringNode *desc = new QoreStringNode();
   desc->sprintf("%s(): ", func);
   switch (rc) {
      case Z_ERRNO:
	 q_strerror(*desc, errno);
         break;

      case Z_STREAM_ERROR:
         desc->concat("inconsistent stream state");
         break;

      case Z_DATA_ERROR:
         desc->set("unable to process input data; data corrupted");
         break;

      case Z_MEM_ERROR:
         desc->set("insufficient memory to complete operation");
         break;

      case Z_BUF_ERROR:
         desc->set("qore buffer-handling error (report as bug to qore developers)");
         break;

      case Z_VERSION_ERROR:
         desc->set("version mismatch on zlib shared library, check library requirements");
         break;

      default:
         desc->sprintf("error code %d encountered", rc);
         break;
   }
   xsink->raiseException("ZLIB-ERROR", desc);
}

static void do_deflate_end(z_stream *c_stream, ExceptionSink *xsink) {
   int rc = deflateEnd(c_stream);
   if (rc != Z_OK)
      do_zlib_exception(rc, "deflateEnd", xsink);
}

static void do_inflate_end(z_stream *d_stream, ExceptionSink *xsink) {
   int rc = inflateEnd(d_stream);
   if (rc != Z_OK)
      do_zlib_exception(rc, "inflateEnd", xsink);
}

BinaryNode *qore_deflate(const void *ptr, unsigned long len, int level, ExceptionSink *xsink) {
   z_stream c_stream; // compression stream
   c_stream.zalloc = Z_NULL;
   c_stream.zfree = Z_NULL;
   c_stream.opaque = Z_NULL;
   
   int rc = deflateInit(&c_stream, level);   
   if (rc != Z_OK) {
      do_zlib_exception(rc, "deflateInit", xsink);
      return 0;
   }

   ON_BLOCK_EXIT(do_deflate_end, &c_stream, xsink);
   
   // allocate new buffer
   unsigned long bsize = len / 5 + 100;
   void *buf = malloc(bsize);
   
   c_stream.next_in = (Bytef *)ptr;
   c_stream.next_out = (Bytef *)buf;
   c_stream.avail_in = len;
   c_stream.avail_out = bsize;

   while (c_stream.avail_in) {
      rc = deflate(&c_stream, Z_NO_FLUSH);
      if (rc != Z_OK && rc != Z_BUF_ERROR) {
         free(buf);
         do_zlib_exception(rc, "deflate", xsink);
         return 0;
      }
      
      if (!c_stream.avail_out) {
         int new_space = ((len / 3) + 100);
         //printd(5, "deflate() Z_BUF_ERROR:1 bsize=%d->%d, new_space=%d avail_out=%d -> %d next_out=%p\n", bsize, bsize + new_space, new_space, c_stream.avail_out, c_stream.avail_out + new_space, c_stream.next_out);
         bsize += new_space;
         c_stream.avail_out += new_space;
         buf = realloc(buf, bsize);
         c_stream.next_out = ((Bytef *)buf) + c_stream.total_out;
      }
      //printd(5, "deflate() Z_BUF_ERROR:1 bsize=%d, avail_out=%d, next_out=%p\n", bsize, c_stream.avail_out, c_stream.next_out);
   }

   while (true) {
      rc = deflate(&c_stream, Z_FINISH);
      if (rc == Z_STREAM_END)
         break;
      if (rc != Z_OK && rc != Z_BUF_ERROR) {
         free(buf);
         do_zlib_exception(rc, "deflate", xsink);
         return 0;
      }
      // resize buffer
      int new_space = 2; //((len / 3) + 100);
      //printd(5, "deflate() Z_BUF_ERROR:2 bsize=%d->%d, new_space=%d avail_out=%d -> %d, next_out=%p\n", bsize, bsize + new_space, new_space, c_stream.avail_out, c_stream.avail_out + new_space, c_stream.next_out);
      
      bsize += new_space;
      c_stream.avail_out += new_space;
      buf = realloc(buf, bsize);
      c_stream.next_out = ((Bytef *)buf) + c_stream.total_out;
   }

   //printd(5, "deflate() buf=%p, bsize=%d, avail_out=%d, size=%d, next_out=%p\n", buf, bsize, c_stream.avail_out, bsize - c_stream.avail_out, c_stream.next_out);
   return new BinaryNode(buf, bsize - c_stream.avail_out);
}

QoreStringNode *qore_inflate_to_string(const BinaryNode *b, const QoreEncoding *enc, ExceptionSink *xsink) {
   z_stream d_stream; // decompression stream
   d_stream.zalloc = Z_NULL;
   d_stream.zfree = Z_NULL;
   d_stream.opaque = Z_NULL;

   int rc = inflateInit(&d_stream);
   if (rc != Z_OK) {
      do_zlib_exception(rc, "inflateInit", xsink);
      return 0;
   }

   ON_BLOCK_EXIT(do_inflate_end, &d_stream, xsink);
   
   int len = b->size();

   // allocate new buffer
   unsigned long bsize = len * 2 + 100;
   void *buf = malloc(bsize);
   
   d_stream.next_in = (Bytef *)b->getPtr();
   d_stream.next_out = (Bytef *)buf;
   d_stream.avail_in = b->size();
   d_stream.avail_out = bsize;

   while (true) {
      rc = inflate(&d_stream, Z_NO_FLUSH);
      if (rc == Z_STREAM_END)
         break;
      if (rc == Z_BUF_ERROR) {
         int new_space = ((len * 3) + 100);
         bsize += new_space;
         d_stream.avail_out += new_space;
         buf = realloc(buf, bsize);
         d_stream.next_out = ((Bytef *)buf) + d_stream.total_out;
      }
      else if (rc != Z_OK) {
         free(buf);
         do_zlib_exception(rc, "inflate", xsink);
         return 0;
      }
   }

   // how much data was decompressed
   len = bsize - d_stream.avail_out;

   // create and return the string
   return new QoreStringNode((char *)buf, len, len, enc);
}

BinaryNode *qore_inflate_to_binary(const BinaryNode *b, ExceptionSink *xsink) {
   z_stream d_stream; // decompression stream
   d_stream.zalloc = Z_NULL;
   d_stream.zfree = Z_NULL;
   d_stream.opaque = Z_NULL;

   int rc = inflateInit(&d_stream);
   if (rc != Z_OK) {
      do_zlib_exception(rc, "inflateInit", xsink);
      return 0;
   }

   ON_BLOCK_EXIT(do_inflate_end, &d_stream, xsink);
   
   int len = b->size();

   // allocate new buffer
   unsigned long bsize = len * 2 + 100;
   void *buf = malloc(bsize);
   
   d_stream.next_in = (Bytef *)b->getPtr();
   d_stream.next_out = (Bytef *)buf;
   d_stream.avail_in = b->size();
   d_stream.avail_out = bsize;

   while (true){
      rc = inflate(&d_stream, Z_NO_FLUSH);
      if (rc == Z_STREAM_END)
         break;
      if (rc == Z_BUF_ERROR) {
         int new_space = ((len * 3) + 100);
         bsize += new_space;
         d_stream.avail_out += new_space;
         buf = realloc(buf, bsize);
         d_stream.next_out = ((Bytef *)buf) + d_stream.total_out;
      }
      else if (rc != Z_OK) {
         free(buf);
         do_zlib_exception(rc, "inflate", xsink);
         return 0;
      }
   }
   
   return new BinaryNode(buf, bsize - d_stream.avail_out);
}

BinaryNode *qore_gzip(const void *ptr, unsigned long len, int level, ExceptionSink *xsink) {
   z_stream c_stream; // compression stream
   c_stream.zalloc = Z_NULL;
   c_stream.zfree = Z_NULL;
   c_stream.opaque = Z_NULL;
   c_stream.next_in = (Bytef *)ptr;
   c_stream.avail_in = len;

   int rc = deflateInit2(&c_stream, level, Z_DEFLATED, 31, 8,  Z_DEFAULT_STRATEGY);   
   if (rc != Z_OK) {
      do_zlib_exception(rc, "deflateInit2", xsink);
      return 0;
   }

   ON_BLOCK_EXIT(do_deflate_end, &c_stream, xsink);
   
   // allocate new buffer
   unsigned long bsize = len / 5 + 100;
   void *buf = malloc(bsize);
   
   c_stream.next_out = (Bytef *)buf;
   c_stream.avail_out = bsize;

   while (c_stream.avail_in) {
      rc = deflate(&c_stream, Z_NO_FLUSH);
      if (rc != Z_OK && rc != Z_BUF_ERROR) {
         free(buf);
         do_zlib_exception(rc, "deflate", xsink);
         return 0;
      }

      if (!c_stream.avail_out) {
         int new_space = ((len / 3) + 100);
         //printd(5, "deflate() Z_BUF_ERROR:1 bsize=%d->%d, new_space=%d avail_out=%d -> %d next_out=%p\n", bsize, bsize + new_space, new_space, c_stream.avail_out, c_stream.avail_out + new_space, c_stream.next_out);

         bsize += new_space;
         c_stream.avail_out += new_space;
         buf = realloc(buf, bsize);
         c_stream.next_out = ((Bytef *)buf) + c_stream.total_out;
      }
      //printd(5, "deflate() Z_BUF_ERROR:1 bsize=%d, avail_out=%d, next_out=%p\n", bsize, c_stream.avail_out, c_stream.next_out);
   }

   while (true) {
      rc = deflate(&c_stream, Z_FINISH);
      if (rc == Z_STREAM_END)
         break;
      if (rc != Z_OK && rc != Z_BUF_ERROR) {
         free(buf);
         do_zlib_exception(rc, "deflate", xsink);
         return 0;
      }
      // resize buffer
      int new_space = 2; //((len / 3) + 100);
      //printd(5, "deflate() Z_BUF_ERROR:2 bsize=%d->%d, new_space=%d avail_out=%d -> %d, next_out=%p\n", bsize, bsize + new_space, new_space, c_stream.avail_out, c_stream.avail_out + new_space, c_stream.next_out);

      bsize += new_space;
      c_stream.avail_out += new_space;
      buf = realloc(buf, bsize);
      c_stream.next_out = ((Bytef *)buf) + c_stream.total_out;
   }

   //printd(5, "deflate() buf=%p, bsize=%d, avail_out=%d, size=%d, next_out=%p\n", buf, bsize, c_stream.avail_out, bsize - c_stream.avail_out, c_stream.next_out);
   return new BinaryNode(buf, bsize - c_stream.avail_out);
}

QoreStringNode *qore_gunzip_to_string(const BinaryNode *bin, const QoreEncoding *enc, ExceptionSink *xsink) {
   z_stream d_stream; // decompression stream
   d_stream.zalloc = Z_NULL;
   d_stream.zfree = Z_NULL;
   d_stream.opaque = Z_NULL;

   d_stream.next_in = (Bytef *)bin->getPtr();
   d_stream.avail_in = bin->size();

   int rc = inflateInit2(&d_stream, 47);
   if (rc != Z_OK) {
      do_zlib_exception(rc, "inflateInit2", xsink);
      return 0;
   }

   ON_BLOCK_EXIT(do_inflate_end, &d_stream, xsink);
   
   int len = bin->size();

   // allocate new buffer
   unsigned long bsize = len * 2 + 100;
   void *buf = malloc(bsize);
   
   d_stream.next_out = (Bytef *)buf;
   d_stream.avail_out = bsize;

   while (true) {
      rc = inflate(&d_stream, Z_NO_FLUSH);
      if (rc == Z_STREAM_END)
         break;
      if (rc == Z_BUF_ERROR) {
         int new_space = ((len * 3) + 100);
         bsize += new_space;
         d_stream.avail_out += new_space;
         buf = realloc(buf, bsize);
         d_stream.next_out = ((Bytef *)buf) + d_stream.total_out;
      }
      else if (rc != Z_OK) {
         free(buf);
         do_zlib_exception(rc, "inflate", xsink);
         return 0;
      }
   }

   // how much data was decompressed
   len = bsize - d_stream.avail_out;

   // create and return the string
   return new QoreStringNode((char *)buf, len, len, enc);
}

BinaryNode *qore_gunzip_to_binary(const BinaryNode *bin, ExceptionSink *xsink) {
   z_stream d_stream; // decompression stream
   d_stream.zalloc = Z_NULL;
   d_stream.zfree = Z_NULL;
   d_stream.opaque = Z_NULL;

   d_stream.next_in = (Bytef *)bin->getPtr();
   d_stream.avail_in = bin->size();

   int rc = inflateInit2(&d_stream, 47);
   if (rc != Z_OK) {
      do_zlib_exception(rc, "inflateInit2", xsink);
      return 0;
   }

   ON_BLOCK_EXIT(do_inflate_end, &d_stream, xsink);
   
   int len = bin->size();

   // allocate new buffer
   unsigned long bsize = len * 2 + 100;
   void *buf = malloc(bsize);
   
   d_stream.next_out = (Bytef *)buf;
   d_stream.avail_out = bsize;

   while (true) {
      rc = inflate(&d_stream, Z_NO_FLUSH);
      if (rc == Z_STREAM_END)
         break;
      if (rc == Z_BUF_ERROR) {
         int new_space = ((len * 3) + 100);
         bsize += new_space;
         d_stream.avail_out += new_space;
         buf = realloc(buf, bsize);
         d_stream.next_out = ((Bytef *)buf) + d_stream.total_out;
      }
      else if (rc != Z_OK) {
         free(buf);
         do_zlib_exception(rc, "inflate", xsink);
         return 0;
      }
   }
   
   return new BinaryNode(buf, bsize - d_stream.avail_out);
}
# 613 "ql_compression.qpp"
# 645 "ql_compression.qpp"
# 673 "ql_compression.qpp"
# 679 "ql_compression.qpp"
# 698 "ql_compression.qpp"
# 704 "ql_compression.qpp"
# 720 "ql_compression.qpp"
# 726 "ql_compression.qpp"
# 754 "ql_compression.qpp"
# 781 "ql_compression.qpp"
# 787 "ql_compression.qpp"
# 805 "ql_compression.qpp"
# 811 "ql_compression.qpp"
# 827 "ql_compression.qpp"
# 833 "ql_compression.qpp"
# 856 "ql_compression.qpp"
# 881 "ql_compression.qpp"
# 887 "ql_compression.qpp"
# 903 "ql_compression.qpp"
# 909 "ql_compression.qpp"
# 927 "ql_compression.qpp"
// binary bunzip2_to_binary(binary bin) {}
static AbstractQoreNode* f_bunzip2_to_binary_Vy(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
# 899 "ql_compression.qpp"
   return qore_bunzip2_to_binary(bin, xsink);
}

// nothing bunzip2_to_binary() {}
static AbstractQoreNode* f_bunzip2_to_binary(const QoreListNode* args, ExceptionSink* xsink) {
# 906 "ql_compression.qpp"

   return 0;
}

// string bunzip2_to_string(binary bin, *string encoding) {}
static AbstractQoreNode* f_bunzip2_to_string_VyNs(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
   const QoreStringNode* encoding = reinterpret_cast<const QoreStringNode*>(get_param(args, 1));
# 922 "ql_compression.qpp"
   const QoreEncoding *qe = encoding ? QEM.findCreate(encoding) : QCS_DEFAULT;
   return qore_bunzip2_to_string(bin, qe, xsink);
}

// nothing bunzip2_to_string() {}
static AbstractQoreNode* f_bunzip2_to_string(const QoreListNode* args, ExceptionSink* xsink) {
# 930 "ql_compression.qpp"

   return 0;
}

// binary bzip2(binary bin, softint level = BZ2_DEFAULT_COMPRESSION) {}
static AbstractQoreNode* f_bzip2_Vyvi(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
   int64 level = HARD_QORE_INT(args, 1);
# 847 "ql_compression.qpp"
   if (!level || level > 9) {
      xsink->raiseException("BZLIB2-LEVEL-ERROR", "level must be between 1 - 9 (value passed: %d)", level);
      return 0;
   }

   return qore_bzip2(bin->getPtr(), bin->size(), level, xsink);
}

// binary bzip2(string str, softint level = BZ2_DEFAULT_COMPRESSION) {}
static AbstractQoreNode* f_bzip2_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 level = HARD_QORE_INT(args, 1);
# 872 "ql_compression.qpp"
   if (!level || level > 9) {
      xsink->raiseException("BZLIB2-LEVEL-ERROR", "level must be between 1 - 9 (value passed: %d)", level);
      return 0;
   }

   return qore_bzip2(str->getBuffer(), str->strlen(), level, xsink);
}

// nothing bzip2() {}
static AbstractQoreNode* f_bzip2(const QoreListNode* args, ExceptionSink* xsink) {
# 884 "ql_compression.qpp"

   return 0;
}

// binary compress(string str, int level = Z_DEFAULT_COMPRESSION) {}
static AbstractQoreNode* f_compress_VsVi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 level = HARD_QORE_INT(args, 1);
# 633 "ql_compression.qpp"
   if ((level < 1 && level != -1) || level > 9) {
      xsink->raiseException("ZLIB-LEVEL-ERROR", "level must be between 1 - 9 or -1 (value passed: %d)", level);
      return 0;
   }
   
   if (!str->strlen())
      return new BinaryNode;
   
   return qore_deflate(str->getBuffer(), str->strlen(), level, xsink);
}

// binary compress(binary bin, int level = Z_DEFAULT_COMPRESSION) {}
static AbstractQoreNode* f_compress_VyVi(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
   int64 level = HARD_QORE_INT(args, 1);
# 661 "ql_compression.qpp"
   if ((level < 1 && level != -1) || level > 9) {
      xsink->raiseException("ZLIB-LEVEL-ERROR", "level must be between 1 - 9 or -1 (value passed: %d)", level);
      return 0;
   }
   
   if (!bin->size())
      return new BinaryNode;
   
   return qore_deflate(bin->getPtr(), bin->size(), level, xsink);
}

// nothing compress() {}
static AbstractQoreNode* f_compress(const QoreListNode* args, ExceptionSink* xsink) {
# 676 "ql_compression.qpp"

   return 0;
}

// binary gunzip_to_binary(binary bin) {}
static AbstractQoreNode* f_gunzip_to_binary_Vy(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
# 823 "ql_compression.qpp"
   return qore_gunzip_to_binary(bin, xsink);
}

// nothing gunzip_to_binary() {}
static AbstractQoreNode* f_gunzip_to_binary(const QoreListNode* args, ExceptionSink* xsink) {
# 830 "ql_compression.qpp"

   return 0;
}

// string gunzip_to_string(binary bin, *string encoding) {}
static AbstractQoreNode* f_gunzip_to_string_VyNs(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
   const QoreStringNode* encoding = reinterpret_cast<const QoreStringNode*>(get_param(args, 1));
# 800 "ql_compression.qpp"
   const QoreEncoding *qe = encoding ? QEM.findCreate(encoding) : QCS_DEFAULT;
   return qore_gunzip_to_string(bin, qe, xsink);
}

// nothing gunzip_to_string() {}
static AbstractQoreNode* f_gunzip_to_string(const QoreListNode* args, ExceptionSink* xsink) {
# 808 "ql_compression.qpp"

   return 0;
}

// binary gzip(string str, int level = Z_DEFAULT_COMPRESSION) {}
static AbstractQoreNode* f_gzip_VsVi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 level = HARD_QORE_INT(args, 1);
# 742 "ql_compression.qpp"
   if (!level || level > 9) {
      xsink->raiseException("ZLIB-LEVEL-ERROR", "level must be between 0 - 9 (value passed: %d)", level);
      return 0;
   }
   
   if (!str->strlen())
      return new BinaryNode;
   
   return qore_gzip(str->getBuffer(), str->strlen(), level, xsink);
}

// binary gzip(binary bin, int level = Z_DEFAULT_COMPRESSION) {}
static AbstractQoreNode* f_gzip_VyVi(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
   int64 level = HARD_QORE_INT(args, 1);
# 769 "ql_compression.qpp"
   if (!level || level > 9) {
      xsink->raiseException("ZLIB-LEVEL-ERROR", "level must be between 0 - 9 (value passed: %d)", level);
      return 0;
   }
   
   if (!bin->size())
      return new BinaryNode;
   
   return qore_gzip(bin->getPtr(), bin->size(), level, xsink);
}

// nothing gzip() {}
static AbstractQoreNode* f_gzip(const QoreListNode* args, ExceptionSink* xsink) {
# 784 "ql_compression.qpp"

   return 0;
}

// binary uncompress_to_binary(binary bin) {}
static AbstractQoreNode* f_uncompress_to_binary_Vy(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
# 716 "ql_compression.qpp"
   return qore_inflate_to_binary(bin, xsink);
}

// nothing uncompress_to_binary() {}
static AbstractQoreNode* f_uncompress_to_binary(const QoreListNode* args, ExceptionSink* xsink) {
# 723 "ql_compression.qpp"

   return 0;
}

// string uncompress_to_string(binary bin, *string encoding) {}
static AbstractQoreNode* f_uncompress_to_string_VyNs(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
   const QoreStringNode* encoding = reinterpret_cast<const QoreStringNode*>(get_param(args, 1));
# 693 "ql_compression.qpp"
   const QoreEncoding *qe = encoding ? QEM.findCreate(encoding) : QCS_DEFAULT;
   return qore_inflate_to_string(bin, qe, xsink);
}

// nothing uncompress_to_string() {}
static AbstractQoreNode* f_uncompress_to_string(const QoreListNode* args, ExceptionSink* xsink) {
# 701 "ql_compression.qpp"

   return 0;
}


DLLLOCAL void init_compression_functions(QoreNamespace& ns) {
   // binary bunzip2_to_binary(binary bin) {}
   ns.addBuiltinVariant("bunzip2_to_binary", (q_func_t)f_bunzip2_to_binary_Vy, QC_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo, 1, binaryTypeInfo, NULL, "bin");
   // nothing bunzip2_to_binary() {}
   ns.addBuiltinVariant("bunzip2_to_binary", (q_func_t)f_bunzip2_to_binary, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string bunzip2_to_string(binary bin, *string encoding) {}
   ns.addBuiltinVariant("bunzip2_to_string", (q_func_t)f_bunzip2_to_string_VyNs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, binaryTypeInfo, NULL, "bin", stringOrNothingTypeInfo, NULL, "encoding");
   // nothing bunzip2_to_string() {}
   ns.addBuiltinVariant("bunzip2_to_string", (q_func_t)f_bunzip2_to_string, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // binary bzip2(binary bin, softint level = BZ2_DEFAULT_COMPRESSION) {}
   ns.addBuiltinVariant("bzip2", (q_func_t)f_bzip2_Vyvi, QC_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo, 2, binaryTypeInfo, NULL, "bin", softBigIntTypeInfo, new QoreBigIntNode(BZ2_DEFAULT_COMPRESSION), "level");
   // binary bzip2(string str, softint level = BZ2_DEFAULT_COMPRESSION) {}
   ns.addBuiltinVariant("bzip2", (q_func_t)f_bzip2_Vsvi, QC_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo, 2, stringTypeInfo, NULL, "str", softBigIntTypeInfo, new QoreBigIntNode(BZ2_DEFAULT_COMPRESSION), "level");
   // nothing bzip2() {}
   ns.addBuiltinVariant("bzip2", (q_func_t)f_bzip2, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // binary compress(string str, int level = Z_DEFAULT_COMPRESSION) {}
   ns.addBuiltinVariant("compress", (q_func_t)f_compress_VsVi, QC_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo, 2, stringTypeInfo, NULL, "str", bigIntTypeInfo, new QoreBigIntNode(Z_DEFAULT_COMPRESSION), "level");
   // binary compress(binary bin, int level = Z_DEFAULT_COMPRESSION) {}
   ns.addBuiltinVariant("compress", (q_func_t)f_compress_VyVi, QC_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo, 2, binaryTypeInfo, NULL, "bin", bigIntTypeInfo, new QoreBigIntNode(Z_DEFAULT_COMPRESSION), "level");
   // nothing compress() {}
   ns.addBuiltinVariant("compress", (q_func_t)f_compress, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // binary gunzip_to_binary(binary bin) {}
   ns.addBuiltinVariant("gunzip_to_binary", (q_func_t)f_gunzip_to_binary_Vy, QC_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo, 1, binaryTypeInfo, NULL, "bin");
   // nothing gunzip_to_binary() {}
   ns.addBuiltinVariant("gunzip_to_binary", (q_func_t)f_gunzip_to_binary, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string gunzip_to_string(binary bin, *string encoding) {}
   ns.addBuiltinVariant("gunzip_to_string", (q_func_t)f_gunzip_to_string_VyNs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, binaryTypeInfo, NULL, "bin", stringOrNothingTypeInfo, NULL, "encoding");
   // nothing gunzip_to_string() {}
   ns.addBuiltinVariant("gunzip_to_string", (q_func_t)f_gunzip_to_string, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // binary gzip(string str, int level = Z_DEFAULT_COMPRESSION) {}
   ns.addBuiltinVariant("gzip", (q_func_t)f_gzip_VsVi, QC_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo, 2, stringTypeInfo, NULL, "str", bigIntTypeInfo, new QoreBigIntNode(Z_DEFAULT_COMPRESSION), "level");
   // binary gzip(binary bin, int level = Z_DEFAULT_COMPRESSION) {}
   ns.addBuiltinVariant("gzip", (q_func_t)f_gzip_VyVi, QC_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo, 2, binaryTypeInfo, NULL, "bin", bigIntTypeInfo, new QoreBigIntNode(Z_DEFAULT_COMPRESSION), "level");
   // nothing gzip() {}
   ns.addBuiltinVariant("gzip", (q_func_t)f_gzip, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // binary uncompress_to_binary(binary bin) {}
   ns.addBuiltinVariant("uncompress_to_binary", (q_func_t)f_uncompress_to_binary_Vy, QC_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo, 1, binaryTypeInfo, NULL, "bin");
   // nothing uncompress_to_binary() {}
   ns.addBuiltinVariant("uncompress_to_binary", (q_func_t)f_uncompress_to_binary, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string uncompress_to_string(binary bin, *string encoding) {}
   ns.addBuiltinVariant("uncompress_to_string", (q_func_t)f_uncompress_to_string_VyNs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, binaryTypeInfo, NULL, "bin", stringOrNothingTypeInfo, NULL, "encoding");
   // nothing uncompress_to_string() {}
   ns.addBuiltinVariant("uncompress_to_string", (q_func_t)f_uncompress_to_string, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);

}

DLLLOCAL void init_compression_constants(QoreNamespace& ns) {
# 603 "ql_compression.qpp"
   ns.addConstant("BZ2_DEFAULT_COMPRESSION", new QoreBigIntNode(BZ2_DEFAULT_COMPRESSION));
   ns.addConstant("Z_DEFAULT_COMPRESSION", new QoreBigIntNode(Z_DEFAULT_COMPRESSION));
# 613 "ql_compression.qpp"
# 645 "ql_compression.qpp"
# 673 "ql_compression.qpp"
# 679 "ql_compression.qpp"
# 698 "ql_compression.qpp"
# 704 "ql_compression.qpp"
# 720 "ql_compression.qpp"
# 726 "ql_compression.qpp"
# 754 "ql_compression.qpp"
# 781 "ql_compression.qpp"
# 787 "ql_compression.qpp"
# 805 "ql_compression.qpp"
# 811 "ql_compression.qpp"
# 827 "ql_compression.qpp"
# 833 "ql_compression.qpp"
# 856 "ql_compression.qpp"
# 881 "ql_compression.qpp"
# 887 "ql_compression.qpp"
# 903 "ql_compression.qpp"
# 909 "ql_compression.qpp"
# 927 "ql_compression.qpp"

}
