/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
  ql_crypto.qpp
  
  Qore Programming Language

  Copyright 2003 - 2013 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>
#include <qore/intern/ql_crypto.h>

#include <stdio.h>
#include <stdlib.h>

#define QCRYPTO_DECRYPT 0
#define QCRYPTO_ENCRYPT 1

// NOTE: the trailing null ('\0') is not included when encrypting strings or calculating digests

// default initialization vector
static unsigned char def_iv[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

#define MD2_ERR "MD2-DIGEST-ERROR"
#define MD4_ERR "MD4-DIGEST-ERROR"
#define MD5_ERR "MD5-DIGEST-ERROR"
#define SHA_ERR "SHA-DIGEST-ERROR"
#define SHA1_ERR "SHA1-DIGEST-ERROR"
static const char SHA224_ERR[] = "SHA224-DIGEST-ERROR";
static const char SHA256_ERR[] = "SHA256-DIGEST-ERROR";
static const char SHA384_ERR[] = "SHA384-DIGEST-ERROR";
static const char SHA512_ERR[] = "SHA512-DIGEST-ERROR";
#define DSS_ERR "DSS-DIGEST-ERROR"
#define DSS1_ERR "DSS1-DIGEST-ERROR"
static const char MDC2_ERR[] = "MDC2-DIGEST-ERROR";
#define RIPEMD160_ERR "RIPEMD160-DIGEST-ERROR"

class CryptoHelper : public BaseHelper {
private:
   unsigned char *iv, *output;
   int output_len;
   const QoreEncoding* qe;

   DLLLOCAL const char *getOrdinal(int n) {
      assert(n > 0 && n <= 3);
      return n == 1 ? "first" : (n == 2 ? "second" : "third");
   }

   DLLLOCAL void setKey(const QoreListNode *args, int n) {
      assert(n > 0 && n <= 3);
      const AbstractQoreNode *pt = get_param(args, n);
      if (pt->getType() == NT_STRING) {
	 const QoreStringNode *str = reinterpret_cast<const QoreStringNode *>(pt);
	 key[n - 1] = (unsigned char *)str->getBuffer();
	 keylen[n - 1] = str->strlen();
	 return;
      }
	 
      const BinaryNode *b = reinterpret_cast<const BinaryNode *>(pt);
      key[n - 1] = (unsigned char *)b->getPtr();
      keylen[n - 1] = b->size();
   }

   DLLLOCAL void setEncoding(const QoreListNode* args, int n) {
      const QoreStringNode* encoding = test_string_param(args, n);
      qe = encoding ? QEM.findCreate(encoding) : QCS_DEFAULT;
   }

   // set the initialization vector (used with hard typing)
   DLLLOCAL int setIV(const char *err, const AbstractQoreNode *pt, ExceptionSink *xsink) {
      assert(pt);
      if (pt->getType() == NT_STRING) {
	 const QoreStringNode *str = reinterpret_cast<const QoreStringNode *>(pt);
	 if (str->strlen() < 8) {
	    xsink->raiseException(err, "the input vector must be at least 8 bytes long (%d bytes passed)", str->strlen());
	    return -1;
	 }
	 iv = (unsigned char *)str->getBuffer();
	 return 0;
      }
      
      assert(pt->getType() == NT_BINARY);
      const BinaryNode *b = reinterpret_cast<const BinaryNode *>(pt);
      if (b->size() < 8) {
	 xsink->raiseException(err, "the input vector must be at least 8 bytes long (%d bytes passed)", b->size());
	 return -1;
      }
      iv = (unsigned char *)b->getPtr();
      return 0;
   }

public:
   unsigned char *key[3];
   int keylen[3];

   DLLLOCAL CryptoHelper() : iv(0), output(0), output_len(0), qe(QCS_DEFAULT) {
   }

   DLLLOCAL ~CryptoHelper() {
      if (output)
	 free(output);
   }

   DLLLOCAL BinaryNode *getBinary() {
      BinaryNode *b = new BinaryNode(output, output_len);
      output = 0;
      return b;
   }

   DLLLOCAL QoreStringNode *getString(const QoreEncoding *enc = QCS_DEFAULT) {
      // create the string
      QoreStringNode *str = new QoreStringNode((char *)output, output_len, output_len, enc);
      // add trailing '\0'
      str->terminate(output_len);

      output = 0;
      return str;
   }

   DLLLOCAL int setSingleKey(const char *err, const QoreListNode *params, ExceptionSink *xsink, bool has_encoding = false) {
      if (setIV(err, get_param(params, 2), xsink))
	 return -1;

      setInput(get_param(params, 0));
      setKey(params, 1);
      if (has_encoding)
	 setEncoding(params, 2);
      return 0;
   }

   DLLLOCAL int doCipher(const EVP_CIPHER *type, const char *cipher, int do_crypt, ExceptionSink *xsink) {
      const char *err = (do_crypt ? "ENCRYPT-ERROR" : "DECRYPT-ERROR");

      EVP_CIPHER_CTX ctx;
      EVP_CIPHER_CTX_init(&ctx);
      EVP_CipherInit_ex(&ctx, type, 0, 0, 0, do_crypt);
      if (key[0]) {
	 if (keylen[0] > EVP_MAX_KEY_LENGTH)
	    keylen[0] = EVP_MAX_KEY_LENGTH;

	 if (!EVP_CIPHER_CTX_set_key_length(&ctx, keylen[0]) || !EVP_CipherInit_ex(&ctx, 0, 0, key[0], iv, -1)) {
	    xsink->raiseException(err, "error setting %s key length=%d", cipher, keylen[0]);
	    EVP_CIPHER_CTX_cleanup(&ctx);
	    return -1;
	 }
      }

      // we allocate 1 byte more than we need in case we return as a string so we can terminate it
      output = (unsigned char *)malloc(sizeof(char) * (input_len + (EVP_MAX_BLOCK_LENGTH * 2)));

      if (!EVP_CipherUpdate(&ctx, output, &output_len, input, input_len)) {
	 xsink->raiseException(err, "error %scrypting %s block", do_crypt ? "en" : "de", cipher);
	 EVP_CIPHER_CTX_cleanup(&ctx);
	 return -1;
      }

      int tmplen;
      // Buffer passed to EVP_EncryptFinal() must be after data just encrypted to avoid overwriting it.
      if (!EVP_CipherFinal_ex(&ctx, output + output_len, &tmplen)) {
	 xsink->raiseException(err, "error %scrypting final %s block", do_crypt ? "en" : "de", cipher);
	 EVP_CIPHER_CTX_cleanup(&ctx);
	 return -1;
      }

      EVP_CIPHER_CTX_cleanup(&ctx);
      output_len += tmplen;
      //printd(5, "cipher_intern() %s: in=%08p (%d) out=%08p (%d)\n", cipher, buf, len, cbuf, *size);
      return 0;
   }

   DLLLOCAL int checkKeyLen(const char *err, int n, int len, ExceptionSink *xsink) {
      if (keylen[n] < len) {
	 xsink->raiseException(err, "key length is not %d bytes long (%d bytes)", len, keylen[n]);
	 return -1;
      }
      keylen[n] = len;
      return 0;
   }

/*
   DLLLOCAL int setDESKey(int n, ExceptionSink *xsink) {
      // force odd parity 
      //DES_set_odd_parity((DES_cblock *)key[n]);
      // populate the schedule structure
      //DES_set_key_unchecked((DES_cblock *)key[n], schedule);
      //key[n] = (unsigned char *)schedule;
      return 0;
   }
*/
};

static void missing_openssl_feature(const char *f, ExceptionSink *xsink) {
   xsink->raiseException("MISSING-FEATURE-ERROR", "the openssl library version that qore was compiled with did not support the %s algorithm; for maximum portability, check Option::HAVE_%s before calling this function", f, f);
}
# 222 "ql_crypto.qpp"
# 249 "ql_crypto.qpp"
# 263 "ql_crypto.qpp"
# 288 "ql_crypto.qpp"
# 313 "ql_crypto.qpp"
# 341 "ql_crypto.qpp"
# 368 "ql_crypto.qpp"
# 395 "ql_crypto.qpp"
# 425 "ql_crypto.qpp"
# 452 "ql_crypto.qpp"
# 479 "ql_crypto.qpp"
# 509 "ql_crypto.qpp"
# 536 "ql_crypto.qpp"
# 563 "ql_crypto.qpp"
# 593 "ql_crypto.qpp"
# 620 "ql_crypto.qpp"
# 647 "ql_crypto.qpp"
# 677 "ql_crypto.qpp"
# 702 "ql_crypto.qpp"
# 727 "ql_crypto.qpp"
# 755 "ql_crypto.qpp"
# 780 "ql_crypto.qpp"
# 805 "ql_crypto.qpp"
# 833 "ql_crypto.qpp"
# 858 "ql_crypto.qpp"
# 883 "ql_crypto.qpp"
# 911 "ql_crypto.qpp"
# 944 "ql_crypto.qpp"
# 977 "ql_crypto.qpp"
// binary blowfish_decrypt_cbc(binary data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_blowfish_decrypt_cbc_VyVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 303 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("BLOWFISH-DECRYPT-PARAM-ERROR", args, xsink)
       || ch.doCipher(EVP_bf_cbc(), "blowfish", QCRYPTO_DECRYPT, xsink))
      return 0;

   return ch.getBinary();
}

// string blowfish_decrypt_cbc_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
static AbstractQoreNode* f_blowfish_decrypt_cbc_to_string_VyVtVtNs(const QoreListNode* args, ExceptionSink* xsink) {
# 331 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("BLOWFISH-DECRYPT-PARAM-ERROR", args, xsink, true)
       || ch.doCipher(EVP_bf_cbc(), "blowfish", QCRYPTO_DECRYPT, xsink))
      return 0;

   return ch.getString();
}

// binary blowfish_encrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_blowfish_encrypt_cbc_VtVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 278 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("BLOWFISH-ENCRYPT-PARAM-ERROR", args, xsink)
       || ch.doCipher(EVP_bf_cbc(), "blowfish", QCRYPTO_ENCRYPT, xsink))
      return 0;

   return ch.getBinary();
}

// binary cast5_decrypt_cbc(binary data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_cast5_decrypt_cbc_VyVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 873 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("CAST5-DECRYPT-PARAM-ERROR", args, xsink)
       || ch.doCipher(EVP_cast5_cbc(), "CAST5", QCRYPTO_DECRYPT, xsink))
      return 0;

   return ch.getBinary();
}

// string cast5_decrypt_cbc_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
static AbstractQoreNode* f_cast5_decrypt_cbc_to_string_VyVtVtNs(const QoreListNode* args, ExceptionSink* xsink) {
# 901 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("CAST5-DECRYPT-PARAM-ERROR", args, xsink, true)
       || ch.doCipher(EVP_cast5_cbc(), "CAST5", QCRYPTO_DECRYPT, xsink))
      return 0;

   return ch.getString();
}

// binary cast5_encrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_cast5_encrypt_cbc_VtVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 848 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("CAST5-ENCRYPT-PARAM-ERROR", args, xsink)
       || ch.doCipher(EVP_cast5_cbc(), "CAST5", QCRYPTO_ENCRYPT, xsink))
      return 0;

   return ch.getBinary();
}

// binary des_decrypt_cbc(binary data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_des_decrypt_cbc_VyVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 384 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("DES-DECRYPT-PARAM-ERROR", args, xsink) 
       || ch.checkKeyLen("DES-KEY-ERROR", 0, 8, xsink)
       || ch.doCipher(EVP_des_cbc(), "DES", QCRYPTO_DECRYPT, xsink))
	    return 0;

   return ch.getBinary();
}

// string des_decrypt_cbc_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
static AbstractQoreNode* f_des_decrypt_cbc_to_string_VyVtVtNs(const QoreListNode* args, ExceptionSink* xsink) {
# 414 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("DES-DECRYPT-PARAM-ERROR", args, xsink, true) 
       || ch.checkKeyLen("DES-KEY-ERROR", 0, 8, xsink)
       || ch.doCipher(EVP_des_cbc(), "DES", QCRYPTO_DECRYPT, xsink))
	    return 0;

   return ch.getString();
}

// binary des_ede3_decrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_des_ede3_decrypt_cbc_VtVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 552 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("DES-DECRYPT-PARAM-ERROR", args, xsink) 
       || ch.checkKeyLen("DES-KEY-ERROR", 0, 24, xsink)
       || ch.doCipher(EVP_des_ede3_cbc(), "DES", QCRYPTO_DECRYPT, xsink))
      return 0;

   return ch.getBinary();
}

// string des_ede3_decrypt_cbc_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
static AbstractQoreNode* f_des_ede3_decrypt_cbc_to_string_VyVtVtNs(const QoreListNode* args, ExceptionSink* xsink) {
# 582 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("DES-DECRYPT-PARAM-ERROR", args, xsink, true) 
       || ch.checkKeyLen("DES-KEY-ERROR", 0, 24, xsink)
       || ch.doCipher(EVP_des_ede3_cbc(), "DES", QCRYPTO_DECRYPT, xsink))
      return 0;

   return ch.getString();
}

// binary des_ede3_encrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_des_ede3_encrypt_cbc_VtVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 525 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("DES-ENCRYPT-PARAM-ERROR", args, xsink) 
       || ch.checkKeyLen("DES-KEY-ERROR", 0, 24, xsink)
       || ch.doCipher(EVP_des_ede3_cbc(), "DES", QCRYPTO_ENCRYPT, xsink))
      return 0;

   return ch.getBinary();
}

// binary des_ede_decrypt_cbc(binary data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_des_ede_decrypt_cbc_VyVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 468 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("DES-DECRYPT-PARAM-ERROR", args, xsink) 
       || ch.checkKeyLen("DES-KEY-ERROR", 0, 16, xsink)
       || ch.doCipher(EVP_des_ede_cbc(), "DES", QCRYPTO_DECRYPT, xsink))
      return 0;

   return ch.getBinary();
}

// string des_ede_decrypt_cbc_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
static AbstractQoreNode* f_des_ede_decrypt_cbc_to_string_VyVtVtNs(const QoreListNode* args, ExceptionSink* xsink) {
# 498 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("DES-DECRYPT-PARAM-ERROR", args, xsink, true) 
       || ch.checkKeyLen("DES-KEY-ERROR", 0, 16, xsink)
       || ch.doCipher(EVP_des_ede_cbc(), "DES", QCRYPTO_DECRYPT, xsink))
      return 0;

   return ch.getString();
}

// binary des_ede_encrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_des_ede_encrypt_cbc_VtVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 441 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("DES-ENCRYPT-PARAM-ERROR", args, xsink) 
       || ch.checkKeyLen("DES-KEY-ERROR", 0, 16, xsink)
       || ch.doCipher(EVP_des_ede_cbc(), "DES", QCRYPTO_ENCRYPT, xsink))
      return 0;

   return ch.getBinary();
}

// binary des_encrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_des_encrypt_cbc_VtVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 357 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("DES-ENCRYPT-PARAM-ERROR", args, xsink) 
       || ch.checkKeyLen("DES-KEY-ERROR", 0, 8, xsink)
       || ch.doCipher(EVP_des_cbc(), "DES", QCRYPTO_ENCRYPT, xsink))
	    return 0;

   return ch.getBinary();
}

// binary des_random_key() {}
static AbstractQoreNode* f_des_random_key(const QoreListNode* args, ExceptionSink* xsink) {
# 257 "ql_crypto.qpp"
   DES_cblock *db = (DES_cblock *)malloc(sizeof(DES_cblock));
   DES_random_key(db);
   return new BinaryNode(db, sizeof(DES_cblock));
}

// binary desx_decrypt_cbc(binary data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_desx_decrypt_cbc_VyVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 636 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("DESX-DECRYPT-PARAM-ERROR", args, xsink) 
       || ch.checkKeyLen("DESX-KEY-ERROR", 0, 24, xsink)
       || ch.doCipher(EVP_desx_cbc(), "DESX", QCRYPTO_DECRYPT, xsink))
      return 0;

   return ch.getBinary();
}

// string desx_decrypt_cbc_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
static AbstractQoreNode* f_desx_decrypt_cbc_to_string_VyVtVtNs(const QoreListNode* args, ExceptionSink* xsink) {
# 666 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("DESX-DECRYPT-PARAM-ERROR", args, xsink, true)
       || ch.checkKeyLen("DESX-KEY-ERROR", 0, 24, xsink)
       || ch.doCipher(EVP_desx_cbc(), "DESX", QCRYPTO_DECRYPT, xsink))
      return 0;

   return ch.getString();
}

// binary desx_encrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_desx_encrypt_cbc_VtVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 609 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("DESX-ENCRYPT-PARAM-ERROR", args, xsink) 
       || ch.checkKeyLen("DESX-KEY-ERROR", 0, 24, xsink)
       || ch.doCipher(EVP_desx_cbc(), "DESX", QCRYPTO_ENCRYPT, xsink))
      return 0;

   return ch.getBinary();
}

// binary rc2_decrypt_cbc(binary data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_rc2_decrypt_cbc_VyVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 795 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("RC2-DECRYPT-PARAM-ERROR", args, xsink)
       || ch.doCipher(EVP_rc2_cbc(), "rc2", QCRYPTO_DECRYPT, xsink))
      return 0;

   return ch.getBinary();
}

// string rc2_decrypt_cbc_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
static AbstractQoreNode* f_rc2_decrypt_cbc_to_string_VyVtVtNs(const QoreListNode* args, ExceptionSink* xsink) {
# 823 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("RC2-DECRYPT-PARAM-ERROR", args, xsink, true)
       || ch.doCipher(EVP_rc2_cbc(), "rc2", QCRYPTO_DECRYPT, xsink))
      return 0;

   return ch.getString();
}

// binary rc2_encrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_rc2_encrypt_cbc_VtVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 770 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("RC2-ENCRYPT-PARAM-ERROR", args, xsink)
       || ch.doCipher(EVP_rc2_cbc(), "rc2", QCRYPTO_ENCRYPT, xsink))
      return 0;

   return ch.getBinary();
}

// binary rc4_decrypt(binary data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_rc4_decrypt_VyVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 717 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("RC4-DECRYPT-PARAM-ERROR", args, xsink)
       || ch.doCipher(EVP_rc4(), "rc4", QCRYPTO_DECRYPT, xsink))
      return 0;

   return ch.getBinary();
}

// string rc4_decrypt_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
static AbstractQoreNode* f_rc4_decrypt_to_string_VyVtVtNs(const QoreListNode* args, ExceptionSink* xsink) {
# 745 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("RC4-DECRYPT-PARAM-ERROR", args, xsink, true)
       || ch.doCipher(EVP_rc4(), "rc4", QCRYPTO_DECRYPT, xsink))
      return 0;

   return ch.getString();
}

// binary rc4_encrypt(data data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_rc4_encrypt_VtVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 692 "ql_crypto.qpp"
   CryptoHelper ch;

   if (ch.setSingleKey("RC4-ENCRYPT-PARAM-ERROR", args, xsink)
       || ch.doCipher(EVP_rc4(), "rc4", QCRYPTO_ENCRYPT, xsink))
      return 0;

   return ch.getBinary();
}

// binary rc5_decrypt_cbc(binary data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_rc5_decrypt_cbc_VyVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 962 "ql_crypto.qpp"
#ifndef OPENSSL_NO_RC5
   CryptoHelper ch;

   if (ch.setSingleKey("RC5-DECRYPT-PARAM-ERROR", args, xsink)
       || ch.doCipher(EVP_rc5_32_12_16_cbc(), "rc5", QCRYPTO_DECRYPT, xsink))
      return 0;

   return ch.getBinary();
#else
   missing_openssl_feature("RC5", xsink);
   return 0;
#endif
   return 0;
}

// string rc5_decrypt_cbc_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
static AbstractQoreNode* f_rc5_decrypt_cbc_to_string_VyVtVtNs(const QoreListNode* args, ExceptionSink* xsink) {
# 998 "ql_crypto.qpp"
#ifndef OPENSSL_NO_RC5
   CryptoHelper ch;

   if (ch.setSingleKey("RC5-DECRYPT-PARAM-ERROR", args, xsink, true)
       || ch.doCipher(EVP_rc5_32_12_16_cbc(), "rc5", QCRYPTO_DECRYPT, xsink))
      return 0;

   return ch.getString();
#else
   missing_openssl_feature("RC5", xsink);
   return 0;
#endif
   return 0;
}

// binary rc5_encrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
static AbstractQoreNode* f_rc5_encrypt_cbc_VtVtVt(const QoreListNode* args, ExceptionSink* xsink) {
# 929 "ql_crypto.qpp"
#ifndef OPENSSL_NO_RC5
   CryptoHelper ch;

   if (ch.setSingleKey("RC5-ENCRYPT-PARAM-ERROR", args, xsink)
       || ch.doCipher(EVP_rc5_32_12_16_cbc(), "rc5", QCRYPTO_ENCRYPT, xsink))
      return 0;

   return ch.getBinary();
#else
   missing_openssl_feature("RC5", xsink);
   return 0;
#endif
   return 0;
}

# 1016 "ql_crypto.qpp"
# 1036 "ql_crypto.qpp"
# 1067 "ql_crypto.qpp"
# 1090 "ql_crypto.qpp"
# 1117 "ql_crypto.qpp"
# 1141 "ql_crypto.qpp"
# 1166 "ql_crypto.qpp"
# 1199 "ql_crypto.qpp"
# 1232 "ql_crypto.qpp"
# 1265 "ql_crypto.qpp"
# 1298 "ql_crypto.qpp"
# 1323 "ql_crypto.qpp"
# 1348 "ql_crypto.qpp"
# 1379 "ql_crypto.qpp"
# 1402 "ql_crypto.qpp"
# 1433 "ql_crypto.qpp"
# 1456 "ql_crypto.qpp"
# 1479 "ql_crypto.qpp"
# 1503 "ql_crypto.qpp"
# 1526 "ql_crypto.qpp"
# 1557 "ql_crypto.qpp"
# 1588 "ql_crypto.qpp"
# 1619 "ql_crypto.qpp"
# 1650 "ql_crypto.qpp"
# 1675 "ql_crypto.qpp"
# 1700 "ql_crypto.qpp"
# 1731 "ql_crypto.qpp"
// string DSS(data data) {}
static AbstractQoreNode* f_DSS_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1315 "ql_crypto.qpp"
   DigestHelper dh(args);
   if (dh.doDigest(DSS_ERR, EVP_dss(), xsink))
      return 0;

   return dh.getString();
}

// string DSS1(data data) {}
static AbstractQoreNode* f_DSS1_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1340 "ql_crypto.qpp"
   DigestHelper dh(args);
   if (dh.doDigest(DSS1_ERR, EVP_dss1(), xsink))
      return 0;

   return dh.getString();
}

// binary DSS1_bin(data data) {}
static AbstractQoreNode* f_DSS1_bin_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1692 "ql_crypto.qpp"
   DigestHelper dh(args);
   if (dh.doDigest(DSS1_ERR, EVP_dss1(), xsink))
      return 0;
   
   return dh.getBinary();
}

// binary DSS_bin(data data) {}
static AbstractQoreNode* f_DSS_bin_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1667 "ql_crypto.qpp"
   DigestHelper dh(args);
   if (dh.doDigest(DSS_ERR, EVP_dss(), xsink))
      return 0;
   
   return dh.getBinary();
}

// string MD2(data data) {}
static AbstractQoreNode* f_MD2_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1054 "ql_crypto.qpp"
#if !defined(OPENSSL_NO_MD2) && !defined(NO_MD2)
   DigestHelper dh(args);
   if (dh.doDigest(MD2_ERR, EVP_md2(), xsink))
      return 0;

   return dh.getString();
#else
   missing_openssl_feature("MD2", xsink);
   return 0;
#endif
   return 0;
}

// binary MD2_bin(data data) {}
static AbstractQoreNode* f_MD2_bin_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1420 "ql_crypto.qpp"
#if !defined(OPENSSL_NO_MD2) && !defined(NO_MD2)
   DigestHelper dh(args);
   if (dh.doDigest(MD2_ERR, EVP_md2(), xsink))
      return 0;
   
   return dh.getBinary();
#else
   missing_openssl_feature("MD2", xsink);
   return 0;
#endif
   return 0;
}

// string MD4(data data) {}
static AbstractQoreNode* f_MD4_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1082 "ql_crypto.qpp"
   DigestHelper dh(args);
   if (dh.doDigest(MD4_ERR, EVP_md4(), xsink))
      return 0;

   return dh.getString();
}

// binary MD4_bin(data data) {}
static AbstractQoreNode* f_MD4_bin_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1448 "ql_crypto.qpp"
   DigestHelper dh(args);
   if (dh.doDigest(MD4_ERR, EVP_md4(), xsink))
      return 0;
   
   return dh.getBinary();
}

// string MD5(data data) {}
static AbstractQoreNode* f_MD5_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1109 "ql_crypto.qpp"
   DigestHelper dh(args);
   if (dh.doDigest(MD5_ERR, EVP_md5(), xsink))
      return 0;

   return dh.getString();
}

// binary MD5_bin(data data) {}
static AbstractQoreNode* f_MD5_bin_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1471 "ql_crypto.qpp"
   DigestHelper dh(args);
   if (dh.doDigest(MD5_ERR, EVP_md5(), xsink))
      return 0;
   
   return dh.getBinary();
}

// string MDC2(data data) {}
static AbstractQoreNode* f_MDC2_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1366 "ql_crypto.qpp"
#ifndef OPENSSL_NO_MDC2
   DigestHelper dh(args);
   if (dh.doDigest(MDC2_ERR, EVP_mdc2(), xsink))
      return 0;

   return dh.getString();
#else
   missing_openssl_feature("MDC2", xsink);
   return 0;
#endif
   return 0;
}

// binary MDC2_bin(data data) {}
static AbstractQoreNode* f_MDC2_bin_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1718 "ql_crypto.qpp"
#ifndef OPENSSL_NO_MDC2
   DigestHelper dh(args);
   if (dh.doDigest(MDC2_ERR, EVP_mdc2(), xsink))
      return 0;
   
   return dh.getBinary();
#else
   missing_openssl_feature("MDC2", xsink);
   return 0;
#endif
   return 0;
}

// string RIPEMD160(data data) {}
static AbstractQoreNode* f_RIPEMD160_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1394 "ql_crypto.qpp"
   DigestHelper dh(args);
   if (dh.doDigest(RIPEMD160_ERR, EVP_ripemd160(), xsink))
      return 0;

   return dh.getString();
}

// binary RIPEMD160_binary(data data) {}
static AbstractQoreNode* f_RIPEMD160_binary_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1746 "ql_crypto.qpp"
   DigestHelper dh(args);
   if (dh.doDigest(RIPEMD160_ERR, EVP_ripemd160(), xsink))
      return 0;
   
   return dh.getBinary();
}

// string SHA(data data) {}
static AbstractQoreNode* f_SHA_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1133 "ql_crypto.qpp"
   DigestHelper dh(args);
   if (dh.doDigest(SHA_ERR, EVP_sha(), xsink))
      return 0;

   return dh.getString();
}

// string SHA1(data data) {}
static AbstractQoreNode* f_SHA1_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1158 "ql_crypto.qpp"
   DigestHelper dh(args);
   if (dh.doDigest(SHA1_ERR, EVP_sha1(), xsink))
      return 0;

   return dh.getString();
}

// binary SHA1_bin(data data) {}
static AbstractQoreNode* f_SHA1_bin_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1518 "ql_crypto.qpp"
   DigestHelper dh(args);
   if (dh.doDigest(SHA1_ERR, EVP_sha1(), xsink))
      return 0;
   
   return dh.getBinary();
}

// string SHA224(data data) {}
static AbstractQoreNode* f_SHA224_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1186 "ql_crypto.qpp"
#if !defined(OPENSSL_NO_SHA256) && defined(HAVE_OPENSSL_SHA512)
   DigestHelper dh(args);
   if (dh.doDigest(SHA224_ERR, EVP_sha224(), xsink))
      return 0;

   return dh.getString();
#else
   missing_openssl_feature("SHA224", xsink);
   return 0;
#endif
   return 0;
}

// binary SHA224_bin(data data) {}
static AbstractQoreNode* f_SHA224_bin_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1544 "ql_crypto.qpp"
#if !defined(OPENSSL_NO_SHA256) && defined(HAVE_OPENSSL_SHA512)
   DigestHelper dh(args);
   if (dh.doDigest(SHA224_ERR, EVP_sha224(), xsink))
      return 0;
   
   return dh.getBinary();
#else
   missing_openssl_feature("SHA224", xsink);
   return 0;
#endif
   return 0;
}

// string SHA256(data data) {}
static AbstractQoreNode* f_SHA256_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1219 "ql_crypto.qpp"
#if !defined(OPENSSL_NO_SHA256) && defined(HAVE_OPENSSL_SHA512)
   DigestHelper dh(args);
   if (dh.doDigest(SHA256_ERR, EVP_sha256(), xsink))
      return 0;

   return dh.getString();
#else
   missing_openssl_feature("SHA256", xsink);
   return 0;
#endif
   return 0;
}

// binary SHA256_bin(data data) {}
static AbstractQoreNode* f_SHA256_bin_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1575 "ql_crypto.qpp"
#if !defined(OPENSSL_NO_SHA256) && defined(HAVE_OPENSSL_SHA512)
   DigestHelper dh(args);
   if (dh.doDigest(SHA256_ERR, EVP_sha256(), xsink))
      return 0;
   
   return dh.getBinary();
#else
   missing_openssl_feature("SHA256", xsink);
   return 0;
#endif
   return 0;
}

// string SHA384(data data) {}
static AbstractQoreNode* f_SHA384_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1252 "ql_crypto.qpp"
#if !defined(OPENSSL_NO_SHA512) && defined(HAVE_OPENSSL_SHA512)
   DigestHelper dh(args);
   if (dh.doDigest(SHA384_ERR, EVP_sha384(), xsink))
      return 0;

   return dh.getString();
#else
   missing_openssl_feature("SHA384", xsink);
   return 0;
#endif
   return 0;
}

// binary SHA384_bin(data data) {}
static AbstractQoreNode* f_SHA384_bin_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1606 "ql_crypto.qpp"
#if !defined(OPENSSL_NO_SHA512) && defined(HAVE_OPENSSL_SHA512)
   DigestHelper dh(args);
   if (dh.doDigest(SHA384_ERR, EVP_sha384(), xsink))
      return 0;
   
   return dh.getBinary();
#else
   missing_openssl_feature("SHA384", xsink);
   return 0;
#endif
   return 0;
}

// string SHA512(data data) {}
static AbstractQoreNode* f_SHA512_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1285 "ql_crypto.qpp"
#if !defined(OPENSSL_NO_SHA512) && defined(HAVE_OPENSSL_SHA512)
   DigestHelper dh(args);
   if (dh.doDigest(SHA512_ERR, EVP_sha512(), xsink))
      return 0;

   return dh.getString();
#else
   missing_openssl_feature("SHA512", xsink);
   return 0;
#endif
   return 0;
}

// binary SHA512_bin(data data) {}
static AbstractQoreNode* f_SHA512_bin_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1637 "ql_crypto.qpp"
#if !defined(OPENSSL_NO_SHA512) && defined(HAVE_OPENSSL_SHA512)
   DigestHelper dh(args);
   if (dh.doDigest(SHA512_ERR, EVP_sha512(), xsink))
      return 0;
   
   return dh.getBinary();
#else
   missing_openssl_feature("SHA512", xsink);
   return 0;
#endif
   return 0;
}

// binary SHA_bin(data data) {}
static AbstractQoreNode* f_SHA_bin_Vt(const QoreListNode* args, ExceptionSink* xsink) {
# 1495 "ql_crypto.qpp"
   DigestHelper dh(args);
   if (dh.doDigest(SHA_ERR, EVP_sha(), xsink))
      return 0;
   
   return dh.getBinary();
}

# 1756 "ql_crypto.qpp"
# 1812 "ql_crypto.qpp"
# 1835 "ql_crypto.qpp"
# 1860 "ql_crypto.qpp"
# 1885 "ql_crypto.qpp"
# 1910 "ql_crypto.qpp"
# 1943 "ql_crypto.qpp"
# 1975 "ql_crypto.qpp"
# 2007 "ql_crypto.qpp"
# 2039 "ql_crypto.qpp"
# 2064 "ql_crypto.qpp"
# 2089 "ql_crypto.qpp"
# 2118 "ql_crypto.qpp"
// string DSS1_hmac(data data, string key) {}
static AbstractQoreNode* f_DSS1_hmac_VtVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* key = HARD_QORE_STRING(args, 1);
# 2081 "ql_crypto.qpp"
   HMACHelper dh(args);
   if (dh.doHMAC("DSS1-HMAC-ERROR", EVP_dss1(), key, xsink))
      return 0;

   return dh.getString();
}

// string DSS_hmac(data data, string key) {}
static AbstractQoreNode* f_DSS_hmac_VtVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* key = HARD_QORE_STRING(args, 1);
# 2056 "ql_crypto.qpp"
   HMACHelper dh(args);
   if (dh.doHMAC("DSS-HMAC-ERROR", EVP_dss(), key, xsink))
      return 0;

   return dh.getString();
}

// string MD2_hmac(data data, string key) {}
static AbstractQoreNode* f_MD2_hmac_VtVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* key = HARD_QORE_STRING(args, 1);
# 1798 "ql_crypto.qpp"
#if !defined(OPENSSL_NO_MD2) && !defined(NO_MD2)
   HMACHelper dh(args);
   if (dh.doHMAC("MD2-HMAC-ERROR", EVP_md2(), key, xsink))
      return 0;

   return dh.getString();
#else
   (void)key;
   missing_openssl_feature("MD2", xsink);
   return 0;
#endif
   return 0;
}

// string MD4_hmac(data data, string key) {}
static AbstractQoreNode* f_MD4_hmac_VtVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* key = HARD_QORE_STRING(args, 1);
# 1827 "ql_crypto.qpp"
   HMACHelper dh(args);
   if (dh.doHMAC("MD4-HMAC-ERROR", EVP_md4(), key, xsink))
      return 0;

   return dh.getString();
}

// string MD5_hmac(data data, string key) {}
static AbstractQoreNode* f_MD5_hmac_VtVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* key = HARD_QORE_STRING(args, 1);
# 1852 "ql_crypto.qpp"
   HMACHelper dh(args);
   if (dh.doHMAC("MD5-HMAC-ERROR", EVP_md5(), key, xsink))
      return 0;

   return dh.getString();
}

// string MDC2_hmac(data data, string key) {}
static AbstractQoreNode* f_MDC2_hmac_VtVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* key = HARD_QORE_STRING(args, 1);
# 2104 "ql_crypto.qpp"
#ifndef OPENSSL_NO_MDC2
   HMACHelper dh(args);
   if (dh.doHMAC("MDC2-HMAC-ERROR", EVP_mdc2(), key, xsink))
      return 0;

   return dh.getString();
#else
   (void)key;
   missing_openssl_feature("MDC2", xsink);
   return 0;
#endif
   return 0;
}

// string RIPEMD160_hmac(data data, string key) {}
static AbstractQoreNode* f_RIPEMD160_hmac_VtVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* key = HARD_QORE_STRING(args, 1);
# 2133 "ql_crypto.qpp"
   HMACHelper dh(args);
   if (dh.doHMAC("RIPEMD160-HMAC-ERROR", EVP_ripemd160(), key, xsink))
      return 0;

   return dh.getString();
}

// string SHA1_hmac(data data, string key) {}
static AbstractQoreNode* f_SHA1_hmac_VtVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* key = HARD_QORE_STRING(args, 1);
# 1902 "ql_crypto.qpp"
   HMACHelper dh(args);
   if (dh.doHMAC("SHA1-HMAC-ERROR", EVP_sha1(), key, xsink))
      return 0;

   return dh.getString();
}

// string SHA224_hmac(data data, string key) {}
static AbstractQoreNode* f_SHA224_hmac_VtVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* key = HARD_QORE_STRING(args, 1);
# 1929 "ql_crypto.qpp"
#if !defined(OPENSSL_NO_SHA256) && defined(HAVE_OPENSSL_SHA512)
   HMACHelper dh(args);
   if (dh.doHMAC("SHA224-HMAC-ERROR", EVP_sha224(), key, xsink))
      return 0;

   return dh.getString();
#else
   (void)key;
   missing_openssl_feature("SHA224", xsink);
   return 0;
#endif
   return 0;
}

// string SHA256_hmac(data data, string key) {}
static AbstractQoreNode* f_SHA256_hmac_VtVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* key = HARD_QORE_STRING(args, 1);
# 1961 "ql_crypto.qpp"
#if !defined(OPENSSL_NO_SHA256) && defined(HAVE_OPENSSL_SHA512)
   HMACHelper dh(args);
   if (dh.doHMAC("SHA256-HMAC-ERROR", EVP_sha256(), key, xsink))
      return 0;

   return dh.getString();
#else
   (void)key;
   missing_openssl_feature("SHA256", xsink);
   return 0;
#endif
   return 0;
}

// string SHA384_hmac(data data, string key) {}
static AbstractQoreNode* f_SHA384_hmac_VtVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* key = HARD_QORE_STRING(args, 1);
# 1993 "ql_crypto.qpp"
#if !defined(OPENSSL_NO_SHA512) && defined(HAVE_OPENSSL_SHA512)
   HMACHelper dh(args);
   if (dh.doHMAC("SHA384-HMAC-ERROR", EVP_sha384(), key, xsink))
      return 0;

   return dh.getString();
#else
   (void)key;
   missing_openssl_feature("SHA384", xsink);
   return 0;
#endif
   return 0;
}

// string SHA512_hmac(data data, string key) {}
static AbstractQoreNode* f_SHA512_hmac_VtVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* key = HARD_QORE_STRING(args, 1);
# 2025 "ql_crypto.qpp"
#if !defined(OPENSSL_NO_SHA512) && defined(HAVE_OPENSSL_SHA512)
   HMACHelper dh(args);
   if (dh.doHMAC("SHA512-HMAC-ERROR", EVP_sha512(), key, xsink))
      return 0;

   return dh.getString();
#else
   (void)key;
   missing_openssl_feature("SHA512", xsink);
   return 0;
#endif
   return 0;
}

// string SHA_hmac(data data, string key) {}
static AbstractQoreNode* f_SHA_hmac_VtVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* key = HARD_QORE_STRING(args, 1);
# 1877 "ql_crypto.qpp"
   HMACHelper dh(args);
   if (dh.doHMAC("SHA-HMAC-ERROR", EVP_sha(), key, xsink))
      return 0;

   return dh.getString();
}


DLLLOCAL void init_crypto_functions(QoreNamespace& ns) {

   // initialization code added by qppinit
   SimpleRefHolder<BinaryNode> DefaultIV(new BinaryNode); DefaultIV->append(def_iv, 8);
   // binary blowfish_decrypt_cbc(binary data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("blowfish_decrypt_cbc", (q_func_t)f_blowfish_decrypt_cbc_VyVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // string blowfish_decrypt_cbc_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
   ns.addBuiltinVariant("blowfish_decrypt_cbc_to_string", (q_func_t)f_blowfish_decrypt_cbc_to_string_VyVtVtNs, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 4, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv", stringOrNothingTypeInfo, NULL, "encoding");
   // binary blowfish_encrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("blowfish_encrypt_cbc", (q_func_t)f_blowfish_encrypt_cbc_VtVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, dataTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // binary cast5_decrypt_cbc(binary data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("cast5_decrypt_cbc", (q_func_t)f_cast5_decrypt_cbc_VyVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // string cast5_decrypt_cbc_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
   ns.addBuiltinVariant("cast5_decrypt_cbc_to_string", (q_func_t)f_cast5_decrypt_cbc_to_string_VyVtVtNs, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 4, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv", stringOrNothingTypeInfo, NULL, "encoding");
   // binary cast5_encrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("cast5_encrypt_cbc", (q_func_t)f_cast5_encrypt_cbc_VtVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, dataTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // binary des_decrypt_cbc(binary data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("des_decrypt_cbc", (q_func_t)f_des_decrypt_cbc_VyVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // string des_decrypt_cbc_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
   ns.addBuiltinVariant("des_decrypt_cbc_to_string", (q_func_t)f_des_decrypt_cbc_to_string_VyVtVtNs, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 4, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv", stringOrNothingTypeInfo, NULL, "encoding");
   // binary des_ede3_decrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("des_ede3_decrypt_cbc", (q_func_t)f_des_ede3_decrypt_cbc_VtVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, dataTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // string des_ede3_decrypt_cbc_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
   ns.addBuiltinVariant("des_ede3_decrypt_cbc_to_string", (q_func_t)f_des_ede3_decrypt_cbc_to_string_VyVtVtNs, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 4, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv", stringOrNothingTypeInfo, NULL, "encoding");
   // binary des_ede3_encrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("des_ede3_encrypt_cbc", (q_func_t)f_des_ede3_encrypt_cbc_VtVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, dataTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // binary des_ede_decrypt_cbc(binary data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("des_ede_decrypt_cbc", (q_func_t)f_des_ede_decrypt_cbc_VyVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // string des_ede_decrypt_cbc_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
   ns.addBuiltinVariant("des_ede_decrypt_cbc_to_string", (q_func_t)f_des_ede_decrypt_cbc_to_string_VyVtVtNs, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 4, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv", stringOrNothingTypeInfo, NULL, "encoding");
   // binary des_ede_encrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("des_ede_encrypt_cbc", (q_func_t)f_des_ede_encrypt_cbc_VtVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, dataTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // binary des_encrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("des_encrypt_cbc", (q_func_t)f_des_encrypt_cbc_VtVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, dataTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // binary des_random_key() {}
   ns.addBuiltinVariant("des_random_key", (q_func_t)f_des_random_key, QC_CONSTANT, QDOM_DEFAULT, binaryTypeInfo);
   // binary desx_decrypt_cbc(binary data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("desx_decrypt_cbc", (q_func_t)f_desx_decrypt_cbc_VyVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // string desx_decrypt_cbc_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
   ns.addBuiltinVariant("desx_decrypt_cbc_to_string", (q_func_t)f_desx_decrypt_cbc_to_string_VyVtVtNs, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 4, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv", stringOrNothingTypeInfo, NULL, "encoding");
   // binary desx_encrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("desx_encrypt_cbc", (q_func_t)f_desx_encrypt_cbc_VtVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, dataTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // binary rc2_decrypt_cbc(binary data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("rc2_decrypt_cbc", (q_func_t)f_rc2_decrypt_cbc_VyVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // string rc2_decrypt_cbc_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
   ns.addBuiltinVariant("rc2_decrypt_cbc_to_string", (q_func_t)f_rc2_decrypt_cbc_to_string_VyVtVtNs, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 4, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv", stringOrNothingTypeInfo, NULL, "encoding");
   // binary rc2_encrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("rc2_encrypt_cbc", (q_func_t)f_rc2_encrypt_cbc_VtVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, dataTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // binary rc4_decrypt(binary data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("rc4_decrypt", (q_func_t)f_rc4_decrypt_VyVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // string rc4_decrypt_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
   ns.addBuiltinVariant("rc4_decrypt_to_string", (q_func_t)f_rc4_decrypt_to_string_VyVtVtNs, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 4, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv", stringOrNothingTypeInfo, NULL, "encoding");
   // binary rc4_encrypt(data data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("rc4_encrypt", (q_func_t)f_rc4_encrypt_VtVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, dataTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // binary rc5_decrypt_cbc(binary data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("rc5_decrypt_cbc", (q_func_t)f_rc5_decrypt_cbc_VyVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // string rc5_decrypt_cbc_to_string(binary data, data key, data iv = Qore::DefaultIV, *string encoding) {}
   ns.addBuiltinVariant("rc5_decrypt_cbc_to_string", (q_func_t)f_rc5_decrypt_cbc_to_string_VyVtVtNs, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 4, binaryTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv", stringOrNothingTypeInfo, NULL, "encoding");
   // binary rc5_encrypt_cbc(data data, data key, data iv = Qore::DefaultIV) {}
   ns.addBuiltinVariant("rc5_encrypt_cbc", (q_func_t)f_rc5_encrypt_cbc_VtVtVt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 3, dataTypeInfo, NULL, "data", dataTypeInfo, NULL, "key", dataTypeInfo, DefaultIV->refSelf(), "iv");
   // string DSS(data data) {}
   ns.addBuiltinVariant("DSS", (q_func_t)f_DSS_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 1, dataTypeInfo, NULL, "data");
   // string DSS1(data data) {}
   ns.addBuiltinVariant("DSS1", (q_func_t)f_DSS1_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 1, dataTypeInfo, NULL, "data");
   // binary DSS1_bin(data data) {}
   ns.addBuiltinVariant("DSS1_bin", (q_func_t)f_DSS1_bin_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 1, dataTypeInfo, NULL, "data");
   // binary DSS_bin(data data) {}
   ns.addBuiltinVariant("DSS_bin", (q_func_t)f_DSS_bin_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 1, dataTypeInfo, NULL, "data");
   // string MD2(data data) {}
   ns.addBuiltinVariant("MD2", (q_func_t)f_MD2_Vt, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 1, dataTypeInfo, NULL, "data");
   // binary MD2_bin(data data) {}
   ns.addBuiltinVariant("MD2_bin", (q_func_t)f_MD2_bin_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 1, dataTypeInfo, NULL, "data");
   // string MD4(data data) {}
   ns.addBuiltinVariant("MD4", (q_func_t)f_MD4_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 1, dataTypeInfo, NULL, "data");
   // binary MD4_bin(data data) {}
   ns.addBuiltinVariant("MD4_bin", (q_func_t)f_MD4_bin_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 1, dataTypeInfo, NULL, "data");
   // string MD5(data data) {}
   ns.addBuiltinVariant("MD5", (q_func_t)f_MD5_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 1, dataTypeInfo, NULL, "data");
   // binary MD5_bin(data data) {}
   ns.addBuiltinVariant("MD5_bin", (q_func_t)f_MD5_bin_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 1, dataTypeInfo, NULL, "data");
   // string MDC2(data data) {}
   ns.addBuiltinVariant("MDC2", (q_func_t)f_MDC2_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 1, dataTypeInfo, NULL, "data");
   // binary MDC2_bin(data data) {}
   ns.addBuiltinVariant("MDC2_bin", (q_func_t)f_MDC2_bin_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 1, dataTypeInfo, NULL, "data");
   // string RIPEMD160(data data) {}
   ns.addBuiltinVariant("RIPEMD160", (q_func_t)f_RIPEMD160_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 1, dataTypeInfo, NULL, "data");
   // binary RIPEMD160_binary(data data) {}
   ns.addBuiltinVariant("RIPEMD160_binary", (q_func_t)f_RIPEMD160_binary_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 1, dataTypeInfo, NULL, "data");
   // string SHA(data data) {}
   ns.addBuiltinVariant("SHA", (q_func_t)f_SHA_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 1, dataTypeInfo, NULL, "data");
   // string SHA1(data data) {}
   ns.addBuiltinVariant("SHA1", (q_func_t)f_SHA1_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 1, dataTypeInfo, NULL, "data");
   // binary SHA1_bin(data data) {}
   ns.addBuiltinVariant("SHA1_bin", (q_func_t)f_SHA1_bin_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 1, dataTypeInfo, NULL, "data");
   // string SHA224(data data) {}
   ns.addBuiltinVariant("SHA224", (q_func_t)f_SHA224_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 1, dataTypeInfo, NULL, "data");
   // binary SHA224_bin(data data) {}
   ns.addBuiltinVariant("SHA224_bin", (q_func_t)f_SHA224_bin_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 1, dataTypeInfo, NULL, "data");
   // string SHA256(data data) {}
   ns.addBuiltinVariant("SHA256", (q_func_t)f_SHA256_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 1, dataTypeInfo, NULL, "data");
   // binary SHA256_bin(data data) {}
   ns.addBuiltinVariant("SHA256_bin", (q_func_t)f_SHA256_bin_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 1, dataTypeInfo, NULL, "data");
   // string SHA384(data data) {}
   ns.addBuiltinVariant("SHA384", (q_func_t)f_SHA384_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 1, dataTypeInfo, NULL, "data");
   // binary SHA384_bin(data data) {}
   ns.addBuiltinVariant("SHA384_bin", (q_func_t)f_SHA384_bin_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 1, dataTypeInfo, NULL, "data");
   // string SHA512(data data) {}
   ns.addBuiltinVariant("SHA512", (q_func_t)f_SHA512_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 1, dataTypeInfo, NULL, "data");
   // binary SHA512_bin(data data) {}
   ns.addBuiltinVariant("SHA512_bin", (q_func_t)f_SHA512_bin_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 1, dataTypeInfo, NULL, "data");
   // binary SHA_bin(data data) {}
   ns.addBuiltinVariant("SHA_bin", (q_func_t)f_SHA_bin_Vt, QC_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 1, dataTypeInfo, NULL, "data");
   // string DSS1_hmac(data data, string key) {}
   ns.addBuiltinVariant("DSS1_hmac", (q_func_t)f_DSS1_hmac_VtVs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, dataTypeInfo, NULL, "data", stringTypeInfo, NULL, "key");
   // string DSS_hmac(data data, string key) {}
   ns.addBuiltinVariant("DSS_hmac", (q_func_t)f_DSS_hmac_VtVs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, dataTypeInfo, NULL, "data", stringTypeInfo, NULL, "key");
   // string MD2_hmac(data data, string key) {}
   ns.addBuiltinVariant("MD2_hmac", (q_func_t)f_MD2_hmac_VtVs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, dataTypeInfo, NULL, "data", stringTypeInfo, NULL, "key");
   // string MD4_hmac(data data, string key) {}
   ns.addBuiltinVariant("MD4_hmac", (q_func_t)f_MD4_hmac_VtVs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, dataTypeInfo, NULL, "data", stringTypeInfo, NULL, "key");
   // string MD5_hmac(data data, string key) {}
   ns.addBuiltinVariant("MD5_hmac", (q_func_t)f_MD5_hmac_VtVs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, dataTypeInfo, NULL, "data", stringTypeInfo, NULL, "key");
   // string MDC2_hmac(data data, string key) {}
   ns.addBuiltinVariant("MDC2_hmac", (q_func_t)f_MDC2_hmac_VtVs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, dataTypeInfo, NULL, "data", stringTypeInfo, NULL, "key");
   // string RIPEMD160_hmac(data data, string key) {}
   ns.addBuiltinVariant("RIPEMD160_hmac", (q_func_t)f_RIPEMD160_hmac_VtVs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, dataTypeInfo, NULL, "data", stringTypeInfo, NULL, "key");
   // string SHA1_hmac(data data, string key) {}
   ns.addBuiltinVariant("SHA1_hmac", (q_func_t)f_SHA1_hmac_VtVs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, dataTypeInfo, NULL, "data", stringTypeInfo, NULL, "key");
   // string SHA224_hmac(data data, string key) {}
   ns.addBuiltinVariant("SHA224_hmac", (q_func_t)f_SHA224_hmac_VtVs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, dataTypeInfo, NULL, "data", stringTypeInfo, NULL, "key");
   // string SHA256_hmac(data data, string key) {}
   ns.addBuiltinVariant("SHA256_hmac", (q_func_t)f_SHA256_hmac_VtVs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, dataTypeInfo, NULL, "data", stringTypeInfo, NULL, "key");
   // string SHA384_hmac(data data, string key) {}
   ns.addBuiltinVariant("SHA384_hmac", (q_func_t)f_SHA384_hmac_VtVs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, dataTypeInfo, NULL, "data", stringTypeInfo, NULL, "key");
   // string SHA512_hmac(data data, string key) {}
   ns.addBuiltinVariant("SHA512_hmac", (q_func_t)f_SHA512_hmac_VtVs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, dataTypeInfo, NULL, "data", stringTypeInfo, NULL, "key");
   // string SHA_hmac(data data, string key) {}
   ns.addBuiltinVariant("SHA_hmac", (q_func_t)f_SHA_hmac_VtVs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, dataTypeInfo, NULL, "data", stringTypeInfo, NULL, "key");

}

DLLLOCAL void init_crypto_constants(QoreNamespace& ns) {
   BinaryNode* bin_DefaultIV = new BinaryNode;
   bin_DefaultIV->append("\x00\x00\x00\x00\x00\x00\x00\x00", 8);

   ns.addConstant("DefaultIV", bin_DefaultIV);
# 222 "ql_crypto.qpp"
# 249 "ql_crypto.qpp"
# 263 "ql_crypto.qpp"
# 288 "ql_crypto.qpp"
# 313 "ql_crypto.qpp"
# 341 "ql_crypto.qpp"
# 368 "ql_crypto.qpp"
# 395 "ql_crypto.qpp"
# 425 "ql_crypto.qpp"
# 452 "ql_crypto.qpp"
# 479 "ql_crypto.qpp"
# 509 "ql_crypto.qpp"
# 536 "ql_crypto.qpp"
# 563 "ql_crypto.qpp"
# 593 "ql_crypto.qpp"
# 620 "ql_crypto.qpp"
# 647 "ql_crypto.qpp"
# 677 "ql_crypto.qpp"
# 702 "ql_crypto.qpp"
# 727 "ql_crypto.qpp"
# 755 "ql_crypto.qpp"
# 780 "ql_crypto.qpp"
# 805 "ql_crypto.qpp"
# 833 "ql_crypto.qpp"
# 858 "ql_crypto.qpp"
# 883 "ql_crypto.qpp"
# 911 "ql_crypto.qpp"
# 944 "ql_crypto.qpp"
# 977 "ql_crypto.qpp"
# 1016 "ql_crypto.qpp"
# 1036 "ql_crypto.qpp"
# 1067 "ql_crypto.qpp"
# 1090 "ql_crypto.qpp"
# 1117 "ql_crypto.qpp"
# 1141 "ql_crypto.qpp"
# 1166 "ql_crypto.qpp"
# 1199 "ql_crypto.qpp"
# 1232 "ql_crypto.qpp"
# 1265 "ql_crypto.qpp"
# 1298 "ql_crypto.qpp"
# 1323 "ql_crypto.qpp"
# 1348 "ql_crypto.qpp"
# 1379 "ql_crypto.qpp"
# 1402 "ql_crypto.qpp"
# 1433 "ql_crypto.qpp"
# 1456 "ql_crypto.qpp"
# 1479 "ql_crypto.qpp"
# 1503 "ql_crypto.qpp"
# 1526 "ql_crypto.qpp"
# 1557 "ql_crypto.qpp"
# 1588 "ql_crypto.qpp"
# 1619 "ql_crypto.qpp"
# 1650 "ql_crypto.qpp"
# 1675 "ql_crypto.qpp"
# 1700 "ql_crypto.qpp"
# 1731 "ql_crypto.qpp"
# 1756 "ql_crypto.qpp"
# 1812 "ql_crypto.qpp"
# 1835 "ql_crypto.qpp"
# 1860 "ql_crypto.qpp"
# 1885 "ql_crypto.qpp"
# 1910 "ql_crypto.qpp"
# 1943 "ql_crypto.qpp"
# 1975 "ql_crypto.qpp"
# 2007 "ql_crypto.qpp"
# 2039 "ql_crypto.qpp"
# 2064 "ql_crypto.qpp"
# 2089 "ql_crypto.qpp"
# 2118 "ql_crypto.qpp"

}
