/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
 ql_file.qpp
 
 Qore Programming Language
 
 Copyright 2003 - 2013 David Nichols
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <qore/Qore.h>
#include <qore/intern/ql_file.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#ifdef HAVE_GLOB_H
#include <glob.h>
#else
#include <qore/intern/glob.h>
#endif

bool check_stat(unsigned code, const QoreStringNode* path) {
   struct stat sbuf;
   int rc;
   
   if ((rc = stat(path->getBuffer(), &sbuf)))
      return false;

   return (sbuf.st_mode & S_IFMT) == code;
}

#ifdef HAVE_LSTAT
bool check_lstat(unsigned code, const QoreStringNode* path) {
   struct stat sbuf;
   int rc;
   
   if ((rc = lstat(path->getBuffer(), &sbuf)))
      return false;
   
   return (sbuf.st_mode & S_IFMT) == code;
}
#endif

bool is_writable_intern(const QoreStringNode* path) {
#ifdef HAVE_ACCESS
   // solution for situation like:
   // $ id
   // uid=5920(user1) gid=6183(group1) groups=6003(group2),6183(group3)
   // $ ls -la /foo/bar/
   // drwxrwsr-x 2  5921 group2 4096 Aug  7 14:37 .
   // where 6183(group1) is an effective group etc so geteuid() != dir group
   return access(path->getBuffer(), W_OK) == 0;

#elif defined HAVE_PWD_H
   struct stat sbuf;
   int rc;
   
   if ((rc = stat(path->getBuffer(), &sbuf)))
      return false;
   
   uid_t euid = geteuid();
   if (!euid || sbuf.st_mode & S_IWOTH 
       || (euid      == sbuf.st_uid && (sbuf.st_mode & S_IWUSR)) 
       || (getegid() == sbuf.st_gid && (sbuf.st_mode & S_IWGRP)))
      return true;
   
   return false;
#else
   // try to open the file with write permissions
   int rc = open(path->getBuffer(), O_WRONLY);
   if (rc == -1)
      return false;
   close(rc);
   return true;
#endif
}
# 95 "ql_file.qpp"
# 166 "ql_file.qpp"
# 181 "ql_file.qpp"
# 203 "ql_file.qpp"
# 218 "ql_file.qpp"
# 233 "ql_file.qpp"
# 248 "ql_file.qpp"
# 273 "ql_file.qpp"
# 295 "ql_file.qpp"
# 312 "ql_file.qpp"
# 329 "ql_file.qpp"
# 344 "ql_file.qpp"
# 377 "ql_file.qpp"
# 411 "ql_file.qpp"
# 439 "ql_file.qpp"
# 463 "ql_file.qpp"
# 469 "ql_file.qpp"
# 498 "ql_file.qpp"
# 504 "ql_file.qpp"
# 528 "ql_file.qpp"
# 534 "ql_file.qpp"
# 562 "ql_file.qpp"
# 568 "ql_file.qpp"
# 596 "ql_file.qpp"
# 602 "ql_file.qpp"
# 619 "ql_file.qpp"
# 625 "ql_file.qpp"
# 640 "ql_file.qpp"
# 646 "ql_file.qpp"
# 664 "ql_file.qpp"
# 679 "ql_file.qpp"
# 695 "ql_file.qpp"
static QoreStringNode *f_getcwd_intern(ExceptionSink *xsink = 0) {
   int bs = 512;
   char *buf = (char *)malloc(sizeof(char) * bs);

   while (true) {
      char *b = getcwd(buf, bs);
      if (!b) {
          if (errno == ERANGE) {
              bs *= 2;
              buf = (char *)q_realloc(buf, sizeof(char) * bs);
              if (!buf) {
                  xsink->outOfMemory();
                  return 0;
              }
              continue;
          }
          if (xsink)
             xsink->raiseErrnoException("GETCWD2-ERROR", errno, "getcwd() failed");
          return 0;
      }
      break;
   }

   return new QoreStringNode(buf, strlen(buf), bs, QCS_DEFAULT);
}
# 739 "ql_file.qpp"
# 753 "ql_file.qpp"
# 766 "ql_file.qpp"
# 790 "ql_file.qpp"
# 817 "ql_file.qpp"
# 844 "ql_file.qpp"
# 876 "ql_file.qpp"
// int chdir(string path) {}
static int64 f_chdir_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 709 "ql_file.qpp"
   return chdir(path->getBuffer());
}

// int chmod(string path, softint mode) {}
static int64 f_chmod_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
   int64 mode = HARD_QORE_INT(args, 1);
# 691 "ql_file.qpp"
   return chmod(path->getBuffer(), (int)mode);
}

// int chown(string path, softint owner = -1, softint group = -1) {}
static int64 f_chown_Vsvivi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
   int64 owner = HARD_QORE_INT(args, 1);
   int64 group = HARD_QORE_INT(args, 2);
# 808 "ql_file.qpp"
#ifdef HAVE_CHOWN
   return chown(path->getBuffer(), (uid_t)owner, (uid_t)group);
#else
   missing_function_error("chown", "UNIX_FILEMGT", xsink);
   return 0;
#endif
   return 0;
}

// string getcwd() {}
static AbstractQoreNode* f_getcwd(const QoreListNode* args, ExceptionSink* xsink) {
# 749 "ql_file.qpp"
   return f_getcwd_intern();
}

// string getcwd2() {}
static AbstractQoreNode* f_getcwd2(const QoreListNode* args, ExceptionSink* xsink) {
# 762 "ql_file.qpp"
   return f_getcwd_intern(xsink);
}

// *list glob(string glob_str) {}
static AbstractQoreNode* f_glob_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* glob_str = HARD_QORE_STRING(args, 0);
# 581 "ql_file.qpp"
   glob_t globbuf;
   if (glob(glob_str->getBuffer(), 0, 0, &globbuf)) {
      globfree(&globbuf);
      return 0;
   }

   QoreListNode *l = new QoreListNode;
   for (unsigned i = 0; i < globbuf.gl_pathc; ++i)
      l->push(new QoreStringNode(globbuf.gl_pathv[i]));
   
   globfree(&globbuf);
   return l;
}

// nothing glob() {}
static AbstractQoreNode* f_glob(const QoreListNode* args, ExceptionSink* xsink) {
# 599 "ql_file.qpp"

   return 0;
}

// *hash hlstat(string path) {}
static AbstractQoreNode* f_hlstat_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 550 "ql_file.qpp"
   struct stat sbuf;
#ifdef HAVE_LSTAT
   if (lstat(path->getBuffer(), &sbuf))
#else
   if (stat(path->getBuffer(), &sbuf))
#endif
      return 0;

   return stat_to_hash(sbuf);
}

// nothing hlstat() {}
static AbstractQoreNode* f_hlstat(const QoreListNode* args, ExceptionSink* xsink) {
# 565 "ql_file.qpp"

   return 0;
}

// *hash hstat(string path) {}
static AbstractQoreNode* f_hstat_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 520 "ql_file.qpp"
   struct stat sbuf;
   if (stat(path->getBuffer(), &sbuf))
      return 0;

   return stat_to_hash(sbuf);
}

// nothing hstat() {}
static AbstractQoreNode* f_hstat(const QoreListNode* args, ExceptionSink* xsink) {
# 531 "ql_file.qpp"

   return 0;
}

// bool is_bdev(string path) {}
static bool f_is_bdev_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 244 "ql_file.qpp"
   return check_stat(S_IFBLK, path);
}

// bool is_cdev(string path) {}
static bool f_is_cdev_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 229 "ql_file.qpp"
   return check_stat(S_IFCHR, path);
}

// bool is_dev(string path) {}
static bool f_is_dev_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 259 "ql_file.qpp"
   struct stat sbuf;
   int rc;
   
#ifdef HAVE_LSTAT
   if ((rc = lstat(path->getBuffer(), &sbuf)))
#else
   if ((rc = stat(path->getBuffer(), &sbuf)))
#endif
      return false;
   
   return ((sbuf.st_mode & S_IFMT) == S_IFCHR) || ((sbuf.st_mode & S_IFMT) == S_IFBLK);
}

// bool is_dir(string path) {}
static bool f_is_dir_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 177 "ql_file.qpp"
   return check_stat(S_IFDIR, path);
}

// bool is_executable(string path) {}
static bool f_is_executable_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 357 "ql_file.qpp"
#ifdef HAVE_PWD_H
   struct stat sbuf;
   int rc;
   
   if ((rc = stat(path->getBuffer(), &sbuf)))
      return 0;
   
   if (sbuf.st_mode & S_IXOTH 
       || (geteuid() == sbuf.st_uid && (sbuf.st_mode & S_IXUSR)) 
       || (getegid() == sbuf.st_gid && (sbuf.st_mode & S_IXGRP)))
      return true;
   
   return false;
#else
   missing_function_error("is_executable", xsink);
   return false;
#endif
   return 0;
}

// bool is_file(string path) {}
static bool f_is_file_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 162 "ql_file.qpp"
   return check_stat(S_IFREG, path);
}

// bool is_link(string path) {}
static bool f_is_link_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 286 "ql_file.qpp"
#ifdef HAVE_LSTAT
   return check_lstat(S_IFLNK, path);
#else
   missing_function_error("is_link", "UNIX_FILEMGT", xsink);
   return false;
#endif
   return 0;
}

// bool is_pipe(string path) {}
static bool f_is_pipe_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 214 "ql_file.qpp"
   return check_stat(S_IFIFO, path);
}

// bool is_readable(string path) {}
static bool f_is_readable_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 308 "ql_file.qpp"
   return q_path_is_readable(path->getBuffer());
}

// bool is_socket(string path) {}
static bool f_is_socket_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 194 "ql_file.qpp"
#ifdef S_IFSOCK
   return check_stat(S_IFSOCK, path);
#else
   missing_function_error("is_socket", "UNIX_FILEMGT", xsink);
   return false;
#endif
   return 0;
}

// bool is_writable(string path) {}
static bool f_is_writable_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 325 "ql_file.qpp"
   return is_writable_intern(path);
}

// bool is_writeable(string path) {}
static bool f_is_writeable_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 340 "ql_file.qpp"
   return is_writable_intern(path);
}

// int lchown(string path, softint uid = -1, softint gid = -1) {}
static int64 f_lchown_Vsvivi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
   int64 uid = HARD_QORE_INT(args, 1);
   int64 gid = HARD_QORE_INT(args, 2);
# 835 "ql_file.qpp"
#ifdef HAVE_LCHOWN
   return lchown(path->getBuffer(), (uid_t)uid, (uid_t)gid);
#else
   missing_function_error("lchown", "UNIX_FILEMGT", xsink);
   return 0;
#endif
   return 0;
}

// *list lstat(string path) {}
static AbstractQoreNode* f_lstat_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 486 "ql_file.qpp"
   struct stat sbuf;
#ifdef HAVE_LSTAT
   if (lstat(path->getBuffer(), &sbuf))
#else
   if (stat(path->getBuffer(), &sbuf))
#endif
      return 0;

   return stat_to_list(sbuf);
}

// nothing lstat() {}
static AbstractQoreNode* f_lstat(const QoreListNode* args, ExceptionSink* xsink) {
# 501 "ql_file.qpp"

   return 0;
}

// int mkdir(string path, softint mode = 0777) {}
static int64 f_mkdir_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
   int64 mode = HARD_QORE_INT(args, 1);
# 660 "ql_file.qpp"
   return mkdir(path->getBuffer(), mode);
}

// int mkfifo(string path, softint mode = 0600) {}
static int64 f_mkfifo_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
   int64 mode = HARD_QORE_INT(args, 1);
# 781 "ql_file.qpp"
#ifdef HAVE_MKFIFO
   return mkfifo(path->getBuffer(), mode);
#else
   missing_function_error("mkfifo", "UNIX_FILEMGT", xsink);
   return 0;
#endif
   return 0;
}

// string readlink(string path) {}
static AbstractQoreNode* f_readlink_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 859 "ql_file.qpp"
#ifdef HAVE_READLINK
   char buf[QORE_PATH_MAX + 1];
   qore_offset_t len = readlink(path->getBuffer(), buf, QORE_PATH_MAX);
   if (len < 0) {
      xsink->raiseErrnoException("READLINK-ERROR", errno, "readlink('%s') failed", path->getBuffer());
      return 0;
   }
   assert(len <= QORE_PATH_MAX);
   buf[len] = '\0';
   return new QoreStringNode(buf);
#else
   missing_function_error("readlink", "UNIX_FILEMGT", xsink);
   return 0;
#endif
   return 0;
}

// nothing rename(string old_path, string new_path) {}
static AbstractQoreNode* f_rename_VsVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* old_path = HARD_QORE_STRING(args, 0);
   const QoreStringNode* new_path = HARD_QORE_STRING(args, 1);
# 392 "ql_file.qpp"
   if (old_path->empty()) {
      xsink->raiseException("RENAME-ERROR", "empty path to current file name given as first argument");
      return 0;
   }

   // new file name
   if (new_path->empty()) {
      xsink->raiseException("RENAME-ERROR", "empty new file path given as second argument");
      return 0;
   }

   int rc = rename(old_path->getBuffer(), new_path->getBuffer());
   if (rc)
      xsink->raiseErrnoException("RENAME-ERROR", errno, "rename('%s', '%s') failed", old_path->getBuffer(), new_path->getBuffer());

   return 0;
}

// int rmdir(string path) {}
static int64 f_rmdir_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 675 "ql_file.qpp"
   return rmdir(path->getBuffer());
}

// *list stat(string path) {}
static AbstractQoreNode* f_stat_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 455 "ql_file.qpp"
   struct stat sbuf;
   if (stat(path->getBuffer(), &sbuf))
      return 0;

   return stat_to_list(sbuf);
}

// nothing stat() {}
static AbstractQoreNode* f_stat(const QoreListNode* args, ExceptionSink* xsink) {
# 466 "ql_file.qpp"

   return 0;
}

// *hash statvfs(string path) {}
static AbstractQoreNode* f_statvfs_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 426 "ql_file.qpp"
#ifdef HAVE_SYS_STATVFS_H
   struct statvfs vfs;
   if (statvfs(path->getBuffer(), &vfs))
      return 0;
   
   return statvfs_to_hash(vfs);
#else
   missing_function_error("statvfs", xsink);
   return 0;
#endif
   return 0;
}

// nothing symlink(string old_path, string new_path) {}
static AbstractQoreNode* f_symlink_VsVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* old_path = HARD_QORE_STRING(args, 0);
   const QoreStringNode* new_path = HARD_QORE_STRING(args, 1);
# 895 "ql_file.qpp"
#ifdef HAVE_SYMLINK
   if (old_path->empty()) {
      xsink->raiseException("SYMLINK-ERROR", "empty path to current file name given as first argument");
      return 0;
   }

   // new file name
   if (new_path->empty()) {
      xsink->raiseException("SYMLINK-ERROR", "empty new file path given as second argument");
      return 0;
   }

   int rc = symlink(old_path->getBuffer(), new_path->getBuffer());
   if (rc)
      xsink->raiseErrnoException("SYMLINK-ERROR", errno, "symlink('%s', '%s') failed", old_path->getBuffer(), new_path->getBuffer());

   return 0;
#else
   missing_function_error("symlink", "UNIX_FILEMGT", xsink);
   return 0;
#endif
   return 0;
}

// int umask(softint mask) {}
static int64 f_umask_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 mask = HARD_QORE_INT(args, 0);
# 636 "ql_file.qpp"
   return umask((int)mask);
}

// nothing umask() {}
static AbstractQoreNode* f_umask(const QoreListNode* args, ExceptionSink* xsink) {
# 643 "ql_file.qpp"

   return 0;
}

// int unlink(string path) {}
static int64 f_unlink_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 615 "ql_file.qpp"
   return unlink(path->getBuffer());
}

// nothing unlink() {}
static AbstractQoreNode* f_unlink(const QoreListNode* args, ExceptionSink* xsink) {
# 622 "ql_file.qpp"

   return 0;
}


DLLLOCAL void init_file_functions(QoreNamespace& ns) {
   // int chdir(string path) {}
   ns.addBuiltinVariant("chdir", (q_func_int64_t)f_chdir_Vs, QC_NO_FLAGS, QDOM_PROCESS|QDOM_FILESYSTEM, bigIntTypeInfo, 1, stringTypeInfo, NULL, "path");
   // int chmod(string path, softint mode) {}
   ns.addBuiltinVariant("chmod", (q_func_int64_t)f_chmod_Vsvi, QC_NO_FLAGS, QDOM_FILESYSTEM, bigIntTypeInfo, 2, stringTypeInfo, NULL, "path", softBigIntTypeInfo, NULL, "mode");
   // int chown(string path, softint owner = -1, softint group = -1) {}
   ns.addBuiltinVariant("chown", (q_func_int64_t)f_chown_Vsvivi, QC_NO_FLAGS, QDOM_FILESYSTEM, bigIntTypeInfo, 3, stringTypeInfo, NULL, "path", softBigIntTypeInfo, new QoreBigIntNode(-1), "owner", softBigIntTypeInfo, new QoreBigIntNode(-1), "group");
   // string getcwd() {}
   ns.addBuiltinVariant("getcwd", (q_func_t)f_getcwd, QC_CONSTANT, QDOM_FILESYSTEM|QDOM_EXTERNAL_INFO, stringTypeInfo);
   // string getcwd2() {}
   ns.addBuiltinVariant("getcwd2", (q_func_t)f_getcwd2, QC_RET_VALUE_ONLY, QDOM_FILESYSTEM|QDOM_EXTERNAL_INFO, stringTypeInfo);
   // *list glob(string glob_str) {}
   ns.addBuiltinVariant("glob", (q_func_t)f_glob_Vs, QC_NO_FLAGS, QDOM_FILESYSTEM, listOrNothingTypeInfo, 1, stringTypeInfo, NULL, "glob_str");
   // nothing glob() {}
   ns.addBuiltinVariant("glob", (q_func_t)f_glob, QC_RUNTIME_NOOP, QDOM_FILESYSTEM, nothingTypeInfo);
   // *hash hlstat(string path) {}
   ns.addBuiltinVariant("hlstat", (q_func_t)f_hlstat_Vs, QC_NO_FLAGS, QDOM_FILESYSTEM, hashOrNothingTypeInfo, 1, stringTypeInfo, NULL, "path");
   // nothing hlstat() {}
   ns.addBuiltinVariant("hlstat", (q_func_t)f_hlstat, QC_RUNTIME_NOOP, QDOM_FILESYSTEM, nothingTypeInfo);
   // *hash hstat(string path) {}
   ns.addBuiltinVariant("hstat", (q_func_t)f_hstat_Vs, QC_NO_FLAGS, QDOM_FILESYSTEM, hashOrNothingTypeInfo, 1, stringTypeInfo, NULL, "path");
   // nothing hstat() {}
   ns.addBuiltinVariant("hstat", (q_func_t)f_hstat, QC_RUNTIME_NOOP, QDOM_FILESYSTEM, nothingTypeInfo);
   // bool is_bdev(string path) {}
   ns.addBuiltinVariant("is_bdev", (q_func_bool_t)f_is_bdev_Vs, QC_CONSTANT, QDOM_FILESYSTEM, boolTypeInfo, 1, stringTypeInfo, NULL, "path");
   // bool is_cdev(string path) {}
   ns.addBuiltinVariant("is_cdev", (q_func_bool_t)f_is_cdev_Vs, QC_CONSTANT, QDOM_FILESYSTEM, boolTypeInfo, 1, stringTypeInfo, NULL, "path");
   // bool is_dev(string path) {}
   ns.addBuiltinVariant("is_dev", (q_func_bool_t)f_is_dev_Vs, QC_CONSTANT, QDOM_FILESYSTEM, boolTypeInfo, 1, stringTypeInfo, NULL, "path");
   // bool is_dir(string path) {}
   ns.addBuiltinVariant("is_dir", (q_func_bool_t)f_is_dir_Vs, QC_CONSTANT, QDOM_FILESYSTEM, boolTypeInfo, 1, stringTypeInfo, NULL, "path");
   // bool is_executable(string path) {}
   ns.addBuiltinVariant("is_executable", (q_func_bool_t)f_is_executable_Vs, QC_CONSTANT, QDOM_FILESYSTEM, boolTypeInfo, 1, stringTypeInfo, NULL, "path");
   // bool is_file(string path) {}
   ns.addBuiltinVariant("is_file", (q_func_bool_t)f_is_file_Vs, QC_CONSTANT, QDOM_FILESYSTEM, boolTypeInfo, 1, stringTypeInfo, NULL, "path");
   // bool is_link(string path) {}
   ns.addBuiltinVariant("is_link", (q_func_bool_t)f_is_link_Vs, QC_CONSTANT, QDOM_FILESYSTEM, boolTypeInfo, 1, stringTypeInfo, NULL, "path");
   // bool is_pipe(string path) {}
   ns.addBuiltinVariant("is_pipe", (q_func_bool_t)f_is_pipe_Vs, QC_CONSTANT, QDOM_FILESYSTEM, boolTypeInfo, 1, stringTypeInfo, NULL, "path");
   // bool is_readable(string path) {}
   ns.addBuiltinVariant("is_readable", (q_func_bool_t)f_is_readable_Vs, QC_CONSTANT, QDOM_FILESYSTEM, boolTypeInfo, 1, stringTypeInfo, NULL, "path");
   // bool is_socket(string path) {}
   ns.addBuiltinVariant("is_socket", (q_func_bool_t)f_is_socket_Vs, QC_CONSTANT, QDOM_FILESYSTEM, boolTypeInfo, 1, stringTypeInfo, NULL, "path");
   // bool is_writable(string path) {}
   ns.addBuiltinVariant("is_writable", (q_func_bool_t)f_is_writable_Vs, QC_CONSTANT, QDOM_FILESYSTEM, boolTypeInfo, 1, stringTypeInfo, NULL, "path");
   // bool is_writeable(string path) {}
   ns.addBuiltinVariant("is_writeable", (q_func_bool_t)f_is_writeable_Vs, QC_CONSTANT, QDOM_FILESYSTEM, boolTypeInfo, 1, stringTypeInfo, NULL, "path");
   // int lchown(string path, softint uid = -1, softint gid = -1) {}
   ns.addBuiltinVariant("lchown", (q_func_int64_t)f_lchown_Vsvivi, QC_NO_FLAGS, QDOM_FILESYSTEM, bigIntTypeInfo, 3, stringTypeInfo, NULL, "path", softBigIntTypeInfo, new QoreBigIntNode(-1), "uid", softBigIntTypeInfo, new QoreBigIntNode(-1), "gid");
   // *list lstat(string path) {}
   ns.addBuiltinVariant("lstat", (q_func_t)f_lstat_Vs, QC_NO_FLAGS, QDOM_FILESYSTEM, listOrNothingTypeInfo, 1, stringTypeInfo, NULL, "path");
   // nothing lstat() {}
   ns.addBuiltinVariant("lstat", (q_func_t)f_lstat, QC_RUNTIME_NOOP, QDOM_FILESYSTEM, nothingTypeInfo);
   // int mkdir(string path, softint mode = 0777) {}
   ns.addBuiltinVariant("mkdir", (q_func_int64_t)f_mkdir_Vsvi, QC_NO_FLAGS, QDOM_FILESYSTEM, bigIntTypeInfo, 2, stringTypeInfo, NULL, "path", softBigIntTypeInfo, new QoreBigIntNode(0777), "mode");
   // int mkfifo(string path, softint mode = 0600) {}
   ns.addBuiltinVariant("mkfifo", (q_func_int64_t)f_mkfifo_Vsvi, QC_NO_FLAGS, QDOM_FILESYSTEM, bigIntTypeInfo, 2, stringTypeInfo, NULL, "path", softBigIntTypeInfo, new QoreBigIntNode(0600), "mode");
   // string readlink(string path) {}
   ns.addBuiltinVariant("readlink", (q_func_t)f_readlink_Vs, QC_NO_FLAGS, QDOM_FILESYSTEM, stringTypeInfo, 1, stringTypeInfo, NULL, "path");
   // nothing rename(string old_path, string new_path) {}
   ns.addBuiltinVariant("rename", (q_func_t)f_rename_VsVs, QC_NO_FLAGS, QDOM_FILESYSTEM, nothingTypeInfo, 2, stringTypeInfo, NULL, "old_path", stringTypeInfo, NULL, "new_path");
   // int rmdir(string path) {}
   ns.addBuiltinVariant("rmdir", (q_func_int64_t)f_rmdir_Vs, QC_NO_FLAGS, QDOM_FILESYSTEM, bigIntTypeInfo, 1, stringTypeInfo, NULL, "path");
   // *list stat(string path) {}
   ns.addBuiltinVariant("stat", (q_func_t)f_stat_Vs, QC_NO_FLAGS, QDOM_FILESYSTEM, listOrNothingTypeInfo, 1, stringTypeInfo, NULL, "path");
   // nothing stat() {}
   ns.addBuiltinVariant("stat", (q_func_t)f_stat, QC_RUNTIME_NOOP, QDOM_FILESYSTEM, nothingTypeInfo);
   // *hash statvfs(string path) {}
   ns.addBuiltinVariant("statvfs", (q_func_t)f_statvfs_Vs, QC_NO_FLAGS, QDOM_FILESYSTEM, hashOrNothingTypeInfo, 1, stringTypeInfo, NULL, "path");
   // nothing symlink(string old_path, string new_path) {}
   ns.addBuiltinVariant("symlink", (q_func_t)f_symlink_VsVs, QC_NO_FLAGS, QDOM_FILESYSTEM, nothingTypeInfo, 2, stringTypeInfo, NULL, "old_path", stringTypeInfo, NULL, "new_path");
   // int umask(softint mask) {}
   ns.addBuiltinVariant("umask", (q_func_int64_t)f_umask_vi, QC_NO_FLAGS, QDOM_PROCESS, bigIntTypeInfo, 1, softBigIntTypeInfo, NULL, "mask");
   // nothing umask() {}
   ns.addBuiltinVariant("umask", (q_func_t)f_umask, QC_RUNTIME_NOOP, QDOM_PROCESS, nothingTypeInfo);
   // int unlink(string path) {}
   ns.addBuiltinVariant("unlink", (q_func_int64_t)f_unlink_Vs, QC_NO_FLAGS, QDOM_FILESYSTEM, bigIntTypeInfo, 1, stringTypeInfo, NULL, "path");
   // nothing unlink() {}
   ns.addBuiltinVariant("unlink", (q_func_t)f_unlink, QC_RUNTIME_NOOP, QDOM_FILESYSTEM, nothingTypeInfo);

}
