/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
  ql_lib.qpp

  Qore Programming Language

  Copyright (C) 2003 - 2014 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>
#include <qore/intern/ql_lib.h>
#include <qore/intern/ExecArgList.h>
#include <qore/intern/QoreSignal.h>
#include <qore/minitest.hpp>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>

#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#include <time.h>

#ifdef HAVE_GLOB_H
#include <glob.h>
#else
#include <qore/intern/glob.h>
#endif

#ifdef HAVE_SYS_TYPES
#include <sys/types.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifndef SIGHUP
#define SIGHUP 0
#endif

#define MAX_GIDS 100

extern bool threads_initialized;

#ifdef DEBUG_TESTS
// Unsorted unit tests are put here
#  include "tests/ReferenceHolder_tests.cpp"
#endif

AbstractQoreNode *missing_function_error(const char *func, ExceptionSink *xsink) {
   QoreString have(func);
   have.toupr();
   xsink->raiseException("MISSING-FEATURE-ERROR", "this system does not implement %s(); for maximum portability use the constant Option::HAVE_%s to check if this function is implemented before calling", func, have.getBuffer());
   return 0;
}

AbstractQoreNode *missing_function_error(const char *func, const char *opt, ExceptionSink *xsink) {
   xsink->raiseException("MISSING-FEATURE-ERROR", "this system does not implement %s(); for maximum portability use the constant Option::HAVE_%s to check if this function is implemented before calling", func, opt);
   return 0;
}

#ifdef HAVE_SYSTEM
static int do_system(const QoreString &str) {
   QoreString c(str);
   // escape backslashes: replace '\' -> '\\'
   c.replaceAll("\\", "\\\\");
   // escape double quotes: replace '"' -> '\"'
   c.replaceAll("\"", "\\\"");

   c.concat('"');
   c.prepend("/bin/sh -c \"");
   
   return system(c.getBuffer());
}
#endif

#ifdef DEBUG_X
static AbstractQoreNode *runQoreTests(const QoreListNode *args, ExceptionSink *xsink) {
   minitest::result res = minitest::execute_all_tests();
   if (res.all_tests_succeeded) {
      printf("Qore runtime: %d tests succeeded\n", res.sucessful_tests_count);
      return 0;
   }
   
   xsink->raiseException("A Qore test failed", "Qore test in file %s, line %d threw an exception.",
			 res.failed_test_file, res.failed_test_line);
   return 0;
}

static AbstractQoreNode *runRecentQoreTests(const QoreListNode *args, ExceptionSink *xsink) {
   minitest::result res = minitest::test_last_changed_files(3); // 3 last modified files
   if (res.all_tests_succeeded) {
      printf("Qore runtime: %d recent tests succeeded\n", res.sucessful_tests_count);
      return 0;
   }
   
   xsink->raiseException("A Qore test failed", "Qore test in file %s, line %d threw an exception.",
			 res.failed_test_file, res.failed_test_line);
   return 0;
}
namespace {
TEST() {
  // just an example of empty test
}
}
#endif

/*
// need an easier to use function here
// usage: mknod(path, mode)
static AbstractQoreNode *f_mknod(const QoreListNode *args, ExceptionSink *xsink) {
   const QoreStringNode *p0 = test_string_param(args, 0);
   if (!p0) {
      xsink->raiseException("MKNOD-PARAMETER-ERROR", "expecting string as first parameter of mknod");
      return 0;
   }

   const AbstractQoreNode *p1 = get_param(args, 1);
   if (is_nothing(p1)) {
      xsink->raiseException("MKNOD-PARAMETER-ERROR", "expecting mode as second parameter of mknod");
      return 0;
   }

   const AbstractQoreNode *p2 = get_param(args, 2);
   if (is_nothing(p2)) {
      xsink->raiseException("MKNOD-PARAMETER-ERROR", "expecting device as second parameter of mknod");
      return 0;
   }

   return new QoreBigIntNode(mknod(p0->getBuffer(), p1->getAsInt(), p2->getAsInt()));
}
*/
# 154 "ql_lib.qpp"
# 190 "ql_lib.qpp"
# 205 "ql_lib.qpp"
# 220 "ql_lib.qpp"
# 275 "ql_lib.qpp"
# 281 "ql_lib.qpp"
# 301 "ql_lib.qpp"
# 321 "ql_lib.qpp"
# 341 "ql_lib.qpp"
# 361 "ql_lib.qpp"
# 393 "ql_lib.qpp"
# 423 "ql_lib.qpp"
# 442 "ql_lib.qpp"
# 448 "ql_lib.qpp"
# 469 "ql_lib.qpp"
# 487 "ql_lib.qpp"
# 493 "ql_lib.qpp"
# 505 "ql_lib.qpp"
# 525 "ql_lib.qpp"
# 588 "ql_lib.qpp"
# 615 "ql_lib.qpp"
# 621 "ql_lib.qpp"
# 642 "ql_lib.qpp"
# 661 "ql_lib.qpp"
# 667 "ql_lib.qpp"
# 685 "ql_lib.qpp"
# 700 "ql_lib.qpp"
# 715 "ql_lib.qpp"
# 730 "ql_lib.qpp"
# 748 "ql_lib.qpp"
# 754 "ql_lib.qpp"
# 772 "ql_lib.qpp"
# 778 "ql_lib.qpp"
# 801 "ql_lib.qpp"
# 824 "ql_lib.qpp"
# 847 "ql_lib.qpp"
# 870 "ql_lib.qpp"
# 890 "ql_lib.qpp"
# 908 "ql_lib.qpp"
# 914 "ql_lib.qpp"
# 937 "ql_lib.qpp"
# 943 "ql_lib.qpp"
# 962 "ql_lib.qpp"
# 968 "ql_lib.qpp"
# 991 "ql_lib.qpp"
# 997 "ql_lib.qpp"
# 1018 "ql_lib.qpp"
// nothing abort() {}
static AbstractQoreNode* f_abort(const QoreListNode* args, ExceptionSink* xsink) {
# 200 "ql_lib.qpp"
   threads_initialized = false;
   abort();
   return 0;
}

// string basename(string path) {}
static AbstractQoreNode* f_basename_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 742 "ql_lib.qpp"
   char *p = q_basename(path->getBuffer());
   int len = strlen(p);
   return new QoreStringNode(p, len, len + 1, path->getEncoding());
}

// nothing basename() {}
static AbstractQoreNode* f_basename(const QoreListNode* args, ExceptionSink* xsink) {
# 751 "ql_lib.qpp"

   return 0;
}

// nothing close_all_fd(*softbool strd) {}
static AbstractQoreNode* f_close_all_fd_nb(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreBoolNode* tmp_strd = reinterpret_cast<const QoreBoolNode*>(get_param(args, 0));
   bool strd = tmp_strd ? tmp_strd->getValue() : false;
# 1031 "ql_lib.qpp"
#if defined(HAVE_UNISTD_H) && defined(_SC_OPEN_MAX)
   int start = strd ? 0 : 2;
   // get max file descriptor
   int max = (int)sysconf(_SC_OPEN_MAX);
   printd(0, "close_all_fd() start: %d max: %d\n", start, max);

   for (;start < max; ++start)
      close(start);
#else
   missing_function_error("close_all_fd", xsink);
#endif // HAVE_UNISTD
   return 0;
}

// string dirname(string path) {}
static AbstractQoreNode* f_dirname_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* path = HARD_QORE_STRING(args, 0);
# 766 "ql_lib.qpp"
   char *p = q_dirname(path->getBuffer());
   int len = strlen(p);
   return new QoreStringNode(p, len, len + 1, path->getEncoding());
}

// nothing dirname() {}
static AbstractQoreNode* f_dirname(const QoreListNode* args, ExceptionSink* xsink) {
# 775 "ql_lib.qpp"

   return 0;
}

// int errno() {}
static int64 f_errno(const QoreListNode* args, ExceptionSink* xsink) {
# 696 "ql_lib.qpp"
   return errno;
}

// nothing exec(string command) {}
static AbstractQoreNode* f_exec_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* command = HARD_QORE_STRING(args, 0);
# 213 "ql_lib.qpp"
   ExecArgList exec_args(command->getBuffer());
   execvp(exec_args.getFile(), exec_args.getArgs());

   xsink->raiseErrnoException("EXEC-ERROR", errno, "execvp() failed in child process for target '%s'", exec_args.getFile());
   return 0;
}

// nothing exit(softint rc = 0) {}
static AbstractQoreNode* f_exit_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 rc = HARD_QORE_INT(args, 0);
# 186 "ql_lib.qpp"
   qore_exit_process(rc);
   return 0;
}

// int fork() {}
static int64 f_fork(const QoreListNode* args, ExceptionSink* xsink) {
# 550 "ql_lib.qpp"
#ifdef HAVE_FORK
#ifdef HAVE_SIGNAL_HANDLING
   int QORE_THREAD_LIMIT = 1 + (QSM.thread_running ? 1 : 0);
#else
#define QORE_THREAD_LIMIT 1
#endif
   if (thread_list.getNumThreads() > (unsigned)(QORE_THREAD_LIMIT)) {
      xsink->raiseException("ILLEGAL-FORK", "cannot fork() when other threads are running");
      return 0;
   }

#ifdef HAVE_SIGNAL_HANDLING
   // we may not fork from within a signal handler
   if (QSM.thread_running && gettid() == QSM.gettid()) {
      xsink->raiseException("ILLEGAL-FORK", "cannot fork() within a signal handler");
      return 0;
   }
   
   // stop signal handling thread and make sure it can't be restarted until fork is done
   QSM.pre_fork_block_and_stop();
#endif

   //printd(5, "stopped signal thread, about to fork pid %d\n", getpid()); fflush(stdout);
   int pid = fork();

#ifdef HAVE_SIGNAL_HANDLING
   // release signal handler lock
   QSM.post_fork_unblock_and_start(!pid, xsink);
#endif

   return pid;
#else
   missing_function_error("fork", xsink);
   return 0;
#endif
   return 0;
}

// list getaddrinfo(*string node, *softstring service, softint family = AF_UNSPEC, softint flags = 0) {}
static AbstractQoreNode* f_getaddrinfo_Nsnsvivi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* node = reinterpret_cast<const QoreStringNode*>(get_param(args, 0));
   const QoreStringNode* service = reinterpret_cast<const QoreStringNode*>(get_param(args, 1));
   int64 family = HARD_QORE_INT(args, 2);
   int64 flags = HARD_QORE_INT(args, 3);
# 1014 "ql_lib.qpp"
   return q_getaddrinfo_to_list(xsink, node ? node->getBuffer() : 0, service ? service->getBuffer() : 0, (int)family, (int)flags);
}

// int getegid() {}
static int64 f_getegid(const QoreListNode* args, ExceptionSink* xsink) {
# 352 "ql_lib.qpp"
#ifdef HAVE_GETEGID
   return getgid();
#else
   missing_function_error("getegid", "UNIX_USERMGT", xsink);
   return 0;
#endif
   return 0;
}

// int geteuid() {}
static int64 f_geteuid(const QoreListNode* args, ExceptionSink* xsink) {
# 312 "ql_lib.qpp"
#ifdef HAVE_GETEUID
   return getuid();
#else
   missing_function_error("geteuid", "UNIX_USERMGT", xsink);
   return 0;
#endif
   return 0;
}

// int getgid() {}
static int64 f_getgid(const QoreListNode* args, ExceptionSink* xsink) {
# 332 "ql_lib.qpp"
#ifdef HAVE_GETGID
   return getgid();
#else
   missing_function_error("getgid", "UNIX_USERMGT", xsink);
   return 0;
#endif
   return 0;
}

// list getgroups() {}
static AbstractQoreNode* f_getgroups(const QoreListNode* args, ExceptionSink* xsink) {
# 374 "ql_lib.qpp"
#ifdef HAVE_GETGROUPS
   gid_t l[MAX_GIDS];

   int n = getgroups(MAX_GIDS, l);
   if (n == -1) {
      xsink->raiseErrnoException("GETGROUPS-ERROR", errno, "error calling getgroups()");
      return 0;
   }

   QoreListNode* rv = new QoreListNode;
   for (int i = 0; i < n; ++i)
      rv->push(new QoreBigIntNode(l[i]));
   return rv;
#else
   missing_function_error("getgroups", "UNIX_USERMGT", xsink);
#endif
   return 0;
}

// *string gethostbyaddr(string addr, softint type = AF_INET) {}
static AbstractQoreNode* f_gethostbyaddr_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* addr = HARD_QORE_STRING(args, 0);
   int64 type = HARD_QORE_INT(args, 1);
# 933 "ql_lib.qpp"
   return q_gethostbyaddr_to_string(xsink, addr->getBuffer(), type);
}

// nothing gethostbyaddr() {}
static AbstractQoreNode* f_gethostbyaddr(const QoreListNode* args, ExceptionSink* xsink) {
# 940 "ql_lib.qpp"

   return 0;
}

// *hash gethostbyaddr_long(string addr, softint type = AF_INET) {}
static AbstractQoreNode* f_gethostbyaddr_long_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* addr = HARD_QORE_STRING(args, 0);
   int64 type = HARD_QORE_INT(args, 1);
# 987 "ql_lib.qpp"
   return q_gethostbyaddr_to_hash(xsink, addr->getBuffer(), type);
}

// nothing gethostbyaddr_long() {}
static AbstractQoreNode* f_gethostbyaddr_long(const QoreListNode* args, ExceptionSink* xsink) {
# 994 "ql_lib.qpp"

   return 0;
}

// *string gethostbyname(string name) {}
static AbstractQoreNode* f_gethostbyname_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* name = HARD_QORE_STRING(args, 0);
# 904 "ql_lib.qpp"
   return q_gethostbyname_to_string(name->getBuffer());
}

// nothing gethostbyname() {}
static AbstractQoreNode* f_gethostbyname(const QoreListNode* args, ExceptionSink* xsink) {
# 911 "ql_lib.qpp"

   return 0;
}

// *hash gethostbyname_long(string name) {}
static AbstractQoreNode* f_gethostbyname_long_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* name = HARD_QORE_STRING(args, 0);
# 958 "ql_lib.qpp"
   return q_gethostbyname_to_hash(name->getBuffer());
}

// nothing gethostbyname_long() {}
static AbstractQoreNode* f_gethostbyname_long(const QoreListNode* args, ExceptionSink* xsink) {
# 965 "ql_lib.qpp"

   return 0;
}

// string gethostname() {}
static AbstractQoreNode* f_gethostname(const QoreListNode* args, ExceptionSink* xsink) {
# 675 "ql_lib.qpp"
   char buf[HOSTNAMEBUFSIZE + 1];

   if (gethostname(buf, HOSTNAMEBUFSIZE)) {
      xsink->raiseErrnoException("GETHOSTNAME-ERROR", errno, "gethostname() failed");
      return 0;
   }
   return new QoreStringNode(buf);
}

// int getpid() {}
static int64 f_getpid(const QoreListNode* args, ExceptionSink* xsink) {
# 501 "ql_lib.qpp"
   return getpid();
}

// int getppid() {}
static int64 f_getppid(const QoreListNode* args, ExceptionSink* xsink) {
# 516 "ql_lib.qpp"
#ifdef HAVE_GETPPID
   return getppid();   
#else
   missing_function_error("getppid", xsink);
   return 0;
#endif
   return 0;
}

// int getuid() {}
static int64 f_getuid(const QoreListNode* args, ExceptionSink* xsink) {
# 292 "ql_lib.qpp"
#ifdef HAVE_GETUID
   return getuid();
#else
   missing_function_error("getuid", "UNIX_USERMGT", xsink);
   return 0;
#endif
   return 0;
}

// int kill(softint pid, softint sig = SIGHUP) {}
static int64 f_kill_vivi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 pid = HARD_QORE_INT(args, 0);
   int64 sig = HARD_QORE_INT(args, 1);
# 606 "ql_lib.qpp"
#ifdef HAVE_KILL
   return kill(pid, sig);
#else
   missing_function_error("kill", xsink);
   return 0;
#endif
   return 0;
}

// nothing kill() {}
static AbstractQoreNode* f_kill(const QoreListNode* args, ExceptionSink* xsink) {
# 618 "ql_lib.qpp"

   return 0;
}

// int rand() {}
static int64 f_rand(const QoreListNode* args, ExceptionSink* xsink) {
# 633 "ql_lib.qpp"
#ifdef HAVE_RANDOM
   // return a random 64-bit integer by calling random() twice
   return random() | (((int64)random()) << 32);
#else
   return rand() | (((int64)rand()) << 32);
#endif
   return 0;
}

// int setegid(softint gid) {}
static int64 f_setegid_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 gid = HARD_QORE_INT(args, 0);
# 861 "ql_lib.qpp"
#ifdef HAVE_SETEGID
   return setegid((int)gid);
#else
   missing_function_error("setegid", xsink);
   return 0;
#endif
   return 0;
}

// int seteuid(softint uid) {}
static int64 f_seteuid_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 uid = HARD_QORE_INT(args, 0);
# 815 "ql_lib.qpp"
#ifdef HAVE_SETEUID
   return seteuid((int)uid);
#else
   missing_function_error("seteuid", xsink);
   return 0;
#endif
   return 0;
}

// int setgid(softint gid) {}
static int64 f_setgid_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 gid = HARD_QORE_INT(args, 0);
# 838 "ql_lib.qpp"
#ifdef HAVE_SETGID
   return setgid((int)gid);
#else
   missing_function_error("setgid", "UNIX_USERMGT", xsink);
   return 0;
#endif
   return 0;
}

// nothing setgroups(softlist gids) {}
static AbstractQoreNode* f_setgroups_vl(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreListNode* gids = HARD_QORE_LIST(args, 0);
# 406 "ql_lib.qpp"
#ifdef HAVE_SETGROUPS
   if (gids->empty())
      return 0;

   gid_t l[gids->size()];
   for (qore_size_t i = 0, e = gids->size(); i < e; ++i) {
      const AbstractQoreNode* n = gids->retrieve_entry(i);
      l[i] = n ? n->getAsInt() : 0;
   }
   if (setgroups(gids->size(), l))
      xsink->raiseErrnoException("SETGROUPS-ERROR", errno, "error calling setgroups()");
#else
   missing_function_error("setgroups", "UNIX_USERMGT", xsink);
#endif
   return 0;
}

// int setsid() {}
static int64 f_setsid(const QoreListNode* args, ExceptionSink* xsink) {
# 881 "ql_lib.qpp"
#ifdef HAVE_SETSID
   return setsid();
#else
   missing_function_error("setsid", xsink);
   return 0;
#endif
   return 0;
}

// int setuid(softint uid) {}
static int64 f_setuid_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 uid = HARD_QORE_INT(args, 0);
# 792 "ql_lib.qpp"
#ifdef HAVE_SETUID
   return setuid((int)uid);
#else
   missing_function_error("setuid", "UNIX_USERMGT", xsink);
   return 0;
#endif
   return 0;
}

// int sleep(softint seconds) {}
static int64 f_sleep_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 seconds = HARD_QORE_INT(args, 0);
# 435 "ql_lib.qpp"
   if (!seconds)
      return 0;

   return qore_usleep(seconds * 1000000ll);
}

// nothing sleep() {}
static AbstractQoreNode* f_sleep(const QoreListNode* args, ExceptionSink* xsink) {
# 445 "ql_lib.qpp"

   return 0;
}

// nothing srand(softint seed) {}
static AbstractQoreNode* f_srand_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 seed = HARD_QORE_INT(args, 0);
# 652 "ql_lib.qpp"
#ifdef HAVE_RANDOM
   srandom((int)seed);
#else
   srand((int)seed);
#endif
   return 0;
}

// nothing srand() {}
static AbstractQoreNode* f_srand(const QoreListNode* args, ExceptionSink* xsink) {
# 664 "ql_lib.qpp"

   return 0;
}

// string strerror(softint err) {}
static AbstractQoreNode* f_strerror_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 err = HARD_QORE_INT(args, 0);
# 711 "ql_lib.qpp"
   return q_strerror(err);
}

// string strerror() {}
static AbstractQoreNode* f_strerror(const QoreListNode* args, ExceptionSink* xsink) {
# 726 "ql_lib.qpp"
   return q_strerror(errno);
}

// int system(string command) {}
static int64 f_system_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* command = HARD_QORE_STRING(args, 0);
# 237 "ql_lib.qpp"
#ifdef HAVE_FORK
   int rc;
   // use system() if shell meta-characters are found
   if (strchrs(command->getBuffer(), "$=*?><;|\"\\")) {
      rc = do_system(*command);
   }
   else { // otherwise fork and exec
      pid_t pid;
      if (!(pid = fork())) {
	 ExecArgList exec_args(command->getBuffer());
	 execvp(exec_args.getFile(), exec_args.getArgs());
	 fprintf(stderr, "execvp() failed in child process for target '%s' with error code %d: %s\n", exec_args.getFile(), errno, strerror(errno));
	 exit(-1);
	 //qore_exit_process(-1);
      }
      if (pid == -1)
	 rc = -1;
      else {
	 int status;
	 wait(&status);
	 if (WIFEXITED(status))
	    rc = WEXITSTATUS(status);
	 else
	    rc = -1;
      }
   }
   return rc;
#else
#ifdef HAVE_SYSTEM
   return do_system(*command);
#else
   missing_function_error("system", xsink);
   return 0;
#endif // HAVE_SYSTEM
#endif // HAVE_FORK
   return 0;
}

// nothing system() {}
static AbstractQoreNode* f_system(const QoreListNode* args, ExceptionSink* xsink) {
# 278 "ql_lib.qpp"

   return 0;
}

// int usleep(softint usecs) {}
static int64 f_usleep_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 usecs = HARD_QORE_INT(args, 0);
# 462 "ql_lib.qpp"
   if (!usecs)
      return 0;

   return qore_usleep(usecs);
}

// int usleep(date d) {}
static int64 f_usleep_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* d = HARD_QORE_DATE(args, 0);
# 483 "ql_lib.qpp"
   return qore_usleep(d->getRelativeMicroseconds());
}

// nothing usleep() {}
static AbstractQoreNode* f_usleep(const QoreListNode* args, ExceptionSink* xsink) {
# 490 "ql_lib.qpp"

   return 0;
}


DLLLOCAL void init_lib_functions(QoreNamespace& ns) {
   // nothing abort() {}
   ns.addBuiltinVariant("abort", (q_func_t)f_abort, QC_NO_FLAGS, QDOM_PROCESS, nothingTypeInfo);
   // string basename(string path) {}
   ns.addBuiltinVariant("basename", (q_func_t)f_basename_Vs, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, NULL, "path");
   // nothing basename() {}
   ns.addBuiltinVariant("basename", (q_func_t)f_basename, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // nothing close_all_fd(*softbool strd) {}
   ns.addBuiltinVariant("close_all_fd", (q_func_t)f_close_all_fd_nb, QC_NO_FLAGS, QDOM_PROCESS, nothingTypeInfo, 1, softBoolOrNothingTypeInfo, NULL, "strd");
   // string dirname(string path) {}
   ns.addBuiltinVariant("dirname", (q_func_t)f_dirname_Vs, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, NULL, "path");
   // nothing dirname() {}
   ns.addBuiltinVariant("dirname", (q_func_t)f_dirname, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int errno() {}
   ns.addBuiltinVariant("errno", (q_func_int64_t)f_errno, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo);
   // nothing exec(string command) {}
   ns.addBuiltinVariant("exec", (q_func_t)f_exec_Vs, QC_NO_FLAGS, QDOM_EXTERNAL_PROCESS|QDOM_PROCESS, nothingTypeInfo, 1, stringTypeInfo, NULL, "command");
   // nothing exit(softint rc = 0) {}
   ns.addBuiltinVariant("exit", (q_func_t)f_exit_vi, QC_NO_FLAGS, QDOM_PROCESS, nothingTypeInfo, 1, softBigIntTypeInfo, zero(), "rc");
   // int fork() {}
   ns.addBuiltinVariant("fork", (q_func_int64_t)f_fork, QC_NO_FLAGS, QDOM_PROCESS, bigIntTypeInfo);
   // list getaddrinfo(*string node, *softstring service, softint family = AF_UNSPEC, softint flags = 0) {}
   ns.addBuiltinVariant("getaddrinfo", (q_func_t)f_getaddrinfo_Nsnsvivi, QC_RET_VALUE_ONLY, QDOM_EXTERNAL_INFO, listTypeInfo, 4, stringOrNothingTypeInfo, NULL, "node", softStringOrNothingTypeInfo, NULL, "service", softBigIntTypeInfo, new QoreBigIntNode(AF_UNSPEC), "family", softBigIntTypeInfo, zero(), "flags");
   // int getegid() {}
   ns.addBuiltinVariant("getegid", (q_func_int64_t)f_getegid, QC_NO_FLAGS, QDOM_EXTERNAL_INFO, bigIntTypeInfo);
   // int geteuid() {}
   ns.addBuiltinVariant("geteuid", (q_func_int64_t)f_geteuid, QC_NO_FLAGS, QDOM_EXTERNAL_INFO, bigIntTypeInfo);
   // int getgid() {}
   ns.addBuiltinVariant("getgid", (q_func_int64_t)f_getgid, QC_NO_FLAGS, QDOM_EXTERNAL_INFO, bigIntTypeInfo);
   // list getgroups() {}
   ns.addBuiltinVariant("getgroups", (q_func_t)f_getgroups, QC_NO_FLAGS, QDOM_EXTERNAL_INFO, listTypeInfo);
   // *string gethostbyaddr(string addr, softint type = AF_INET) {}
   ns.addBuiltinVariant("gethostbyaddr", (q_func_t)f_gethostbyaddr_Vsvi, QC_CONSTANT, QDOM_EXTERNAL_INFO, stringOrNothingTypeInfo, 2, stringTypeInfo, NULL, "addr", softBigIntTypeInfo, new QoreBigIntNode(AF_INET), "type");
   // nothing gethostbyaddr() {}
   ns.addBuiltinVariant("gethostbyaddr", (q_func_t)f_gethostbyaddr, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // *hash gethostbyaddr_long(string addr, softint type = AF_INET) {}
   ns.addBuiltinVariant("gethostbyaddr_long", (q_func_t)f_gethostbyaddr_long_Vsvi, QC_CONSTANT, QDOM_EXTERNAL_INFO, hashOrNothingTypeInfo, 2, stringTypeInfo, NULL, "addr", softBigIntTypeInfo, new QoreBigIntNode(AF_INET), "type");
   // nothing gethostbyaddr_long() {}
   ns.addBuiltinVariant("gethostbyaddr_long", (q_func_t)f_gethostbyaddr_long, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // *string gethostbyname(string name) {}
   ns.addBuiltinVariant("gethostbyname", (q_func_t)f_gethostbyname_Vs, QC_CONSTANT, QDOM_EXTERNAL_INFO, stringOrNothingTypeInfo, 1, stringTypeInfo, NULL, "name");
   // nothing gethostbyname() {}
   ns.addBuiltinVariant("gethostbyname", (q_func_t)f_gethostbyname, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // *hash gethostbyname_long(string name) {}
   ns.addBuiltinVariant("gethostbyname_long", (q_func_t)f_gethostbyname_long_Vs, QC_CONSTANT, QDOM_EXTERNAL_INFO, hashOrNothingTypeInfo, 1, stringTypeInfo, NULL, "name");
   // nothing gethostbyname_long() {}
   ns.addBuiltinVariant("gethostbyname_long", (q_func_t)f_gethostbyname_long, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string gethostname() {}
   ns.addBuiltinVariant("gethostname", (q_func_t)f_gethostname, QC_CONSTANT, QDOM_EXTERNAL_INFO, stringTypeInfo);
   // int getpid() {}
   ns.addBuiltinVariant("getpid", (q_func_int64_t)f_getpid, QC_RET_VALUE_ONLY, QDOM_EXTERNAL_INFO, bigIntTypeInfo);
   // int getppid() {}
   ns.addBuiltinVariant("getppid", (q_func_int64_t)f_getppid, QC_RET_VALUE_ONLY, QDOM_EXTERNAL_INFO, bigIntTypeInfo);
   // int getuid() {}
   ns.addBuiltinVariant("getuid", (q_func_int64_t)f_getuid, QC_NO_FLAGS, QDOM_EXTERNAL_INFO, bigIntTypeInfo);
   // int kill(softint pid, softint sig = SIGHUP) {}
   ns.addBuiltinVariant("kill", (q_func_int64_t)f_kill_vivi, QC_NO_FLAGS, QDOM_EXTERNAL_PROCESS, bigIntTypeInfo, 2, softBigIntTypeInfo, NULL, "pid", softBigIntTypeInfo, new QoreBigIntNode(SIGHUP), "sig");
   // nothing kill() {}
   ns.addBuiltinVariant("kill", (q_func_t)f_kill, QC_RUNTIME_NOOP, QDOM_EXTERNAL_PROCESS, nothingTypeInfo);
   // int rand() {}
   ns.addBuiltinVariant("rand", (q_func_int64_t)f_rand, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);
   // int setegid(softint gid) {}
   ns.addBuiltinVariant("setegid", (q_func_int64_t)f_setegid_vi, QC_NO_FLAGS, QDOM_PROCESS, bigIntTypeInfo, 1, softBigIntTypeInfo, NULL, "gid");
   // int seteuid(softint uid) {}
   ns.addBuiltinVariant("seteuid", (q_func_int64_t)f_seteuid_vi, QC_NO_FLAGS, QDOM_PROCESS, bigIntTypeInfo, 1, softBigIntTypeInfo, NULL, "uid");
   // int setgid(softint gid) {}
   ns.addBuiltinVariant("setgid", (q_func_int64_t)f_setgid_vi, QC_NO_FLAGS, QDOM_PROCESS, bigIntTypeInfo, 1, softBigIntTypeInfo, NULL, "gid");
   // nothing setgroups(softlist gids) {}
   ns.addBuiltinVariant("setgroups", (q_func_t)f_setgroups_vl, QC_NO_FLAGS, QDOM_PROCESS, nothingTypeInfo, 1, softListTypeInfo, NULL, "gids");
   // int setsid() {}
   ns.addBuiltinVariant("setsid", (q_func_int64_t)f_setsid, QC_NO_FLAGS, QDOM_PROCESS, bigIntTypeInfo);
   // int setuid(softint uid) {}
   ns.addBuiltinVariant("setuid", (q_func_int64_t)f_setuid_vi, QC_NO_FLAGS, QDOM_PROCESS, bigIntTypeInfo, 1, softBigIntTypeInfo, NULL, "uid");
   // int sleep(softint seconds) {}
   ns.addBuiltinVariant("sleep", (q_func_int64_t)f_sleep_vi, QC_NO_FLAGS, QDOM_PROCESS, bigIntTypeInfo, 1, softBigIntTypeInfo, NULL, "seconds");
   // nothing sleep() {}
   ns.addBuiltinVariant("sleep", (q_func_t)f_sleep, QC_RUNTIME_NOOP, QDOM_PROCESS, nothingTypeInfo);
   // nothing srand(softint seed) {}
   ns.addBuiltinVariant("srand", (q_func_t)f_srand_vi, QC_NO_FLAGS, QDOM_PROCESS, nothingTypeInfo, 1, softBigIntTypeInfo, NULL, "seed");
   // nothing srand() {}
   ns.addBuiltinVariant("srand", (q_func_t)f_srand, QC_RUNTIME_NOOP, QDOM_PROCESS, nothingTypeInfo);
   // string strerror(softint err) {}
   ns.addBuiltinVariant("strerror", (q_func_t)f_strerror_vi, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 1, softBigIntTypeInfo, NULL, "err");
   // string strerror() {}
   ns.addBuiltinVariant("strerror", (q_func_t)f_strerror, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);
   // int system(string command) {}
   ns.addBuiltinVariant("system", (q_func_int64_t)f_system_Vs, QC_NO_FLAGS, QDOM_EXTERNAL_PROCESS, bigIntTypeInfo, 1, stringTypeInfo, NULL, "command");
   // nothing system() {}
   ns.addBuiltinVariant("system", (q_func_t)f_system, QC_RUNTIME_NOOP, QDOM_EXTERNAL_PROCESS, nothingTypeInfo);
   // int usleep(softint usecs) {}
   ns.addBuiltinVariant("usleep", (q_func_int64_t)f_usleep_vi, QC_NO_FLAGS, QDOM_PROCESS, bigIntTypeInfo, 1, softBigIntTypeInfo, NULL, "usecs");
   // int usleep(date d) {}
   ns.addBuiltinVariant("usleep", (q_func_int64_t)f_usleep_Vd, QC_NO_FLAGS, QDOM_PROCESS, bigIntTypeInfo, 1, dateTypeInfo, NULL, "d");
   // nothing usleep() {}
   ns.addBuiltinVariant("usleep", (q_func_t)f_usleep, QC_RUNTIME_NOOP, QDOM_PROCESS, nothingTypeInfo);

}
