/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
  ql_misc.qpp

  Qore Programming Language

  Copyright (C) 2003 - 2014 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>
#include <qore/intern/ql_misc.h>
#include <qore/intern/QoreSignal.h>
#include <qore/intern/QoreNamespaceIntern.h>

#include <string.h>
#include <time.h>
#include <errno.h>

static const QoreFunction *get_builtin_func(const QoreStringNode *str, ExceptionSink *xsink) {
   const qore_ns_private* ns;
   const QoreFunction *f = qore_root_ns_private::runtimeFindFunction(*(getRootNS()), str->getBuffer(), ns);
   if (f)
      return f;

   xsink->raiseException("NO-FUNCTION", "cannot find any builtin function '%s()'", str->getBuffer());
   return 0;
}

static const char *tlist[] = { "OPTION", "ALGORITHM", "FUNCTION" };

// signal constants - if they are not defined define them here
#ifndef SIGABRT
#define SIGABRT 0
#endif
#ifndef SIGALRM
#define SIGALRM 0
#endif
#ifndef SIGBUS
#define SIGBUS 0
#endif
#ifndef SIGCANCEL
#define SIGCANCEL 0
#endif
#ifndef SIGCHLD
#define SIGCHLD 0
#endif
#ifndef SIGCLD
#define SIGCLD 0
#endif
#ifndef SIGCONT
#define SIGCONT 0
#endif
#ifndef SIGEMT
#define SIGEMT 0
#endif
#ifndef SIGFPE
#define SIGFPE 0
#endif
#ifndef SIGFREEZE
#define SIGFREEZE 0
#endif
#ifndef SIGHUP
#define SIGHUP 0
#endif
#ifndef SIGILL
#define SIGILL 0
#endif
#ifndef SIGINFO
#define SIGINFO 0
#endif
#ifndef SIGINT
#define SIGINT 0
#endif
#ifndef SIGIO
#define SIGIO 0
#endif
#ifndef SIGIOT
#define SIGIOT 0
#endif
#ifndef SIGJVM1
#define SIGJVM1 0
#endif
#ifndef SIGJVM2
#define SIGJVM2 0
#endif
#ifndef SIGKILL
#define SIGKILL 0
#endif
#ifndef SIGLOST
#define SIGLOST 0
#endif
#ifndef SIGLWP
#define SIGLWP 0
#endif
#ifndef SIGPIPE
#define SIGPIPE 0
#endif
#ifndef SIGPOLL
#define SIGPOLL 0
#endif
#ifndef SIGPROF
#define SIGPROF 0
#endif
#ifndef SIGPWR
#define SIGPWR 0
#endif
#ifndef SIGQUIT
#define SIGQUIT 0
#endif
#ifndef SIGSEGV
#define SIGSEGV 0
#endif
#ifndef SIGSTKFLT
#define SIGSTKFLT 0
#endif
#ifndef SIGSTOP
#define SIGSTOP 0
#endif
#ifndef SIGSYS
#define SIGSYS 0
#endif
#ifndef SIGTERM
#define SIGTERM 0
#endif
#ifndef SIGTHAW
#define SIGTHAW 0
#endif
#ifndef SIGTRAP
#define SIGTRAP 0
#endif
#ifndef SIGTSTP
#define SIGTSTP 0
#endif
#ifndef SIGTTIN
#define SIGTTIN 0
#endif
#ifndef SIGTTOU
#define SIGTTOU 0
#endif
#ifndef SIGURG
#define SIGURG 0
#endif
#ifndef SIGUSR1
#define SIGUSR1 0
#endif
#ifndef SIGUSR2
#define SIGUSR2 0
#endif
#ifndef SIGVTALRM
#define SIGVTALRM 0
#endif
#ifndef SIGWAITING
#define SIGWAITING 0
#endif
#ifndef SIGWINCH
#define SIGWINCH 0
#endif
#ifndef SIGXCPU
#define SIGXCPU 0
#endif
#ifndef SIGXFSZ
#define SIGXFSZ 0
#endif
#ifndef SIGXRES
#define SIGXRES 0
#endif
# 376 "ql_misc.qpp"
# 412 "ql_misc.qpp"
// nothing remove_signal_handler(softint signal) {}
static AbstractQoreNode* f_remove_signal_handler_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 signal = HARD_QORE_INT(args, 0);
# 428 "ql_misc.qpp"
#ifdef HAVE_SIGNAL_HANDLING
   if (!signal || signal > QORE_SIGNAL_MAX) {
      xsink->raiseException("REMOVE-SIGNAL-HANDLER-ERROR", "%d is not a valid signal", signal);
      return 0;
   }
   QSM.removeHandler(signal, xsink);
#else
   xsink->raiseException("MISSING-FEATURE-ERROR", "this platform does not support signal handling, therefore the set_signal_handler() and remove_signal_handler() functions are not available in Qore; for maximum portability, use the constant Option::HAVE_SIGNAL_HANDLING to check if this function is implemented before calling");
#endif
   return 0;
}

// nothing set_signal_handler(softint signal, code f) {}
static AbstractQoreNode* f_set_signal_handler_viVc(const QoreListNode* args, ExceptionSink* xsink) {
   int64 signal = HARD_QORE_INT(args, 0);
   const ResolvedCallReferenceNode* f = reinterpret_cast<const ResolvedCallReferenceNode*>(get_param(args, 1));
# 398 "ql_misc.qpp"
#ifdef HAVE_SIGNAL_HANDLING
   if (!signal || signal > QORE_SIGNAL_MAX) {
      xsink->raiseException("SET-SIGNAL-HANDLER-ERROR", "%d is not a valid signal", signal);
      return 0;
   }

   QSM.setHandler(signal, f, xsink);
#else
   xsink->raiseException("MISSING-FEATURE-ERROR", "this platform does not support signal handling, therefore the set_signal_handler() and remove_signal_handler() functions are not available in Qore; for maximum portability, use the constant Option::HAVE_SIGNAL_HANDLING to check if this function is implemented before calling");
#endif
   return 0;
}

# 442 "ql_misc.qpp"
# 472 "ql_misc.qpp"
# 493 "ql_misc.qpp"
# 514 "ql_misc.qpp"
# 530 "ql_misc.qpp"
# 562 "ql_misc.qpp"
# 589 "ql_misc.qpp"
# 605 "ql_misc.qpp"
# 621 "ql_misc.qpp"
# 632 "ql_misc.qpp"
# 638 "ql_misc.qpp"
# 657 "ql_misc.qpp"
# 663 "ql_misc.qpp"
# 679 "ql_misc.qpp"
# 685 "ql_misc.qpp"
# 701 "ql_misc.qpp"
# 707 "ql_misc.qpp"
# 719 "ql_misc.qpp"
# 759 "ql_misc.qpp"
# 765 "ql_misc.qpp"
# 779 "ql_misc.qpp"
# 785 "ql_misc.qpp"
# 818 "ql_misc.qpp"
# 844 "ql_misc.qpp"
# 850 "ql_misc.qpp"
# 884 "ql_misc.qpp"
# 890 "ql_misc.qpp"
# 915 "ql_misc.qpp"
# 938 "ql_misc.qpp"
# 944 "ql_misc.qpp"
# 964 "ql_misc.qpp"
# 970 "ql_misc.qpp"
# 994 "ql_misc.qpp"
# 1000 "ql_misc.qpp"
# 1023 "ql_misc.qpp"
# 1046 "ql_misc.qpp"
# 1058 "ql_misc.qpp"
# 1082 "ql_misc.qpp"
# 1088 "ql_misc.qpp"
# 1108 "ql_misc.qpp"
# 1125 "ql_misc.qpp"
# 1150 "ql_misc.qpp"
# 1175 "ql_misc.qpp"
# 1202 "ql_misc.qpp"
# 1208 "ql_misc.qpp"
# 1235 "ql_misc.qpp"
# 1261 "ql_misc.qpp"
# 1288 "ql_misc.qpp"
# 1316 "ql_misc.qpp"
# 1345 "ql_misc.qpp"
# 1373 "ql_misc.qpp"
# 1402 "ql_misc.qpp"
# 1408 "ql_misc.qpp"
# 1434 "ql_misc.qpp"
# 1461 "ql_misc.qpp"
# 1487 "ql_misc.qpp"
# 1513 "ql_misc.qpp"
# 1541 "ql_misc.qpp"
# 1568 "ql_misc.qpp"
# 1594 "ql_misc.qpp"
# 1620 "ql_misc.qpp"
# 1631 "ql_misc.qpp"
# 1652 "ql_misc.qpp"
# 1675 "ql_misc.qpp"
# 1694 "ql_misc.qpp"
# 1717 "ql_misc.qpp"
# 1723 "ql_misc.qpp"
# 1744 "ql_misc.qpp"
# 1765 "ql_misc.qpp"
# 1771 "ql_misc.qpp"
# 1790 "ql_misc.qpp"
# 1796 "ql_misc.qpp"
# 1840 "ql_misc.qpp"
# 1846 "ql_misc.qpp"
# 1867 "ql_misc.qpp"
# 1873 "ql_misc.qpp"
# 1895 "ql_misc.qpp"
# 1901 "ql_misc.qpp"
# 1921 "ql_misc.qpp"
# 1927 "ql_misc.qpp"
# 1948 "ql_misc.qpp"
# 1960 "ql_misc.qpp"
# 1972 "ql_misc.qpp"
# 1984 "ql_misc.qpp"
# 2012 "ql_misc.qpp"
# 2043 "ql_misc.qpp"
# 2083 "ql_misc.qpp"
# 2097 "ql_misc.qpp"
// string backquote(string cmd, *reference rc) {}
static AbstractQoreNode* f_backquote_VsNr(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* cmd = HARD_QORE_STRING(args, 0);
   const ReferenceNode* rc = reinterpret_cast<const ReferenceNode*>(get_param(args, 1));
# 871 "ql_misc.qpp"
   int prc;
   SimpleRefHolder<QoreStringNode> str(backquoteEval(cmd->getBuffer(), prc, xsink));
   if (!*xsink && rc) {
      QoreTypeSafeReferenceHelper rh(rc, xsink);
      if (!rh)
         return 0;

      rh.assignBigInt(prc);
   }
   return str.release();
}

// nothing backquote() {}
static AbstractQoreNode* f_backquote(const QoreListNode* args, ExceptionSink* xsink) {
# 887 "ql_misc.qpp"

   return 0;
}

// any call_builtin_function(string name, ...) {}
static AbstractQoreNode* f_call_builtin_function_VsVV(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* name = HARD_QORE_STRING(args, 0);
# 549 "ql_misc.qpp"
   const QoreFunction *f = get_builtin_func(name, xsink);
   if (!f)
      return 0;

   ReferenceHolder<QoreListNode> vargs(xsink);
   // if there are arguments to pass, create argument list by copying current list
   if (num_args(args) > 1)
      vargs = args->copyListFrom(1);

   return f->evalDynamic(*vargs, xsink);
}

// any call_builtin_function_args(string name, *softlist vargs) {}
static AbstractQoreNode* f_call_builtin_function_args_Vsnl(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* name = HARD_QORE_STRING(args, 0);
   const QoreListNode* vargs = reinterpret_cast<const QoreListNode*>(get_param(args, 1));
# 581 "ql_misc.qpp"
   const QoreFunction *f = get_builtin_func(name, xsink);
   if (!f)
      return 0;

   return f->evalDynamic(vargs, xsink);
}

// any call_function(string name, ...) {}
static AbstractQoreNode* f_call_function_VsVV(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* name = HARD_QORE_STRING(args, 0);
# 463 "ql_misc.qpp"
   ReferenceHolder<QoreListNode> vargs(xsink);
   // if there are arguments to pass, create argument list by copying current list
   if (num_args(args) > 1)
      vargs = args->copyListFrom(1);

   return getProgram()->callFunction(name->getBuffer(), *vargs, xsink);
}

// any call_function(code f, ...) {}
static AbstractQoreNode* f_call_function_VcVV(const QoreListNode* args, ExceptionSink* xsink) {
   const ResolvedCallReferenceNode* f = reinterpret_cast<const ResolvedCallReferenceNode*>(get_param(args, 0));
# 484 "ql_misc.qpp"
   ReferenceHolder<QoreListNode> vargs(xsink);
   // if there are arguments to pass, create argument list by copying current list
   if (num_args(args) > 1)
      vargs = args->copyListFrom(1);

   return f->exec(*vargs, xsink);
}

// any call_function_args(string name, *softlist vargs) {}
static AbstractQoreNode* f_call_function_args_Vsnl(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* name = HARD_QORE_STRING(args, 0);
   const QoreListNode* vargs = reinterpret_cast<const QoreListNode*>(get_param(args, 1));
# 510 "ql_misc.qpp"
   return getProgram()->callFunction(name->getBuffer(), vargs, xsink);
}

// any call_function_args(code f, *softlist vargs) {}
static AbstractQoreNode* f_call_function_args_Vcnl(const QoreListNode* args, ExceptionSink* xsink) {
   const ResolvedCallReferenceNode* f = reinterpret_cast<const ResolvedCallReferenceNode*>(get_param(args, 0));
   const QoreListNode* vargs = reinterpret_cast<const QoreListNode*>(get_param(args, 1));
# 526 "ql_misc.qpp"
   return f->exec(vargs, xsink);
}

// string decode_url(string url) {}
static AbstractQoreNode* f_decode_url_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* url = HARD_QORE_STRING(args, 0);
# 1915 "ql_misc.qpp"
   QoreStringNodeHolder str(new QoreStringNode(QCS_UTF8));
   str->concatDecodeUrl(*url, xsink);
   return *xsink ? 0 : str.release();
}

// nothing decode_url() {}
static AbstractQoreNode* f_decode_url(const QoreListNode* args, ExceptionSink* xsink) {
# 1924 "ql_misc.qpp"

   return 0;
}

// string encode_url(string url, softbool encode_all = False) {}
static AbstractQoreNode* f_encode_url_Vsvb(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* url = HARD_QORE_STRING(args, 0);
   bool encode_all = HARD_QORE_BOOL(args, 1);
# 1942 "ql_misc.qpp"
   QoreStringNodeHolder str(new QoreStringNode(url->getEncoding()));
   str->concatEncodeUrl(xsink, url, encode_all);
   return *xsink ? 0 : str.release();
}

// bool exists(...) {}
static bool f_exists_VV(const QoreListNode* args, ExceptionSink* xsink) {
# 599 "ql_misc.qpp"
   // to emulate the exists operator, we must return True if more than one argument is passed
   // as this will appear to be a list to the exists operator, which is different from NOTHING
   return (num_args(args) <= 1) ? !is_nothing(get_param(args, 0)) : true;
}

// bool existsFunction(string name) {}
static bool f_existsFunction_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* name = HARD_QORE_STRING(args, 0);
# 615 "ql_misc.qpp"
   const char *str = name->getBuffer();
   
   return getProgram()->existsFunction(str);
}

// bool existsFunction(code c) {}
static bool f_existsFunction_Vc(const QoreListNode* args, ExceptionSink* xsink) {
# 628 "ql_misc.qpp"
   return true;
}

// nothing existsFunction() {}
static AbstractQoreNode* f_existsFunction(const QoreListNode* args, ExceptionSink* xsink) {
# 635 "ql_misc.qpp"

   return 0;
}

// *string functionType(string name) {}
static AbstractQoreNode* f_functionType_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* name = HARD_QORE_STRING(args, 0);
# 648 "ql_misc.qpp"
   const qore_ns_private* ns;
   const QoreFunction *f = qore_root_ns_private::runtimeFindFunction(*(getRootNS()), name->getBuffer(), ns);
   if (!f)
      return 0;

   return new QoreStringNode(f->hasBuiltin() ? "builtin" : "user");
}

// nothing functionType() {}
static AbstractQoreNode* f_functionType(const QoreListNode* args, ExceptionSink* xsink) {
# 660 "ql_misc.qpp"

   return 0;
}

// *int getByte(string str, softint offset = 0) {}
static AbstractQoreNode* f_getByte_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1192 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)str->getBuffer();
   int size = str->strlen();

   if (offset >= size || offset < 0)
      return 0;

   return new QoreBigIntNode(ptr[offset]);
}

// nothing getByte() {}
static AbstractQoreNode* f_getByte(const QoreListNode* args, ExceptionSink* xsink) {
# 1205 "ql_misc.qpp"

   return 0;
}

// *int getByte(binary b, softint offset = 0) {}
static AbstractQoreNode* f_getByte_Vyvi(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* b = HARD_QORE_BINARY(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1225 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)b->getPtr();
   int size = b->size();

   if (offset >= size || offset < 0)
      return 0;

   return new QoreBigIntNode(ptr[offset]);
}

// string getClassName(object obj) {}
static AbstractQoreNode* f_getClassName_Vo(const QoreListNode* args, ExceptionSink* xsink) {
   QoreObject* obj = HARD_QORE_OBJECT(args, 0);
# 775 "ql_misc.qpp"
   return new QoreStringNode(obj->getClass()->getName());
}

// nothing getClassName() {}
static AbstractQoreNode* f_getClassName(const QoreListNode* args, ExceptionSink* xsink) {
# 782 "ql_misc.qpp"

   return 0;
}

// list getFeatureList() {}
static AbstractQoreNode* f_getFeatureList(const QoreListNode* args, ExceptionSink* xsink) {
# 1054 "ql_misc.qpp"
   return getProgram()->getFeatureList();
}

// hash getModuleHash() {}
static AbstractQoreNode* f_getModuleHash(const QoreListNode* args, ExceptionSink* xsink) {
# 1042 "ql_misc.qpp"
   return MM.getModuleHash();
}

// list getModuleList() {}
static AbstractQoreNode* f_getModuleList(const QoreListNode* args, ExceptionSink* xsink) {
# 1019 "ql_misc.qpp"
   return MM.getModuleList();
}

// *int getWord32(string str, softint offset = 0) {}
static AbstractQoreNode* f_getWord32_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1362 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)str->getBuffer();
   int size = str->strlen();

   if (offset >= (size - 3) || offset < 0)
      return 0;

   int64 val = ntohl(*((unsigned int *)&ptr[offset]));
   return new QoreBigIntNode(val);
}

// *int getWord32(binary b, softint offset = 0) {}
static AbstractQoreNode* f_getWord32_Vyvi(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* b = HARD_QORE_BINARY(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1390 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)b->getPtr();
   int size = b->size();

   if (offset >= (size - 3) || offset < 0)
      return 0;

   int64 val = ntohl(*((unsigned int *)&ptr[offset]));

   return new QoreBigIntNode(val);
}

// nothing getWord32() {}
static AbstractQoreNode* f_getWord32(const QoreListNode* args, ExceptionSink* xsink) {
# 1405 "ql_misc.qpp"

   return 0;
}

// *int get_byte(string str, softint offset = 0) {}
static AbstractQoreNode* f_get_byte_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1140 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)str->getBuffer();
   int size = str->strlen();

   if (offset >= size || offset < 0)
      return 0;

   return new QoreBigIntNode(ptr[offset]);
}

// *int get_byte(binary b, softint offset = 0) {}
static AbstractQoreNode* f_get_byte_Vyvi(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* b = HARD_QORE_BINARY(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1165 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)b->getPtr();
   int size = b->size();

   if (offset >= size || offset < 0)
      return 0;

   return new QoreBigIntNode(ptr[offset]);
}

// string get_default_encoding() {}
static AbstractQoreNode* f_get_default_encoding(const QoreListNode* args, ExceptionSink* xsink) {
# 715 "ql_misc.qpp"
   return new QoreStringNode(QCS_DEFAULT->getCode());
}

// string get_ex_pos(hash ex) {}
static AbstractQoreNode* f_get_ex_pos_Vh(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreHashNode* ex = HARD_QORE_HASH(args, 0);
# 2114 "ql_misc.qpp"
   const AbstractQoreNode* n = ex->getKeyValue("file");
   QoreStringNode* str = new QoreStringNode(get_node_type(n) == NT_STRING ? reinterpret_cast<const QoreStringNode*>(n)->getBuffer() : "<unknown>");
   bool found = false;
   int ln = (int)ex->getKeyAsBigInt("line", found);
   str->sprintf(":%d", ln);
   n = ex->getKeyValue("source");
   if (get_node_type(n) == NT_STRING && !reinterpret_cast<const QoreStringNode*>(n)->empty()) {
      int offset = (int)ex->getKeyAsBigInt("offset", found);
      str->sprintf(" (source \"%s\":%d)", reinterpret_cast<const QoreStringNode*>(n)->getBuffer(), ln + offset);
   }

   return str;
}

// int get_parse_options() {}
static int64 f_get_parse_options(const QoreListNode* args, ExceptionSink* xsink) {
# 2093 "ql_misc.qpp"
   return runtime_get_parse_options();
}

// hash get_qore_library_info() {}
static AbstractQoreNode* f_get_qore_library_info(const QoreListNode* args, ExceptionSink* xsink) {
# 2063 "ql_misc.qpp"
   QoreHashNode *h = new QoreHashNode;

   h->setKeyValue("PlatformOS", new QoreStringNode(qore_target_os), xsink);
   h->setKeyValue("PlatformCPU", new QoreStringNode(qore_target_arch), xsink);
   h->setKeyValue("VersionString", new QoreStringNode(qore_version_string), xsink);   
   h->setKeyValue("VersionMajor", new QoreBigIntNode(qore_version_major), xsink);
   h->setKeyValue("VersionMinor", new QoreBigIntNode(qore_version_minor), xsink);
   h->setKeyValue("VersionSub", new QoreBigIntNode(qore_version_sub), xsink);
   h->setKeyValue("Build", new QoreStringNode(qore_build_number), xsink);
   h->setKeyValue("ModuleDir", new QoreStringNode(qore_module_dir), xsink);

   h->setKeyValue("BuildHost", new QoreStringNode(qore_build_host), xsink);
   h->setKeyValue("Compiler", new QoreStringNode(qore_cplusplus_compiler), xsink);
   h->setKeyValue("CFLAGS", new QoreStringNode(qore_cflags), xsink);
   h->setKeyValue("LDFLAGS", new QoreStringNode(qore_ldflags), xsink);

   return h;
}

// hash get_qore_option_hash() {}
static AbstractQoreNode* f_get_qore_option_hash(const QoreListNode* args, ExceptionSink* xsink) {
# 2028 "ql_misc.qpp"
   QoreHashNode *rv = new QoreHashNode;

   for (unsigned j = 0; j < qore_option_list_size; ++j) {
      QoreHashNode *h = new QoreHashNode;
      h->setKeyValue("option", new QoreStringNode(qore_option_list[j].option), xsink);
      h->setKeyValue("constant", new QoreStringNode(qore_option_list[j].constant), xsink);
      h->setKeyValue("type", new QoreStringNode(tlist[qore_option_list[j].type]), xsink);
      h->setKeyValue("value", get_bool_node(qore_option_list[j].value), xsink);

      rv->setKeyValue(qore_option_list[j].constant, h, 0);
   }
   return rv;
}

// list get_qore_option_list() {}
static AbstractQoreNode* f_get_qore_option_list(const QoreListNode* args, ExceptionSink* xsink) {
# 1998 "ql_misc.qpp"
   QoreListNode *l = new QoreListNode;

   for (unsigned j = 0; j < qore_option_list_size; ++j) {
      QoreHashNode *h = new QoreHashNode;
      h->setKeyValue("option", new QoreStringNode(qore_option_list[j].option), xsink);
      h->setKeyValue("constant", new QoreStringNode(qore_option_list[j].constant), xsink);
      h->setKeyValue("type", new QoreStringNode(tlist[qore_option_list[j].type]), xsink);
      h->setKeyValue("value", get_bool_node(qore_option_list[j].value), xsink);
      l->push(h);
   }
   return l;
}

// *string get_script_dir() {}
static AbstractQoreNode* f_get_script_dir(const QoreListNode* args, ExceptionSink* xsink) {
# 1968 "ql_misc.qpp"
   return getProgram()->getScriptDir();
}

// *string get_script_name() {}
static AbstractQoreNode* f_get_script_name(const QoreListNode* args, ExceptionSink* xsink) {
# 1980 "ql_misc.qpp"
   return getProgram()->getScriptName();
}

// *string get_script_path() {}
static AbstractQoreNode* f_get_script_path(const QoreListNode* args, ExceptionSink* xsink) {
# 1956 "ql_misc.qpp"
   return getProgram()->getScriptPath();
}

// *int get_word_16(string str, softint offset = 0) {}
static AbstractQoreNode* f_get_word_16_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1250 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)str->getBuffer();
   int size = str->strlen();

   if (offset >= (size - 1) || offset < 0)
      return 0;

   short val = ntohs(*((unsigned short *)&ptr[offset]));
   return new QoreBigIntNode(val);
}

// *int get_word_16(binary b, softint offset = 0) {}
static AbstractQoreNode* f_get_word_16_Vyvi(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* b = HARD_QORE_BINARY(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1276 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)b->getPtr();
   int size = b->size();

   if (offset >= (size - 1) || offset < 0)
      return 0;

   short val = ntohs(*((unsigned short *)&ptr[offset]));

   return new QoreBigIntNode(val);
}

// *int get_word_16_lsb(string str, softint offset = 0) {}
static AbstractQoreNode* f_get_word_16_lsb_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1476 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)str->getBuffer();
   int size = str->strlen();

   if (offset >= (size - 1) || offset < 0)
      return 0;

   short val = LSBi2(*((unsigned short *)&ptr[offset]));
   return new QoreBigIntNode(val);
}

// *int get_word_16_lsb(binary b, softint offset = 0) {}
static AbstractQoreNode* f_get_word_16_lsb_Vyvi(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* b = HARD_QORE_BINARY(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1502 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)b->getPtr();
   int size = b->size();

   if (offset >= (size - 1) || offset < 0)
      return 0;

   short val = LSBi2(*((unsigned short *)&ptr[offset]));
   return new QoreBigIntNode(val);
}

// *int get_word_32(string str, softint offset = 0) {}
static AbstractQoreNode* f_get_word_32_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1305 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)str->getBuffer();
   int size = str->strlen();

   if (offset >= (size - 3) || offset < 0)
      return 0;

   int64 val = ntohl(*((unsigned int *)&ptr[offset]));
   return new QoreBigIntNode(val);
}

// *int get_word_32(binary b, softint offset = 0) {}
static AbstractQoreNode* f_get_word_32_Vyvi(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* b = HARD_QORE_BINARY(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1333 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)b->getPtr();
   int size = b->size();

   if (offset >= (size - 3) || offset < 0)
      return 0;

   int64 val = ntohl(*((unsigned int *)&ptr[offset]));

   return new QoreBigIntNode(val);
}

// *int get_word_32_lsb(string str, softint offset = 0) {}
static AbstractQoreNode* f_get_word_32_lsb_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1530 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)str->getBuffer();
   int size = str->strlen();

   if (offset >= (size - 3) || offset < 0)
      return 0;

   int64 val = LSBi4(*((unsigned int *)&ptr[offset]));
   return new QoreBigIntNode(val);
}

// *int get_word_32_lsb(binary b, softint offset = 0) {}
static AbstractQoreNode* f_get_word_32_lsb_Vyvi(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* b = HARD_QORE_BINARY(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1556 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)b->getPtr();
   int size = b->size();

   if (offset >= (size - 3) || offset < 0)
      return 0;

   int64 val = LSBi4(*((unsigned int *)&ptr[offset]));

   return new QoreBigIntNode(val);
}

// *int get_word_64(string str, softint offset = 0) {}
static AbstractQoreNode* f_get_word_64_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1423 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)str->getBuffer();
   int size = str->strlen();

   if (offset >= (size - 7) || offset < 0)
      return 0;

   int64 val = MSBi8(*((int64 *)&ptr[offset]));
   return new QoreBigIntNode(val);
}

// *int get_word_64(binary b, softint offset = 0) {}
static AbstractQoreNode* f_get_word_64_Vyvi(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* b = HARD_QORE_BINARY(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1449 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)b->getPtr();
   int size = b->size();

   if (offset >= (size - 7) || offset < 0)
      return 0;

   int64 val = MSBi8(*((int64 *)&ptr[offset]));

   return new QoreBigIntNode(val);
}

// *int get_word_64_lsb(string str, softint offset = 0) {}
static AbstractQoreNode* f_get_word_64_lsb_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1583 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)str->getBuffer();
   int size = str->strlen();

   if (offset >= (size - 7) || offset < 0)
      return 0;

   int64 val = LSBi8(*((int64 *)&ptr[offset]));
   return new QoreBigIntNode(val);
}

// *int get_word_64_lsb(binary b, softint offset = 0) {}
static AbstractQoreNode* f_get_word_64_lsb_Vyvi(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* b = HARD_QORE_BINARY(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 1609 "ql_misc.qpp"
   unsigned char *ptr = (unsigned char *)b->getPtr();
   int size = b->size();

   if (offset >= (size - 7) || offset < 0)
      return 0;

   int64 val = LSBi8(*((int64 *)&ptr[offset]));
   return new QoreBigIntNode(val);
}

// bool has_key(hash h, string key) {}
static bool f_has_key_VhVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreHashNode* h = HARD_QORE_HASH(args, 0);
   const QoreStringNode* key = HARD_QORE_STRING(args, 1);
# 1102 "ql_misc.qpp"
   bool exists;
   h->getKeyValueExistence(key, exists, xsink);
   return exists;
}

// bool has_key(object obj, string key) {}
static bool f_has_key_VoVs(const QoreListNode* args, ExceptionSink* xsink) {
   QoreObject* obj = HARD_QORE_OBJECT(args, 0);
   const QoreStringNode* key = HARD_QORE_STRING(args, 1);
# 1119 "ql_misc.qpp"
   AutoVLock vl(xsink);
   AbstractQoreNode **ptr = obj->getExistingValuePtr(key, &vl, xsink);
   return (bool)ptr;
}

// list hash_values(hash h) {}
static AbstractQoreNode* f_hash_values_Vh(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreHashNode* h = HARD_QORE_HASH(args, 0);
# 1072 "ql_misc.qpp"
   QoreListNode *l = new QoreListNode;

   ConstHashIterator hi(h);
   while (hi.next())
      l->push(hi.getReferencedValue());

   return l;
}

// nothing hash_values() {}
static AbstractQoreNode* f_hash_values(const QoreListNode* args, ExceptionSink* xsink) {
# 1085 "ql_misc.qpp"

   return 0;
}

// int hextoint(string str) {}
static int64 f_hextoint_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
# 1808 "ql_misc.qpp"
   if (!str->strlen())
      return 0;

   int64 rc = 0;
   int64 pow = 0;
   const char *buf = str->getBuffer();
   qore_size_t len = str->size();
   if (*buf == '0' && *(buf + 1) == 'x') {
      buf += 2;
      len -= 2;
   }
   else if (*buf == 'x') {
      buf++;
      --len;
   }
   for (const char *p = buf + len - 1; p >= buf; p--) {
      int n = get_nibble(*p, xsink);
      if (xsink->isException())
	 return 0;
      if (!pow) {
	 rc = n;
	 pow = 16;
      }
      else {
	 rc += n * pow;
	 pow *= 16;
      }
   }
   return rc;
}

// nothing hextoint() {}
static AbstractQoreNode* f_hextoint(const QoreListNode* args, ExceptionSink* xsink) {
# 1843 "ql_misc.qpp"

   return 0;
}

// string html_decode(string str) {}
static AbstractQoreNode* f_html_decode_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
# 695 "ql_misc.qpp"
   QoreStringNode *ns = new QoreStringNode(str->getEncoding());
   ns->concatAndHTMLDecode(str);
   return ns;
}

// nothing html_decode() {}
static AbstractQoreNode* f_html_decode(const QoreListNode* args, ExceptionSink* xsink) {
# 704 "ql_misc.qpp"

   return 0;
}

// string html_encode(string str) {}
static AbstractQoreNode* f_html_encode_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
# 673 "ql_misc.qpp"
   QoreStringNode *ns = new QoreStringNode(str->getEncoding());
   ns->concatAndHTMLEncode(str->getBuffer());
   return ns;
}

// nothing html_encode() {}
static AbstractQoreNode* f_html_encode(const QoreListNode* args, ExceptionSink* xsink) {
# 682 "ql_misc.qpp"

   return 0;
}

// nothing load_module(string name) {}
static AbstractQoreNode* f_load_module_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* name = HARD_QORE_STRING(args, 0);
# 1891 "ql_misc.qpp"
   MM.runTimeLoadModule(name->getBuffer(), xsink);
   return 0;
}

// nothing load_module() {}
static AbstractQoreNode* f_load_module(const QoreListNode* args, ExceptionSink* xsink) {
# 1898 "ql_misc.qpp"

   return 0;
}

// string makeBase64String(string str, softint maxlinelen = -1) {}
static AbstractQoreNode* f_makeBase64String_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 maxlinelen = HARD_QORE_INT(args, 1);
# 909 "ql_misc.qpp"
   QoreStringNode *rv = new QoreStringNode;
   rv->concatBase64(str, maxlinelen);
   return rv;
}

// string makeBase64String(binary bin, softint maxlinelen = -1) {}
static AbstractQoreNode* f_makeBase64String_Vyvi(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
   int64 maxlinelen = HARD_QORE_INT(args, 1);
# 934 "ql_misc.qpp"
   return new QoreStringNode(bin, maxlinelen);
}

// nothing makeBase64String() {}
static AbstractQoreNode* f_makeBase64String(const QoreListNode* args, ExceptionSink* xsink) {
# 941 "ql_misc.qpp"

   return 0;
}

// string makeHexString(string str) {}
static AbstractQoreNode* f_makeHexString_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
# 1738 "ql_misc.qpp"
   QoreStringNode *rv = new QoreStringNode;
   rv->concatHex(str);
   return rv;
}

// string makeHexString(binary bin) {}
static AbstractQoreNode* f_makeHexString_Vy(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
# 1759 "ql_misc.qpp"
   QoreStringNode *str = new QoreStringNode;
   str->concatHex(bin);
   return str;
}

// nothing makeHexString() {}
static AbstractQoreNode* f_makeHexString(const QoreListNode* args, ExceptionSink* xsink) {
# 1768 "ql_misc.qpp"

   return 0;
}

// *hash parse(string code, string label, *softint warning_mask, *string source, *softint offset, softbool format_label = True) {}
static AbstractQoreNode* f_parse_VsVsniNsnivb(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* code = HARD_QORE_STRING(args, 0);
   const QoreStringNode* label = HARD_QORE_STRING(args, 1);
   const QoreBigIntNode* tmp_warning_mask = reinterpret_cast<const QoreBigIntNode*>(get_param(args, 2));
   int64 warning_mask = tmp_warning_mask ? tmp_warning_mask->val : 0;
   const QoreStringNode* source = reinterpret_cast<const QoreStringNode*>(get_param(args, 3));
   const QoreBigIntNode* tmp_offset = reinterpret_cast<const QoreBigIntNode*>(get_param(args, 4));
   int64 offset = tmp_offset ? tmp_offset->val : 0;
   bool format_label = HARD_QORE_BOOL(args, 5);
# 743 "ql_misc.qpp"
   QoreStringMaker nlabel(label->getEncoding(), format_label ? "<run-time-loaded: %s>" : "%s", label->getBuffer());

   if (warning_mask) {
      ExceptionSink wsink;
      getProgram()->parse(code, &nlabel, xsink, &wsink, warning_mask, source, offset);
      if (!wsink.isException())
         return 0;

      QoreException *e = wsink.catchException();
      return e->makeExceptionObjectAndDelete(xsink);
   }

   getProgram()->parse(code, &nlabel, xsink, 0, 0, source, offset);
   return 0;
}

// nothing parse() {}
static AbstractQoreNode* f_parse(const QoreListNode* args, ExceptionSink* xsink) {
# 762 "ql_misc.qpp"

   return 0;
}

// binary parseBase64String(string str) {}
static AbstractQoreNode* f_parseBase64String_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
# 960 "ql_misc.qpp"
   return str->parseBase64(xsink);
}

// nothing parseBase64String() {}
static AbstractQoreNode* f_parseBase64String(const QoreListNode* args, ExceptionSink* xsink) {
# 967 "ql_misc.qpp"

   return 0;
}

// string parseBase64StringToString(string str, *string encoding) {}
static AbstractQoreNode* f_parseBase64StringToString_VsNs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   const QoreStringNode* encoding = reinterpret_cast<const QoreStringNode*>(get_param(args, 1));
# 989 "ql_misc.qpp"
   const QoreEncoding *qe = encoding ? QEM.findCreate(encoding) : QCS_DEFAULT;
   return str->parseBase64ToString(qe, xsink);
}

// nothing parseBase64StringToString() {}
static AbstractQoreNode* f_parseBase64StringToString(const QoreListNode* args, ExceptionSink* xsink) {
# 997 "ql_misc.qpp"

   return 0;
}

// binary parseHexString(string hexstr) {}
static AbstractQoreNode* f_parseHexString_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* hexstr = HARD_QORE_STRING(args, 0);
# 1786 "ql_misc.qpp"
   return hexstr->parseHex(xsink);
}

// nothing parseHexString() {}
static AbstractQoreNode* f_parseHexString(const QoreListNode* args, ExceptionSink* xsink) {
# 1793 "ql_misc.qpp"

   return 0;
}

// *hash parseURL(string url, bool keep_brackets = False) {}
static AbstractQoreNode* f_parseURL_VsVb(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* url = HARD_QORE_STRING(args, 0);
   bool keep_brackets = HARD_QORE_BOOL(args, 1);
# 839 "ql_misc.qpp"
   QoreURL qurl(url, keep_brackets);
   return qurl.isValid() ? qurl.getHash() : 0;
}

// nothing parseURL() {}
static AbstractQoreNode* f_parseURL(const QoreListNode* args, ExceptionSink* xsink) {
# 847 "ql_misc.qpp"

   return 0;
}

// hash parse_url(string url, bool keep_brackets = False) {}
static AbstractQoreNode* f_parse_url_VsVb(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* url = HARD_QORE_STRING(args, 0);
   bool keep_brackets = HARD_QORE_BOOL(args, 1);
# 808 "ql_misc.qpp"
   QoreURL qurl(url, keep_brackets);
   if (!qurl.isValid()) {
      xsink->raiseException("PARSE-URL-ERROR", "URL '%s' cannot be parsed", url->getBuffer());
      return 0;
   }

   return qurl.getHash();
}

// string splice(string str) {}
static AbstractQoreNode* f_splice_Vs(const QoreListNode* args, ExceptionSink* xsink) {
# 1627 "ql_misc.qpp"
   return new QoreStringNode;
}

// string splice(string str, softint start) {}
static AbstractQoreNode* f_splice_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 start = HARD_QORE_INT(args, 1);
# 1646 "ql_misc.qpp"
   QoreStringNodeHolder pstr(str->copy());
   pstr->splice(start, xsink);
   return *xsink ? 0 : pstr.release();
}

// string splice(string str, softint start, softint len, *string nstr) {}
static AbstractQoreNode* f_splice_VsviviNs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 start = HARD_QORE_INT(args, 1);
   int64 len = HARD_QORE_INT(args, 2);
   const QoreStringNode* nstr = reinterpret_cast<const QoreStringNode*>(get_param(args, 3));
# 1669 "ql_misc.qpp"
   QoreStringNodeHolder pstr(str->copy());
   pstr->splice(start, len, nstr, xsink);
   return *xsink ? 0 : pstr.release();
}

// list splice(list l, softint start) {}
static AbstractQoreNode* f_splice_Vlvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreListNode* l = HARD_QORE_LIST(args, 0);
   int64 start = HARD_QORE_INT(args, 1);
# 1688 "ql_misc.qpp"
   ReferenceHolder<QoreListNode> lst(l->copy(), xsink);
   lst->splice(start, xsink);
   return *xsink ? 0 : lst.release();
}

// list splice(list l, softint start, softint len, *softlist nlist) {}
static AbstractQoreNode* f_splice_Vlvivinl(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreListNode* l = HARD_QORE_LIST(args, 0);
   int64 start = HARD_QORE_INT(args, 1);
   int64 len = HARD_QORE_INT(args, 2);
   const QoreListNode* nlist = reinterpret_cast<const QoreListNode*>(get_param(args, 3));
# 1711 "ql_misc.qpp"
   ReferenceHolder<QoreListNode> lst(l->copy(), xsink);
   lst->splice(start, len, nlist, xsink);
   return *xsink ? 0 : lst.release();
}

// nothing splice() {}
static AbstractQoreNode* f_splice(const QoreListNode* args, ExceptionSink* xsink) {
# 1720 "ql_misc.qpp"

   return 0;
}

// int strtoint(string num, softint base = 10) {}
static int64 f_strtoint_Vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* num = HARD_QORE_STRING(args, 0);
   int64 base = HARD_QORE_INT(args, 1);
# 1859 "ql_misc.qpp"
   errno = 0;
   int64 rv = strtoll(num->getBuffer(), 0, base);
   if (errno == EINVAL || errno == ERANGE)
      xsink->raiseException("STRTOINT-ERROR", "could not parse '%s' to an integer", num->getBuffer());
   return rv;
}

// nothing strtoint() {}
static AbstractQoreNode* f_strtoint(const QoreListNode* args, ExceptionSink* xsink) {
# 1870 "ql_misc.qpp"

   return 0;
}


DLLLOCAL void init_misc_functions(QoreNamespace& ns) {
   // nothing remove_signal_handler(softint signal) {}
   ns.addBuiltinVariant("remove_signal_handler", (q_func_t)f_remove_signal_handler_vi, QC_NO_FLAGS, QDOM_PROCESS, nothingTypeInfo, 1, softBigIntTypeInfo, NULL, "signal");
   // nothing set_signal_handler(softint signal, code f) {}
   ns.addBuiltinVariant("set_signal_handler", (q_func_t)f_set_signal_handler_viVc, QC_NO_FLAGS, QDOM_PROCESS, nothingTypeInfo, 2, softBigIntTypeInfo, NULL, "signal", codeTypeInfo, NULL, "f");
   // string backquote(string cmd, *reference rc) {}
   ns.addBuiltinVariant("backquote", (q_func_t)f_backquote_VsNr, QC_NO_FLAGS, QDOM_EXTERNAL_PROCESS, stringTypeInfo, 2, stringTypeInfo, NULL, "cmd", anyTypeInfo, NULL, "rc");
   // nothing backquote() {}
   ns.addBuiltinVariant("backquote", (q_func_t)f_backquote, QC_RUNTIME_NOOP, QDOM_EXTERNAL_PROCESS, nothingTypeInfo);
   // any call_builtin_function(string name, ...) {}
   ns.addBuiltinVariant("call_builtin_function", (q_func_t)f_call_builtin_function_VsVV, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, anyTypeInfo, 1, stringTypeInfo, NULL, "name");
   // any call_builtin_function_args(string name, *softlist vargs) {}
   ns.addBuiltinVariant("call_builtin_function_args", (q_func_t)f_call_builtin_function_args_Vsnl, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 2, stringTypeInfo, NULL, "name", softListOrNothingTypeInfo, NULL, "vargs");
   // any call_function(string name, ...) {}
   ns.addBuiltinVariant("call_function", (q_func_t)f_call_function_VsVV, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, anyTypeInfo, 1, stringTypeInfo, NULL, "name");
   // any call_function(code f, ...) {}
   ns.addBuiltinVariant("call_function", (q_func_t)f_call_function_VcVV, QC_USES_EXTRA_ARGS, QDOM_DEFAULT, anyTypeInfo, 1, codeTypeInfo, NULL, "f");
   // any call_function_args(string name, *softlist vargs) {}
   ns.addBuiltinVariant("call_function_args", (q_func_t)f_call_function_args_Vsnl, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 2, stringTypeInfo, NULL, "name", softListOrNothingTypeInfo, NULL, "vargs");
   // any call_function_args(code f, *softlist vargs) {}
   ns.addBuiltinVariant("call_function_args", (q_func_t)f_call_function_args_Vcnl, QC_NO_FLAGS, QDOM_DEFAULT, anyTypeInfo, 2, codeTypeInfo, NULL, "f", softListOrNothingTypeInfo, NULL, "vargs");
   // string decode_url(string url) {}
   ns.addBuiltinVariant("decode_url", (q_func_t)f_decode_url_Vs, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, NULL, "url");
   // nothing decode_url() {}
   ns.addBuiltinVariant("decode_url", (q_func_t)f_decode_url, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string encode_url(string url, softbool encode_all = False) {}
   ns.addBuiltinVariant("encode_url", (q_func_t)f_encode_url_Vsvb, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 2, stringTypeInfo, NULL, "url", softBoolTypeInfo, &False, "encode_all");
   // bool exists(...) {}
   ns.addBuiltinVariant("exists", (q_func_bool_t)f_exists_VV, QC_CONSTANT|QC_USES_EXTRA_ARGS, QDOM_DEFAULT, boolTypeInfo);
   // bool existsFunction(string name) {}
   ns.addBuiltinVariant("existsFunction", (q_func_bool_t)f_existsFunction_Vs, QC_CONSTANT, QDOM_DEFAULT, boolTypeInfo, 1, stringTypeInfo, NULL, "name");
   // bool existsFunction(code c) {}
   ns.addBuiltinVariant("existsFunction", (q_func_bool_t)f_existsFunction_Vc, QC_NOOP, QDOM_DEFAULT, boolTypeInfo, 1, codeTypeInfo, NULL, "c");
   // nothing existsFunction() {}
   ns.addBuiltinVariant("existsFunction", (q_func_t)f_existsFunction, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // *string functionType(string name) {}
   ns.addBuiltinVariant("functionType", (q_func_t)f_functionType_Vs, QC_NO_FLAGS, QDOM_DEFAULT, stringOrNothingTypeInfo, 1, stringTypeInfo, NULL, "name");
   // nothing functionType() {}
   ns.addBuiltinVariant("functionType", (q_func_t)f_functionType, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // *int getByte(string str, softint offset = 0) {}
   ns.addBuiltinVariant("getByte", (q_func_t)f_getByte_Vsvi, QC_CONSTANT|QC_DEPRECATED, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, stringTypeInfo, NULL, "str", softBigIntTypeInfo, zero(), "offset");
   // nothing getByte() {}
   ns.addBuiltinVariant("getByte", (q_func_t)f_getByte, QC_RUNTIME_NOOP|QC_DEPRECATED, QDOM_DEFAULT, nothingTypeInfo);
   // *int getByte(binary b, softint offset = 0) {}
   ns.addBuiltinVariant("getByte", (q_func_t)f_getByte_Vyvi, QC_CONSTANT|QC_DEPRECATED, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, binaryTypeInfo, NULL, "b", softBigIntTypeInfo, zero(), "offset");
   // string getClassName(object obj) {}
   ns.addBuiltinVariant("getClassName", (q_func_t)f_getClassName_Vo, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 1, objectTypeInfo, NULL, "obj");
   // nothing getClassName() {}
   ns.addBuiltinVariant("getClassName", (q_func_t)f_getClassName, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // list getFeatureList() {}
   ns.addBuiltinVariant("getFeatureList", (q_func_t)f_getFeatureList, QC_CONSTANT, QDOM_DEFAULT, listTypeInfo);
   // hash getModuleHash() {}
   ns.addBuiltinVariant("getModuleHash", (q_func_t)f_getModuleHash, QC_CONSTANT, QDOM_DEFAULT, hashTypeInfo);
   // list getModuleList() {}
   ns.addBuiltinVariant("getModuleList", (q_func_t)f_getModuleList, QC_CONSTANT, QDOM_DEFAULT, listTypeInfo);
   // *int getWord32(string str, softint offset = 0) {}
   ns.addBuiltinVariant("getWord32", (q_func_t)f_getWord32_Vsvi, QC_CONSTANT|QC_DEPRECATED, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, stringTypeInfo, NULL, "str", softBigIntTypeInfo, zero(), "offset");
   // *int getWord32(binary b, softint offset = 0) {}
   ns.addBuiltinVariant("getWord32", (q_func_t)f_getWord32_Vyvi, QC_CONSTANT|QC_DEPRECATED, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, binaryTypeInfo, NULL, "b", softBigIntTypeInfo, zero(), "offset");
   // nothing getWord32() {}
   ns.addBuiltinVariant("getWord32", (q_func_t)f_getWord32, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // *int get_byte(string str, softint offset = 0) {}
   ns.addBuiltinVariant("get_byte", (q_func_t)f_get_byte_Vsvi, QC_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, stringTypeInfo, NULL, "str", softBigIntTypeInfo, zero(), "offset");
   // *int get_byte(binary b, softint offset = 0) {}
   ns.addBuiltinVariant("get_byte", (q_func_t)f_get_byte_Vyvi, QC_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, binaryTypeInfo, NULL, "b", softBigIntTypeInfo, zero(), "offset");
   // string get_default_encoding() {}
   ns.addBuiltinVariant("get_default_encoding", (q_func_t)f_get_default_encoding, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo);
   // string get_ex_pos(hash ex) {}
   ns.addBuiltinVariant("get_ex_pos", (q_func_t)f_get_ex_pos_Vh, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 1, hashTypeInfo, NULL, "ex");
   // int get_parse_options() {}
   ns.addBuiltinVariant("get_parse_options", (q_func_int64_t)f_get_parse_options, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);
   // hash get_qore_library_info() {}
   ns.addBuiltinVariant("get_qore_library_info", (q_func_t)f_get_qore_library_info, QC_CONSTANT, QDOM_DEFAULT, hashTypeInfo);
   // hash get_qore_option_hash() {}
   ns.addBuiltinVariant("get_qore_option_hash", (q_func_t)f_get_qore_option_hash, QC_CONSTANT, QDOM_DEFAULT, hashTypeInfo);
   // list get_qore_option_list() {}
   ns.addBuiltinVariant("get_qore_option_list", (q_func_t)f_get_qore_option_list, QC_CONSTANT, QDOM_DEFAULT, listTypeInfo);
   // *string get_script_dir() {}
   ns.addBuiltinVariant("get_script_dir", (q_func_t)f_get_script_dir, QC_NO_FLAGS, QDOM_DEFAULT, stringOrNothingTypeInfo);
   // *string get_script_name() {}
   ns.addBuiltinVariant("get_script_name", (q_func_t)f_get_script_name, QC_NO_FLAGS, QDOM_DEFAULT, stringOrNothingTypeInfo);
   // *string get_script_path() {}
   ns.addBuiltinVariant("get_script_path", (q_func_t)f_get_script_path, QC_NO_FLAGS, QDOM_DEFAULT, stringOrNothingTypeInfo);
   // *int get_word_16(string str, softint offset = 0) {}
   ns.addBuiltinVariant("get_word_16", (q_func_t)f_get_word_16_Vsvi, QC_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, stringTypeInfo, NULL, "str", softBigIntTypeInfo, zero(), "offset");
   // *int get_word_16(binary b, softint offset = 0) {}
   ns.addBuiltinVariant("get_word_16", (q_func_t)f_get_word_16_Vyvi, QC_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, binaryTypeInfo, NULL, "b", softBigIntTypeInfo, zero(), "offset");
   // *int get_word_16_lsb(string str, softint offset = 0) {}
   ns.addBuiltinVariant("get_word_16_lsb", (q_func_t)f_get_word_16_lsb_Vsvi, QC_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, stringTypeInfo, NULL, "str", softBigIntTypeInfo, zero(), "offset");
   // *int get_word_16_lsb(binary b, softint offset = 0) {}
   ns.addBuiltinVariant("get_word_16_lsb", (q_func_t)f_get_word_16_lsb_Vyvi, QC_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, binaryTypeInfo, NULL, "b", softBigIntTypeInfo, zero(), "offset");
   // *int get_word_32(string str, softint offset = 0) {}
   ns.addBuiltinVariant("get_word_32", (q_func_t)f_get_word_32_Vsvi, QC_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, stringTypeInfo, NULL, "str", softBigIntTypeInfo, zero(), "offset");
   // *int get_word_32(binary b, softint offset = 0) {}
   ns.addBuiltinVariant("get_word_32", (q_func_t)f_get_word_32_Vyvi, QC_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, binaryTypeInfo, NULL, "b", softBigIntTypeInfo, zero(), "offset");
   // *int get_word_32_lsb(string str, softint offset = 0) {}
   ns.addBuiltinVariant("get_word_32_lsb", (q_func_t)f_get_word_32_lsb_Vsvi, QC_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, stringTypeInfo, NULL, "str", softBigIntTypeInfo, zero(), "offset");
   // *int get_word_32_lsb(binary b, softint offset = 0) {}
   ns.addBuiltinVariant("get_word_32_lsb", (q_func_t)f_get_word_32_lsb_Vyvi, QC_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, binaryTypeInfo, NULL, "b", softBigIntTypeInfo, zero(), "offset");
   // *int get_word_64(string str, softint offset = 0) {}
   ns.addBuiltinVariant("get_word_64", (q_func_t)f_get_word_64_Vsvi, QC_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, stringTypeInfo, NULL, "str", softBigIntTypeInfo, zero(), "offset");
   // *int get_word_64(binary b, softint offset = 0) {}
   ns.addBuiltinVariant("get_word_64", (q_func_t)f_get_word_64_Vyvi, QC_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, binaryTypeInfo, NULL, "b", softBigIntTypeInfo, zero(), "offset");
   // *int get_word_64_lsb(string str, softint offset = 0) {}
   ns.addBuiltinVariant("get_word_64_lsb", (q_func_t)f_get_word_64_lsb_Vsvi, QC_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, stringTypeInfo, NULL, "str", softBigIntTypeInfo, zero(), "offset");
   // *int get_word_64_lsb(binary b, softint offset = 0) {}
   ns.addBuiltinVariant("get_word_64_lsb", (q_func_t)f_get_word_64_lsb_Vyvi, QC_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo, 2, binaryTypeInfo, NULL, "b", softBigIntTypeInfo, zero(), "offset");
   // bool has_key(hash h, string key) {}
   ns.addBuiltinVariant("has_key", (q_func_bool_t)f_has_key_VhVs, QC_NO_FLAGS, QDOM_DEFAULT, boolTypeInfo, 2, hashTypeInfo, NULL, "h", stringTypeInfo, NULL, "key");
   // bool has_key(object obj, string key) {}
   ns.addBuiltinVariant("has_key", (q_func_bool_t)f_has_key_VoVs, QC_NO_FLAGS, QDOM_DEFAULT, boolTypeInfo, 2, objectTypeInfo, NULL, "obj", stringTypeInfo, NULL, "key");
   // list hash_values(hash h) {}
   ns.addBuiltinVariant("hash_values", (q_func_t)f_hash_values_Vh, QC_CONSTANT, QDOM_DEFAULT, listTypeInfo, 1, hashTypeInfo, NULL, "h");
   // nothing hash_values() {}
   ns.addBuiltinVariant("hash_values", (q_func_t)f_hash_values, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int hextoint(string str) {}
   ns.addBuiltinVariant("hextoint", (q_func_int64_t)f_hextoint_Vs, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, stringTypeInfo, NULL, "str");
   // nothing hextoint() {}
   ns.addBuiltinVariant("hextoint", (q_func_t)f_hextoint, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string html_decode(string str) {}
   ns.addBuiltinVariant("html_decode", (q_func_t)f_html_decode_Vs, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, NULL, "str");
   // nothing html_decode() {}
   ns.addBuiltinVariant("html_decode", (q_func_t)f_html_decode, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string html_encode(string str) {}
   ns.addBuiltinVariant("html_encode", (q_func_t)f_html_encode_Vs, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, NULL, "str");
   // nothing html_encode() {}
   ns.addBuiltinVariant("html_encode", (q_func_t)f_html_encode, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // nothing load_module(string name) {}
   ns.addBuiltinVariant("load_module", (q_func_t)f_load_module_Vs, QC_NO_FLAGS, QDOM_MODULES, nothingTypeInfo, 1, stringTypeInfo, NULL, "name");
   // nothing load_module() {}
   ns.addBuiltinVariant("load_module", (q_func_t)f_load_module, QC_RUNTIME_NOOP, QDOM_MODULES, nothingTypeInfo);
   // string makeBase64String(string str, softint maxlinelen = -1) {}
   ns.addBuiltinVariant("makeBase64String", (q_func_t)f_makeBase64String_Vsvi, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 2, stringTypeInfo, NULL, "str", softBigIntTypeInfo, new QoreBigIntNode(-1), "maxlinelen");
   // string makeBase64String(binary bin, softint maxlinelen = -1) {}
   ns.addBuiltinVariant("makeBase64String", (q_func_t)f_makeBase64String_Vyvi, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 2, binaryTypeInfo, NULL, "bin", softBigIntTypeInfo, new QoreBigIntNode(-1), "maxlinelen");
   // nothing makeBase64String() {}
   ns.addBuiltinVariant("makeBase64String", (q_func_t)f_makeBase64String, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string makeHexString(string str) {}
   ns.addBuiltinVariant("makeHexString", (q_func_t)f_makeHexString_Vs, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, NULL, "str");
   // string makeHexString(binary bin) {}
   ns.addBuiltinVariant("makeHexString", (q_func_t)f_makeHexString_Vy, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 1, binaryTypeInfo, NULL, "bin");
   // nothing makeHexString() {}
   ns.addBuiltinVariant("makeHexString", (q_func_t)f_makeHexString, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // *hash parse(string code, string label, *softint warning_mask, *string source, *softint offset, softbool format_label = True) {}
   ns.addBuiltinVariant("parse", (q_func_t)f_parse_VsVsniNsnivb, QC_NO_FLAGS, QDOM_EMBEDDED_LOGIC|QDOM_IN_MODULE, hashOrNothingTypeInfo, 6, stringTypeInfo, NULL, "code", stringTypeInfo, NULL, "label", softBigIntOrNothingTypeInfo, NULL, "warning_mask", stringOrNothingTypeInfo, NULL, "source", softBigIntOrNothingTypeInfo, NULL, "offset", softBoolTypeInfo, &True, "format_label");
   // nothing parse() {}
   ns.addBuiltinVariant("parse", (q_func_t)f_parse, QC_RUNTIME_NOOP, QDOM_EMBEDDED_LOGIC|QDOM_IN_MODULE, nothingTypeInfo);
   // binary parseBase64String(string str) {}
   ns.addBuiltinVariant("parseBase64String", (q_func_t)f_parseBase64String_Vs, QC_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo, 1, stringTypeInfo, NULL, "str");
   // nothing parseBase64String() {}
   ns.addBuiltinVariant("parseBase64String", (q_func_t)f_parseBase64String, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string parseBase64StringToString(string str, *string encoding) {}
   ns.addBuiltinVariant("parseBase64StringToString", (q_func_t)f_parseBase64StringToString_VsNs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, stringTypeInfo, NULL, "str", stringOrNothingTypeInfo, NULL, "encoding");
   // nothing parseBase64StringToString() {}
   ns.addBuiltinVariant("parseBase64StringToString", (q_func_t)f_parseBase64StringToString, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // binary parseHexString(string hexstr) {}
   ns.addBuiltinVariant("parseHexString", (q_func_t)f_parseHexString_Vs, QC_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo, 1, stringTypeInfo, NULL, "hexstr");
   // nothing parseHexString() {}
   ns.addBuiltinVariant("parseHexString", (q_func_t)f_parseHexString, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // *hash parseURL(string url, bool keep_brackets = False) {}
   ns.addBuiltinVariant("parseURL", (q_func_t)f_parseURL_VsVb, QC_NO_FLAGS, QDOM_DEFAULT, hashOrNothingTypeInfo, 2, stringTypeInfo, NULL, "url", boolTypeInfo, &False, "keep_brackets");
   // nothing parseURL() {}
   ns.addBuiltinVariant("parseURL", (q_func_t)f_parseURL, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // hash parse_url(string url, bool keep_brackets = False) {}
   ns.addBuiltinVariant("parse_url", (q_func_t)f_parse_url_VsVb, QC_NO_FLAGS, QDOM_DEFAULT, hashTypeInfo, 2, stringTypeInfo, NULL, "url", boolTypeInfo, &False, "keep_brackets");
   // string splice(string str) {}
   ns.addBuiltinVariant("splice", (q_func_t)f_splice_Vs, QC_NOOP, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, NULL, "str");
   // string splice(string str, softint start) {}
   ns.addBuiltinVariant("splice", (q_func_t)f_splice_Vsvi, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, stringTypeInfo, NULL, "str", softBigIntTypeInfo, NULL, "start");
   // string splice(string str, softint start, softint len, *string nstr) {}
   ns.addBuiltinVariant("splice", (q_func_t)f_splice_VsviviNs, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 4, stringTypeInfo, NULL, "str", softBigIntTypeInfo, NULL, "start", softBigIntTypeInfo, NULL, "len", stringOrNothingTypeInfo, NULL, "nstr");
   // list splice(list l, softint start) {}
   ns.addBuiltinVariant("splice", (q_func_t)f_splice_Vlvi, QC_NO_FLAGS, QDOM_DEFAULT, listTypeInfo, 2, listTypeInfo, NULL, "l", softBigIntTypeInfo, NULL, "start");
   // list splice(list l, softint start, softint len, *softlist nlist) {}
   ns.addBuiltinVariant("splice", (q_func_t)f_splice_Vlvivinl, QC_NO_FLAGS, QDOM_DEFAULT, listTypeInfo, 4, listTypeInfo, NULL, "l", softBigIntTypeInfo, NULL, "start", softBigIntTypeInfo, NULL, "len", softListOrNothingTypeInfo, NULL, "nlist");
   // nothing splice() {}
   ns.addBuiltinVariant("splice", (q_func_t)f_splice, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int strtoint(string num, softint base = 10) {}
   ns.addBuiltinVariant("strtoint", (q_func_int64_t)f_strtoint_Vsvi, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, stringTypeInfo, NULL, "num", softBigIntTypeInfo, new QoreBigIntNode(10), "base");
   // nothing strtoint() {}
   ns.addBuiltinVariant("strtoint", (q_func_t)f_strtoint, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);

}

DLLLOCAL void init_misc_constants(QoreNamespace& ns) {
# 183 "ql_misc.qpp"
# 278 "ql_misc.qpp"
# 327 "ql_misc.qpp"
   QoreHashNode* hash_NameToSignal = new QoreHashNode;
   hash_NameToSignal->setKeyValue("SIGABRT", new QoreBigIntNode(SIGABRT), 0);
   hash_NameToSignal->setKeyValue("SIGALRM", new QoreBigIntNode(SIGALRM), 0);
   hash_NameToSignal->setKeyValue("SIGBUS", new QoreBigIntNode(SIGBUS), 0);
   hash_NameToSignal->setKeyValue("SIGCANCEL", new QoreBigIntNode(SIGCANCEL), 0);
   hash_NameToSignal->setKeyValue("SIGCHLD", new QoreBigIntNode(SIGCHLD), 0);
   hash_NameToSignal->setKeyValue("SIGCLD", new QoreBigIntNode(SIGCLD), 0);
   hash_NameToSignal->setKeyValue("SIGCONT", new QoreBigIntNode(SIGCONT), 0);
   hash_NameToSignal->setKeyValue("SIGEMT", new QoreBigIntNode(SIGEMT), 0);
   hash_NameToSignal->setKeyValue("SIGFPE", new QoreBigIntNode(SIGFPE), 0);
   hash_NameToSignal->setKeyValue("SIGFREEZE", new QoreBigIntNode(SIGFREEZE), 0);
   hash_NameToSignal->setKeyValue("SIGHUP", new QoreBigIntNode(SIGHUP), 0);
   hash_NameToSignal->setKeyValue("SIGILL", new QoreBigIntNode(SIGILL), 0);
   hash_NameToSignal->setKeyValue("SIGINFO", new QoreBigIntNode(SIGINFO), 0);
   hash_NameToSignal->setKeyValue("SIGINT", new QoreBigIntNode(SIGINT), 0);
   hash_NameToSignal->setKeyValue("SIGIO", new QoreBigIntNode(SIGIO), 0);
   hash_NameToSignal->setKeyValue("SIGIOT", new QoreBigIntNode(SIGIOT), 0);
   hash_NameToSignal->setKeyValue("SIGJVM1", new QoreBigIntNode(SIGJVM1), 0);
   hash_NameToSignal->setKeyValue("SIGJVM2", new QoreBigIntNode(SIGJVM2), 0);
   hash_NameToSignal->setKeyValue("SIGKILL", new QoreBigIntNode(SIGKILL), 0);
   hash_NameToSignal->setKeyValue("SIGLOST", new QoreBigIntNode(SIGLOST), 0);
   hash_NameToSignal->setKeyValue("SIGLWP", new QoreBigIntNode(SIGLWP), 0);
   hash_NameToSignal->setKeyValue("SIGPIPE", new QoreBigIntNode(SIGPIPE), 0);
   hash_NameToSignal->setKeyValue("SIGPOLL", new QoreBigIntNode(SIGPOLL), 0);
   hash_NameToSignal->setKeyValue("SIGPROF", new QoreBigIntNode(SIGPROF), 0);
   hash_NameToSignal->setKeyValue("SIGPWR", new QoreBigIntNode(SIGPWR), 0);
   hash_NameToSignal->setKeyValue("SIGQUIT", new QoreBigIntNode(SIGQUIT), 0);
   hash_NameToSignal->setKeyValue("SIGSEGV", new QoreBigIntNode(SIGSEGV), 0);
   hash_NameToSignal->setKeyValue("SIGSTKFLT", new QoreBigIntNode(SIGSTKFLT), 0);
   hash_NameToSignal->setKeyValue("SIGSTOP", new QoreBigIntNode(SIGSTOP), 0);
   hash_NameToSignal->setKeyValue("SIGSYS", new QoreBigIntNode(SIGSYS), 0);
   hash_NameToSignal->setKeyValue("SIGTERM", new QoreBigIntNode(SIGTERM), 0);
   hash_NameToSignal->setKeyValue("SIGTHAW", new QoreBigIntNode(SIGTHAW), 0);
   hash_NameToSignal->setKeyValue("SIGTRAP", new QoreBigIntNode(SIGTRAP), 0);
   hash_NameToSignal->setKeyValue("SIGTSTP", new QoreBigIntNode(SIGTSTP), 0);
   hash_NameToSignal->setKeyValue("SIGTTIN", new QoreBigIntNode(SIGTTIN), 0);
   hash_NameToSignal->setKeyValue("SIGTTOU", new QoreBigIntNode(SIGTTOU), 0);
   hash_NameToSignal->setKeyValue("SIGURG", new QoreBigIntNode(SIGURG), 0);
   hash_NameToSignal->setKeyValue("SIGUSR1", new QoreBigIntNode(SIGUSR1), 0);
   hash_NameToSignal->setKeyValue("SIGUSR2", new QoreBigIntNode(SIGUSR2), 0);
   hash_NameToSignal->setKeyValue("SIGVTALRM", new QoreBigIntNode(SIGVTALRM), 0);
   hash_NameToSignal->setKeyValue("SIGWAITING", new QoreBigIntNode(SIGWAITING), 0);
   hash_NameToSignal->setKeyValue("SIGWINCH", new QoreBigIntNode(SIGWINCH), 0);
   hash_NameToSignal->setKeyValue("SIGXCPU", new QoreBigIntNode(SIGXCPU), 0);
   hash_NameToSignal->setKeyValue("SIGXFSZ", new QoreBigIntNode(SIGXFSZ), 0);
   hash_NameToSignal->setKeyValue("SIGXRES", new QoreBigIntNode(SIGXRES), 0);

   ns.addConstant("NameToSignal", hash_NameToSignal);
   ns.addConstant("SIGABRT", new QoreBigIntNode(SIGABRT));
   ns.addConstant("SIGALRM", new QoreBigIntNode(SIGALRM));
   ns.addConstant("SIGBUS", new QoreBigIntNode(SIGBUS));
   ns.addConstant("SIGCANCEL", new QoreBigIntNode(SIGCANCEL));
   ns.addConstant("SIGCHLD", new QoreBigIntNode(SIGCHLD));
   ns.addConstant("SIGCLD", new QoreBigIntNode(SIGCLD));
   ns.addConstant("SIGCONT", new QoreBigIntNode(SIGCONT));
   ns.addConstant("SIGEMT", new QoreBigIntNode(SIGEMT));
   ns.addConstant("SIGFPE", new QoreBigIntNode(SIGFPE));
   ns.addConstant("SIGFREEZE", new QoreBigIntNode(SIGFREEZE));
   ns.addConstant("SIGHUP", new QoreBigIntNode(SIGHUP));
   ns.addConstant("SIGILL", new QoreBigIntNode(SIGILL));
   ns.addConstant("SIGINFO", new QoreBigIntNode(SIGINFO));
   ns.addConstant("SIGINT", new QoreBigIntNode(SIGINT));
   ns.addConstant("SIGIO", new QoreBigIntNode(SIGIO));
   ns.addConstant("SIGIOT", new QoreBigIntNode(SIGIOT));
   ns.addConstant("SIGJVM1", new QoreBigIntNode(SIGJVM1));
   ns.addConstant("SIGJVM2", new QoreBigIntNode(SIGJVM2));
   ns.addConstant("SIGKILL", new QoreBigIntNode(SIGKILL));
   ns.addConstant("SIGLOST", new QoreBigIntNode(SIGLOST));
   ns.addConstant("SIGLWP", new QoreBigIntNode(SIGLWP));
   ns.addConstant("SIGPIPE", new QoreBigIntNode(SIGPIPE));
   ns.addConstant("SIGPOLL", new QoreBigIntNode(SIGPOLL));
   ns.addConstant("SIGPROF", new QoreBigIntNode(SIGPROF));
   ns.addConstant("SIGPWR", new QoreBigIntNode(SIGPWR));
   ns.addConstant("SIGQUIT", new QoreBigIntNode(SIGQUIT));
   ns.addConstant("SIGSEGV", new QoreBigIntNode(SIGSEGV));
   ns.addConstant("SIGSTKFLT", new QoreBigIntNode(SIGSTKFLT));
   ns.addConstant("SIGSTOP", new QoreBigIntNode(SIGSTOP));
   ns.addConstant("SIGSYS", new QoreBigIntNode(SIGSYS));
   ns.addConstant("SIGTERM", new QoreBigIntNode(SIGTERM));
   ns.addConstant("SIGTHAW", new QoreBigIntNode(SIGTHAW));
   ns.addConstant("SIGTRAP", new QoreBigIntNode(SIGTRAP));
   ns.addConstant("SIGTSTP", new QoreBigIntNode(SIGTSTP));
   ns.addConstant("SIGTTIN", new QoreBigIntNode(SIGTTIN));
   ns.addConstant("SIGTTOU", new QoreBigIntNode(SIGTTOU));
   ns.addConstant("SIGURG", new QoreBigIntNode(SIGURG));
   ns.addConstant("SIGUSR1", new QoreBigIntNode(SIGUSR1));
   ns.addConstant("SIGUSR2", new QoreBigIntNode(SIGUSR2));
   ns.addConstant("SIGVTALRM", new QoreBigIntNode(SIGVTALRM));
   ns.addConstant("SIGWAITING", new QoreBigIntNode(SIGWAITING));
   ns.addConstant("SIGWINCH", new QoreBigIntNode(SIGWINCH));
   ns.addConstant("SIGXCPU", new QoreBigIntNode(SIGXCPU));
   ns.addConstant("SIGXFSZ", new QoreBigIntNode(SIGXFSZ));
   ns.addConstant("SIGXRES", new QoreBigIntNode(SIGXRES));
   QoreHashNode* hash_SignalToName = new QoreHashNode;
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGABRT), new QoreStringNode("SIGABRT"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGALRM), new QoreStringNode("SIGALRM"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGBUS), new QoreStringNode("SIGBUS"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGCANCEL), new QoreStringNode("SIGCANCEL"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGCHLD), new QoreStringNode("SIGCHLD"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGCONT), new QoreStringNode("SIGCONT"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGEMT), new QoreStringNode("SIGEMT"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGFPE), new QoreStringNode("SIGFPE"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGFREEZE), new QoreStringNode("SIGFREEZE"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGHUP), new QoreStringNode("SIGHUP"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGILL), new QoreStringNode("SIGILL"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGINFO), new QoreStringNode("SIGINFO"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGINT), new QoreStringNode("SIGINT"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGIO), new QoreStringNode("SIGIO"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGIOT), new QoreStringNode("SIGIOT"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGJVM1), new QoreStringNode("SIGJVM1"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGJVM2), new QoreStringNode("SIGJVM2"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGKILL), new QoreStringNode("SIGKILL"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGLOST), new QoreStringNode("SIGLOST"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGLWP), new QoreStringNode("SIGLWP"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGPIPE), new QoreStringNode("SIGPIPE"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGPOLL), new QoreStringNode("SIGPOLL"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGPROF), new QoreStringNode("SIGPROF"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGPWR), new QoreStringNode("SIGPWR"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGQUIT), new QoreStringNode("SIGQUIT"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGSEGV), new QoreStringNode("SIGSEGV"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGSTKSZ), new QoreStringNode("SIGSTKSZ"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGSTOP), new QoreStringNode("SIGSTOP"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGSYS), new QoreStringNode("SIGSYS"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGTERM), new QoreStringNode("SIGTERM"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGTHAW), new QoreStringNode("SIGTHAW"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGTRAP), new QoreStringNode("SIGTRAP"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGTSTP), new QoreStringNode("SIGTSTP"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGTTIN), new QoreStringNode("SIGTTIN"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGTTOU), new QoreStringNode("SIGTTOU"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGURG), new QoreStringNode("SIGURG"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGUSR1), new QoreStringNode("SIGUSR1"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGUSR2), new QoreStringNode("SIGUSR2"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGVTALRM), new QoreStringNode("SIGVTALRM"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGWAITING), new QoreStringNode("SIGWAITING"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGWINCH), new QoreStringNode("SIGWINCH"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGXCPU), new QoreStringNode("SIGXCPU"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGXFSZ), new QoreStringNode("SIGXFSZ"), 0);
   hash_SignalToName->setKeyValue(MAKE_STRING_FROM_SYMBOL(SIGXRES), new QoreStringNode("SIGXRES"), 0);

   ns.addConstant("SignalToName", hash_SignalToName);
# 376 "ql_misc.qpp"
# 412 "ql_misc.qpp"
# 442 "ql_misc.qpp"
# 472 "ql_misc.qpp"
# 493 "ql_misc.qpp"
# 514 "ql_misc.qpp"
# 530 "ql_misc.qpp"
# 562 "ql_misc.qpp"
# 589 "ql_misc.qpp"
# 605 "ql_misc.qpp"
# 621 "ql_misc.qpp"
# 632 "ql_misc.qpp"
# 638 "ql_misc.qpp"
# 657 "ql_misc.qpp"
# 663 "ql_misc.qpp"
# 679 "ql_misc.qpp"
# 685 "ql_misc.qpp"
# 701 "ql_misc.qpp"
# 707 "ql_misc.qpp"
# 719 "ql_misc.qpp"
# 759 "ql_misc.qpp"
# 765 "ql_misc.qpp"
# 779 "ql_misc.qpp"
# 785 "ql_misc.qpp"
# 818 "ql_misc.qpp"
# 844 "ql_misc.qpp"
# 850 "ql_misc.qpp"
# 884 "ql_misc.qpp"
# 890 "ql_misc.qpp"
# 915 "ql_misc.qpp"
# 938 "ql_misc.qpp"
# 944 "ql_misc.qpp"
# 964 "ql_misc.qpp"
# 970 "ql_misc.qpp"
# 994 "ql_misc.qpp"
# 1000 "ql_misc.qpp"
# 1023 "ql_misc.qpp"
# 1046 "ql_misc.qpp"
# 1058 "ql_misc.qpp"
# 1082 "ql_misc.qpp"
# 1088 "ql_misc.qpp"
# 1108 "ql_misc.qpp"
# 1125 "ql_misc.qpp"
# 1150 "ql_misc.qpp"
# 1175 "ql_misc.qpp"
# 1202 "ql_misc.qpp"
# 1208 "ql_misc.qpp"
# 1235 "ql_misc.qpp"
# 1261 "ql_misc.qpp"
# 1288 "ql_misc.qpp"
# 1316 "ql_misc.qpp"
# 1345 "ql_misc.qpp"
# 1373 "ql_misc.qpp"
# 1402 "ql_misc.qpp"
# 1408 "ql_misc.qpp"
# 1434 "ql_misc.qpp"
# 1461 "ql_misc.qpp"
# 1487 "ql_misc.qpp"
# 1513 "ql_misc.qpp"
# 1541 "ql_misc.qpp"
# 1568 "ql_misc.qpp"
# 1594 "ql_misc.qpp"
# 1620 "ql_misc.qpp"
# 1631 "ql_misc.qpp"
# 1652 "ql_misc.qpp"
# 1675 "ql_misc.qpp"
# 1694 "ql_misc.qpp"
# 1717 "ql_misc.qpp"
# 1723 "ql_misc.qpp"
# 1744 "ql_misc.qpp"
# 1765 "ql_misc.qpp"
# 1771 "ql_misc.qpp"
# 1790 "ql_misc.qpp"
# 1796 "ql_misc.qpp"
# 1840 "ql_misc.qpp"
# 1846 "ql_misc.qpp"
# 1867 "ql_misc.qpp"
# 1873 "ql_misc.qpp"
# 1895 "ql_misc.qpp"
# 1901 "ql_misc.qpp"
# 1921 "ql_misc.qpp"
# 1927 "ql_misc.qpp"
# 1948 "ql_misc.qpp"
# 1960 "ql_misc.qpp"
# 1972 "ql_misc.qpp"
# 1984 "ql_misc.qpp"
# 2012 "ql_misc.qpp"
# 2043 "ql_misc.qpp"
# 2083 "ql_misc.qpp"
# 2097 "ql_misc.qpp"

}
