/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
  ql_string.qpp

  Qore Programming Language

  Copyright 2003 - 2013 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>
#include <qore/intern/ql_string.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

// finds the last occurrence of needle in haystack at or before position pos
// pos must be a non-negative valid byte offset in haystack
static inline int rindex_simple_intern(const char *haystack, int hlen, const char *needle, int nlen, int pos) {
   // if the offset does not allow for the needle string to be present, then adjust
   if ((pos + nlen) > hlen) {
      pos = hlen - nlen;
      if (pos < 0)
	 return -1;
   }

   while (pos != -1) {
      if (!strncmp(haystack + pos, needle, nlen))
	 return pos;
      pos--;
   }
   return -1;
}

static const char *memstr(const char *str, const char *pattern, qore_size_t pl, qore_size_t len) {
   while (true) {
      const char *p = (const char *)memchr(str, pattern[0], len);
      if (!p)
	 return 0;

      //printd(5, "memstr() pattern=%s str=%p p=%p len=%d pl=%d remaining=%d (%c %c %c)\n", pattern, str, p, len, pl, len-(p-str), p[1], p[2], p[3]);

      // if there is not enough string left for the pattern, then return
      if ((len - (p - str)) < pl)
	 return 0;

      bool found = true;
      for (qore_size_t i = 1; i < pl; ++i) {
	 if (pattern[i] != p[i]) {
	    len -= (p - str + 1);
	    str = p + 1;
	    found = false;
	    break;
	 }
      }
      if (!found)
	 continue;
     
      // found
      //printd(5, "memstr() got it! p=%p\n", p);
      return p;
   }
   return 0;
}

static void split_add_element(QoreListNode *l, const char *str, unsigned len, const QoreEncoding *enc) {
   if (enc)
      l->push(new QoreStringNode(str, len, enc));
   else {
      BinaryNode *b = new BinaryNode;
      b->append(str, len);
      l->push(b);
   }
}

QoreListNode *split_intern(const char *pattern, qore_size_t pl, const char *str, qore_size_t sl, const QoreEncoding *enc, bool with_separator) {
   QoreListNode *l = new QoreListNode();
   const char *ostr = str;
   while (const char *p = memstr(str, pattern, pl, sl - (str - ostr))) {
      split_add_element(l, str, p - str + (with_separator ? pl : 0), enc);
      str = p + pl;
   }
   // add last field if there is data remaining
   if (sl - (str - ostr))
      split_add_element(l, str, sl - (str - ostr), enc);

   return l;
}

QoreListNode* split_with_quote(const QoreString* sep, const QoreString* str, const QoreString* quote, bool trim_unquoted, ExceptionSink* xsink) {
   // convert pattern encoding to string if necessary
   TempEncodingHelper pat(sep, str->getEncoding(), xsink);
   if (*xsink)
      return 0;

   // convert quote to string if necessary
   TempEncodingHelper tquote(quote, str->getEncoding(), xsink);
   if (*xsink)
      return 0;

   //printd(5, "split_with_quote() sep: %s str: %s quote: %s trim_unquoted: %d\n", pat->getBuffer(), str->getBuffer(), tquote->getBuffer(), trim_unquoted);

   if (!tquote->strlen() || tquote->strlen() > sep->strlen())
      return split_intern(pat->getBuffer(), pat->strlen(), str->getBuffer(), str->strlen(), str->getEncoding());

   ReferenceHolder<QoreListNode> l(new QoreListNode, xsink);
   const char *ostr = str->getBuffer();
   qore_size_t sl = str->strlen();
   const char *tpattern = pat->getBuffer();
   qore_size_t pl = pat->strlen();

   const char *ststr = ostr;

   // remaining byte length
   qore_size_t len = sl;

   while (len > 0) {
      // see if the field begins with the quote string
      // and if the remaining string length is at least big enough for two quote strings
      if ((tquote->strlen() * 2) <= len
          && !memcmp(tquote->getBuffer(), ststr, tquote->strlen())) {
         // advance pointer past quote
         ststr += tquote->strlen();
         // find next quote character, ignore escaped quotes
         const char *tstr = ststr;
         const char *p;
         while (true) {
            p = memstr(tstr, tquote->getBuffer(), tquote->strlen(), len);
            if (!p) {
               xsink->raiseException("SPLIT-ERROR", "cannot find closing quote '%s' in field "QSD, tquote->getBuffer(), l->size() + 1);
               return 0;
            }
            if (p == tstr)
               break;
            if (*(p - 1) != '\\')
               break;
            tstr = p + 1;
         }
         // optimistically add the field to the list
         l->push(new QoreStringNode(ststr, p - ststr, sep->getEncoding()));

         ststr = p + tquote->strlen();
         // see if we are at the end of the string
         len = sl + (ostr - ststr);

         if (!len)
            break;

         // or a separator string comes next
         if (len < pl || memcmp(tpattern, ststr, pl)) {
            xsink->raiseException("SPLIT-ERROR", "separator pattern '%s' does not follow end quote in field "QSD, tpattern, l->size());
            return 0;
         }
         ststr += pl;
         len -= pl;
         continue;
      }

      const char *p = memstr(ststr, tpattern, pl, sl - (ststr - ostr));
      if (!p) {
         QoreStringNode* se = new QoreStringNode(ststr, sl - (ststr - ostr), sep->getEncoding());
         if (trim_unquoted)
            se->trim();
         l->push(se);
         break;
      }

      QoreStringNode* se = new QoreStringNode(ststr, p - ststr, sep->getEncoding());
      if (trim_unquoted)
         se->trim();
      l->push(se);
      len -= (p - ststr);
      ststr = p + pl;
   }

   return l.release();
}

QoreStringNode *join_intern(const QoreStringNode *p0, const QoreListNode *l, int offset, ExceptionSink* xsink) {
   SimpleRefHolder<QoreStringNode> str(new QoreStringNode(p0->getEncoding()));

   for (unsigned i = offset; i < l->size(); i++) {
      const AbstractQoreNode *p = l->retrieve_entry(i);
      if (p) {
	 QoreStringValueHelper t(p, p0->getEncoding(), xsink);	 
	 if (*xsink)
	    return 0;

	 str->concat(*t);
      }
      if (i < (l->size() - 1))
	 str->concat(p0);
   }

   return str.release();
}

QoreStringNode* format_float_intern(const QoreString& fmt, double num) {
   int decimals = 0, neg = 1, len;
   int64 tr, bi, mi, th, val;
   char thousands_sep, decimal_sep = '.', chr[40], str[40], dec[20];

   len = fmt.strlen();
   if ((len != 1) && (len != 3))
      return new QoreStringNode;

   thousands_sep = fmt.getBuffer()[0];
   if (len == 3) {
      decimal_sep = fmt.getBuffer()[1];
      decimals = atoi(fmt.getBuffer() + 2);
   }

   if (num < 0) {
      neg = -1;
      num *= -1;
   }
   val = (int64)num;
   if (len == 3) {
      num -= (double)val;
      sprintf(chr, "%%.%df", decimals);
      sprintf(dec, chr, num);
   }
   tr = val / 1000000000000ll;
   val -= tr * 1000000000000ll;
   bi = val / 1000000000ll;
   val -= bi * 1000000000ll;
   mi = val / 1000000ll;
   val -= mi * 1000000ll;
   th = val / 1000ll;
   val -= th * 1000ll;
   //printd(0, "tr=%lld bi=%lld mi=%lld th=%lld val=%lld\n", tr, bi, mi, th, val);

   if (tr) {
      if (len == 3)
         //sprintf(str, "%lld%c%03lld%c%03lld%c%03lld%c%03lld%c%s",
         sprintf(str, QLLD "%c" QLLDx(03) "%c" QLLDx(03) "%c" QLLDx(03) "%c" QLLDx(03) "%c%s",
               neg * tr, thousands_sep,
               bi, thousands_sep,
               mi, thousands_sep,
               th, thousands_sep,
               val, decimal_sep,
               dec + 2);
      else
         //sprintf(str, "%lld%c%03lld%c%03lld%c%03lld%c%03lld",
         sprintf(str, QLLD "%c" QLLDx(03) "%c" QLLDx(03) "%c" QLLDx(03) "%c" QLLDx(03),
               neg * tr, thousands_sep,
               bi, thousands_sep,
               mi, thousands_sep,
               th, thousands_sep,
               val);
   }
   else if (bi)
      if (len == 3)
         //sprintf(str, "%lld%c%03lld%c%03lld%c%03lld%c%s",
         sprintf(str, QLLD "%c" QLLDx(03) "%c" QLLDx(03) "%c" QLLDx(03) "%c%s",
               neg * bi, thousands_sep,
               mi, thousands_sep,
               th, thousands_sep,
               val, decimal_sep,
               dec + 2);
      else
         //sprintf(str, "%lld%c%03lld%c%03lld%c%03lld",
         sprintf(str, QLLD "%c" QLLDx(03) "%c" QLLDx(03) "%c" QLLDx(03),
               neg * bi, thousands_sep, mi, thousands_sep,
               th, thousands_sep, val);
   else if (mi)
      if (len == 3)
         //sprintf(str, "%lld%c%03lld%c%03lld%c%s", neg * mi, thousands_sep,
         sprintf(str, QLLD "%c" QLLDx(03) "%c" QLLDx(03) "%c%s", neg * mi, thousands_sep,
               th, thousands_sep, val, decimal_sep, dec + 2);
      else
         //sprintf(str, "%lld%c%03lld%c%03lld", neg * mi, thousands_sep,
         sprintf(str, QLLD "%c" QLLDx(03) "%c" QLLDx(03), neg * mi, thousands_sep,
               th, thousands_sep, val);
   else if (th)
      if (len == 3)
         //sprintf(str, "%lld%c%03lld%c%s", neg * th, thousands_sep,
         sprintf(str, QLLD "%c" QLLDx(03) "%c%s", neg * th, thousands_sep,
               val, decimal_sep,
               dec + 2);
      else
         //sprintf(str, "%lld%c%03lld", neg * th, thousands_sep, val);
         sprintf(str, QLLD "%c" QLLDx(03), neg * th, thousands_sep, val);
   else
      if (len == 3)
         //sprintf(str, "%lld%c%s", neg * val, decimal_sep, &dec[2]);
         sprintf(str, QLLD "%c%s", neg * val, decimal_sep, &dec[2]);
      else
         //sprintf(str, "%lld", neg * val);
         sprintf(str, QLLD, neg * val);

   return new QoreStringNode(str);
}
# 335 "ql_string.qpp"
# 393 "ql_string.qpp"
# 418 "ql_string.qpp"
# 434 "ql_string.qpp"
# 440 "ql_string.qpp"
# 447 "ql_string.qpp"
# 469 "ql_string.qpp"
# 475 "ql_string.qpp"
# 482 "ql_string.qpp"
# 512 "ql_string.qpp"
# 518 "ql_string.qpp"
# 548 "ql_string.qpp"
# 554 "ql_string.qpp"
# 579 "ql_string.qpp"
# 605 "ql_string.qpp"
# 629 "ql_string.qpp"
# 654 "ql_string.qpp"
# 660 "ql_string.qpp"
# 691 "ql_string.qpp"
# 698 "ql_string.qpp"
# 727 "ql_string.qpp"
# 734 "ql_string.qpp"
# 765 "ql_string.qpp"
# 772 "ql_string.qpp"
# 801 "ql_string.qpp"
# 808 "ql_string.qpp"
# 829 "ql_string.qpp"
# 836 "ql_string.qpp"
# 856 "ql_string.qpp"
# 863 "ql_string.qpp"
# 869 "ql_string.qpp"
# 895 "ql_string.qpp"
# 921 "ql_string.qpp"
# 939 "ql_string.qpp"
# 946 "ql_string.qpp"
# 963 "ql_string.qpp"
# 969 "ql_string.qpp"
# 987 "ql_string.qpp"
# 993 "ql_string.qpp"
# 1010 "ql_string.qpp"
# 1016 "ql_string.qpp"
# 1046 "ql_string.qpp"
# 1052 "ql_string.qpp"
# 1107 "ql_string.qpp"
# 1113 "ql_string.qpp"
# 1145 "ql_string.qpp"
# 1151 "ql_string.qpp"
# 1231 "ql_string.qpp"
# 1237 "ql_string.qpp"
# 1255 "ql_string.qpp"
# 1275 "ql_string.qpp"
# 1281 "ql_string.qpp"
# 1301 "ql_string.qpp"
# 1326 "ql_string.qpp"
# 1332 "ql_string.qpp"
# 1354 "ql_string.qpp"
# 1385 "ql_string.qpp"
# 1391 "ql_string.qpp"
# 1437 "ql_string.qpp"
# 1476 "ql_string.qpp"
# 1497 "ql_string.qpp"
# 1503 "ql_string.qpp"
# 1521 "ql_string.qpp"
# 1541 "ql_string.qpp"
# 1563 "ql_string.qpp"
# 1570 "ql_string.qpp"
# 1592 "ql_string.qpp"
# 1599 "ql_string.qpp"
# 1623 "ql_string.qpp"
# 1645 "ql_string.qpp"
# 1652 "ql_string.qpp"
# 1676 "ql_string.qpp"
# 1683 "ql_string.qpp"
# 1707 "ql_string.qpp"
# 1714 "ql_string.qpp"
# 1740 "ql_string.qpp"
# 1764 "ql_string.qpp"
# 1771 "ql_string.qpp"
# 1782 "ql_string.qpp"
# 1796 "ql_string.qpp"
// int bindex(softstring str, softstring substr, softint pos = 0) {}
static int64 f_bindex_vsvsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   const QoreStringNode* substr = HARD_QORE_STRING(args, 1);
   int64 pos = HARD_QORE_INT(args, 2);
# 723 "ql_string.qpp"
   return str->bindex(*substr, (qore_offset_t)pos);
}

// int bindex() {}
static int64 f_bindex(const QoreListNode* args, ExceptionSink* xsink) {
# 730 "ql_string.qpp"
   return -1;
}

// int brindex(softstring str, softstring substr, softint pos = -1) {}
static int64 f_brindex_vsvsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   const QoreStringNode* substr = HARD_QORE_STRING(args, 1);
   int64 pos = HARD_QORE_INT(args, 2);
# 797 "ql_string.qpp"
   return str->brindex(*substr, (qore_offset_t)pos);
}

// int brindex() {}
static int64 f_brindex(const QoreListNode* args, ExceptionSink* xsink) {
# 804 "ql_string.qpp"
   return -1;
}

// string chomp(string str) {}
static AbstractQoreNode* f_chomp_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
# 1295 "ql_string.qpp"
   QoreStringNode *rv = str->copy();
   rv->chomp();
   return rv;
}

// *string chomp(reference str) {}
static AbstractQoreNode* f_chomp_Vr(const QoreListNode* args, ExceptionSink* xsink) {
   const ReferenceNode* str = HARD_QORE_REF(args, 0);
# 1313 "ql_string.qpp"
   QoreTypeSafeReferenceHelper ref(str, xsink);
   if (!ref || ref.getType() != NT_STRING)
      return 0;

   QoreStringNode *rv = reinterpret_cast<QoreStringNode *>(ref.getUnique(xsink));
   if (*xsink)
      return 0;

   rv->chomp();
   return rv->refSelf();
}

// nothing chomp() {}
static AbstractQoreNode* f_chomp(const QoreListNode* args, ExceptionSink* xsink) {
# 1329 "ql_string.qpp"

   return 0;
}

// string chr(softint val, *string encoding) {}
static AbstractQoreNode* f_chr_viNs(const QoreListNode* args, ExceptionSink* xsink) {
   int64 val = HARD_QORE_INT(args, 0);
   const QoreStringNode* encoding = reinterpret_cast<const QoreStringNode*>(get_param(args, 1));
# 849 "ql_string.qpp"
   const QoreEncoding *qe = encoding ? QEM.findCreate(encoding) : QCS_DEFAULT;
   QoreStringNode* rv = new QoreStringNode(qe);
   rv->concat((char)val);
   return rv;
}

// string chr(any arg) {}
static AbstractQoreNode* f_chr_Va(const QoreListNode* args, ExceptionSink* xsink) {
# 859 "ql_string.qpp"
   return new QoreStringNode('\0');
}

// nothing chr() {}
static AbstractQoreNode* f_chr(const QoreListNode* args, ExceptionSink* xsink) {
# 866 "ql_string.qpp"

   return 0;
}

// string convert_encoding(string str, string encoding) {}
static AbstractQoreNode* f_convert_encoding_VsVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   const QoreStringNode* encoding = HARD_QORE_STRING(args, 1);
# 983 "ql_string.qpp"
   return str->convertEncoding(QEM.findCreate(encoding), xsink);
}

// nothing convert_encoding() {}
static AbstractQoreNode* f_convert_encoding(const QoreListNode* args, ExceptionSink* xsink) {
# 990 "ql_string.qpp"

   return 0;
}

// string f_printf(string fmt, ...) {}
static AbstractQoreNode* f_f_printf_VsVV(const QoreListNode* args, ExceptionSink* xsink) {
# 1670 "ql_string.qpp"
   AbstractQoreNode *node;
   print_node(stdout, node = q_sprintf(args, 1, 0, xsink));
   return node;
}

// string f_printf() {}
static AbstractQoreNode* f_f_printf(const QoreListNode* args, ExceptionSink* xsink) {
# 1679 "ql_string.qpp"
   return null_string();
}

// string f_sprintf(string fmt, ...) {}
static AbstractQoreNode* f_f_sprintf_VsVV(const QoreListNode* args, ExceptionSink* xsink) {
# 1559 "ql_string.qpp"
   return q_sprintf(args, 1, 0, xsink);
}

// string f_sprintf() {}
static AbstractQoreNode* f_f_sprintf(const QoreListNode* args, ExceptionSink* xsink) {
# 1566 "ql_string.qpp"
   return null_string();
}

// string f_vprintf(string fmt, any varg) {}
static AbstractQoreNode* f_f_vprintf_VsVa(const QoreListNode* args, ExceptionSink* xsink) {
# 1734 "ql_string.qpp"
   AbstractQoreNode *node;
   print_node(stdout, node = q_vsprintf(args, 1, 0, xsink));
   return node;
}

// string f_vsprintf(string fmt, any varg) {}
static AbstractQoreNode* f_f_vsprintf_VsVa(const QoreListNode* args, ExceptionSink* xsink) {
# 1619 "ql_string.qpp"
   return q_vsprintf(args, 1, 0, xsink);
}

// nothing flush() {}
static AbstractQoreNode* f_flush(const QoreListNode* args, ExceptionSink* xsink) {
# 1778 "ql_string.qpp"
   fflush(stdout);
   return 0;
}

// string force_encoding(string str, string encoding) {}
static AbstractQoreNode* f_force_encoding_VsVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   const QoreStringNode* encoding = HARD_QORE_STRING(args, 1);
# 1006 "ql_string.qpp"
   return new QoreStringNode(str->getBuffer(), str->strlen(), QEM.findCreate(encoding));
}

// nothing force_encoding() {}
static AbstractQoreNode* f_force_encoding(const QoreListNode* args, ExceptionSink* xsink) {
# 1013 "ql_string.qpp"

   return 0;
}

// string format_number(string fmt, softfloat num) {}
static AbstractQoreNode* f_format_number_Vsvf(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* fmt = HARD_QORE_STRING(args, 0);
   double num = HARD_QORE_FLOAT(args, 1);
# 1493 "ql_string.qpp"
   return format_float_intern(*fmt, num);
}

// nothing format_number() {}
static AbstractQoreNode* f_format_number(const QoreListNode* args, ExceptionSink* xsink) {
# 1500 "ql_string.qpp"

   return 0;
}

// string get_encoding(string str) {}
static AbstractQoreNode* f_get_encoding_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
# 959 "ql_string.qpp"
   return new QoreStringNode(str->getEncoding()->getCode());
}

// nothing get_encoding() {}
static AbstractQoreNode* f_get_encoding(const QoreListNode* args, ExceptionSink* xsink) {
# 966 "ql_string.qpp"

   return 0;
}

// int index(softstring str, softstring substr, softint pos = 0) {}
static int64 f_index_vsvsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   const QoreStringNode* substr = HARD_QORE_STRING(args, 1);
   int64 pos = HARD_QORE_INT(args, 2);
# 687 "ql_string.qpp"
   return str->index(*substr, pos, xsink);
}

// int index() {}
static int64 f_index(const QoreListNode* args, ExceptionSink* xsink) {
# 694 "ql_string.qpp"
   return -1;
}

// string join(string str, ...) {}
static AbstractQoreNode* f_join_VsVV(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
# 1251 "ql_string.qpp"
   return join_intern(str, args, 1, xsink);
}

// string join(string str, list l) {}
static AbstractQoreNode* f_join_VsVl(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   const QoreListNode* l = HARD_QORE_LIST(args, 1);
# 1271 "ql_string.qpp"
   return join_intern(str, l, 0, xsink);
}

// nothing join() {}
static AbstractQoreNode* f_join(const QoreListNode* args, ExceptionSink* xsink) {
# 1278 "ql_string.qpp"

   return 0;
}

// int length(softstring str) {}
static int64 f_length_vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
# 414 "ql_string.qpp"
   return str->length();
}

// int length(binary bin) {}
static int64 f_length_Vy(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
# 430 "ql_string.qpp"
   return bin->size();
}

// nothing length() {}
static AbstractQoreNode* f_length(const QoreListNode* args, ExceptionSink* xsink) {
# 437 "ql_string.qpp"

   return 0;
}

// int length(any arg) {}
static int64 f_length_Va(const QoreListNode* args, ExceptionSink* xsink) {
# 443 "ql_string.qpp"
   return 0;
}

// int ord(softstring str, softint offset = 0) {}
static int64 f_ord_vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 offset = HARD_QORE_INT(args, 1);
# 822 "ql_string.qpp"
   if (offset < 0 || (qore_size_t)offset >= str->strlen())
      return -1;

   return str->getBuffer()[offset];
}

// int ord() {}
static int64 f_ord(const QoreListNode* args, ExceptionSink* xsink) {
# 832 "ql_string.qpp"
   return 0;
}

// bool parse_boolean(string str) {}
static bool f_parse_boolean_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
# 1792 "ql_string.qpp"
   return q_parse_bool(str->getBuffer());
}

// bool parse_boolean(...) {}
static bool f_parse_boolean_VV(const QoreListNode* args, ExceptionSink* xsink) {
# 1804 "ql_string.qpp"
   const AbstractQoreNode* p = get_param(args, 0);
   return p ? p->getAsBool() : false;
}

// nothing print(...) {}
static AbstractQoreNode* f_print_VV(const QoreListNode* args, ExceptionSink* xsink) {
# 1532 "ql_string.qpp"
   if (!args)
      return 0;

   for (unsigned i = 0; i < args->size(); i++)
      print_node(stdout, get_param(args, i));
   return 0;
}

// string printf(string fmt, ...) {}
static AbstractQoreNode* f_printf_VsVV(const QoreListNode* args, ExceptionSink* xsink) {
# 1701 "ql_string.qpp"
   AbstractQoreNode *node;
   print_node(stdout, node = q_sprintf(args, 0, 0, xsink));
   return node;
}

// string printf() {}
static AbstractQoreNode* f_printf(const QoreListNode* args, ExceptionSink* xsink) {
# 1710 "ql_string.qpp"
   return null_string();
}

// bool regex(string str, string regex, int options = 0) {}
static bool f_regex_VsVsVi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   const QoreStringNode* regex = HARD_QORE_STRING(args, 1);
   int64 options = HARD_QORE_INT(args, 2);
# 1038 "ql_string.qpp"
   QoreRegexNode qr(*regex, options, xsink);
   if (*xsink)
      return 0;

   return qr.exec(str, xsink);
}

// nothing regex() {}
static AbstractQoreNode* f_regex(const QoreListNode* args, ExceptionSink* xsink) {
# 1049 "ql_string.qpp"

   return 0;
}

// *list regex_extract(string str, string regex, int options = 0) {}
static AbstractQoreNode* f_regex_extract_VsVsVi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   const QoreStringNode* regex = HARD_QORE_STRING(args, 1);
   int64 options = HARD_QORE_INT(args, 2);
# 1137 "ql_string.qpp"
   QoreRegexNode qr(*regex, options, xsink);
   if (*xsink)
      return 0;

   return qr.extractSubstrings(str, xsink);
}

// nothing regex_extract() {}
static AbstractQoreNode* f_regex_extract(const QoreListNode* args, ExceptionSink* xsink) {
# 1148 "ql_string.qpp"

   return 0;
}

// string regex_subst(string str, string regex, string subst, int options = 0) {}
static AbstractQoreNode* f_regex_subst_VsVsVsVi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   const QoreStringNode* regex = HARD_QORE_STRING(args, 1);
   const QoreStringNode* subst = HARD_QORE_STRING(args, 2);
   int64 options = HARD_QORE_INT(args, 3);
# 1089 "ql_string.qpp"
   bool global;
   if (options & QRE_GLOBAL) {
      global = true;
      options &= 0xffffffff;
   }
   else
      global = false;

   RegexSubstNode qrs(regex, (int)options, xsink);
   if (*xsink)
      return 0;

   if (global)
      qrs.setGlobal();
   return qrs.exec(str, subst, xsink);
}

// nothing regex_subst() {}
static AbstractQoreNode* f_regex_subst(const QoreListNode* args, ExceptionSink* xsink) {
# 1110 "ql_string.qpp"

   return 0;
}

// string replace(string str, string source, string target, int start = 0, int end = -1) {}
static AbstractQoreNode* f_replace_VsVsVsViVi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   const QoreStringNode* source = HARD_QORE_STRING(args, 1);
   const QoreStringNode* target = HARD_QORE_STRING(args, 2);
   int64 start = HARD_QORE_INT(args, 3);
   int64 end = HARD_QORE_INT(args, 4);
# 1172 "ql_string.qpp"
   const QoreEncoding *ccs = str->getEncoding();

   TempEncodingHelper t1(source, ccs, xsink);
   if (*xsink)
      return 0;

   if (t1->empty() || !end || (end > 0 && end < start))
      return str->refSelf();

   TempEncodingHelper t2(target, ccs, xsink);
   if (*xsink)
      return 0;

   QoreStringNode *nstr = new QoreStringNode(ccs);   

   const char *cstr, *pattern;
   cstr = str->getBuffer();
   pattern = t1->getBuffer();
   int plen = strlen(pattern);

   if (start) {
      const char* cn = cstr + str->size();
      qore_size_t i = ccs->getByteLen(cstr, cn, start, xsink);
      cstr = cn;
      if (*xsink || (i == str->size()))
         return 0;

      cstr += i;
   }

   if (end > 0) {
      const char* cn = cstr + str->size();
      qore_size_t i = ccs->getByteLen(cstr, cn, end, xsink);
      cstr = cn;
      if (*xsink || i == str->size())
         return 0;

      cstr += i;
   }

   while (const char* p = strstr(cstr, pattern)) {
      //printd(5, "str=%08p p=%08p '%s' '%s'->'%s'\n", cstr, p, cstr, pattern, t1->getBuffer());
      if ((end > 0) && (p - str->getBuffer() + plen) > end)
         break;

      if (p != cstr)
	 nstr->concat(cstr, p - cstr);
      nstr->concat(*t2);

      cstr = p + plen;
   }
   // add last field
   if (*cstr)
      nstr->concat(cstr);
   
   return nstr;
}

// nothing replace() {}
static AbstractQoreNode* f_replace(const QoreListNode* args, ExceptionSink* xsink) {
# 1234 "ql_string.qpp"

   return 0;
}

// string reverse(softstring str) {}
static AbstractQoreNode* f_reverse_vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
# 1517 "ql_string.qpp"
   return str->reverse();
}

// int rindex(softstring str, softstring substr, softint pos = -1) {}
static int64 f_rindex_vsvsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   const QoreStringNode* substr = HARD_QORE_STRING(args, 1);
   int64 pos = HARD_QORE_INT(args, 2);
# 761 "ql_string.qpp"
   return str->rindex(*substr, (qore_offset_t)pos, xsink);
}

// int rindex() {}
static int64 f_rindex(const QoreListNode* args, ExceptionSink* xsink) {
# 768 "ql_string.qpp"
   return -1;
}

// list split(string sep, string str, bool with_separator = False) {}
static AbstractQoreNode* f_split_VsVsVb(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sep = HARD_QORE_STRING(args, 0);
   const QoreStringNode* str = HARD_QORE_STRING(args, 1);
   bool with_separator = HARD_QORE_BOOL(args, 2);
# 886 "ql_string.qpp"
   // convert pattern encoding to string if necessary
   TempEncodingHelper temp(sep, str->getEncoding(), xsink);
   if (*xsink)
      return 0;
   
   return split_intern(temp->getBuffer(), temp->strlen(), str->getBuffer(), str->strlen(), str->getEncoding(), with_separator);
}

// list split(string sep, string str, string quote, bool trim_unquoted = False) {}
static AbstractQoreNode* f_split_VsVsVsVb(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* sep = HARD_QORE_STRING(args, 0);
   const QoreStringNode* str = HARD_QORE_STRING(args, 1);
   const QoreStringNode* quote = HARD_QORE_STRING(args, 2);
   bool trim_unquoted = HARD_QORE_BOOL(args, 3);
# 917 "ql_string.qpp"
   return split_with_quote(sep, str, quote, trim_unquoted, xsink);
}

// list split(binary sep, binary data) {}
static AbstractQoreNode* f_split_VyVy(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* sep = HARD_QORE_BINARY(args, 0);
   const BinaryNode* data = HARD_QORE_BINARY(args, 1);
# 935 "ql_string.qpp"
   return split_intern((const char*)sep->getPtr(), sep->size(), (const char*)data->getPtr(), data->size(), 0);
}

// list split() {}
static AbstractQoreNode* f_split(const QoreListNode* args, ExceptionSink* xsink) {
# 942 "ql_string.qpp"
   return new QoreListNode;
}

// string sprintf(string fmt, ...) {}
static AbstractQoreNode* f_sprintf_VsVV(const QoreListNode* args, ExceptionSink* xsink) {
# 1588 "ql_string.qpp"
   return q_sprintf(args, 0, 0, xsink);
}

// string sprintf() {}
static AbstractQoreNode* f_sprintf(const QoreListNode* args, ExceptionSink* xsink) {
# 1595 "ql_string.qpp"
   return null_string();
}

// int strlen(softstring str) {}
static int64 f_strlen_vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
# 465 "ql_string.qpp"
   return str->strlen();
}

// nothing strlen() {}
static AbstractQoreNode* f_strlen(const QoreListNode* args, ExceptionSink* xsink) {
# 472 "ql_string.qpp"

   return 0;
}

// int strlen(any arg) {}
static int64 f_strlen_Va(const QoreListNode* args, ExceptionSink* xsink) {
# 478 "ql_string.qpp"
   return 0;
}

// string strmul(softstring str, softint smul, *softint offset) {}
static AbstractQoreNode* f_strmul_vsvini(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 smul = HARD_QORE_INT(args, 1);
   const QoreBigIntNode* tmp_offset = reinterpret_cast<const QoreBigIntNode*>(get_param(args, 2));
   int64 offset = tmp_offset ? tmp_offset->val : 0;
# 1450 "ql_string.qpp"
   if (offset < 0) {
      xsink->raiseException("STRMUL-ERROR", "Offset argument has to be >= 0; value: "QSD, offset);
      return 0;
   }
   
   if (smul < 1) {
      xsink->raiseException("STRMUL-ERROR", "Multiple argument has to be > 0; value: "QSD, smul);
      return 0;
   }

   QoreStringNode* ret = new QoreStringNode(str);
   while (smul != 1) {
      ret->concat(str);
      --smul;
   }
   
   if (offset) {
      offset = ret->length() - offset;
      SimpleRefHolder<QoreStringNode> t(ret);
      return ret->substr(0, offset, xsink);
   }
       
   return ret;
}

// string substr(softstring str, softint start) {}
static AbstractQoreNode* f_substr_vsvi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 start = HARD_QORE_INT(args, 1);
# 572 "ql_string.qpp"
   QoreStringNode* rv = str->substr(start, xsink);
   if (!rv)
      rv = new QoreStringNode(str->getEncoding());
   return rv;
}

// string substr(softstring str, softint start, softint len) {}
static AbstractQoreNode* f_substr_vsvivi(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 start = HARD_QORE_INT(args, 1);
   int64 len = HARD_QORE_INT(args, 2);
# 598 "ql_string.qpp"
   QoreStringNode* rv = str->substr(start, len, xsink);
   if (!rv)
      rv = new QoreStringNode(str->getEncoding());
   return rv;
}

// binary substr(binary b, softint start) {}
static AbstractQoreNode* f_substr_Vyvi(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* b = HARD_QORE_BINARY(args, 0);
   int64 start = HARD_QORE_INT(args, 1);
# 623 "ql_string.qpp"
   BinaryNode* b1 = new BinaryNode;
   b->substr(*b1, start);
   return b1;
}

// binary substr(binary b, softint start, softint len) {}
static AbstractQoreNode* f_substr_Vyvivi(const QoreListNode* args, ExceptionSink* xsink) {
   const BinaryNode* b = HARD_QORE_BINARY(args, 0);
   int64 start = HARD_QORE_INT(args, 1);
   int64 len = HARD_QORE_INT(args, 2);
# 648 "ql_string.qpp"
   BinaryNode* b1 = new BinaryNode;
   b->substr(*b1, start, len);
   return b1;
}

// nothing substr() {}
static AbstractQoreNode* f_substr(const QoreListNode* args, ExceptionSink* xsink) {
# 657 "ql_string.qpp"

   return 0;
}

// string tolower(string str) {}
static AbstractQoreNode* f_tolower_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
# 504 "ql_string.qpp"
   SimpleRefHolder<QoreStringNode> rv(new QoreStringNode(str->getEncoding()));
   if (do_tolower(*(*rv), *str, xsink))
      return 0;

   return rv.release();
}

// nothing tolower() {}
static AbstractQoreNode* f_tolower(const QoreListNode* args, ExceptionSink* xsink) {
# 515 "ql_string.qpp"

   return 0;
}

// string toupper(string str) {}
static AbstractQoreNode* f_toupper_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
# 540 "ql_string.qpp"
   SimpleRefHolder<QoreStringNode> rv(new QoreStringNode(str->getEncoding()));
   if (do_toupper(*(*rv), *str, xsink))
      return 0;

   return rv.release();
}

// nothing toupper() {}
static AbstractQoreNode* f_toupper(const QoreListNode* args, ExceptionSink* xsink) {
# 551 "ql_string.qpp"

   return 0;
}

// string trim(string str, *string chars) {}
static AbstractQoreNode* f_trim_VsNs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   const QoreStringNode* chars = reinterpret_cast<const QoreStringNode*>(get_param(args, 1));
# 1347 "ql_string.qpp"
   const char *tchars = chars && chars->strlen() ? chars->getBuffer() : 0;
   QoreStringNode *rv = str->copy();
   rv->trim(tchars);
   return rv;
}

// *string trim(reference str, *string chars) {}
static AbstractQoreNode* f_trim_VrNs(const QoreListNode* args, ExceptionSink* xsink) {
   const ReferenceNode* str = HARD_QORE_REF(args, 0);
   const QoreStringNode* chars = reinterpret_cast<const QoreStringNode*>(get_param(args, 1));
# 1371 "ql_string.qpp"
   QoreTypeSafeReferenceHelper ref(str, xsink);
   if (!ref || ref.getType() != NT_STRING)
      return 0;

   QoreStringNode *rv = reinterpret_cast<QoreStringNode *>(ref.getUnique(xsink));
   if (*xsink)
      return 0;

   const char *tchars = chars && chars->strlen() ? chars->getBuffer() : 0;
   rv->trim(tchars);
   return rv->refSelf();
}

// nothing trim() {}
static AbstractQoreNode* f_trim(const QoreListNode* args, ExceptionSink* xsink) {
# 1388 "ql_string.qpp"

   return 0;
}

// string trunc_str(softstring str, softint len, *string encoding) {}
static AbstractQoreNode* f_trunc_str_vsviNs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* str = HARD_QORE_STRING(args, 0);
   int64 len = HARD_QORE_INT(args, 1);
   const QoreStringNode* encoding = reinterpret_cast<const QoreStringNode*>(get_param(args, 2));
# 1406 "ql_string.qpp"
   const QoreEncoding *enc = encoding ? QEM.findCreate(encoding) : str->getEncoding();
   if (len <= 0)
      return new QoreStringNode(enc);

   TempEncodingHelper tmp(str, enc, xsink);
   if (!tmp)
      return 0;

   if (tmp->strlen() <= (qore_size_t)len) {
      len = tmp->strlen();
      return new QoreStringNode(tmp.giveBuffer(), len, len + 1, enc);
   }

   if (!enc->isMultiByte())
      return new QoreStringNode(tmp->getBuffer(), len, enc);

   // find position of last character fitting in len bytes
   const char *p = tmp->getBuffer();
   int64 sl = 0;
   while (true) {
      qore_size_t size = enc->getCharLen(p, len - sl);
      if ((sl + size) > (qore_size_t)len)
	 break;
      sl += size;
      p += size;
   }

   return new QoreStringNode(tmp->getBuffer(), sl, enc);
}

// string vprintf(string fmt, ...) {}
static AbstractQoreNode* f_vprintf_VsVV(const QoreListNode* args, ExceptionSink* xsink) {
# 1758 "ql_string.qpp"
   AbstractQoreNode *node;
   print_node(stdout, node = q_vsprintf(args, 0, 0, xsink));
   return node;
}

// string vprintf() {}
static AbstractQoreNode* f_vprintf(const QoreListNode* args, ExceptionSink* xsink) {
# 1767 "ql_string.qpp"
   return null_string();
}

// string vsprintf(string fmt, any varg) {}
static AbstractQoreNode* f_vsprintf_VsVa(const QoreListNode* args, ExceptionSink* xsink) {
# 1641 "ql_string.qpp"
   return q_vsprintf(args, 0, 0, xsink);
}

// string vsprintf() {}
static AbstractQoreNode* f_vsprintf(const QoreListNode* args, ExceptionSink* xsink) {
# 1648 "ql_string.qpp"
   return null_string();
}


DLLLOCAL void init_string_functions(QoreNamespace& ns) {
   // int bindex(softstring str, softstring substr, softint pos = 0) {}
   ns.addBuiltinVariant("bindex", (q_func_int64_t)f_bindex_vsvsvi, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 3, softStringTypeInfo, NULL, "str", softStringTypeInfo, NULL, "substr", softBigIntTypeInfo, zero(), "pos");
   // int bindex() {}
   ns.addBuiltinVariant("bindex", (q_func_int64_t)f_bindex, QC_RUNTIME_NOOP, QDOM_DEFAULT, bigIntTypeInfo);
   // int brindex(softstring str, softstring substr, softint pos = -1) {}
   ns.addBuiltinVariant("brindex", (q_func_int64_t)f_brindex_vsvsvi, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 3, softStringTypeInfo, NULL, "str", softStringTypeInfo, NULL, "substr", softBigIntTypeInfo, new QoreBigIntNode(-1), "pos");
   // int brindex() {}
   ns.addBuiltinVariant("brindex", (q_func_int64_t)f_brindex, QC_RUNTIME_NOOP, QDOM_DEFAULT, bigIntTypeInfo);
   // string chomp(string str) {}
   ns.addBuiltinVariant("chomp", (q_func_t)f_chomp_Vs, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, NULL, "str");
   // *string chomp(reference str) {}
   ns.addBuiltinVariant("chomp", (q_func_t)f_chomp_Vr, QC_NO_FLAGS, QDOM_DEFAULT, stringOrNothingTypeInfo, 1, referenceTypeInfo, NULL, "str");
   // nothing chomp() {}
   ns.addBuiltinVariant("chomp", (q_func_t)f_chomp, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string chr(softint val, *string encoding) {}
   ns.addBuiltinVariant("chr", (q_func_t)f_chr_viNs, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 2, softBigIntTypeInfo, NULL, "val", stringOrNothingTypeInfo, NULL, "encoding");
   // string chr(any arg) {}
   ns.addBuiltinVariant("chr", (q_func_t)f_chr_Va, QC_NOOP, QDOM_DEFAULT, stringTypeInfo, 1, anyTypeInfo, NULL, "arg");
   // nothing chr() {}
   ns.addBuiltinVariant("chr", (q_func_t)f_chr, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string convert_encoding(string str, string encoding) {}
   ns.addBuiltinVariant("convert_encoding", (q_func_t)f_convert_encoding_VsVs, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 2, stringTypeInfo, NULL, "str", stringTypeInfo, NULL, "encoding");
   // nothing convert_encoding() {}
   ns.addBuiltinVariant("convert_encoding", (q_func_t)f_convert_encoding, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string f_printf(string fmt, ...) {}
   ns.addBuiltinVariant("f_printf", (q_func_t)f_f_printf_VsVV, QC_USES_EXTRA_ARGS, QDOM_TERMINAL_IO, stringTypeInfo, 1, stringTypeInfo, NULL, "fmt");
   // string f_printf() {}
   ns.addBuiltinVariant("f_printf", (q_func_t)f_f_printf, QC_NOOP, QDOM_DEFAULT, stringTypeInfo);
   // string f_sprintf(string fmt, ...) {}
   ns.addBuiltinVariant("f_sprintf", (q_func_t)f_f_sprintf_VsVV, QC_CONSTANT|QC_USES_EXTRA_ARGS, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, NULL, "fmt");
   // string f_sprintf() {}
   ns.addBuiltinVariant("f_sprintf", (q_func_t)f_f_sprintf, QC_NOOP, QDOM_DEFAULT, stringTypeInfo);
   // string f_vprintf(string fmt, any varg) {}
   ns.addBuiltinVariant("f_vprintf", (q_func_t)f_f_vprintf_VsVa, QC_NO_FLAGS, QDOM_TERMINAL_IO, stringTypeInfo, 2, stringTypeInfo, NULL, "fmt", anyTypeInfo, NULL, "varg");
   // string f_vsprintf(string fmt, any varg) {}
   ns.addBuiltinVariant("f_vsprintf", (q_func_t)f_f_vsprintf_VsVa, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 2, stringTypeInfo, NULL, "fmt", anyTypeInfo, NULL, "varg");
   // nothing flush() {}
   ns.addBuiltinVariant("flush", (q_func_t)f_flush, QC_NO_FLAGS, QDOM_TERMINAL_IO, nothingTypeInfo);
   // string force_encoding(string str, string encoding) {}
   ns.addBuiltinVariant("force_encoding", (q_func_t)f_force_encoding_VsVs, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 2, stringTypeInfo, NULL, "str", stringTypeInfo, NULL, "encoding");
   // nothing force_encoding() {}
   ns.addBuiltinVariant("force_encoding", (q_func_t)f_force_encoding, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string format_number(string fmt, softfloat num) {}
   ns.addBuiltinVariant("format_number", (q_func_t)f_format_number_Vsvf, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 2, stringTypeInfo, NULL, "fmt", softFloatTypeInfo, NULL, "num");
   // nothing format_number() {}
   ns.addBuiltinVariant("format_number", (q_func_t)f_format_number, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string get_encoding(string str) {}
   ns.addBuiltinVariant("get_encoding", (q_func_t)f_get_encoding_Vs, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, NULL, "str");
   // nothing get_encoding() {}
   ns.addBuiltinVariant("get_encoding", (q_func_t)f_get_encoding, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int index(softstring str, softstring substr, softint pos = 0) {}
   ns.addBuiltinVariant("index", (q_func_int64_t)f_index_vsvsvi, QC_RET_VALUE_ONLY, QDOM_DEFAULT, bigIntTypeInfo, 3, softStringTypeInfo, NULL, "str", softStringTypeInfo, NULL, "substr", softBigIntTypeInfo, zero(), "pos");
   // int index() {}
   ns.addBuiltinVariant("index", (q_func_int64_t)f_index, QC_RUNTIME_NOOP, QDOM_DEFAULT, bigIntTypeInfo);
   // string join(string str, ...) {}
   ns.addBuiltinVariant("join", (q_func_t)f_join_VsVV, QC_RET_VALUE_ONLY|QC_USES_EXTRA_ARGS, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, NULL, "str");
   // string join(string str, list l) {}
   ns.addBuiltinVariant("join", (q_func_t)f_join_VsVl, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 2, stringTypeInfo, NULL, "str", listTypeInfo, NULL, "l");
   // nothing join() {}
   ns.addBuiltinVariant("join", (q_func_t)f_join, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int length(softstring str) {}
   ns.addBuiltinVariant("length", (q_func_int64_t)f_length_vs, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, softStringTypeInfo, NULL, "str");
   // int length(binary bin) {}
   ns.addBuiltinVariant("length", (q_func_int64_t)f_length_Vy, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, binaryTypeInfo, NULL, "bin");
   // nothing length() {}
   ns.addBuiltinVariant("length", (q_func_t)f_length, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int length(any arg) {}
   ns.addBuiltinVariant("length", (q_func_int64_t)f_length_Va, QC_NOOP, QDOM_DEFAULT, bigIntTypeInfo, 1, anyTypeInfo, NULL, "arg");
   // int ord(softstring str, softint offset = 0) {}
   ns.addBuiltinVariant("ord", (q_func_int64_t)f_ord_vsvi, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 2, softStringTypeInfo, NULL, "str", softBigIntTypeInfo, zero(), "offset");
   // int ord() {}
   ns.addBuiltinVariant("ord", (q_func_int64_t)f_ord, QC_RUNTIME_NOOP, QDOM_DEFAULT, bigIntTypeInfo);
   // bool parse_boolean(string str) {}
   ns.addBuiltinVariant("parse_boolean", (q_func_bool_t)f_parse_boolean_Vs, QC_CONSTANT, QDOM_DEFAULT, boolTypeInfo, 1, stringTypeInfo, NULL, "str");
   // bool parse_boolean(...) {}
   ns.addBuiltinVariant("parse_boolean", (q_func_bool_t)f_parse_boolean_VV, QC_CONSTANT|QC_USES_EXTRA_ARGS, QDOM_DEFAULT, boolTypeInfo);
   // nothing print(...) {}
   ns.addBuiltinVariant("print", (q_func_t)f_print_VV, QC_USES_EXTRA_ARGS, QDOM_TERMINAL_IO, nothingTypeInfo);
   // string printf(string fmt, ...) {}
   ns.addBuiltinVariant("printf", (q_func_t)f_printf_VsVV, QC_USES_EXTRA_ARGS, QDOM_TERMINAL_IO, stringTypeInfo, 1, stringTypeInfo, NULL, "fmt");
   // string printf() {}
   ns.addBuiltinVariant("printf", (q_func_t)f_printf, QC_NOOP, QDOM_TERMINAL_IO, stringTypeInfo);
   // bool regex(string str, string regex, int options = 0) {}
   ns.addBuiltinVariant("regex", (q_func_bool_t)f_regex_VsVsVi, QC_RET_VALUE_ONLY, QDOM_DEFAULT, boolTypeInfo, 3, stringTypeInfo, NULL, "str", stringTypeInfo, NULL, "regex", bigIntTypeInfo, zero(), "options");
   // nothing regex() {}
   ns.addBuiltinVariant("regex", (q_func_t)f_regex, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // *list regex_extract(string str, string regex, int options = 0) {}
   ns.addBuiltinVariant("regex_extract", (q_func_t)f_regex_extract_VsVsVi, QC_RET_VALUE_ONLY, QDOM_DEFAULT, listOrNothingTypeInfo, 3, stringTypeInfo, NULL, "str", stringTypeInfo, NULL, "regex", bigIntTypeInfo, zero(), "options");
   // nothing regex_extract() {}
   ns.addBuiltinVariant("regex_extract", (q_func_t)f_regex_extract, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string regex_subst(string str, string regex, string subst, int options = 0) {}
   ns.addBuiltinVariant("regex_subst", (q_func_t)f_regex_subst_VsVsVsVi, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 4, stringTypeInfo, NULL, "str", stringTypeInfo, NULL, "regex", stringTypeInfo, NULL, "subst", bigIntTypeInfo, zero(), "options");
   // nothing regex_subst() {}
   ns.addBuiltinVariant("regex_subst", (q_func_t)f_regex_subst, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string replace(string str, string source, string target, int start = 0, int end = -1) {}
   ns.addBuiltinVariant("replace", (q_func_t)f_replace_VsVsVsViVi, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 5, stringTypeInfo, NULL, "str", stringTypeInfo, NULL, "source", stringTypeInfo, NULL, "target", bigIntTypeInfo, zero(), "start", bigIntTypeInfo, new QoreBigIntNode(-1), "end");
   // nothing replace() {}
   ns.addBuiltinVariant("replace", (q_func_t)f_replace, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string reverse(softstring str) {}
   ns.addBuiltinVariant("reverse", (q_func_t)f_reverse_vs, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 1, softStringTypeInfo, NULL, "str");
   // int rindex(softstring str, softstring substr, softint pos = -1) {}
   ns.addBuiltinVariant("rindex", (q_func_int64_t)f_rindex_vsvsvi, QC_RET_VALUE_ONLY, QDOM_DEFAULT, bigIntTypeInfo, 3, softStringTypeInfo, NULL, "str", softStringTypeInfo, NULL, "substr", softBigIntTypeInfo, new QoreBigIntNode(-1), "pos");
   // int rindex() {}
   ns.addBuiltinVariant("rindex", (q_func_int64_t)f_rindex, QC_RUNTIME_NOOP, QDOM_DEFAULT, bigIntTypeInfo);
   // list split(string sep, string str, bool with_separator = False) {}
   ns.addBuiltinVariant("split", (q_func_t)f_split_VsVsVb, QC_RET_VALUE_ONLY, QDOM_DEFAULT, listTypeInfo, 3, stringTypeInfo, NULL, "sep", stringTypeInfo, NULL, "str", boolTypeInfo, &False, "with_separator");
   // list split(string sep, string str, string quote, bool trim_unquoted = False) {}
   ns.addBuiltinVariant("split", (q_func_t)f_split_VsVsVsVb, QC_RET_VALUE_ONLY, QDOM_DEFAULT, listTypeInfo, 4, stringTypeInfo, NULL, "sep", stringTypeInfo, NULL, "str", stringTypeInfo, NULL, "quote", boolTypeInfo, &False, "trim_unquoted");
   // list split(binary sep, binary data) {}
   ns.addBuiltinVariant("split", (q_func_t)f_split_VyVy, QC_CONSTANT, QDOM_DEFAULT, listTypeInfo, 2, binaryTypeInfo, NULL, "sep", binaryTypeInfo, NULL, "data");
   // list split() {}
   ns.addBuiltinVariant("split", (q_func_t)f_split, QC_RUNTIME_NOOP, QDOM_DEFAULT, listTypeInfo);
   // string sprintf(string fmt, ...) {}
   ns.addBuiltinVariant("sprintf", (q_func_t)f_sprintf_VsVV, QC_CONSTANT|QC_USES_EXTRA_ARGS, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, NULL, "fmt");
   // string sprintf() {}
   ns.addBuiltinVariant("sprintf", (q_func_t)f_sprintf, QC_NOOP, QDOM_DEFAULT, stringTypeInfo);
   // int strlen(softstring str) {}
   ns.addBuiltinVariant("strlen", (q_func_int64_t)f_strlen_vs, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, softStringTypeInfo, NULL, "str");
   // nothing strlen() {}
   ns.addBuiltinVariant("strlen", (q_func_t)f_strlen, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int strlen(any arg) {}
   ns.addBuiltinVariant("strlen", (q_func_int64_t)f_strlen_Va, QC_NOOP, QDOM_DEFAULT, bigIntTypeInfo, 1, anyTypeInfo, NULL, "arg");
   // string strmul(softstring str, softint smul, *softint offset) {}
   ns.addBuiltinVariant("strmul", (q_func_t)f_strmul_vsvini, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 3, softStringTypeInfo, NULL, "str", softBigIntTypeInfo, NULL, "smul", softBigIntOrNothingTypeInfo, NULL, "offset");
   // string substr(softstring str, softint start) {}
   ns.addBuiltinVariant("substr", (q_func_t)f_substr_vsvi, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 2, softStringTypeInfo, NULL, "str", softBigIntTypeInfo, NULL, "start");
   // string substr(softstring str, softint start, softint len) {}
   ns.addBuiltinVariant("substr", (q_func_t)f_substr_vsvivi, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 3, softStringTypeInfo, NULL, "str", softBigIntTypeInfo, NULL, "start", softBigIntTypeInfo, NULL, "len");
   // binary substr(binary b, softint start) {}
   ns.addBuiltinVariant("substr", (q_func_t)f_substr_Vyvi, QC_CONSTANT, QDOM_DEFAULT, binaryTypeInfo, 2, binaryTypeInfo, NULL, "b", softBigIntTypeInfo, NULL, "start");
   // binary substr(binary b, softint start, softint len) {}
   ns.addBuiltinVariant("substr", (q_func_t)f_substr_Vyvivi, QC_CONSTANT, QDOM_DEFAULT, binaryTypeInfo, 3, binaryTypeInfo, NULL, "b", softBigIntTypeInfo, NULL, "start", softBigIntTypeInfo, NULL, "len");
   // nothing substr() {}
   ns.addBuiltinVariant("substr", (q_func_t)f_substr, QC_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string tolower(string str) {}
   ns.addBuiltinVariant("tolower", (q_func_t)f_tolower_Vs, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, NULL, "str");
   // nothing tolower() {}
   ns.addBuiltinVariant("tolower", (q_func_t)f_tolower, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string toupper(string str) {}
   ns.addBuiltinVariant("toupper", (q_func_t)f_toupper_Vs, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, NULL, "str");
   // nothing toupper() {}
   ns.addBuiltinVariant("toupper", (q_func_t)f_toupper, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string trim(string str, *string chars) {}
   ns.addBuiltinVariant("trim", (q_func_t)f_trim_VsNs, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 2, stringTypeInfo, NULL, "str", stringOrNothingTypeInfo, NULL, "chars");
   // *string trim(reference str, *string chars) {}
   ns.addBuiltinVariant("trim", (q_func_t)f_trim_VrNs, QC_NO_FLAGS, QDOM_DEFAULT, stringOrNothingTypeInfo, 2, referenceTypeInfo, NULL, "str", stringOrNothingTypeInfo, NULL, "chars");
   // nothing trim() {}
   ns.addBuiltinVariant("trim", (q_func_t)f_trim, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string trunc_str(softstring str, softint len, *string encoding) {}
   ns.addBuiltinVariant("trunc_str", (q_func_t)f_trunc_str_vsviNs, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 3, softStringTypeInfo, NULL, "str", softBigIntTypeInfo, NULL, "len", stringOrNothingTypeInfo, NULL, "encoding");
   // string vprintf(string fmt, ...) {}
   ns.addBuiltinVariant("vprintf", (q_func_t)f_vprintf_VsVV, QC_USES_EXTRA_ARGS, QDOM_TERMINAL_IO, stringTypeInfo, 1, stringTypeInfo, NULL, "fmt");
   // string vprintf() {}
   ns.addBuiltinVariant("vprintf", (q_func_t)f_vprintf, QC_NOOP, QDOM_DEFAULT, stringTypeInfo);
   // string vsprintf(string fmt, any varg) {}
   ns.addBuiltinVariant("vsprintf", (q_func_t)f_vsprintf_VsVa, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 2, stringTypeInfo, NULL, "fmt", anyTypeInfo, NULL, "varg");
   // string vsprintf() {}
   ns.addBuiltinVariant("vsprintf", (q_func_t)f_vsprintf, QC_NOOP, QDOM_DEFAULT, stringTypeInfo);

}

DLLLOCAL void init_string_constants(QoreNamespace& ns) {
# 314 "ql_string.qpp"
# 322 "ql_string.qpp"
# 325 "ql_string.qpp"
# 328 "ql_string.qpp"
   ns.addConstant("RE_Caseless", new QoreBigIntNode(PCRE_CASELESS));
   ns.addConstant("RE_DotAll", new QoreBigIntNode(PCRE_DOTALL));
   ns.addConstant("RE_Extended", new QoreBigIntNode(PCRE_EXTENDED));
   ns.addConstant("RE_Global", new QoreBigIntNode(QRE_GLOBAL));
   ns.addConstant("RE_MultiLine", new QoreBigIntNode(PCRE_MULTILINE));
# 335 "ql_string.qpp"
# 393 "ql_string.qpp"
# 418 "ql_string.qpp"
# 434 "ql_string.qpp"
# 440 "ql_string.qpp"
# 447 "ql_string.qpp"
# 469 "ql_string.qpp"
# 475 "ql_string.qpp"
# 482 "ql_string.qpp"
# 512 "ql_string.qpp"
# 518 "ql_string.qpp"
# 548 "ql_string.qpp"
# 554 "ql_string.qpp"
# 579 "ql_string.qpp"
# 605 "ql_string.qpp"
# 629 "ql_string.qpp"
# 654 "ql_string.qpp"
# 660 "ql_string.qpp"
# 691 "ql_string.qpp"
# 698 "ql_string.qpp"
# 727 "ql_string.qpp"
# 734 "ql_string.qpp"
# 765 "ql_string.qpp"
# 772 "ql_string.qpp"
# 801 "ql_string.qpp"
# 808 "ql_string.qpp"
# 829 "ql_string.qpp"
# 836 "ql_string.qpp"
# 856 "ql_string.qpp"
# 863 "ql_string.qpp"
# 869 "ql_string.qpp"
# 895 "ql_string.qpp"
# 921 "ql_string.qpp"
# 939 "ql_string.qpp"
# 946 "ql_string.qpp"
# 963 "ql_string.qpp"
# 969 "ql_string.qpp"
# 987 "ql_string.qpp"
# 993 "ql_string.qpp"
# 1010 "ql_string.qpp"
# 1016 "ql_string.qpp"
# 1046 "ql_string.qpp"
# 1052 "ql_string.qpp"
# 1107 "ql_string.qpp"
# 1113 "ql_string.qpp"
# 1145 "ql_string.qpp"
# 1151 "ql_string.qpp"
# 1231 "ql_string.qpp"
# 1237 "ql_string.qpp"
# 1255 "ql_string.qpp"
# 1275 "ql_string.qpp"
# 1281 "ql_string.qpp"
# 1301 "ql_string.qpp"
# 1326 "ql_string.qpp"
# 1332 "ql_string.qpp"
# 1354 "ql_string.qpp"
# 1385 "ql_string.qpp"
# 1391 "ql_string.qpp"
# 1437 "ql_string.qpp"
# 1476 "ql_string.qpp"
# 1497 "ql_string.qpp"
# 1503 "ql_string.qpp"
# 1521 "ql_string.qpp"
# 1541 "ql_string.qpp"
# 1563 "ql_string.qpp"
# 1570 "ql_string.qpp"
# 1592 "ql_string.qpp"
# 1599 "ql_string.qpp"
# 1623 "ql_string.qpp"
# 1645 "ql_string.qpp"
# 1652 "ql_string.qpp"
# 1676 "ql_string.qpp"
# 1683 "ql_string.qpp"
# 1707 "ql_string.qpp"
# 1714 "ql_string.qpp"
# 1740 "ql_string.qpp"
# 1764 "ql_string.qpp"
# 1771 "ql_string.qpp"
# 1782 "ql_string.qpp"
# 1796 "ql_string.qpp"

}
