/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
  ql_thread.qpp

  POSIX thread library for Qore

  Qore Programming Language

  Copyright 2003 - 2013 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>
#include <qore/intern/ql_thread.h>
#include <qore/intern/qore_program_private.h>

#include <pthread.h>

static void delete_thread_data_intern(const QoreListNode *args, ExceptionSink *xsink) {
   // get thread data hash
   QoreHashNode *data = getProgram()->getThreadData();
      
   // iterate through arguments and delete each key
   for (unsigned i = 0; i < args->size(); i++) {
      const AbstractQoreNode *p = get_param(args, i);
      if (p) {
         QoreStringValueHelper t(p, QCS_DEFAULT, xsink);
	 if (*xsink)
	    break;

         data->deleteKey(*t, xsink);
         if (*xsink)
            break;
      }
   }
}

static QoreHashNode *remove_thread_data_intern(const QoreListNode *args, ExceptionSink *xsink) {
   // get thread data hash
   QoreHashNode *data = getProgram()->getThreadData();
      
   QoreHashNode* rv = new QoreHashNode;

   // iterate through arguments and remove each key
   for (unsigned i = 0; i < args->size(); i++) {
      const AbstractQoreNode *p = get_param(args, i);
      if (p) {
         QoreStringValueHelper t(p, QCS_DEFAULT, xsink);
	 if (*xsink)
	    break;

	 rv->setKeyValue(*t, data->takeKeyValue((*t)->getBuffer()), xsink);
	 assert(!*xsink);
      }
   }

   return rv;
}
# 74 "ql_thread.qpp"
# 91 "ql_thread.qpp"
# 104 "ql_thread.qpp"
# 118 "ql_thread.qpp"
# 133 "ql_thread.qpp"
# 149 "ql_thread.qpp"
# 155 "ql_thread.qpp"
# 174 "ql_thread.qpp"
# 191 "ql_thread.qpp"
# 208 "ql_thread.qpp"
# 225 "ql_thread.qpp"
# 238 "ql_thread.qpp"
# 255 "ql_thread.qpp"
# 261 "ql_thread.qpp"
# 275 "ql_thread.qpp"
# 305 "ql_thread.qpp"
# 328 "ql_thread.qpp"
# 358 "ql_thread.qpp"
# 387 "ql_thread.qpp"
# 403 "ql_thread.qpp"
# 428 "ql_thread.qpp"
# 448 "ql_thread.qpp"
// nothing delete_all_thread_data() {}
static AbstractQoreNode* f_delete_all_thread_data(const QoreListNode* args, ExceptionSink* xsink) {
# 234 "ql_thread.qpp"
   qore_program_private::clearThreadData(*getProgram(), xsink);
   return 0;
}

// nothing delete_thread_data(...) {}
static AbstractQoreNode* f_delete_thread_data_VV(const QoreListNode* args, ExceptionSink* xsink) {
# 168 "ql_thread.qpp"
   if (num_args(args))
      delete_thread_data_intern(args, xsink);
   return 0;
}

// nothing delete_thread_data(list l) {}
static AbstractQoreNode* f_delete_thread_data_Vl(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreListNode* l = HARD_QORE_LIST(args, 0);
# 187 "ql_thread.qpp"
   delete_thread_data_intern(l, xsink);
   return 0;
}

// hash getAllThreadCallStacks() {}
static AbstractQoreNode* f_getAllThreadCallStacks(const QoreListNode* args, ExceptionSink* xsink) {
# 296 "ql_thread.qpp"
#ifdef QORE_RUNTIME_THREAD_STACK_TRACE
   return getAllCallStacks();
#else
   xsink->raiseException("MISSING-FEATURE-ERROR", "this version of the Qore library was built without support for runtime thread stack tracing; check Qore::Option::HAVE_RUNTIME_THREAD_STACK_TRACE before calling");
   return 0;
#endif
   return 0;
}

// hash get_all_thread_data() {}
static AbstractQoreNode* f_get_all_thread_data(const QoreListNode* args, ExceptionSink* xsink) {
# 270 "ql_thread.qpp"
   const QoreHashNode* h = getProgram()->getThreadData();
   return h->copy();
}

// any get_thread_data(string key) {}
static AbstractQoreNode* f_get_thread_data_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* key = HARD_QORE_STRING(args, 0);
# 247 "ql_thread.qpp"
   TempEncodingHelper tmp(key, QCS_DEFAULT, xsink);
   if (!tmp)
      return 0;
   const QoreHashNode* h = getProgram()->getThreadData();
   return h->getReferencedKeyValue(tmp->getBuffer());
}

// nothing get_thread_data() {}
static AbstractQoreNode* f_get_thread_data(const QoreListNode* args, ExceptionSink* xsink) {
# 258 "ql_thread.qpp"

   return 0;
}

// *TimeZone get_thread_tz() {}
static QoreObject* f_get_thread_tz(const QoreListNode* args, ExceptionSink* xsink) {
# 464 "ql_thread.qpp"
   bool set;
   const AbstractQoreZoneInfo* tz = get_thread_tz(set);
   if (!set)
      return 0;
   return new QoreObject(QC_TIMEZONE, 0, new TimeZoneData(tz));
}

// int gettid() {}
static int64 f_gettid(const QoreListNode* args, ExceptionSink* xsink) {
# 87 "ql_thread.qpp"
   return gettid();
}

// nothing mark_thread_resources() {}
static AbstractQoreNode* f_mark_thread_resources(const QoreListNode* args, ExceptionSink* xsink) {
# 383 "ql_thread.qpp"
   mark_thread_resources();
   return 0;
}

// int num_threads() {}
static int64 f_num_threads(const QoreListNode* args, ExceptionSink* xsink) {
# 100 "ql_thread.qpp"
   return thread_list.getNumThreads();
}

// *hash remove_thread_data(...) {}
static AbstractQoreNode* f_remove_thread_data_VV(const QoreListNode* args, ExceptionSink* xsink) {
# 204 "ql_thread.qpp"
   return num_args(args) ? remove_thread_data_intern(args, xsink) : 0;
}

// hash remove_thread_data(list l) {}
static AbstractQoreNode* f_remove_thread_data_Vl(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreListNode* l = HARD_QORE_LIST(args, 0);
# 221 "ql_thread.qpp"
   return remove_thread_data_intern(l, xsink);
}

// nothing save_thread_data(hash h) {}
static AbstractQoreNode* f_save_thread_data_Vh(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreHashNode* h = HARD_QORE_HASH(args, 0);
# 128 "ql_thread.qpp"
   QoreHashNode* data = getProgram()->getThreadData();
   data->merge(h, xsink);
   return 0;
}

// nothing save_thread_data(string key, any value) {}
static AbstractQoreNode* f_save_thread_data_VsVa(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* key = HARD_QORE_STRING(args, 0);
   const AbstractQoreNode* value = get_param(args, 1);
# 144 "ql_thread.qpp"
   QoreHashNode* data = getProgram()->getThreadData();
   data->setKeyValue(key, value ? value->refSelf() : 0, xsink);
   return 0;
}

// nothing save_thread_data() {}
static AbstractQoreNode* f_save_thread_data(const QoreListNode* args, ExceptionSink* xsink) {
# 152 "ql_thread.qpp"

   return 0;
}

// bool set_thread_init(code init) {}
static bool f_set_thread_init_Vc(const QoreListNode* args, ExceptionSink* xsink) {
   const ResolvedCallReferenceNode* init = reinterpret_cast<const ResolvedCallReferenceNode*>(get_param(args, 0));
# 399 "ql_thread.qpp"
   return qore_program_private::setThreadInit(*getProgram(), init, xsink);
}

// nothing set_thread_tz(TimeZone zone) {}
static AbstractQoreNode* f_set_thread_tz_C8TimeZone(const QoreListNode* args, ExceptionSink* xsink) {
   HARD_QORE_OBJ_DATA(zone, TimeZoneData, args, 0, CID_TIMEZONE, "set_thread_tz()", "TimeZone", xsink);
   if (*xsink)
      return 0;
# 423 "ql_thread.qpp"
   ReferenceHolder<TimeZoneData> holder(zone, xsink);
   set_thread_tz(zone->get());
   return 0;
}

// nothing set_thread_tz() {}
static AbstractQoreNode* f_set_thread_tz(const QoreListNode* args, ExceptionSink* xsink) {
# 444 "ql_thread.qpp"
   clear_thread_tz();
   return 0;
}

// list thread_list() {}
static AbstractQoreNode* f_thread_list(const QoreListNode* args, ExceptionSink* xsink) {
# 114 "ql_thread.qpp"
   return get_thread_list();
}

// nothing throwThreadResourceExceptions() {}
static AbstractQoreNode* f_throwThreadResourceExceptions(const QoreListNode* args, ExceptionSink* xsink) {
# 324 "ql_thread.qpp"
   purge_thread_resources(xsink);
   return 0;
}

// bool throw_thread_resource_exceptions_to_mark() {}
static bool f_throw_thread_resource_exceptions_to_mark(const QoreListNode* args, ExceptionSink* xsink) {
# 354 "ql_thread.qpp"
   return (bool)purge_thread_resources_to_mark(xsink);
}


DLLLOCAL void init_thread_functions(QoreNamespace& ns) {
   // nothing delete_all_thread_data() {}
   ns.addBuiltinVariant("delete_all_thread_data", (q_func_t)f_delete_all_thread_data, QC_NO_FLAGS, QDOM_THREAD_CONTROL, nothingTypeInfo);
   // nothing delete_thread_data(...) {}
   ns.addBuiltinVariant("delete_thread_data", (q_func_t)f_delete_thread_data_VV, QC_USES_EXTRA_ARGS, QDOM_THREAD_CONTROL, nothingTypeInfo);
   // nothing delete_thread_data(list l) {}
   ns.addBuiltinVariant("delete_thread_data", (q_func_t)f_delete_thread_data_Vl, QC_NO_FLAGS, QDOM_THREAD_CONTROL, nothingTypeInfo, 1, listTypeInfo, NULL, "l");
   // hash getAllThreadCallStacks() {}
   ns.addBuiltinVariant("getAllThreadCallStacks", (q_func_t)f_getAllThreadCallStacks, QC_NO_FLAGS, QDOM_THREAD_CONTROL|QDOM_THREAD_INFO, hashTypeInfo);
   // hash get_all_thread_data() {}
   ns.addBuiltinVariant("get_all_thread_data", (q_func_t)f_get_all_thread_data, QC_CONSTANT, QDOM_THREAD_CONTROL|QDOM_THREAD_INFO, hashTypeInfo);
   // any get_thread_data(string key) {}
   ns.addBuiltinVariant("get_thread_data", (q_func_t)f_get_thread_data_Vs, QC_CONSTANT, QDOM_THREAD_CONTROL|QDOM_THREAD_INFO, anyTypeInfo, 1, stringTypeInfo, NULL, "key");
   // nothing get_thread_data() {}
   ns.addBuiltinVariant("get_thread_data", (q_func_t)f_get_thread_data, QC_RUNTIME_NOOP, QDOM_THREAD_CONTROL|QDOM_THREAD_INFO, nothingTypeInfo);
   // *TimeZone get_thread_tz() {}
   ns.addBuiltinVariant("get_thread_tz", (q_func_t)f_get_thread_tz, QC_NO_FLAGS, QDOM_LOCALE_CONTROL, QC_TIMEZONE->getOrNothingTypeInfo());
   // int gettid() {}
   ns.addBuiltinVariant("gettid", (q_func_int64_t)f_gettid, QC_CONSTANT, QDOM_THREAD_INFO, bigIntTypeInfo);
   // nothing mark_thread_resources() {}
   ns.addBuiltinVariant("mark_thread_resources", (q_func_t)f_mark_thread_resources, QC_NO_FLAGS, QDOM_THREAD_CONTROL, nothingTypeInfo);
   // int num_threads() {}
   ns.addBuiltinVariant("num_threads", (q_func_int64_t)f_num_threads, QC_CONSTANT, QDOM_THREAD_INFO, bigIntTypeInfo);
   // *hash remove_thread_data(...) {}
   ns.addBuiltinVariant("remove_thread_data", (q_func_t)f_remove_thread_data_VV, QC_USES_EXTRA_ARGS, QDOM_THREAD_CONTROL, hashOrNothingTypeInfo);
   // hash remove_thread_data(list l) {}
   ns.addBuiltinVariant("remove_thread_data", (q_func_t)f_remove_thread_data_Vl, QC_NO_FLAGS, QDOM_THREAD_CONTROL, hashTypeInfo, 1, listTypeInfo, NULL, "l");
   // nothing save_thread_data(hash h) {}
   ns.addBuiltinVariant("save_thread_data", (q_func_t)f_save_thread_data_Vh, QC_NO_FLAGS, QDOM_THREAD_CONTROL, nothingTypeInfo, 1, hashTypeInfo, NULL, "h");
   // nothing save_thread_data(string key, any value) {}
   ns.addBuiltinVariant("save_thread_data", (q_func_t)f_save_thread_data_VsVa, QC_NO_FLAGS, QDOM_THREAD_CONTROL, nothingTypeInfo, 2, stringTypeInfo, NULL, "key", anyTypeInfo, NULL, "value");
   // nothing save_thread_data() {}
   ns.addBuiltinVariant("save_thread_data", (q_func_t)f_save_thread_data, QC_RUNTIME_NOOP, QDOM_THREAD_CONTROL, nothingTypeInfo);
   // bool set_thread_init(code init) {}
   ns.addBuiltinVariant("set_thread_init", (q_func_bool_t)f_set_thread_init_Vc, QC_NO_FLAGS, QDOM_THREAD_CONTROL, boolTypeInfo, 1, codeTypeInfo, NULL, "init");
   // nothing set_thread_tz(TimeZone zone) {}
   ns.addBuiltinVariant("set_thread_tz", (q_func_t)f_set_thread_tz_C8TimeZone, QC_NO_FLAGS, QDOM_LOCALE_CONTROL, nothingTypeInfo, 1, QC_TIMEZONE->getTypeInfo(), NULL, "zone");
   // nothing set_thread_tz() {}
   ns.addBuiltinVariant("set_thread_tz", (q_func_t)f_set_thread_tz, QC_NO_FLAGS, QDOM_LOCALE_CONTROL, nothingTypeInfo);
   // list thread_list() {}
   ns.addBuiltinVariant("thread_list", (q_func_t)f_thread_list, QC_CONSTANT, QDOM_THREAD_INFO, listTypeInfo);
   // nothing throwThreadResourceExceptions() {}
   ns.addBuiltinVariant("throwThreadResourceExceptions", (q_func_t)f_throwThreadResourceExceptions, QC_NO_FLAGS, QDOM_THREAD_CONTROL, nothingTypeInfo);
   // bool throw_thread_resource_exceptions_to_mark() {}
   ns.addBuiltinVariant("throw_thread_resource_exceptions_to_mark", (q_func_bool_t)f_throw_thread_resource_exceptions_to_mark, QC_NO_FLAGS, QDOM_THREAD_CONTROL, boolTypeInfo);

}
