/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
  ql_time.qpp

  Qore Programming Language

  Copyright 2003 - 2013 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>
#include <qore/intern/ql_time.h>
#include <qore/intern/QC_TimeZone.h>
#include <qore/intern/qore_date_private.h>

#include <stdio.h>
#include <time.h>
#include <sys/time.h>

static int qd_get_digit(unsigned& rv, const char*& p, unsigned factor) {
   if (*p < '0' || *p > '9')
      return -1;

   rv += (*p - '0') * factor;
   ++p;
   return 0;
}

// returns 0 if invalid data is encountered
static unsigned parse_int_2(const char *p) {
   unsigned rv = 0;
   if (qd_get_digit(rv, p, 10))
      return 0;
   if (qd_get_digit(rv, p, 1))
      return 0;
   return rv;
}

// returns 0 if invalid data is encountered
static unsigned parse_int_3(const char *p) {
   unsigned rv = 0;
   if (qd_get_digit(rv, p, 100))
      return 0;
   if (qd_get_digit(rv, p, 10))
      return 0;
   if (qd_get_digit(rv, p, 1))
      return 0;
   return rv;
}

// returns 0 if invalid data is encountered
static unsigned parse_int_4(const char *p) {
   unsigned rv = 0;
   if (qd_get_digit(rv, p, 1000))
      return 0;
   if (qd_get_digit(rv, p, 100))
      return 0;
   if (qd_get_digit(rv, p, 10))
      return 0;
   if (qd_get_digit(rv, p, 1))
      return 0;
   return rv;
}

// returns 0 if invalid data is encountered
static unsigned parse_int_6(const char *p) {
   unsigned rv = 0;
   if (qd_get_digit(rv, p, 100000))
      return 0;
   if (qd_get_digit(rv, p, 10000))
      return 0;
   if (qd_get_digit(rv, p, 1000))
      return 0;
   if (qd_get_digit(rv, p, 100))
      return 0;
   if (qd_get_digit(rv, p, 10))
      return 0;
   if (qd_get_digit(rv, p, 1))
      return 0;
   return rv;
}

QoreHashNode* date_info(const DateTime& d) {
   qore_tm info;
   d.getInfo(info);

   QoreHashNode *h = new QoreHashNode;
   h->setKeyValue("relative", get_bool_node(d.isRelative()), 0);
   h->setKeyValue("year", new QoreBigIntNode(info.year), 0);
   h->setKeyValue("month", new QoreBigIntNode(info.month), 0);
   h->setKeyValue("day", new QoreBigIntNode(info.day), 0);
   h->setKeyValue("hour", new QoreBigIntNode(info.hour), 0);
   h->setKeyValue("minute", new QoreBigIntNode(info.minute), 0);
   h->setKeyValue("second", new QoreBigIntNode(info.second), 0);
   h->setKeyValue("microsecond", new QoreBigIntNode(info.us), 0);

   if (d.isAbsolute()) {
      h->setKeyValue("dow", new QoreBigIntNode(d.getDayOfWeek()), 0);
      h->setKeyValue("doy", new QoreBigIntNode(d.getDayNumber()), 0);
      h->setKeyValue("utc_secs_east", new QoreBigIntNode(info.utc_secs_east), 0);
      h->setKeyValue("dst", get_bool_node(info.dst), 0);
      h->setKeyValue("zone_name", new QoreStringNode(info.zone_name), 0);
      h->setKeyValue("zone", new QoreObject(QC_TIMEZONE, 0, new TimeZoneData(info.zone)), 0);
   }

   return h;
}

DateTimeNode* make_date_with_mask(const AbstractQoreZoneInfo* tz, const QoreString& dtstr, const QoreString& mask, ExceptionSink* xsink) {
   qore_tm dt;
   dt.clear();

   const char *d = dtstr.getBuffer();
   const char *de = d + dtstr.strlen();

   const char *s = mask.getBuffer();
   const char *se = s + mask.strlen();

   while (*s) {
      // break if we run out of data in either the mask or the date string
      if (d >= de || s >= se)
         break;
      switch (*s) {
         case 'Y':
            if (s[1] != 'Y') {
               xsink->raiseException("DATE-CONVERT-ERROR", "'Y' has to be used as 'YY' or 'YYYY'");
               return 0;
               break;
            }
            if (s[2] == 'Y' && s[3] == 'Y') {
               dt.year = parse_int_4(d);
               s += 3;
               d += 3;
            }
            else {
               ++s;
               // obtain the current century
               {
                  DateTime tmpdt(q_epoch());
                  dt.year = parse_int_2(d) + (tmpdt.getYear() / 100 * 100);
               }
               ++d;
            }
            break;

         case 'M':
            if (s[1] == 'M') {
               dt.month = parse_int_2(d);
               s++;
               d++;
               break;
            }
            // 'M' is not supported because there is no clear way how to get eg. 1 or 11
            if ((s[1] == 'o' || s[1] == 'O') && (s[2] == 'n' || s[2] == 'N')) {
               QoreString str(d, 3);
               dt.month = str.strlen() == 3 ? qore_date_info::getMonthIxFromAbbr(str.getBuffer()) : -1;
               if (dt.month < 0 || dt.month > 11) {
                  xsink->raiseException("DATE-CONVERT-ERROR", "Invalid 'Mon' string: '%s'", !str.empty() ? str.getBuffer() : "<none>");
                  return 0;
               }

               ++dt.month;
               s += 2;
               d += 2;
               break;
            }
            break;

         case 'm':
            if (s[1] == 'm') {
               dt.minute = parse_int_2(d);
               s++;
               d++;
               break;
            }
            if (s[1] == 's') {
               dt.us = parse_int_3(d) * 1000;
               ++s;
               d += 2;
               break;
            }
            break;

         case 'D':
         case 'd':
            if (s[1] == 'D' || s[1] == 'd') {
               dt.day = parse_int_2(d);
               s++;
               d++;
            }
           break;

         case 'H':
         case 'h':
            if (s[1] == 'H' || s[1] == 'h') {
               dt.hour = parse_int_2(d);
               s++;
               d++;
            }
            break;

         case 's':
            if (s[1] == 's' && s[2] == 's') {
               if (s[3] == 's' && s[4] == 's' && s[5] == 's') {
                  dt.us = parse_int_6(d);
                  s += 5;
                  d += 5;
               }
               else {
                  dt.us = parse_int_3(d) * 1000;
                  s += 2;
                  d += 2;
               }
            }
            break;

         case 'S':
            if (s[1] == 'S') {
               dt.second = parse_int_2(d);
               s++;
               d++;
            }
            break;

         case 'u':
            if (s[1] == 's') {
               dt.us = parse_int_6(d);
               ++s;
               d += 5;
               break;
            }
      }
      s++;
      d++;
   }

   if (*xsink)
      return 0;

   return DateTimeNode::makeAbsolute(tz, dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.us);
}
# 257 "ql_time.qpp"
# 336 "ql_time.qpp"
# 359 "ql_time.qpp"
# 382 "ql_time.qpp"
# 399 "ql_time.qpp"
# 420 "ql_time.qpp"
# 426 "ql_time.qpp"
# 446 "ql_time.qpp"
# 463 "ql_time.qpp"
# 481 "ql_time.qpp"
# 498 "ql_time.qpp"
# 515 "ql_time.qpp"
# 533 "ql_time.qpp"
# 550 "ql_time.qpp"
# 567 "ql_time.qpp"
# 573 "ql_time.qpp"
# 592 "ql_time.qpp"
# 598 "ql_time.qpp"
# 617 "ql_time.qpp"
# 623 "ql_time.qpp"
# 647 "ql_time.qpp"
# 653 "ql_time.qpp"
# 677 "ql_time.qpp"
# 683 "ql_time.qpp"
# 707 "ql_time.qpp"
# 713 "ql_time.qpp"
# 737 "ql_time.qpp"
# 743 "ql_time.qpp"
# 767 "ql_time.qpp"
# 773 "ql_time.qpp"
# 797 "ql_time.qpp"
# 803 "ql_time.qpp"
# 834 "ql_time.qpp"
# 840 "ql_time.qpp"
# 871 "ql_time.qpp"
# 888 "ql_time.qpp"
# 894 "ql_time.qpp"
# 911 "ql_time.qpp"
# 917 "ql_time.qpp"
# 934 "ql_time.qpp"
# 940 "ql_time.qpp"
# 957 "ql_time.qpp"
# 963 "ql_time.qpp"
# 980 "ql_time.qpp"
# 986 "ql_time.qpp"
# 1005 "ql_time.qpp"
# 1011 "ql_time.qpp"
# 1030 "ql_time.qpp"
# 1036 "ql_time.qpp"
# 1055 "ql_time.qpp"
# 1074 "ql_time.qpp"
# 1080 "ql_time.qpp"
# 1095 "ql_time.qpp"
# 1101 "ql_time.qpp"
# 1118 "ql_time.qpp"
# 1124 "ql_time.qpp"
# 1142 "ql_time.qpp"
# 1148 "ql_time.qpp"
# 1173 "ql_time.qpp"
# 1179 "ql_time.qpp"
# 1200 "ql_time.qpp"
# 1206 "ql_time.qpp"
# 1227 "ql_time.qpp"
# 1244 "ql_time.qpp"
# 1261 "ql_time.qpp"
# 1278 "ql_time.qpp"
# 1294 "ql_time.qpp"
# 1300 "ql_time.qpp"
# 1316 "ql_time.qpp"
# 1340 "ql_time.qpp"
# 1364 "ql_time.qpp"
# 1387 "ql_time.qpp"
# 1404 "ql_time.qpp"
# 1421 "ql_time.qpp"
# 1438 "ql_time.qpp"
# 1445 "ql_time.qpp"
# 1462 "ql_time.qpp"
# 1469 "ql_time.qpp"
# 1480 "ql_time.qpp"
# 1504 "ql_time.qpp"
# 1519 "ql_time.qpp"
# 1534 "ql_time.qpp"
# 1543 "ql_time.qpp"
# 1552 "ql_time.qpp"
// int clock_getmicros() {}
static int64 f_clock_getmicros(const QoreListNode* args, ExceptionSink* xsink) {
# 1274 "ql_time.qpp"
   return q_clock_getmicros();
}

// int clock_getmillis() {}
static int64 f_clock_getmillis(const QoreListNode* args, ExceptionSink* xsink) {
# 1240 "ql_time.qpp"
   return q_clock_getmillis();
}

// int clock_getnanos() {}
static int64 f_clock_getnanos(const QoreListNode* args, ExceptionSink* xsink) {
# 1257 "ql_time.qpp"
   return q_clock_getnanos();
}

// date date(date dt) {}
static AbstractQoreNode* f_date_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 1476 "ql_time.qpp"
  return dt->refSelf();
}

// date date(string dtstr) {}
static AbstractQoreNode* f_date_Vs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* dtstr = HARD_QORE_STRING(args, 0);
# 1499 "ql_time.qpp"
   DateTimeNodeValueHelper date(dtstr);
   return date.getReferencedValue();
}

// date date(float f) {}
static AbstractQoreNode* f_date_Vf(const QoreListNode* args, ExceptionSink* xsink) {
   double f = HARD_QORE_FLOAT(args, 0);
# 1515 "ql_time.qpp"
   return DateTimeNode::makeAbsoluteLocal(currentTZ(), (int64)f, (int)((f - (float)((int)f)) * 1000000));
}

// date date(softint i) {}
static AbstractQoreNode* f_date_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 i = HARD_QORE_INT(args, 0);
# 1530 "ql_time.qpp"
   return new DateTimeNode(i);
}

// date date() {}
static AbstractQoreNode* f_date(const QoreListNode* args, ExceptionSink* xsink) {
# 1539 "ql_time.qpp"
   return zero_date();
}

// date date(null null) {}
static AbstractQoreNode* f_date_Vu(const QoreListNode* args, ExceptionSink* xsink) {
# 1548 "ql_time.qpp"
   return zero_date();
}

// date date(string dtstr, string mask) {}
static AbstractQoreNode* f_date_VsVs(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* dtstr = HARD_QORE_STRING(args, 0);
   const QoreStringNode* mask = HARD_QORE_STRING(args, 1);
# 1568 "ql_time.qpp"
   return make_date_with_mask(currentTZ(), *dtstr, *mask, xsink);
}

// hash date_info(date dt) {}
static AbstractQoreNode* f_date_info_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 1400 "ql_time.qpp"
   return date_info(*dt);
}

// hash date_info() {}
static AbstractQoreNode* f_date_info(const QoreListNode* args, ExceptionSink* xsink) {
# 1413 "ql_time.qpp"
   int us;
   int64 seconds = q_epoch_us(us);
   DateTime d;
   d.setDate(currentTZ(), seconds, us);
   return date_info(d);
}

// date date_ms(softint ms) {}
static AbstractQoreNode* f_date_ms_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 ms = HARD_QORE_INT(args, 0);
# 1290 "ql_time.qpp"
   return new DateTimeNode(ms / 1000, (int)(ms % 1000));
}

// nothing date_ms() {}
static AbstractQoreNode* f_date_ms(const QoreListNode* args, ExceptionSink* xsink) {
# 1297 "ql_time.qpp"

   return 0;
}

// date date_us(softint us) {}
static AbstractQoreNode* f_date_us_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 us = HARD_QORE_INT(args, 0);
# 1312 "ql_time.qpp"
   return DateTimeNode::makeAbsoluteLocal(currentTZ(), us / 1000000, (int)(us % 1000000));
}

// date days(softint days) {}
static AbstractQoreNode* f_days_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 days = HARD_QORE_INT(args, 0);
# 703 "ql_time.qpp"
   return new DateTimeNode(0, 0, (int)days, 0, 0, 0, 0, true);
}

// nothing days() {}
static AbstractQoreNode* f_days(const QoreListNode* args, ExceptionSink* xsink) {
# 710 "ql_time.qpp"

   return 0;
}

// string format_date(string format, date dt) {}
static AbstractQoreNode* f_format_date_VsVd(const QoreListNode* args, ExceptionSink* xsink) {
   const QoreStringNode* format = HARD_QORE_STRING(args, 0);
   const DateTimeNode* dt = HARD_QORE_DATE(args, 1);
# 414 "ql_time.qpp"
   QoreStringNode *rv = new QoreStringNode;
   dt->format(*rv, format->getBuffer());
   return rv;
}

// nothing format_date() {}
static AbstractQoreNode* f_format_date(const QoreListNode* args, ExceptionSink* xsink) {
# 423 "ql_time.qpp"

   return 0;
}

// date getDateFromISOWeek(softint year, softint week, softint day = 1) {}
static AbstractQoreNode* f_getDateFromISOWeek_vivivi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 year = HARD_QORE_INT(args, 0);
   int64 week = HARD_QORE_INT(args, 1);
   int64 day = HARD_QORE_INT(args, 2);
# 1223 "ql_time.qpp"
   return DateTimeNode::getDateFromISOWeek(year, week, day, xsink);
}

// int getDayNumber(date dt) {}
static int64 f_getDayNumber_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 1091 "ql_time.qpp"
   return dt->getDayNumber();
}

// nothing getDayNumber() {}
static AbstractQoreNode* f_getDayNumber(const QoreListNode* args, ExceptionSink* xsink) {
# 1098 "ql_time.qpp"

   return 0;
}

// int getDayOfWeek(date dt) {}
static int64 f_getDayOfWeek_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 1114 "ql_time.qpp"
   return dt->getDayOfWeek();
}

// nothing getDayOfWeek() {}
static AbstractQoreNode* f_getDayOfWeek(const QoreListNode* args, ExceptionSink* xsink) {
# 1121 "ql_time.qpp"

   return 0;
}

// int getISODayOfWeek(date dt) {}
static int64 f_getISODayOfWeek_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 1137 "ql_time.qpp"
   int d = dt->getDayOfWeek();
   return (!d ? 7 : d);
}

// nothing getISODayOfWeek() {}
static AbstractQoreNode* f_getISODayOfWeek(const QoreListNode* args, ExceptionSink* xsink) {
# 1145 "ql_time.qpp"

   return 0;
}

// hash getISOWeekHash(date dt) {}
static AbstractQoreNode* f_getISOWeekHash_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 1162 "ql_time.qpp"
   int year, week, day;
   dt->getISOWeek(year, week, day);
   QoreHashNode *h = new QoreHashNode;
   h->setKeyValue("year", new QoreBigIntNode(year), 0);
   h->setKeyValue("week", new QoreBigIntNode(week), 0);
   h->setKeyValue("day", new QoreBigIntNode(day), 0);

   return h;
}

// nothing getISOWeekHash() {}
static AbstractQoreNode* f_getISOWeekHash(const QoreListNode* args, ExceptionSink* xsink) {
# 1176 "ql_time.qpp"

   return 0;
}

// string getISOWeekString(date dt) {}
static AbstractQoreNode* f_getISOWeekString_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 1192 "ql_time.qpp"
   int year, week, day;
   dt->getISOWeek(year, week, day);
   QoreStringNode *str = new QoreStringNode();
   str->sprintf("%04d-W%02d-%d", year, week, day);
   return str;
}

// nothing getISOWeekString() {}
static AbstractQoreNode* f_getISOWeekString(const QoreListNode* args, ExceptionSink* xsink) {
# 1203 "ql_time.qpp"

   return 0;
}

// int get_days(date dt) {}
static int64 f_get_days_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 930 "ql_time.qpp"
   return dt->getDay();
}

// nothing get_days() {}
static AbstractQoreNode* f_get_days(const QoreListNode* args, ExceptionSink* xsink) {
# 937 "ql_time.qpp"

   return 0;
}

// int get_duration_microseconds(date dt) {}
static int64 f_get_duration_microseconds_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 1383 "ql_time.qpp"
   return dt->getRelativeMicroseconds();
}

// int get_duration_milliseconds(date dt) {}
static int64 f_get_duration_milliseconds_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 1360 "ql_time.qpp"
   return dt->getRelativeMilliseconds();
}

// int get_duration_seconds(date dt) {}
static int64 f_get_duration_seconds_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 1336 "ql_time.qpp"
   return dt->getRelativeSeconds();
}

// int get_epoch_seconds(date dt) {}
static int64 f_get_epoch_seconds_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 588 "ql_time.qpp"
   return dt->getEpochSecondsUTC();
}

// nothing get_epoch_seconds() {}
static AbstractQoreNode* f_get_epoch_seconds(const QoreListNode* args, ExceptionSink* xsink) {
# 595 "ql_time.qpp"

   return 0;
}

// int get_hours(date dt) {}
static int64 f_get_hours_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 953 "ql_time.qpp"
   return dt->getHour();
}

// nothing get_hours() {}
static AbstractQoreNode* f_get_hours(const QoreListNode* args, ExceptionSink* xsink) {
# 960 "ql_time.qpp"

   return 0;
}

// int get_microseconds(date dt) {}
static int64 f_get_microseconds_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 1051 "ql_time.qpp"
   return dt->getMicrosecond();
}

// date get_midnight(date dt) {}
static AbstractQoreNode* f_get_midnight_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 1068 "ql_time.qpp"
   DateTimeNode *rv = new DateTimeNode(*dt);
   rv->setTime(0, 0, 0);
   return rv;
}

// nothing get_midnight() {}
static AbstractQoreNode* f_get_midnight(const QoreListNode* args, ExceptionSink* xsink) {
# 1077 "ql_time.qpp"

   return 0;
}

// int get_milliseconds(date dt) {}
static int64 f_get_milliseconds_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 1026 "ql_time.qpp"
   return dt->getMillisecond();
}

// nothing get_milliseconds() {}
static AbstractQoreNode* f_get_milliseconds(const QoreListNode* args, ExceptionSink* xsink) {
# 1033 "ql_time.qpp"

   return 0;
}

// int get_minutes(date dt) {}
static int64 f_get_minutes_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 976 "ql_time.qpp"
   return dt->getMinute();
}

// nothing get_minutes() {}
static AbstractQoreNode* f_get_minutes(const QoreListNode* args, ExceptionSink* xsink) {
# 983 "ql_time.qpp"

   return 0;
}

// int get_months(date dt) {}
static int64 f_get_months_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 907 "ql_time.qpp"
   return dt->getMonth();
}

// nothing get_months() {}
static AbstractQoreNode* f_get_months(const QoreListNode* args, ExceptionSink* xsink) {
# 914 "ql_time.qpp"

   return 0;
}

// int get_seconds(date dt) {}
static int64 f_get_seconds_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 1001 "ql_time.qpp"
   return dt->getSecond();
}

// nothing get_seconds() {}
static AbstractQoreNode* f_get_seconds(const QoreListNode* args, ExceptionSink* xsink) {
# 1008 "ql_time.qpp"

   return 0;
}

// int get_years(date dt) {}
static int64 f_get_years_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 884 "ql_time.qpp"
   return dt->getYear();
}

// nothing get_years() {}
static AbstractQoreNode* f_get_years(const QoreListNode* args, ExceptionSink* xsink) {
# 891 "ql_time.qpp"

   return 0;
}

// date gmtime() {}
static AbstractQoreNode* f_gmtime(const QoreListNode* args, ExceptionSink* xsink) {
# 511 "ql_time.qpp"
   return DateTimeNode::makeAbsolute(NULL, (int64)time(0), 0);
}

// date gmtime(softint secs, softint us = 0) {}
static AbstractQoreNode* f_gmtime_vivi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 secs = HARD_QORE_INT(args, 0);
   int64 us = HARD_QORE_INT(args, 1);
# 529 "ql_time.qpp"
   return DateTimeNode::makeAbsolute(NULL, secs, (int)us);
}

// date gmtime(date dt) {}
static AbstractQoreNode* f_gmtime_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 546 "ql_time.qpp"
   return DateTimeNode::makeAbsolute(NULL, dt->getEpochSecondsUTC(), dt->getMicrosecond());
}

// date hours(softint hours) {}
static AbstractQoreNode* f_hours_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 hours = HARD_QORE_INT(args, 0);
# 733 "ql_time.qpp"
   return new DateTimeNode(0, 0, 0, (int)hours, 0, 0, 0, true);
}

// nothing hours() {}
static AbstractQoreNode* f_hours(const QoreListNode* args, ExceptionSink* xsink) {
# 740 "ql_time.qpp"

   return 0;
}

// bool is_date_absolute(date dt) {}
static bool f_is_date_absolute_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 1458 "ql_time.qpp"
   return dt->isAbsolute();
}

// bool is_date_absolute() {}
static bool f_is_date_absolute(const QoreListNode* args, ExceptionSink* xsink) {
# 1465 "ql_time.qpp"
   return false;
}

// bool is_date_relative(date dt) {}
static bool f_is_date_relative_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 1434 "ql_time.qpp"
   return dt->isRelative();
}

// bool is_date_relative() {}
static bool f_is_date_relative(const QoreListNode* args, ExceptionSink* xsink) {
# 1441 "ql_time.qpp"
   return false;
}

// date localtime() {}
static AbstractQoreNode* f_localtime(const QoreListNode* args, ExceptionSink* xsink) {
# 459 "ql_time.qpp"
   return DateTimeNode::makeAbsolute(currentTZ(), (int64)time(0), 0);
}

// date localtime(softint secs, softint us = 0) {}
static AbstractQoreNode* f_localtime_vivi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 secs = HARD_QORE_INT(args, 0);
   int64 us = HARD_QORE_INT(args, 1);
# 477 "ql_time.qpp"
   return DateTimeNode::makeAbsolute(currentTZ(), secs, (int)us);
}

// date localtime(date dt) {}
static AbstractQoreNode* f_localtime_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 494 "ql_time.qpp"
   return DateTimeNode::makeAbsolute(currentTZ(), dt->getEpochSecondsUTC(), dt->getMicrosecond());
}

// date microseconds(softint us) {}
static AbstractQoreNode* f_microseconds_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 us = HARD_QORE_INT(args, 0);
# 860 "ql_time.qpp"
   int64 secs = us / 10000000;
   us -= (secs * 10000000);
   int64 minutes = secs / 60;
   secs -= (minutes * 60);
   int64 hours = minutes / 60;
   minutes -= (hours * 60);

   return DateTimeNode::makeRelative(0, 0, 0, hours, minutes, secs, (int)us);
}

// date milliseconds(softint ms) {}
static AbstractQoreNode* f_milliseconds_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 ms = HARD_QORE_INT(args, 0);
# 823 "ql_time.qpp"
   int64 secs = ms / 1000;
   ms -= (secs * 1000);
   int64 minutes = secs / 60;
   secs -= (minutes * 60);
   int64 hours = minutes / 60;
   minutes -= (hours * 60);

   return new DateTimeNode(0, 0, 0, (int)hours, (int)minutes, (int)secs, (int)ms, true);
}

// nothing milliseconds() {}
static AbstractQoreNode* f_milliseconds(const QoreListNode* args, ExceptionSink* xsink) {
# 837 "ql_time.qpp"

   return 0;
}

// date minutes(softint minutes) {}
static AbstractQoreNode* f_minutes_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 minutes = HARD_QORE_INT(args, 0);
# 763 "ql_time.qpp"
   return new DateTimeNode(0, 0, 0, 0, (int)minutes, 0, 0, true);
}

// nothing minutes() {}
static AbstractQoreNode* f_minutes(const QoreListNode* args, ExceptionSink* xsink) {
# 770 "ql_time.qpp"

   return 0;
}

// int mktime(date dt) {}
static int64 f_mktime_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 613 "ql_time.qpp"
   return dt->getEpochSecondsUTC();
}

// nothing mktime() {}
static AbstractQoreNode* f_mktime(const QoreListNode* args, ExceptionSink* xsink) {
# 620 "ql_time.qpp"

   return 0;
}

// date months(softint months) {}
static AbstractQoreNode* f_months_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 months = HARD_QORE_INT(args, 0);
# 673 "ql_time.qpp"
   return new DateTimeNode(0, (int)months, 0, 0, 0, 0, 0, true);
}

// nothing months() {}
static AbstractQoreNode* f_months(const QoreListNode* args, ExceptionSink* xsink) {
# 680 "ql_time.qpp"

   return 0;
}

// date now() {}
static AbstractQoreNode* f_now(const QoreListNode* args, ExceptionSink* xsink) {
# 442 "ql_time.qpp"
   return DateTimeNode::makeAbsolute(currentTZ(), (int64)time(0), 0);
}

// date now_ms() {}
static AbstractQoreNode* f_now_ms(const QoreListNode* args, ExceptionSink* xsink) {
# 352 "ql_time.qpp"
   int us;
   int64 seconds = q_epoch_us(us);

   return DateTimeNode::makeAbsolute(currentTZ(), seconds, (us / 1000) * 1000);
}

// date now_us() {}
static AbstractQoreNode* f_now_us(const QoreListNode* args, ExceptionSink* xsink) {
# 375 "ql_time.qpp"
   int us;
   int64 seconds = q_epoch_us(us);

   return DateTimeNode::makeAbsolute(currentTZ(), seconds, us);
}

// date now_utc() {}
static AbstractQoreNode* f_now_utc(const QoreListNode* args, ExceptionSink* xsink) {
# 393 "ql_time.qpp"
   int us;
   int64 seconds = q_epoch_us(us);
   return DateTimeNode::makeAbsolute(NULL, seconds, us);
}

// date seconds(softint seconds) {}
static AbstractQoreNode* f_seconds_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 seconds = HARD_QORE_INT(args, 0);
# 793 "ql_time.qpp"
   return new DateTimeNode(0, 0, 0, 0, 0, (int)seconds, 0, true);
}

// nothing seconds() {}
static AbstractQoreNode* f_seconds(const QoreListNode* args, ExceptionSink* xsink) {
# 800 "ql_time.qpp"

   return 0;
}

// int timegm(date dt) {}
static int64 f_timegm_Vd(const QoreListNode* args, ExceptionSink* xsink) {
   const DateTimeNode* dt = HARD_QORE_DATE(args, 0);
# 563 "ql_time.qpp"
   return dt->getEpochSeconds();
}

// nothing timegm() {}
static AbstractQoreNode* f_timegm(const QoreListNode* args, ExceptionSink* xsink) {
# 570 "ql_time.qpp"

   return 0;
}

// date years(softint years) {}
static AbstractQoreNode* f_years_vi(const QoreListNode* args, ExceptionSink* xsink) {
   int64 years = HARD_QORE_INT(args, 0);
# 643 "ql_time.qpp"
   return new DateTimeNode((int)years, 0, 0, 0, 0, 0, 0, true);
}

// nothing years() {}
static AbstractQoreNode* f_years(const QoreListNode* args, ExceptionSink* xsink) {
# 650 "ql_time.qpp"

   return 0;
}


DLLLOCAL void init_time_functions(QoreNamespace& ns) {
   // int clock_getmicros() {}
   ns.addBuiltinVariant("clock_getmicros", (q_func_int64_t)f_clock_getmicros, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);
   // int clock_getmillis() {}
   ns.addBuiltinVariant("clock_getmillis", (q_func_int64_t)f_clock_getmillis, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);
   // int clock_getnanos() {}
   ns.addBuiltinVariant("clock_getnanos", (q_func_int64_t)f_clock_getnanos, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);
   // date date(date dt) {}
   ns.addBuiltinVariant("date", (q_func_t)f_date_Vd, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // date date(string dtstr) {}
   ns.addBuiltinVariant("date", (q_func_t)f_date_Vs, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, stringTypeInfo, NULL, "dtstr");
   // date date(float f) {}
   ns.addBuiltinVariant("date", (q_func_t)f_date_Vf, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, floatTypeInfo, NULL, "f");
   // date date(softint i) {}
   ns.addBuiltinVariant("date", (q_func_t)f_date_vi, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, softBigIntTypeInfo, NULL, "i");
   // date date() {}
   ns.addBuiltinVariant("date", (q_func_t)f_date, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo);
   // date date(null null) {}
   ns.addBuiltinVariant("date", (q_func_t)f_date_Vu, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, nullTypeInfo, NULL, "null");
   // date date(string dtstr, string mask) {}
   ns.addBuiltinVariant("date", (q_func_t)f_date_VsVs, QC_RET_VALUE_ONLY, QDOM_DEFAULT, dateTypeInfo, 2, stringTypeInfo, NULL, "dtstr", stringTypeInfo, NULL, "mask");
   // hash date_info(date dt) {}
   ns.addBuiltinVariant("date_info", (q_func_t)f_date_info_Vd, QC_CONSTANT, QDOM_DEFAULT, hashTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // hash date_info() {}
   ns.addBuiltinVariant("date_info", (q_func_t)f_date_info, QC_CONSTANT, QDOM_DEFAULT, hashTypeInfo);
   // date date_ms(softint ms) {}
   ns.addBuiltinVariant("date_ms", (q_func_t)f_date_ms_vi, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, softBigIntTypeInfo, NULL, "ms");
   // nothing date_ms() {}
   ns.addBuiltinVariant("date_ms", (q_func_t)f_date_ms, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // date date_us(softint us) {}
   ns.addBuiltinVariant("date_us", (q_func_t)f_date_us_vi, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, softBigIntTypeInfo, NULL, "us");
   // date days(softint days) {}
   ns.addBuiltinVariant("days", (q_func_t)f_days_vi, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, softBigIntTypeInfo, NULL, "days");
   // nothing days() {}
   ns.addBuiltinVariant("days", (q_func_t)f_days, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string format_date(string format, date dt) {}
   ns.addBuiltinVariant("format_date", (q_func_t)f_format_date_VsVd, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 2, stringTypeInfo, NULL, "format", dateTypeInfo, NULL, "dt");
   // nothing format_date() {}
   ns.addBuiltinVariant("format_date", (q_func_t)f_format_date, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // date getDateFromISOWeek(softint year, softint week, softint day = 1) {}
   ns.addBuiltinVariant("getDateFromISOWeek", (q_func_t)f_getDateFromISOWeek_vivivi, QC_RET_VALUE_ONLY, QDOM_DEFAULT, dateTypeInfo, 3, softBigIntTypeInfo, NULL, "year", softBigIntTypeInfo, NULL, "week", softBigIntTypeInfo, new QoreBigIntNode(1), "day");
   // int getDayNumber(date dt) {}
   ns.addBuiltinVariant("getDayNumber", (q_func_int64_t)f_getDayNumber_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // nothing getDayNumber() {}
   ns.addBuiltinVariant("getDayNumber", (q_func_t)f_getDayNumber, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int getDayOfWeek(date dt) {}
   ns.addBuiltinVariant("getDayOfWeek", (q_func_int64_t)f_getDayOfWeek_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // nothing getDayOfWeek() {}
   ns.addBuiltinVariant("getDayOfWeek", (q_func_t)f_getDayOfWeek, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int getISODayOfWeek(date dt) {}
   ns.addBuiltinVariant("getISODayOfWeek", (q_func_int64_t)f_getISODayOfWeek_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // nothing getISODayOfWeek() {}
   ns.addBuiltinVariant("getISODayOfWeek", (q_func_t)f_getISODayOfWeek, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // hash getISOWeekHash(date dt) {}
   ns.addBuiltinVariant("getISOWeekHash", (q_func_t)f_getISOWeekHash_Vd, QC_CONSTANT, QDOM_DEFAULT, hashTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // nothing getISOWeekHash() {}
   ns.addBuiltinVariant("getISOWeekHash", (q_func_t)f_getISOWeekHash, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // string getISOWeekString(date dt) {}
   ns.addBuiltinVariant("getISOWeekString", (q_func_t)f_getISOWeekString_Vd, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // nothing getISOWeekString() {}
   ns.addBuiltinVariant("getISOWeekString", (q_func_t)f_getISOWeekString, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int get_days(date dt) {}
   ns.addBuiltinVariant("get_days", (q_func_int64_t)f_get_days_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // nothing get_days() {}
   ns.addBuiltinVariant("get_days", (q_func_t)f_get_days, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int get_duration_microseconds(date dt) {}
   ns.addBuiltinVariant("get_duration_microseconds", (q_func_int64_t)f_get_duration_microseconds_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // int get_duration_milliseconds(date dt) {}
   ns.addBuiltinVariant("get_duration_milliseconds", (q_func_int64_t)f_get_duration_milliseconds_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // int get_duration_seconds(date dt) {}
   ns.addBuiltinVariant("get_duration_seconds", (q_func_int64_t)f_get_duration_seconds_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // int get_epoch_seconds(date dt) {}
   ns.addBuiltinVariant("get_epoch_seconds", (q_func_int64_t)f_get_epoch_seconds_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // nothing get_epoch_seconds() {}
   ns.addBuiltinVariant("get_epoch_seconds", (q_func_t)f_get_epoch_seconds, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int get_hours(date dt) {}
   ns.addBuiltinVariant("get_hours", (q_func_int64_t)f_get_hours_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // nothing get_hours() {}
   ns.addBuiltinVariant("get_hours", (q_func_t)f_get_hours, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int get_microseconds(date dt) {}
   ns.addBuiltinVariant("get_microseconds", (q_func_int64_t)f_get_microseconds_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // date get_midnight(date dt) {}
   ns.addBuiltinVariant("get_midnight", (q_func_t)f_get_midnight_Vd, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // nothing get_midnight() {}
   ns.addBuiltinVariant("get_midnight", (q_func_t)f_get_midnight, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int get_milliseconds(date dt) {}
   ns.addBuiltinVariant("get_milliseconds", (q_func_int64_t)f_get_milliseconds_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // nothing get_milliseconds() {}
   ns.addBuiltinVariant("get_milliseconds", (q_func_t)f_get_milliseconds, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int get_minutes(date dt) {}
   ns.addBuiltinVariant("get_minutes", (q_func_int64_t)f_get_minutes_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // nothing get_minutes() {}
   ns.addBuiltinVariant("get_minutes", (q_func_t)f_get_minutes, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int get_months(date dt) {}
   ns.addBuiltinVariant("get_months", (q_func_int64_t)f_get_months_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // nothing get_months() {}
   ns.addBuiltinVariant("get_months", (q_func_t)f_get_months, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int get_seconds(date dt) {}
   ns.addBuiltinVariant("get_seconds", (q_func_int64_t)f_get_seconds_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // nothing get_seconds() {}
   ns.addBuiltinVariant("get_seconds", (q_func_t)f_get_seconds, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int get_years(date dt) {}
   ns.addBuiltinVariant("get_years", (q_func_int64_t)f_get_years_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // nothing get_years() {}
   ns.addBuiltinVariant("get_years", (q_func_t)f_get_years, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // date gmtime() {}
   ns.addBuiltinVariant("gmtime", (q_func_t)f_gmtime, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo);
   // date gmtime(softint secs, softint us = 0) {}
   ns.addBuiltinVariant("gmtime", (q_func_t)f_gmtime_vivi, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 2, softBigIntTypeInfo, NULL, "secs", softBigIntTypeInfo, zero(), "us");
   // date gmtime(date dt) {}
   ns.addBuiltinVariant("gmtime", (q_func_t)f_gmtime_Vd, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // date hours(softint hours) {}
   ns.addBuiltinVariant("hours", (q_func_t)f_hours_vi, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, softBigIntTypeInfo, NULL, "hours");
   // nothing hours() {}
   ns.addBuiltinVariant("hours", (q_func_t)f_hours, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // bool is_date_absolute(date dt) {}
   ns.addBuiltinVariant("is_date_absolute", (q_func_bool_t)f_is_date_absolute_Vd, QC_CONSTANT, QDOM_DEFAULT, boolTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // bool is_date_absolute() {}
   ns.addBuiltinVariant("is_date_absolute", (q_func_bool_t)f_is_date_absolute, QC_NOOP, QDOM_DEFAULT, boolTypeInfo);
   // bool is_date_relative(date dt) {}
   ns.addBuiltinVariant("is_date_relative", (q_func_bool_t)f_is_date_relative_Vd, QC_CONSTANT, QDOM_DEFAULT, boolTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // bool is_date_relative() {}
   ns.addBuiltinVariant("is_date_relative", (q_func_bool_t)f_is_date_relative, QC_NOOP, QDOM_DEFAULT, boolTypeInfo);
   // date localtime() {}
   ns.addBuiltinVariant("localtime", (q_func_t)f_localtime, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo);
   // date localtime(softint secs, softint us = 0) {}
   ns.addBuiltinVariant("localtime", (q_func_t)f_localtime_vivi, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 2, softBigIntTypeInfo, NULL, "secs", softBigIntTypeInfo, zero(), "us");
   // date localtime(date dt) {}
   ns.addBuiltinVariant("localtime", (q_func_t)f_localtime_Vd, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // date microseconds(softint us) {}
   ns.addBuiltinVariant("microseconds", (q_func_t)f_microseconds_vi, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, softBigIntTypeInfo, NULL, "us");
   // date milliseconds(softint ms) {}
   ns.addBuiltinVariant("milliseconds", (q_func_t)f_milliseconds_vi, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, softBigIntTypeInfo, NULL, "ms");
   // nothing milliseconds() {}
   ns.addBuiltinVariant("milliseconds", (q_func_t)f_milliseconds, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // date minutes(softint minutes) {}
   ns.addBuiltinVariant("minutes", (q_func_t)f_minutes_vi, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, softBigIntTypeInfo, NULL, "minutes");
   // nothing minutes() {}
   ns.addBuiltinVariant("minutes", (q_func_t)f_minutes, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int mktime(date dt) {}
   ns.addBuiltinVariant("mktime", (q_func_int64_t)f_mktime_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // nothing mktime() {}
   ns.addBuiltinVariant("mktime", (q_func_t)f_mktime, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // date months(softint months) {}
   ns.addBuiltinVariant("months", (q_func_t)f_months_vi, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, softBigIntTypeInfo, NULL, "months");
   // nothing months() {}
   ns.addBuiltinVariant("months", (q_func_t)f_months, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // date now() {}
   ns.addBuiltinVariant("now", (q_func_t)f_now, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo);
   // date now_ms() {}
   ns.addBuiltinVariant("now_ms", (q_func_t)f_now_ms, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo);
   // date now_us() {}
   ns.addBuiltinVariant("now_us", (q_func_t)f_now_us, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo);
   // date now_utc() {}
   ns.addBuiltinVariant("now_utc", (q_func_t)f_now_utc, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo);
   // date seconds(softint seconds) {}
   ns.addBuiltinVariant("seconds", (q_func_t)f_seconds_vi, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, softBigIntTypeInfo, NULL, "seconds");
   // nothing seconds() {}
   ns.addBuiltinVariant("seconds", (q_func_t)f_seconds, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // int timegm(date dt) {}
   ns.addBuiltinVariant("timegm", (q_func_int64_t)f_timegm_Vd, QC_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, NULL, "dt");
   // nothing timegm() {}
   ns.addBuiltinVariant("timegm", (q_func_t)f_timegm, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);
   // date years(softint years) {}
   ns.addBuiltinVariant("years", (q_func_t)f_years_vi, QC_CONSTANT, QDOM_DEFAULT, dateTypeInfo, 1, softBigIntTypeInfo, NULL, "years");
   // nothing years() {}
   ns.addBuiltinVariant("years", (q_func_t)f_years, QC_RUNTIME_NOOP, QDOM_DEFAULT, nothingTypeInfo);

}
